Package: growthcleanr
Type: Package
Title: Data Cleaner for Anthropometric Measurements
Version: 2.2.3
Authors@R: c(
    person("Carrie", "Daymont", email = "cdaymont@pennstatehealth.psu.edu", role = c("ctb","cre")),
    person("Robert", "Grundmeier", role = "aut"),
    person("Jeffrey", "Miller", role = "aut"),
    person("Diego", "Campos", role = "aut"),
    person("Dan", "Chudnov", email = "dlchudnov@mitre.org", role = "ctb"),
    person("Hannah", "De los Santos", email = "hdelossantos@mitre.org", role = c("ctb")),
    person("Lusha", "Cao", role = "ctb"),
    person("Steffani", "Silva", role = "ctb"),
    person("Hanzhe", "Zhang", role = "ctb"),
    person("Matt", "Boyas", role = "ctb"),
    person("David", "Freedman", role = "ctb"),
    person("Andreas", "Achilleos", role = "ctb"),
    person("Jessica", "Butts", role = "ctb"),
    person("Sheila", "Nguyen", role = "ctb"),
    person("Taraneh", "Soleymani", role = "ctb"),
    person("Max", "Olivier", role = "ctb")
  )
Description: Identifies implausible anthropometric (e.g., height,
    weight) measurements in irregularly spaced longitudinal datasets, such as those from electronic health records.
URL: https://carriedaymont.github.io/growthcleanr/index.html,
        https://github.com/carriedaymont/growthcleanr
BugReports: https://github.com/carriedaymont/growthcleanr/issues
Imports: R.utils (>= 2.11.0), data.table (>= 1.13.0), tidyr (>= 1.1.0),
        plyr (>= 1.8.6), dplyr (>= 1.0.1), foreach (>= 1.5.0),
        doParallel (>= 1.0.15), labelled (>= 2.5.0), magrittr (>= 1.5),
        rlang (>= 0.4.0)
Depends: R (>= 2.10)
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: true
Suggests: argparser (>= 0.6), bit64 (>= 4.0.2), knitr (>= 1.29),
        rmarkdown (>= 2.3), testthat (>= 2.3.2)
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-05-09 14:44:19 UTC; carriedaymont
Author: Carrie Daymont [ctb, cre],
  Robert Grundmeier [aut],
  Jeffrey Miller [aut],
  Diego Campos [aut],
  Dan Chudnov [ctb],
  Hannah De los Santos [ctb],
  Lusha Cao [ctb],
  Steffani Silva [ctb],
  Hanzhe Zhang [ctb],
  Matt Boyas [ctb],
  David Freedman [ctb],
  Andreas Achilleos [ctb],
  Jessica Butts [ctb],
  Sheila Nguyen [ctb],
  Taraneh Soleymani [ctb],
  Max Olivier [ctb]
Maintainer: Carrie Daymont <cdaymont@pennstatehealth.psu.edu>
Repository: CRAN
Date/Publication: 2026-05-09 15:20:02 UTC
Built: R 4.6.0; ; 2026-05-09 18:00:08 UTC; unix

# Generated by roxygen2: do not edit by hand

export(acf_answers)
export(adjustcarryforward)
export(cleangrowth)
export(ewma)
export(ext_bmiz)
export(longwide)
export(read_anthro)
export(recode_sex)
export(sd_median)
export(simple_bmi)
export(splitinput)
export(testacf)
import(R.utils, except = c(extract))
import(data.table)
import(doParallel)
import(dplyr, except = c(last, first, summarize, src, between))
import(foreach)
import(labelled)
import(magrittr)
import(parallel)
import(plyr, except = c(failwith, id, summarize, count, desc, mutate, arrange, rename, is.discrete, summarise, summarize))
import(tidyr, except = extract)
importFrom(rlang,.data)
importFrom(stats,approx)
importFrom(stats,embed)
importFrom(stats,median)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,runif)
importFrom(stats,setNames)
importFrom(utils,read.csv)
importFrom(utils,read.table)
importFrom(utils,tail)
importFrom(utils,write.csv)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# growthcleanr

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/growthcleanr)](https://cran.r-project.org/package=growthcleanr)
[![R build
status](https://github.com/carriedaymont/growthcleanr/workflows/R-CMD-check/badge.svg)](https://github.com/carriedaymont/growthcleanr/actions)
[![Docker](https://github.com/mitre/growthcleanr/actions/workflows/build-and-publish-image-tag.yml/badge.svg)](https://github.com/mitre/growthcleanr/actions/workflows/build-and-publish-image-tag.yml)

<!-- badges: end -->

R package for cleaning data from Electronic Health Record systems,
focused on cleaning height and weight measurements.

<a name="cite"></a> This package implements the [Daymont et
al. algorithm](https://academic.oup.com/jamia/article/24/6/1080/3767271),
as specified in Supplemental File 3 within the [Supplementary
Material](https://academic.oup.com/jamia/article/24/6/1080/3767271#97610899)
published with that paper.

> Carrie Daymont, Michelle E Ross, A Russell Localio, Alexander G Fiks,
> Richard C Wasserman, Robert W Grundmeier, Automated identification of
> implausible values in growth data from pediatric electronic health
> records, Journal of the American Medical Informatics Association,
> Volume 24, Issue 6, November 2017, Pages 1080–1087,
> <https://doi.org/10.1093/jamia/ocx037>

This package also includes an R version of the [SAS macro published by
the
CDC](https://www.cdc.gov/growth-chart-training/hcp/computer-programs/sas.html)
for calculating percentiles and Z-scores of pediatric growth
observations and utilities for working with both functions. As of summer
2021, it also supports cleaning anthropometric measurements for adults
up to age 65. The adult algorithm has not yet been published in a
peer-reviewed publication, but is described in detail at [Adult
algorithm](https://carriedaymont.github.io/growthcleanr/articles/adult-algorithm.html).

## Installation

To install the stable version from CRAN:

``` r
install.packages("growthcleanr")
```

## Summary

The `growthcleanr` package processes data prepared in a specific format
to identify biologically implausible height and weight measurements. It
bases these evaluations on techniques which use patient-specific
longitudinal analysis and variations from published growth trajectory
charts for pediatric subjects. These techniques are performed in a
specific order which refines and improves results throughout the
process.

Results from `growthcleanr` include a flag for each measurement
indicating whether it is to be included or excluded based on
plausibility, with a variety of specific types of exclusions identified
distinctly. These flags can be analyzed further by researchers studying
anthropometric EHR data to determine which measurements to include or
exclude in their own studies. No values are deleted or otherwise
removed; each is only flagged in a new column.

To start running `growthcleanr`, an R installation with a variety of
additional packages is required, as is a growth measurement dataset
prepared for use in `growthcleanr`.

The rest of this documentation includes:

### Getting started:

- [Quickstart](https://carriedaymont.github.io/growthcleanr/articles/quickstart.html),
  a brief tour of using growthcleanr, including data preparation
- [Installation](https://carriedaymont.github.io/growthcleanr/articles/installation.html),
  options for installing growthcleanr, with notes on specific platforms
  and source-level installation for developers
- [Usage](https://carriedaymont.github.io/growthcleanr/articles/usage.html),
  examples of cleaning data, multiple options, example data

### Advanced topics:

- [Configuration
  options](https://carriedaymont.github.io/growthcleanr/articles/configuration.html),
  changing growthcleanr operational settings
- [Understanding growthcleanr
  output](https://carriedaymont.github.io/growthcleanr/articles/output.html),
  the exclusion types growthcleanr identifies
- [Adult
  algorithm](https://carriedaymont.github.io/growthcleanr/articles/adult-algorithm.html),
  a detailed description of how growthcleanr assesses observations from
  adult subjects
- [Computing BMI percentiles and
  Z-scores](https://carriedaymont.github.io/growthcleanr/articles/utilities.html),
  additional functions for common data transforms and determining
  percentiles and Z-scores using the CDC method
- [Working with large
  datasets](https://carriedaymont.github.io/growthcleanr/articles/large-data-sets.html),
  notes and suggestions for running `growthcleanr` with large data
  sources
- [Next
  steps](https://carriedaymont.github.io/growthcleanr/articles/next-steps.html),
  notes on potential enhancements to the pediatric and adult algorithms
- [Developer
  guidelines](https://carriedaymont.github.io/growthcleanr/articles/developer-guidelines.html),
  advice for contributors to this package, including a CRAN release
  checklist

## Changes

For a detailed history of released versions, see the
[Changelog](https://carriedaymont.github.io/growthcleanr/news/index.html)
or`NEWS.md`. Tagged releases, starting with 1.2.3 in January 2021, are
listed [at
GitHub](https://github.com/carriedaymont/growthcleanr/releases).

#!/usr/bin/env Rscript

library(argparser)
library(data.table)

library(growthcleanr)

parser <- arg_parser("CLI driver for growthcleanr")

parser <- add_argument(
  parser,
  "infile",
  type = "character",
  nargs = 1,
  help = "input file"
)
parser <- add_argument(
  parser,
  "outfile",
  type = "character",
  nargs = 1,
  help = "output file"
)
parser <- add_argument(
  parser,
  "--sdrecenter",
  type = "character",
  nargs = 1,
  default = "",
  help = "sd.recenter data file"
)
parser <- add_argument(
  parser,
  "--adult_cutpoint",
  type = "numeric",
  default = 20,
  help = "adult cutpoint"
)
parser <- add_argument(
  parser,
  "--weightcap",
  type = "numeric",
  default = Inf,
  help = "weight cap"
)
parser <- add_argument(
  parser,
  "--numbatches",
  type = "numeric",
  nargs = 1,
  default = 1,
  help = "Number of batches"
)
parser <- add_argument(
  parser,
  "--quietly",
  flag = TRUE,
  help = "Disable verbose output"
)
parser <- add_argument(
  parser,
  "--adult_split",
  type = "numeric",
  default = Inf,
  help = "Number of splits to run data on"
)

argv <- parse_args(parser)
print(argv)

log.path <- sprintf("output/log/%s/%s", Sys.Date(), basename(argv$infile))

if (argv$sdrecenter != "") {
  if (argv$sdrecenter == "nhanes") {
    sdrecenter = "nhanes"
  } else {
    sdrecenter <- fread(argv$sdrecenter)
  }
} else {
  sdrecenter <- ""
}

parallel <- argv$numbatches > 1
num.batches <- argv$numbatches

df_in <- fread(argv$infile)
df_in$id_order <- 1:nrow(df_in)

# handle adult split arguments
adult_split <- argv$adult_split

# we'll process the adult data using a split, then do the work
if (adult_split < Inf & length(unique(df_in$subjid)) < adult_split){
  adult_split <- length(unique(df_in$subjid))
}

if (adult_split < Inf & adult_split > 1){
  # will warn if they're not exact multiples
  # split by subject
  subj_split <- suppressWarnings(split(unique(df_in$subjid), 1:adult_split))
  # map indices to subjects
  # subj_split <- data.frame(
  #   entry = rep(seq_along(subj_split), length(subj_split)),
  #   subjid = as.character(unlist(subj_split))
  # )
  subj_split_df <- data.frame()
  for (spl in 1:length(subj_split)){
    subj_split_df <- rbind(
      subj_split_df,
      data.frame(
        "entry" = spl,
        "subjid" = as.character(subj_split[[spl]])
      ))
  }
  df_in$subjid <- as.character(df_in$subjid)

  # add batch to df_in
  df_in <- merge(df_in, subj_split_df, by = "subjid")

  # split based on subject id
  split.list <- suppressWarnings(split(df_in, df_in$entry))

} else {
  # no need for copy, they can refer to the same thing
  split.df <- df_in
  adult_split <- 1 # for Inf, just converting for later
}

# do split based on number of adult splits
df_out <- lapply(1:adult_split, function(x){
  if (!argv$quietly){
    cat(sprintf(
      "[%s] Processing adult data split %g, using %d batch(es)...\n",
      Sys.time(), x, num.batches))
  }

  if (adult_split > 1){
    split.df <- split.list[[x]]
  } # otherwise, split adult has already been created

  # Separate the logs or they'll overwrite each other
  split.log.path <- sprintf("%s/split-%03d", log.path, x)

  split.df$gcr_result<- cleangrowth(
    split.df$subjid,
    split.df$param,
    split.df$agedays,
    split.df$sex,
    split.df$measurement,
    sd.recenter = sdrecenter,
    adult_cutpoint = argv$adult_cutpoint,
    weight_cap = argv$weightcap,
    log.path = split.log.path,
    num.batches = num.batches,
    parallel = parallel,
    quietly = argv$quietly
  )

  return(split.df)
})

# put all the splits together
df_out <- rbindlist(df_out)
# get it in the original order
df_out <- df_out[order(df_out$id_order),]
# remove indexing variables
idx_var <- if (adult_split == 1){
  c("id_order")
} else {
  c("id_order", "entry")
}
df_out <- df_out[, !idx_var, with = FALSE]

fwrite(df_out, argv$outfile, row.names = FALSE)

#!/usr/bin/env Rscript

### CARRY FORWARD ADJUSTMENT SCRIPT
# The goal of this script is to consider the height values that growthcleanr
# excludes as “carried forward” for potential re-inclusion by using a reverse
# absolute height velocity check based on step 15 of the Daymont et al.
# algorithm.

# Specify libraries ----

library(argparser, quietly = TRUE)
# Load plyr before dplyr intentionally
library(plyr, quietly = T)
library(dplyr, quietly = T)

library(data.table, quietly = T)

library(growthcleanr)

# Specify functions ----

# Function to make a grid vector for future sweep
# inputs:
# - low: number to start vector
# - high: number to end vector
# - grid.length: length of vector
# - searchtype: type of vector to make ("random","line-grid","full-grid")
# - is_included:
# - fg_val: value to use for non-used param in full-grid search
# outputs
# - numeric vector with specified qualities
make_grid_vect <- function(low,
                           high,
                           grid.length,
                           searchtype,
                           is_included = T,
                           fg_val = (low + high) / 2) {
  if (searchtype == "random") {
    # random parameter sweep
    mid <- (low + high) / 2
    # return half random below, midpoint, half random above
    gv <- c(
      runif(floor(grid.length / 2), low, mid),
      mid,
      runif(floor(grid.length / 2), mid, high)
    )

    return(gv)
  } else if (searchtype == "line-grid") {
    # line-grid parameter sweep
    return(seq(low, high, length = grid.length))
  } else if (searchtype == "full-grid") {
    if (is_included) {
      # full grid vectors are put together outside the list
      return(seq(low, high, length = grid.length))
    } else {
      if (is.na(fg_val)) {
        return(c((low + high) / 2))
      } else {
        return(c(fg_val))
      }
    }
  } else {
    stop("Invalid search type")
  }
}

# Function to execute sweep of parameter values for adjustcarryforward
# inputs:
# - grid_df: data frame of parameters to sweep
# outputs:
# - combined input file and result of each run
exec_sweep <- function(grid_df, exclude_opt) {
  for (index in 1:nrow(grid_df)) {
    if (!quietly) {
      cat(
        sprintf(
          "[%s] Calling adjustcarryforward(), run %s of %s\n",
          Sys.time(),
          index,
          nrow(grid_df)
        )
      )
    }
    out <-
      adjustcarryforward(
        subjid,
        param,
        agedays,
        sex,
        measurement,
        orig.exclude,
        exclude_opt = exclude_opt,
        quietly = quietly,
        minfactor = grid_df$minfactor[index],
        maxfactor = grid_df$maxfactor[index],
        banddiff = grid_df$banddiff[index],
        banddiff_plus = grid_df$banddiff_plus[index],
        min_ht.exp_under = grid_df$min_ht.exp_under[index],
        min_ht.exp_over = grid_df$min_ht.exp_over[index],
        max_ht.exp_under = grid_df$max_ht.exp_under[index],
        max_ht.exp_over = grid_df$max_ht.exp_over[index]
      )

    setnames(out, "adjustcarryforward", sprintf("run-%s", index))
    export <-
      merge(as.data.frame(dm_filt),
        out,
        by = "n",
        all = T
      )

    # Combine into one wide set for simplicity
    if (index == 1) {
      combo <- export
    } else {
      combo <- merge(combo, out, by = "n", all = T) # %>% select(-n)
    }
  }

  return(combo)
}

# Specify parser ----

parser <- arg_parser("CLI driver for carry forward adjustments")
parser <- add_argument(parser,
  "infile",
  type = "character",
  nargs = 1,
  help = "Input file, cleaned by cleangrowth()"
)
parser <- add_argument(
  parser,
  "--searchtype",
  default = "random",
  type = "character",
  help = "Type of search to perform: random (default), line-grid, full-grid"
)
parser <- add_argument(
  parser,
  "--gridlength",
  default = 9,
  type = "integer",
  help = "Number of steps in grid to search"
)
parser <- add_argument(parser,
  "--seed",
  default = 7,
  type = "integer",
  help = "Random seed, used only when performing random search"
)
parser <- add_argument(parser,
  "--param",
  default = "none",
  type = "character",
  help = "CSV to specify which parameters to run full search on, and values to use if not, used only when performing full-grid search"
)
parser <- add_argument(parser,
  "--quietly",
  flag = TRUE,
  help = "Verbose debugging info"
)
parser <- add_argument(parser,
  "--debug",
  flag = TRUE,
  help = "Produce extra data files for debugging; use w/--outdir"
)
parser <- add_argument(
  parser,
  "--maxrecs",
  default = 0,
  type = "integer",
  help = "Limit to specified # recs, default 0 (no limit)"
)
parser <- add_argument(parser, "--outfile",
  default = paste0(
    "test_adjustcarryforward_",
    format(Sys.time(), "%m-%d-%Y_%H-%M-%S")
  ),
  type = "character",
  help = "Output file name, default 'test_adjustcarrforward_DATE_TIME', where DATE is the current system date and time"
)
parser <- add_argument(
  parser,
  "--exclude_opt",
  default = 0,
  type = "integer",
  help = "Type of exclusion method for carried forward strings, 0 to 3. See adjustcarryforward documentation for more information"
)
parser <- add_argument(
  parser,
  "--add_answers",
  default = TRUE,
  type = "logical",
  help = "TRUE or FALSE, indicating whether or not to add answers (definely include/exclude) for the given dataset. Defaults to TRUE"
)

# Parse arguments for ease ----

argv <- parse_args(parser)

seed <- argv$seed
searchtype <- argv$searchtype
grid.length <- argv$gridlength
outfile <- argv$outfile
quietly <- argv$quietly
param_choice <- argv$param
exclude_opt <- argv$exclude_opt
add_answers <- argv$add_answers

if (searchtype == "random") {
  # if the line search isn't odd, add one so that it includes the midpoint
  grid.length <-
    if (grid.length %% 2 == 0) {
      grid.length + 1
    } else {
      grid.length
    }
}

# Prepare the data ----

# Read in data
dm <- fread(argv$infile)
setkey(dm, subjid, param, agedays)

# limit dataset for debugging/performance reasons if specified
if (argv$maxrecs > 0) {
  if (!quietly) {
    cat(sprintf(
      "[%s] Taking max %s recs for debugging\n",
      Sys.time(),
      argv$maxrecs
    ))
  }
  dm_filt <- dm[subjid %in% unique(dm[, subjid])[1:argv$maxrecs], ]
} else {
  dm_filt <- dm
}

dm_filt$n <- as.numeric(rownames(dm_filt))

subjid <- with(dm_filt, subjid)
param <- with(dm_filt, param)
sex <- with(dm_filt, sex)
agedays <- with(dm_filt, agedays)
measurement <- with(dm_filt, measurement)
orig.exclude <- with(dm_filt, gcr_result)

# Find out answers ----

if (add_answers){
  if (!quietly) {
    cat(
      sprintf(
        "[%s] Finding answers for cleaned dataset\n",
        Sys.time()
      )
    )
  }

  acf_answer_df <- acf_answers(subjid,
                               param,
                               agedays,
                               sex,
                               measurement,
                               orig.exclude,
                               quietly = quietly)

  dm_filt <- merge(as.data.frame(dm_filt),
        acf_answer_df,
        by = "n",
        all = T
  )
}

# Execute parameter sweep ----

if (!quietly) {
  cat(
    sprintf(
      "[%s] Running parameter sweep of adjustcarryforward on cleaned dataset\n",
      Sys.time()
    )
  )
}

# Specify seed, if using random
if (searchtype == "random") {
  set.seed(seed)
}

# Have a data frame of ranges
param_n <- c(
  "minfactor",
  "maxfactor",
  "banddiff",
  "banddiff_plus",
  "min_ht.exp_under",
  "min_ht.exp_over",
  "max_ht.exp_under",
  "max_ht.exp_over"
)

p_range <- data.frame(
  "param" = param_n,
  low = c(0, 0, 0, 0, 0, -1, 0, 0),
  high = c(1, 4, 6, 11, 4, 1, 0.66, 3)
)
rownames(p_range) <- p_range$param

# Define the params to test
if (searchtype == "full-grid") {
  # specify parameter choices -- default to choosing all
  if (param_choice != "none") {
    param_df <- read.csv(param_choice)
    colnames(param_df) <- c("param", "include", "value")
  } else {
    param_df <- data.frame(
      "param" = param_n,
      "include" = T,
      "value" = NA
    )
  }

  if (searchtype == "full-grid") {
    cat(
      paste0(
        "[",
        Sys.time(),
        "] Note: due to the search type specified (full-grid), there will be ",
        grid.length^sum(param_df$include),
        " runs\n"
      )
    )
  }

  # now pass them in one by one
  grid_list <- list()
  for (i in 1:nrow(param_df)) {
    grid_list[[param_df$param[i]]] <-
      make_grid_vect(
        p_range[param_df$param[i], "low"],
        p_range[param_df$param[i], "high"],
        grid.length,
        searchtype,
        as.logical(param_df$include[i]),
        as.numeric(param_df$value[i])
      )
  }

  # make the full grid based on the list
  grid_df <- expand.grid(grid_list)
} else {
  # for random and line-grid options
  grid_df <-
    as.data.frame(matrix(NA, nrow = grid.length, ncol = nrow(p_range)))
  colnames(grid_df) <- p_range$param
  for (i in 1:ncol(grid_df)) {
    grid_df[, p_range$param[i]] <-
      make_grid_vect(
        p_range[p_range$param[i], "low"],
        p_range[p_range$param[i], "high"],
        grid.length,
        searchtype
      )
  }
}


# Execute
combo <- exec_sweep(grid_df, exclude_opt)

# Write out results ----

# Combined adjusted set
fwrite(combo %>% select(-n), paste0(outfile, ".csv"), row.names = F)

# Record the sweep parameters for review
grid <- cbind("run" = 1:nrow(grid_df), grid_df)
fwrite(grid, paste0(outfile, "_parameters.csv"), row.names = F)

# Any additional debugging output
if (argv$debug) {
  if (!quietly) {
    cat(sprintf("[%s] Writing debugging output files\n", Sys.time()))
  }
  # Filtered data
  fwrite(dm_filt, paste0(outfile, "_debug_filtered_data.csv"), row.names = F)
}

### Define the helper function and the carryforward adjustment function ###
# Segments labeled with ADJUSTCF EDIT in comments are areas where substantive
# changes were made to the original cleangrowth logic. There are other changes
# as well that are not specifically called out.

# helper function to treat NA values as FALSE
na.as.false = function(v) {
  v[is.na(v)] = FALSE
  v
}

# function to calculate temporary exclusion for step 15 (a - q)
calc_temp_exclusion_15 <- function(
  df,
  ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
  ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
  min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over){

  # avoid "no visible binding" warnings
  abs.tbc.sd <- abs.tbc.sd.next <- abs.tbc.sd.prev <- aft.g.befp1 <- NULL
  agedays <- agedays.next <- bef.g.aftm1 <- delta.agedays.next <- NULL
  delta.next.ht <- delta.prev.ht <- dewma.after <- dewma.after.prev <- NULL
  dewma.before <- dewma.before.next <- ewma.after <- ewma.all <- ewma.before <- NULL
  ht.exp <- index <- max.ht.vel <- max.whoinc.1.ht <- max.whoinc.2.ht <- max.whoinc.3.ht <- NULL
  max.whoinc.4.ht <- max.whoinc.6.ht <- maxdiff.next.ht <- maxdiff.prev.ht <- NULL
  mid.agedays <- min.ht.vel <- mindiff.next.ht <- mindiff.prev.ht <- pair <- NULL
  pair.next <- pair.prev <- sex <- tanner.months <- tbc.sd <- temp.diff <- temp.exclude <- NULL
  v <- v.next <- v.prev <- who.maxdiff.next.ht <- who.mindiff.next.ht <- whoagegrp.ht <- NULL
  whoinc.1.ht <- whoinc.2.ht <- whoinc.3.ht <- whoinc.4.ht <- whoinc.6.ht <- NULL
  whoinc.age.ht <- NULL


  # initialize fields
  df[, (ewma.fields) := as.double(NaN)]
  df[, `:=`(
    v.prev = as.double(NaN),
    v.next = as.double(NaN),
    dewma.after.prev = as.double(NaN),
    dewma.before.next = as.double(NaN),
    abs.tbc.sd.prev = as.double(NaN),
    abs.tbc.sd.next = as.double(NaN),
    agedays.next = as.integer(NaN),
    abs.2ndlast.sd = as.double(NaN),
    mindiff.prev.ht = as.double(NaN),
    mindiff.next.ht = as.double(NaN),
    maxdiff.prev.ht = as.double(NaN),
    maxdiff.next.ht = as.double(NaN),
    pair.prev = FALSE,
    pair.next = FALSE
  )]

  # ewma fields are needed later -- calculate now for efficiency
  df[, (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE)]

  # calculate some useful values (e.g. dewma values and tbc.sd) for use in later steps
  df[, `:=`(
    dewma.all = tbc.sd - ewma.all,
    dewma.before = tbc.sd - ewma.before,
    dewma.after = tbc.sd - ewma.after,
    abs.tbc.sd = abs(tbc.sd)
  )]

  # 15a.  As with steps 11 and 14, only one value will be excluded per round, and the step will be repeated until there are no more values to exclude
  # b.  For each height, calculate the d_age=agedays of next value-agedays of current value
  # NOTE: obtain next measurement, ewma.before and abs.tbc.sd as well since they are needed later

  # for efficiency, bring get.next inline here (working on valid rows within a single parameter for a single subject)
  # structure c(field.name[-1], NA) == get.next
  df[, `:=`(
    agedays.next = c(agedays[-1], NA),
    v.next = c(v[-1], NA),
    dewma.before.next = c(dewma.before[-1], NA),
    abs.tbc.sd.next = c(abs.tbc.sd[-1], NA)
  )]
  df[, delta.agedays.next := agedays.next - agedays]

  # 15c.	For each height, calculate mid_agedays=0.5*(agedays of next value + agedays of current value)
  df[, mid.agedays := 0.5 * (agedays.next + agedays)]

  # 15d.	Generate variable tanner_months= 6+12*(round(mid_agedays/365.25))
  # only calculate for rows that relate to height (may speed up subsequent processing)
  df[, tanner.months := 6 + 12 * (round(mid.agedays / 365.25))]

  # 15e.	Merge with dataset tanner_ht_vel using sex and tanner_months – this will give you min_ht_vel and max_ht_vel
  setkey(df, sex, tanner.months)
  df = tanner.ht.vel[df]

  # 15f.	Calculate the following:
  #   i.	mindiff_ht=0.5*min_ht_vel*(d_agedays/365.25)^2-3 if d_agedays<365.25
  #   ii.	replace mindiff_ht=0.5*min_ht_vel-3 if d_agedays>365.25
  df[, ht.exp := ifelse(delta.agedays.next < 365.25,
                        min_ht.exp_under,
                        min_ht.exp_over)]
  df[, `:=`(maxdiff.next.ht = as.double(NA),
            mindiff.next.ht = as.double(NaN))]
  df[, mindiff.next.ht := minfactor * min.ht.vel * (delta.agedays.next /
                                                      365.25) ^ ht.exp - banddiff]

  # 15f.iii.	maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^1.5+5.5 if d_agedays>365.25
  #   iv.	replace maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^0.33+5.5 if d_agedays<365.25
  df[, ht.exp := ifelse(delta.agedays.next < 365.25,
                        max_ht.exp_under,
                        max_ht.exp_over)]
  df[, maxdiff.next.ht := maxfactor * max.ht.vel * (delta.agedays.next /
                                                      365.25) ^ ht.exp + banddiff_plus]

  # 15g.	Generate variable whoagegrp_ht=agedays/30.4375 rounded to the nearest integer
  df[, whoagegrp.ht := round(agedays / 30.4375)]

  # 15h.	Generate variable whoinc_age_ht based on values of d_agedays_ht according the the following table
  #   d_agedays_ht	whoinc_age_ht
  #   20-45	        1
  #   46-75	        2
  #   76-106	      3
  #   107-152	      4
  #   153-198	      6
  #   All others	  missing
  df[, whoinc.age.ht :=
       ifelse(delta.agedays.next < 20 ,
              NA,
              ifelse(
                delta.agedays.next <= 45,
                1,
                ifelse(
                  delta.agedays.next <= 75,
                  2,
                  ifelse(
                    delta.agedays.next <= 106,
                    3,
                    ifelse(
                      delta.agedays.next <= 152,
                      4,
                      ifelse(delta.agedays.next <= 198, 6, NA)
                    )
                  )
                )
              ))]

  # i.	Merge using sex and whoagegrp_ht using who_ht_vel_3sd and who_ht_maxvel_3sd; this will give you varaibles whoinc_i_ht and maxwhoinc_i_ht
  #     for various intervals where i is 1,2, 3,4, 6 and corresponds to whoinc_age_ht.
  setkey(df, sex, whoagegrp.ht)
  df = who.ht.vel[df]

  # restore original sort order (ensures valid.rows variable applies to correct rows)
  setkey(df, index)

  # 15j.	Generate variable who_mindiff_ht=whoinc_i_ht according to the value if whoinc_age_ht; make who_mindiff_ht missing if whoinc_i_ht or whoinc_age_ht is missing.
  df[, who.mindiff.next.ht := ifelse(
    delta.agedays.next < 20 ,
    NA,
    ifelse(
      delta.agedays.next <= 45,
      whoinc.1.ht,
      ifelse(
        delta.agedays.next <= 75,
        whoinc.2.ht,
        ifelse(
          delta.agedays.next <= 106,
          whoinc.3.ht,
          ifelse(
            delta.agedays.next <= 152,
            whoinc.4.ht,
            ifelse(delta.agedays.next <= 198, whoinc.6.ht, NA)
          )
        )
      )
    )
  )]

  # 15k.	Generate variable who_maxdiff_ht=max_whoinc_i_ht according to the value if whoinc_age_ht; make who_maxdiff_ht missing if max_whoinc_i_ht or
  #     whoinc_age_ht is missing.
  df[, who.maxdiff.next.ht := ifelse(
    delta.agedays.next < 20 ,
    NA,
    ifelse(
      delta.agedays.next <= 45,
      max.whoinc.1.ht,
      ifelse(
        delta.agedays.next <= 75,
        max.whoinc.2.ht,
        ifelse(
          delta.agedays.next <= 106,
          max.whoinc.3.ht,
          ifelse(
            delta.agedays.next <= 152,
            max.whoinc.4.ht,
            ifelse(delta.agedays.next <= 198, max.whoinc.6.ht, NA)
          )
        )
      )
    )
  )]

  # 15l.	Scale allowed value based on d_agedays_ht:
  #   1.	replace who_mindiff_`p'=who_mindiff_`p'*d_agedays_`p'/(whoinc_age_`p'*30.4375) if d_agedays_`p'<(whoinc_age_`p'*30.4375)
  #   2.	replace who_maxdiff_`p'=who_maxdiff_`p'*d_agedays_`p'/(whoinc_age_`p'*30.4375) if d_agedays_`p'>(whoinc_age_`p'*30.4375)
  df[delta.agedays.next < whoinc.age.ht * 30.4375,
     `:=`(
       who.mindiff.next.ht = who.mindiff.next.ht * delta.agedays.next / (whoinc.age.ht *
                                                                           30.4375),
       who.maxdiff.next.ht = who.maxdiff.next.ht * delta.agedays.next /
         (whoinc.age.ht * 30.4375)
     )]

  # 15m.	Replace mindiff_ht/maxdiff_ht with adjusted WHO value if Tanner value is missing or if both are present and age difference is < 9 months:
  #   1.	replace mindiff_`p'=0.5*who_mindiff_`p'-3 if who_mindiff_`p' is not missing & d_agedays_`p'<(9*30.4375)
  #   2.	replace maxdiff_`p'=2*who_maxdiff_`p'+3 if who_maxdiff_`p' is not missing & d_agedays_`p'<(9*30.4375)
  #   3.	replace mindiff_`p'=0.5*who_mindiff_`p'-3 if mindiff_`p' is missing & who_mindiff_`p' is not missing
  #   4.	replace maxdiff_`p'=2*who_maxdiff_`p'+3 if maxdiff_`p is missing & who_maxdiff_`p' is not missing

  # refactored logic slightly for efficiency
  df[!is.na(who.mindiff.next.ht) &
       (delta.agedays.next < 9 * 30.4375 |
          is.na(mindiff.next.ht)),
     `:=`(
       mindiff.next.ht = minfactor * who.mindiff.next.ht - banddiff,
       maxdiff.next.ht = maxfactor * who.maxdiff.next.ht + banddiff
     )]

  # 15m.5.  replace mindiff_`p'=-3 if mindiff_`p' is missing
  df[is.na(mindiff.next.ht), mindiff.next.ht := -3]

  # 15n.	Determine the min/maxdiffs for the previous age: mindiff_prev_ht, maxdiff_prev_ht
  # NOTE: obtain previous height, ewma.after value and abs.tbc.sd as well since they are needed in next steps

  # for efficiency, bring get.prev inline here (working on valid rows within a single parameter for a single subject)
  # structure c(NA, tbc.sd[-.N]) == get.prev
  df[, `:=`(
    v.prev = c(NA, v[-.N]),
    dewma.after.prev = c(NA, dewma.after[-.N]),
    abs.tbc.sd.prev = c(NA, abs.tbc.sd[-.N]),
    mindiff.prev.ht = c(NA, mindiff.next.ht[-.N]),
    maxdiff.prev.ht = c(NA, maxdiff.next.ht[-.N])
  )]

  # 15o.	Determine d_prev_ht=ht-htprev (set to missing for the first value for a subject) and d_next_ht=htnext-ht (set to missing for the last value for a subject)
  df[, `:=`(delta.prev.ht = v - v.prev,
            delta.next.ht = v.next - v)]

  # 15p.  Perform a EWMA calculation with the following modifications:
  #  i.	  Generate a variable pair=1 if (d_prev_ht<mindiff_prev_ht OR d_ht<mindiff_ht OR d_prev_ht>maxdiff_prev_ht  OR d_ht>maxdiff_ht) AND exc_ht==0
  df[, pair := na.as.false(
    delta.prev.ht < mindiff.prev.ht |
      delta.next.ht < mindiff.next.ht |
      delta.prev.ht > maxdiff.prev.ht |
      delta.next.ht > maxdiff.next.ht
  )]

  # for efficiency, bring get.prev and get.next inline here (working on valid rows within a single parameter for a single subject)
  # structure c(NA, field.name[-.N]) == get.prev
  # structure c(field.name[-1], NA) == get.next
  df[, `:=`(pair.prev = c(FALSE, pair[-.N]),
            pair.next = c(pair[-1], FALSE))]

  #  ii.	Generate bef_g_aftm1=1 if |Δewma_htbef| for the value of interest is greater than |Δewma_htaft| for the previous value
  #       AND the value of interest is not the first height value for that subject AND pair==1 AND pair for the previous value==1

  #  iii.	Generate aft_g_befp1=1 if |Δewma_htaft| for the value of interest is greater than |Δewma_htbef| for the next value
  #       AND the value of interest is not the last height value for that subject AND pair==1 AND pair for the next value==1
  # NOTE: pair.next will be NA last height, which will result in a FALSE value below
  df[, `:=`(
    bef.g.aftm1 = na.as.false(
      abs(dewma.before) > abs(dewma.after.prev)  & pair & pair.prev
    ),
    aft.g.befp1 = na.as.false(
      abs(dewma.after)  > abs(dewma.before.next) & pair & pair.next
    )
  )]

  #  iv.	Determine tbchtsd for each value as well as the one before prev_tbchtsd and after next_tbchtsd it
  # NOTE: done previously for efficiency

  # 15p.v.  Determine the total number of ht values for each subject (tot_ht)
  # NOTE: all rows are valid due to constraint in subj.df[...] statement
  num.valid = .N

  # 15q.	Identify a value for possible exclusion if one of the following sets of criteria are met. For values identified by each set of criteria determine
  #       the value of temp_diff using the formula given
  #   i.	d_prev_ht<mindiff_prev_ht & bef_g_aftm1_ht==1 & exc_ht==0 & mindiff_prev_ht is not missing
  #     a.  (temp_diff=|dewma_ht_bef|)
  df[, temp.diff := as.double(NaN)]
  df[, temp.exclude := factor(NA, levels = exclude.levels, ordered =
                                TRUE)]
  df[delta.prev.ht < mindiff.prev.ht & bef.g.aftm1,
     `:=`(temp.diff = abs(dewma.before),
          temp.exclude = 'Exclude-Min-Height-Change')]

  #   ii.	d_ht<mindiff_ht & aft_g_befp1_ht==1 & exc_ht==0 & mindiff_ht is not missing
  #     a.	(temp_diff=|dewma_ht_aft|)
  df[delta.next.ht < mindiff.next.ht & aft.g.befp1,
     `:=`(temp.diff = abs(dewma.after),
          temp.exclude = 'Exclude-Min-Height-Change')]

  #   iii.	d_prev_ht>maxdiff_prev_ht & bef_g_aftm1_ht==1 & exc_ht==0 & mindiff_prev_ht is not missing
  #     a.  (temp_diff=|dewma_ht_bef|)
  df[delta.prev.ht > maxdiff.prev.ht & bef.g.aftm1,
     `:=`(temp.diff = abs(dewma.before),
          temp.exclude = 'Exclude-Max-Height-Change')]

  #   iv.	d_ht>maxdiff_ht & aft_g_befp1_ht==1 & exc_ht==0 & mindiff_ht is not missing
  #     a.  (temp_diff=|dewma_ht_aft|)
  df[delta.next.ht > maxdiff.next.ht & aft.g.befp1,
     `:=`(temp.diff = abs(dewma.after),
          temp.exclude = 'Exclude-Max-Height-Change')]

  #   v.	d_prev_ht<mindiff_prev_ht & tot_ht==2 & |tbchtsd|>|prev_tbchtsd|
  #     a. for v-viii temp_diff is kept as missing
  #   vi. d_ht<mindiff_ht & tot_ht==2 & |tbchtsd|>|next_tbchtsd|
  df[delta.prev.ht < mindiff.prev.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.prev
     |
       delta.next.ht < mindiff.next.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.next,
     temp.exclude := 'Exclude-Min-Height-Change']

  #   vii.	d_prev_ht>maxdiff_prev_ht & tot_ht==2 & |tbchtsd|>|prev_tbchtsd|
  #   viii. d_ht>maxdiff_ht & tot_ht==2 & |tbchtsd|>|next_tbchtsd|
  df[delta.prev.ht > maxdiff.prev.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.prev
     |
       delta.next.ht > maxdiff.next.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.next,
     temp.exclude := 'Exclude-Max-Height-Change']

  return(df)
}

# function to check every carried forward in a string one by one
check_cf_string <- function(
  eval_df, wh_exclude,
  ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
  ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
  min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over){

  # find the next include OR the row end
  next_incl <- which(
    eval_df$orig.exclude[(wh_exclude+1):nrow(eval_df)] == "Include"
  )[1]+wh_exclude
  if (is.na(next_incl)){
    next_incl <- nrow(eval_df)+1
  }

  # find the include this is based on
  first_incl <- which(
    eval_df$orig.exclude[1:wh_exclude] == "Include"
  )
  first_incl <- first_incl[length(first_incl)]

  # now we want to test all of the ones in the string
  verdict <- "Include"
  cf_ind <- first_incl+1
  while(verdict == "Include" & cf_ind < next_incl){
    # build the dataframe for evaluation
    sub_df <- copy(
      if (next_incl <= nrow(eval_df)){
        eval_df[c(first_incl, cf_ind, next_incl),]
      } else {
        eval_df[c(first_incl, cf_ind),]
      }
    )

    eval_sub <- calc_temp_exclusion_15(
      copy(sub_df),
      ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
      ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
      min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over)

    verdict <-
      if (is.na(eval_sub$temp.exclude[2])){
        "Include"
      } else {
        "Exclude"
      }

    cf_ind <- cf_ind + 1
  }

  return(list("verdict" = verdict,
              "cf_ind" = cf_ind,
              "next_incl" = next_incl,
              "first_incl" = first_incl))
}

# function to calculate step 15 as in the original algorithm (no parameters)
# eval type: definitely "exclude" or "include"
calc_step_15_no_param <- function(
  df,
  eval_type = "exclude",
  ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
  ewma.exp){

  # avoid "no visible binding for global variable" warnings
  agedays <- tbc.sd <- ewma.all <- ewma.before <- ewma.after <- v <- dewma.before <- NULL
  abs.tbc.sd <- sex <- tanner.months <- ht.exp <- delta.agedays.next <- mindiff.next.ht <- NULL
  min.ht.vel <- maxdiff.next.ht <- max.ht.vel <- minhtvel.exp <- min.ht.vel.2sd <- NULL
  max.ht.vel.2sd <- whoagegrp.ht <- whoinc.age.ht <- index <- who.mindiff.next.ht <- NULL
  whoinc.1.ht <- whoinc.2.ht <- whoinc.3.ht <- whoinc.4.ht <- whoinc.6.ht <- NULL
  who.maxdiff.next.ht <- max.whoinc.1.ht <- max.whoinc.2.ht <- max.whoinc.3.ht <- NULL
  max.whoinc.4.ht <- max.whoinc.6.ht <- dewma.after <- v.prev <- v.next <- pair <- NULL
  delta.prev.ht <- mindiff.prev.ht <- delta.next.ht <- maxdiff.prev.ht <- NULL
  dewma.after.prev <- pair.prev <- dewma.before.next <- pair.next <- temp.diff <- NULL
  bef.g.aftm1 <- aft.g.befp1 <- abs.tbc.sd.prev <- abs.tbc.sd.next <- temp.exclude <- NULL

  # initialize fields
  df[, (ewma.fields) := as.double(NaN)]
  df[, `:=`(
    v.prev = as.double(NaN),
    v.next = as.double(NaN),
    dewma.after.prev = as.double(NaN),
    dewma.before.next = as.double(NaN),
    abs.tbc.sd.prev = as.double(NaN),
    abs.tbc.sd.next = as.double(NaN),
    agedays.next = as.integer(NaN),
    abs.2ndlast.sd = as.double(NaN),
    mindiff.prev.ht = as.double(NaN),
    mindiff.next.ht = as.double(NaN),
    maxdiff.prev.ht = as.double(NaN),
    maxdiff.next.ht = as.double(NaN),
    pair.prev = FALSE,
    pair.next = FALSE
  )]

  # ewma fields are needed later -- calculate now for efficiency
  df[, (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE)]

  # calculate some usefule values (e.g. dewma values and tbc.sd) for use in later steps
  df[, `:=`(
    dewma.all = tbc.sd - ewma.all,
    dewma.before = tbc.sd - ewma.before,
    dewma.after = tbc.sd - ewma.after,
    abs.tbc.sd = abs(tbc.sd)
  )]

  # 15a.  As with steps 11 and 14, only one value will be excluded per round, and the step will be repeated until there are no more values to exclude
  # b.  For each height, calculate the d_age=agedays of next value-agedays of current value
  # NOTE: obtain next measurement, ewma.before and abs.tbc.sd as well since they are needed later

  # for efficiency, bring get.next inline here (working on valid rows within a single parameter for a single subject)
  # structure c(field.name[-1], NA) == get.next
  df[, `:=`(
    agedays.next = c(agedays[-1], NA),
    v.next = c(v[-1], NA),
    dewma.before.next = c(dewma.before[-1], NA),
    abs.tbc.sd.next = c(abs.tbc.sd[-1], NA)
  )]
  df$delta.agedays.next <- with(df, agedays.next - agedays)

  # 15c.	For each height, calculate mid_agedays=0.5*(agedays of next value + agedays of current value)
  df$mid.agedays <- 0.5 * (df$agedays.next + df$agedays)

  # 15d.	Generate variable tanner_months= 6+12*(round(mid_agedays/365.25))
  # only calculate for rows that relate to height (may speed up subsequent processing)
  df$tanner.months <-
    with(df, 6 + 12 * (round(mid.agedays / 365.25)))

  # 15e.	Merge with dataset tanner_ht_vel using sex and tanner_months – this will give you min_ht_vel and max_ht_vel
  setkey(df, sex, tanner.months)
  df <- tanner.ht.vel[df]

  if (eval_type == "exclude"){
    # 15f.	Calculate the following:
    #   i.	mindiff_ht=0.5*min_ht_vel*(d_agedays/365.25)^2-3 if d_agedays<365.25
    #   ii.	replace mindiff_ht=0.5*min_ht_vel-3 if d_agedays>365.25
    df[, ht.exp := ifelse(delta.agedays.next < 365.25, 2, 0)]
    df[, `:=`(maxdiff.next.ht = as.double(NA),
              mindiff.next.ht = as.double(NaN))]
    df[, mindiff.next.ht := 0.5 * min.ht.vel * (delta.agedays.next /
                                                  365.25) ^ ht.exp - 3]

    # 15f.iii.	maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^1.5+5.5 if d_agedays>365.25
    #   iv.	replace maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^0.33+5.5 if d_agedays<365.25
    df[, ht.exp := ifelse(delta.agedays.next < 365.25, 0.33, 1.5)]
    df[, maxdiff.next.ht := 2 * max.ht.vel * (delta.agedays.next /
                                                365.25) ^ ht.exp + 5.5]
  } else { # we're evaluating ones we want to keep
    # 15f.	Calculate the following:
    #   i.	mindiff_ht=0.5*min_ht_vel*(d_agedays/365.25)^2-3 if d_agedays<365.25
    #   ii.	replace mindiff_ht=0.5*min_ht_vel-3 if d_agedays>365.25
    df[, ht.exp := ifelse(delta.agedays.next < 365.25, 1, 0)]
    df[, minhtvel.exp := ifelse(delta.agedays.next < 365.25, 0, 1)]
    df[, `:=`(maxdiff.next.ht = as.double(NA),
              mindiff.next.ht = as.double(NaN))]
    df[, mindiff.next.ht := min.ht.vel.2sd * (min.ht.vel) ^ minhtvel.exp *
         (delta.agedays.next /365.25) ^ ht.exp]

    # 15f.iii.	maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^1.5+5.5 if d_agedays>365.25
    #   iv.	replace maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^0.33+5.5 if d_agedays<365.25
    df[, maxdiff.next.ht := max.ht.vel.2sd * (delta.agedays.next / 365.25)]
  }

  # 15g.	Generate variable whoagegrp_ht=agedays/30.4375 rounded to the nearest integer
  df[, whoagegrp.ht := round(agedays / 30.4375)]

  # 15h.	Generate variable whoinc_age_ht based on values of d_agedays_ht according the the following table
  #   d_agedays_ht	whoinc_age_ht
  #   20-45	        1
  #   46-75	        2
  #   76-106	      3
  #   107-152	      4
  #   153-198	      6
  #   All others	  missing
  df[, whoinc.age.ht := ifelse(delta.agedays.next < 20 ,
                               NA,
                               ifelse(
                                 delta.agedays.next <= 45,
                                 1,
                                 ifelse(
                                   delta.agedays.next <= 75,
                                   2,
                                   ifelse(
                                     delta.agedays.next <= 106,
                                     3,
                                     ifelse(
                                       delta.agedays.next <= 152,
                                       4,
                                       ifelse(delta.agedays.next <= 198, 6, NA)
                                     )
                                   )
                                 )
                               ))]

  # i.	Merge using sex and whoagegrp_ht using who_ht_vel_3sd and who_ht_maxvel_3sd; this will give you varaibles whoinc_i_ht and maxwhoinc_i_ht
  #     for various intervals where i is 1,2, 3,4, 6 and corresponds to whoinc_age_ht.
  setkey(df, sex, whoagegrp.ht)
  df <- who.ht.vel[df]

  # restore original sort order (ensures valid.rows variable applies to correct rows)
  setkey(df, index)

  # 15j.	Generate variable who_mindiff_ht=whoinc_i_ht according to the value if whoinc_age_ht; make who_mindiff_ht missing if whoinc_i_ht or whoinc_age_ht is missing.
  df[, who.mindiff.next.ht := ifelse(
    delta.agedays.next < 20 ,
    NA,
    ifelse(
      delta.agedays.next <= 45,
      whoinc.1.ht,
      ifelse(
        delta.agedays.next <= 75,
        whoinc.2.ht,
        ifelse(
          delta.agedays.next <= 106,
          whoinc.3.ht,
          ifelse(
            delta.agedays.next <= 152,
            whoinc.4.ht,
            ifelse(delta.agedays.next <= 198, whoinc.6.ht, NA)
          )
        )
      )
    )
  )]

  # 15k.	Generate variable who_maxdiff_ht=max_whoinc_i_ht according to the value if whoinc_age_ht; make who_maxdiff_ht missing if max_whoinc_i_ht or
  #     whoinc_age_ht is missing.
  df[, who.maxdiff.next.ht := ifelse(
    delta.agedays.next < 20 ,
    NA,
    ifelse(
      delta.agedays.next <= 45,
      max.whoinc.1.ht,
      ifelse(
        delta.agedays.next <= 75,
        max.whoinc.2.ht,
        ifelse(
          delta.agedays.next <= 106,
          max.whoinc.3.ht,
          ifelse(
            delta.agedays.next <= 152,
            max.whoinc.4.ht,
            ifelse(delta.agedays.next <= 198, max.whoinc.6.ht, NA)
          )
        )
      )
    )
  )]

  # 15l.	Scale allowed value based on d_agedays_ht:
  #   1.	replace who_mindiff_`p'=who_mindiff_`p'*d_agedays_`p'/(whoinc_age_`p'*30.4375) if d_agedays_`p'<(whoinc_age_`p'*30.4375)
  #   2.	replace who_maxdiff_`p'=who_maxdiff_`p'*d_agedays_`p'/(whoinc_age_`p'*30.4375) if d_agedays_`p'>(whoinc_age_`p'*30.4375)
  df[delta.agedays.next < whoinc.age.ht * 30.4375,
     `:=`(
       who.mindiff.next.ht = who.mindiff.next.ht * delta.agedays.next / (whoinc.age.ht *
                                                                           30.4375),
       who.maxdiff.next.ht = who.maxdiff.next.ht * delta.agedays.next /
         (whoinc.age.ht * 30.4375)
     )]

  # 15m.	Replace mindiff_ht/maxdiff_ht with adjusted WHO value if Tanner value is missing or if both are present and age difference is < 9 months:
  #   1.	replace mindiff_`p'=0.5*who_mindiff_`p'-3 if who_mindiff_`p' is not missing & d_agedays_`p'<(9*30.4375)
  #   2.	replace maxdiff_`p'=2*who_maxdiff_`p'+3 if who_maxdiff_`p' is not missing & d_agedays_`p'<(9*30.4375)
  #   3.	replace mindiff_`p'=0.5*who_mindiff_`p'-3 if mindiff_`p' is missing & who_mindiff_`p' is not missing
  #   4.	replace maxdiff_`p'=2*who_maxdiff_`p'+3 if maxdiff_`p is missing & who_maxdiff_`p' is not missing

  if (eval_type == "exclude"){
    # refactored logic slightly for efficiency
    df[!is.na(who.mindiff.next.ht) &
         (delta.agedays.next < 9 * 30.4375 |
            is.na(mindiff.next.ht)),
       `:=`(
         mindiff.next.ht = 0.5 * who.mindiff.next.ht - 3,
         maxdiff.next.ht = 2.0 * who.maxdiff.next.ht + 3
       )]
  } else { # we're evaluating if we want to keep or not
    # should only be used if age is less than 24 months
    # refactored logic slightly for efficiency
    df[!is.na(who.mindiff.next.ht) &
         agedays < 24 * 30.4375 &
         (delta.agedays.next < 9 * 30.4375 |
            is.na(mindiff.next.ht)),
       `:=`(
         mindiff.next.ht = who.mindiff.next.ht,
         maxdiff.next.ht = who.maxdiff.next.ht
       )]
  }

  # 15m.5.  replace mindiff_`p'=-3 if mindiff_`p' is missing
  df[is.na(mindiff.next.ht), mindiff.next.ht := -3]

  # 15n.	Determine the min/maxdiffs for the previous age: mindiff_prev_ht, maxdiff_prev_ht
  # NOTE: obtain previous height, ewma.after value and abs.tbc.sd as well since they are needed in next steps

  # for efficiency, bring get.prev inline here (working on valid rows within a single parameter for a single subject)
  # structure c(NA, tbc.sd[-.N]) == get.prev
  df[, `:=`(
    v.prev = c(NA, v[-.N]),
    dewma.after.prev = c(NA, dewma.after[-.N]),
    abs.tbc.sd.prev = c(NA, abs.tbc.sd[-.N]),
    mindiff.prev.ht = c(NA, mindiff.next.ht[-.N]),
    maxdiff.prev.ht = c(NA, maxdiff.next.ht[-.N])
  )]

  # 15o.	Determine d_prev_ht=ht-htprev (set to missing for the first value for a subject) and d_next_ht=htnext-ht (set to missing for the last value for a subject)
  df[, `:=`(delta.prev.ht = v - v.prev,
            delta.next.ht = v.next - v)]

  # 15p.  Perform a EWMA calculation with the following modifications:
  #  i.	  Generate a variable pair=1 if (d_prev_ht<mindiff_prev_ht OR d_ht<mindiff_ht OR d_prev_ht>maxdiff_prev_ht  OR d_ht>maxdiff_ht) AND exc_ht==0
  df[, pair := na_as_false(
    delta.prev.ht < mindiff.prev.ht |
      delta.next.ht < mindiff.next.ht |
      delta.prev.ht > maxdiff.prev.ht |
      delta.next.ht > maxdiff.next.ht
  )]

  # for efficiency, bring get.prev and get.next inline here (working on valid rows within a single parameter for a single subject)
  # structure c(NA, field.name[-.N]) == get.prev
  # structure c(field.name[-1], NA) == get.next
  df[, `:=`(pair.prev = c(FALSE, pair[-.N]),
            pair.next = c(pair[-1], FALSE))]

  #  ii.	Generate bef_g_aftm1=1 if |Δewma_htbef| for the value of interest is greater than |Δewma_htaft| for the previous value
  #       AND the value of interest is not the first height value for that subject AND pair==1 AND pair for the previous value==1

  #  iii.	Generate aft_g_befp1=1 if |Δewma_htaft| for the value of interest is greater than |Δewma_htbef| for the next value
  #       AND the value of interest is not the last height value for that subject AND pair==1 AND pair for the next value==1
  # NOTE: pair.next will be NA last height, which will result in a FALSE value below
  df[, `:=`(
    bef.g.aftm1 = na_as_false(
      abs(dewma.before) > abs(dewma.after.prev)  & pair & pair.prev
    ),
    aft.g.befp1 = na_as_false(
      abs(dewma.after)  > abs(dewma.before.next) & pair & pair.next
    )
  )]

  #  iv.	Determine tbchtsd for each value as well as the one before prev_tbchtsd and after next_tbchtsd it
  # NOTE: done previously for efficiency

  # 15p.v.  Determine the total number of ht values for each subject (tot_ht)
  # NOTE: all rows are valid due to constraint in subj.df[...] statement
  num.valid <- .N

  # 15q.	Identify a value for possible exclusion if one of the following sets of criteria are met. For values identified by each set of criteria determine
  #       the value of temp_diff using the formula given
  #   i.	d_prev_ht<mindiff_prev_ht & bef_g_aftm1_ht==1 & exc_ht==0 & mindiff_prev_ht is not missing
  #     a.  (temp_diff=|dewma_ht_bef|)
  df[, temp.diff := as.double(NaN)]
  df$temp.exclude <-
    factor(NA, levels = exclude.levels, ordered = TRUE)
  df[delta.prev.ht < mindiff.prev.ht & bef.g.aftm1,
     `:=`(temp.diff = abs(dewma.before),
          temp.exclude = 'Exclude-Min-Height-Change')]

  #   ii.	d_ht<mindiff_ht & aft_g_befp1_ht==1 & exc_ht==0 & mindiff_ht is not missing
  #     a.	(temp_diff=|dewma_ht_aft|)
  df[delta.next.ht < mindiff.next.ht & aft.g.befp1,
     `:=`(temp.diff = abs(dewma.after),
          temp.exclude = 'Exclude-Min-Height-Change')]

  #   iii.	d_prev_ht>maxdiff_prev_ht & bef_g_aftm1_ht==1 & exc_ht==0 & mindiff_prev_ht is not missing
  #     a.  (temp_diff=|dewma_ht_bef|)
  df[delta.prev.ht > maxdiff.prev.ht & bef.g.aftm1,
     `:=`(temp.diff = abs(dewma.before),
          temp.exclude = 'Exclude-Max-Height-Change')]

  #   iv.	d_ht>maxdiff_ht & aft_g_befp1_ht==1 & exc_ht==0 & mindiff_ht is not missing
  #     a.  (temp_diff=|dewma_ht_aft|)
  df[delta.next.ht > maxdiff.next.ht & aft.g.befp1,
     `:=`(temp.diff = abs(dewma.after),
          temp.exclude = 'Exclude-Max-Height-Change')]

  #   v.	d_prev_ht<mindiff_prev_ht & tot_ht==2 & |tbchtsd|>|prev_tbchtsd|
  #     a. for v-viii temp_diff is kept as missing
  #   vi. d_ht<mindiff_ht & tot_ht==2 & |tbchtsd|>|next_tbchtsd|
  df[delta.prev.ht < mindiff.prev.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.prev
     |
       delta.next.ht < mindiff.next.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.next,
     temp.exclude := 'Exclude-Min-Height-Change']

  #   vii.	d_prev_ht>maxdiff_prev_ht & tot_ht==2 & |tbchtsd|>|prev_tbchtsd|
  #   viii. d_ht>maxdiff_ht & tot_ht==2 & |tbchtsd|>|next_tbchtsd|
  df[delta.prev.ht > maxdiff.prev.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.prev
     |
       delta.next.ht > maxdiff.next.ht &
       num.valid == 2 & abs.tbc.sd > abs.tbc.sd.next,
     temp.exclude := 'Exclude-Max-Height-Change']

  return(df$temp.exclude)
}

#' Answers for adjustcarryforward
#'
#' Determines what should absolutely be reincluded or definitely excluded
#' for a given dataset, already run through \code{cleangrowth}.
#'
#' @param subjid Vector of unique identifiers for each subject in the database.
#' @param param Vector identifying each measurement, may be 'WEIGHTKG', 'HEIGHTCM', or 'LENGTHCM'
#'   'HEIGHTCM' vs. 'LENGTHCM' only affects z-score calculations between ages 24 to 35 months (730 to 1095 days).
#'   All linear measurements below 731 days of life (age 0-23 months) are interpreted as supine length, and
#'   all linear measurements above 1095 days of life (age 36+ months) are interpreted as standing height.
#'   Note: at the moment, all LENGTHCM will be converted to HEIGHTCM. In the future, the algorithm will be updated to consider this difference.
#' @param agedays Numeric vector containing the age in days at each measurement.
#' @param sex Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F',
#'  'f' or 1 for females.
#' @param measurement Numeric vector containing the actual measurement data.  Weight must be in
#'   kilograms (kg), and linear measurements (height vs. length) in centimeters (cm).
#' @param orig.exclude Vector of exclusion assessment results from cleangrowth()
#' @param sd.recenter Data frame or table with median SD-scores per day of life
#' @param ewma.exp Exponent to use for weighting measurements in the exponentially weighted moving
#'  average calculations. Defaults to -1.5. This exponent should be negative in order to weight growth
#'  measurements closer to the measurement being evaluated more strongly. Exponents that are further from
#'  zero (e.g. -3) will increase the relative influence of measurements close in time to the measurement
#'  being evaluated compared to using the default exponent.
#' @param ref.data.path Path to reference data. If not supplied, the year 2000
#' Centers for Disease Control (CDC) reference data will be used.
#' @param quietly Determines if function messages are to be displayed and if log files (parallel only)
#' are to be generated. Defaults to TRUE.
#'
#' @return A data frame, containing an index "n" of rows, corresponding to the
#' original order of the input vectors, and "acf_answers", containing the answers
#' on whether a height value should be kept or excluded (returns "Definitely
#' Exclude", "Definitely Include", or "Unknown" for height values, NA for weight
#' values).
#'
#' @export
#' @rawNamespace import(plyr, except = c(failwith, id, summarize, count, desc, mutate, arrange, rename, is.discrete, summarise, summarize))
#' @rawNamespace import(dplyr, except = c(last, first, summarize, src, between))
#' @import data.table
#' @rawNamespace import(R.utils, except = c(extract))
# NOTE: no examples, since this is a temporary function
acf_answers <- function(subjid,
                        param,
                        agedays,
                        sex,
                        measurement,
                        orig.exclude,
                        sd.recenter = NA,
                        ewma.exp = -1.5,
                        ref.data.path = "",
                        quietly = TRUE){

  # avoid "no visible binding for global variable" warnings
  tanner.months <- whoagegrp_ht <- whoagegrp.ht <- z.orig <- z.orig <- v <- sd.orig <- NULL
  index <- exclude <- tbc.sd <- sd.median <- acf_answer <- NULL

  # process ----
  # organize data into a dataframe along with a line "index" so the original data order can be recovered
  data.all <- data.table(
    line = seq_along(measurement),
    subjid = as.factor(subjid),
    param,
    agedays = as.integer(agedays),
    v = ifelse(measurement == 0, NaN, measurement),
    sex = as.integer(ifelse(
      sex %in% c(0, 'm', 'M'), 0, ifelse(sex %in% c(1, 'f', 'F'), 1, NA)
    )),
    orig.exclude = as.factor(orig.exclude)
  )

  ### ADJUSTCF EDIT
  data.all <- data.all[, n := 1:.N]
  ### ENDEDIT

  data.orig <- data.all

  setkey(data.all, subjid)

  subjid.unique <- unique(data.all$subjid)

  # tanner/who 3 SD ----

  # load tanner height velocity data. sex variable is defined such that 0=male and 1=female
  # recode column names to match syntactic style ("." rather than "_" in variable names)
  tanner_ht_vel_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/tanner_ht_vel.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "tanner_ht_vel.csv.gz", sep =
            "")
  )
  tanner.ht.vel <- fread(tanner_ht_vel_path)

  setnames(tanner.ht.vel,
           colnames(tanner.ht.vel),
           gsub('_', '.', colnames(tanner.ht.vel)))
  setkey(tanner.ht.vel, sex, tanner.months)
  # keep track of column names in the tanner data
  tanner.fields <- colnames(tanner.ht.vel)
  tanner.fields <- tanner.fields[!tanner.fields %in% c('sex', 'tanner.months')]

  who_max_ht_vel_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/who_ht_maxvel_3sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "who_ht_maxvel_3sd.csv.gz", sep =
            "")
  )

  who_ht_vel_3sd_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/who_ht_vel_3sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "who_ht_vel_3sd.csv.gz", sep =
            "")
  )
  who.max.ht.vel <- fread(who_max_ht_vel_path)
  who.ht.vel <- fread(who_ht_vel_3sd_path)
  setkey(who.max.ht.vel, sex, whoagegrp_ht)
  setkey(who.ht.vel, sex, whoagegrp_ht)
  who.ht.vel <- as.data.table(dplyr::full_join(who.ht.vel, who.max.ht.vel, by =
                                                 c('sex', 'whoagegrp_ht')))

  setnames(who.ht.vel, colnames(who.ht.vel), gsub('_', '.', colnames(who.ht.vel)))
  setkey(who.ht.vel, sex, whoagegrp.ht)
  # keep track of column names in the who growth velocity data
  who.fields <- colnames(who.ht.vel)
  who.fields <- who.fields[!who.fields %in% c('sex', 'whoagegrp.ht')]

  # tanner/who 2 SD ----

  # load tanner height velocity data. sex variable is defined such that 0=male and 1=female
  # recode column names to match syntactic style ("." rather than "_" in variable names)
  tanner_ht_vel_2sd_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/tanner_ht_vel_with_2sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "tanner_ht_vel_with_2sd.csv.gz", sep = "")
  )
  tanner.ht.vel.2sd <- fread(tanner_ht_vel_2sd_path)

  setnames(tanner.ht.vel.2sd,
           colnames(tanner.ht.vel.2sd),
           gsub('_', '.', colnames(tanner.ht.vel.2sd)))
  setkey(tanner.ht.vel.2sd, sex, tanner.months)
  # keep track of column names in the tanner data
  tanner.fields.2sd <- colnames(tanner.ht.vel.2sd)
  tanner.fields.2sd <-
    tanner.fields.2sd[!tanner.fields.2sd %in% c('sex', 'tanner.months')]

  who_max_ht_vel_2sd_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/who_ht_maxvel_2sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "who_ht_maxvel_2sd.csv.gz", sep =
            "")
  )

  who_ht_vel_2sd_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/who_ht_vel_2sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "who_ht_vel_2sd.csv.gz", sep =
            "")
  )
  who.max.ht.vel.2sd <- fread(who_max_ht_vel_2sd_path)
  who.ht.vel.2sd <- fread(who_ht_vel_2sd_path)
  setkey(who.max.ht.vel.2sd, sex, whoagegrp_ht)
  setkey(who.ht.vel.2sd, sex, whoagegrp_ht)
  who.ht.vel.2sd <-
    as.data.table(dplyr::full_join(who.ht.vel.2sd, who.max.ht.vel.2sd, by =
                                     c('sex', 'whoagegrp_ht')))

  setnames(who.ht.vel.2sd,
           colnames(who.ht.vel.2sd),
           gsub(
             ".2sd", "", # replace 2sd with nothing, for ease of use in calculations
             gsub('_', '.', colnames(who.ht.vel.2sd))
            ))
  setkey(who.ht.vel.2sd, sex, whoagegrp.ht)
  # keep track of column names in the who growth velocity data
  who.fields.2sd <- colnames(who.ht.vel.2sd)
  who.fields.2sd <- who.fields.2sd[!who.fields.2sd %in% c('sex', 'whoagegrp.ht')]

  # getting zscores ----

  # recategorize linear parameters as 'HEIGHTCM'
  # NOTE: this will be changed in future to consider this difference
  data.all[param == 'LENGTHCM', param := 'HEIGHTCM']

  # calculate z scores
  if (!quietly)
    cat(sprintf("[%s] Calculating z-scores...\n", Sys.time()))
  measurement.to.z <- read_anthro(ref.data.path, cdc.only = TRUE)
  data.all[, z.orig := measurement.to.z(param, agedays, sex, v)]

  # calculate "standard deviation" scores
  if (!quietly)
    cat(sprintf("[%s] Calculating SD-scores...\n", Sys.time()))
  data.all[, sd.orig := measurement.to.z(param, agedays, sex, v, TRUE)]

  # sort by subjid, param, agedays
  setkey(data.all, subjid, param, agedays)

  # add a new convenience index for bookkeeping
  data.all[, index := 1:.N]

  # enumerate the different exclusion levels
  exclude.levels <- c(
    'Missing',
    'No Change',
    'Include',
    'Exclude-Min-Height-Change',
    'Exclude-Max-Height-Change'
  )

  # Mark missing values for exclusion
  data.all[, exclude := factor(with(data.all, ifelse(
    is.na(v) |
      agedays < 0, 'Missing', 'No Change'
  )),
  levels = exclude.levels,
  ordered = TRUE)] # why is this ordered??

  # define field names needed by helper functions
  ewma.fields <- c('ewma.all', 'ewma.before', 'ewma.after')

  # 3.  SD-score recentering: Because the basis of the method is comparing SD-scores over time, we need to account for the fact that
  #     the mean SD-score for the population changes with age.
  # a.	Determine the median cdc*sd for each parameter by year of age (with sexes combined): median*sd.
  # b.	The median*sd should be considered to apply to midyear-age, defined as the age in days with the same value as the integer
  #     portion of (365.25*year + 365.25/2).
  # c.	Linearly interpolate median*sd for each parameter between each midyear-age, naming the interpolated values rc*sd.
  # d.	For ages below the first midyear-age, let rc*sd equal the median*sd for the earliest year.
  #     For ages above the last midyear_age, let rc*sd equal the median*sd for the last year.
  # e.	Subtract rcsd_* from SDorig to create the recentered SD-score.  This recentered SD-score, labeled tbc*sd
  #     (stands for "to be cleaned") will be used for most of the rest of the analyses.
  # f.	In future steps I will sometimes refer to measprev and measnext which refer to the previous or next wt or ht measurement
  #     for which exc_*==0 for the subject and parameter, when the data are sorted by subject, parameter, and agedays. SDprev and SDnext refer to the tbc*sd of the previous or next measurement.

  if (!quietly)
    cat(sprintf("[%s] Re-centering data...\n", Sys.time()))

  # see function definition below for explanation of the re-centering process
  # returns a data table indexed by param, sex, agedays
  if (!is.data.table(sd.recenter)) {
    # ADJUSTCF EDIT - be explicit about levels to keep
    keep.levels <- c(
      "Include",
      "Unit-Error-High",
      "Unit-Error-Low",
      "Unit-Error-Possible",
      "Swapped-Measurements"
    )
    sd.recenter <- data.all[orig.exclude %in% keep.levels, sd_median(param, sex, agedays, sd.orig)]
    # END EDIT
  }
  # add sd.recenter to data, and recenter
  setkey(data.all, param, sex, agedays)
  data.all <- sd.recenter[data.all]
  setkey(data.all, subjid, param, agedays)
  data.all[, tbc.sd := sd.orig - sd.median]

  # safety check: treat observations where tbc.sd cannot be calculated as missing
  data.all[is.na(tbc.sd), exclude := 'Missing']

  # start evaluation ----

  if (!quietly)
    cat(sprintf("[%s] Calculating definitely exclude/include...\n", Sys.time()))

  data.all[param == 'HEIGHTCM', acf_answer := (function(subj.df) {
    # assign some book keeping variables
    #subj.df[, `:=`(subjid = subjid, param='HEIGHTCM',index=1:.N)]
    subj.df[, index := 1:.N]

    # use a closure to discard all the extra fields added to df with each iteration
    subj.df[,
      acf_answer := (function (df) {
        # calculate definitely exclude
        # do steps 15a - 15q (functionalized for ease)
        def_excl <- calc_step_15_no_param(
          copy(df),
          eval_type = "exclude",
          ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
          ewma.exp)

        def_incl <- calc_step_15_no_param(
          copy(df),
          eval_type = "include",
          ewma.fields, tanner.ht.vel.2sd, who.ht.vel.2sd, exclude.levels,
          ewma.exp)

        # calculate which should definitely be included and excluded
        verdict <- rep("Unknown", length(def_incl))
        verdict[!is.na(def_excl)] <- "Definitely Exclude"
        verdict[is.na(def_incl)] <- "Definitely Include"

        return(verdict)

      })(copy(.SD))]

    setkey(subj.df, index)
    return(subj.df$acf_answer)
  })(copy(.SD)), by = .(subjid), .SDcols = c('sex', 'agedays', 'v', 'tbc.sd', 'exclude', 'orig.exclude')]

  # formulating results
  acf_answer_df <- data.frame(n = data.all$n, acf_answer = data.all$acf_answer)
  # sort for user ease
  acf_answer_df <- acf_answer_df[order(acf_answer_df$n),]

  return(acf_answer_df)
}

#' adjustcarryforward
#' \code{adjustcarryforward} Uses absolute height velocity to identify values
#' excluded as carried forward values for reinclusion.
#' @param subjid Vector of unique identifiers for each subject in the database.
#' @param param Vector identifying each measurement, may be 'WEIGHTKG', 'HEIGHTCM', or 'LENGTHCM'
#'   'HEIGHTCM' vs. 'LENGTHCM' only affects z-score calculations between ages 24 to 35 months (730 to 1095 days).
#'   All linear measurements below 731 days of life (age 0-23 months) are interpreted as supine length, and
#'   all linear measurements above 1095 days of life (age 36+ months) are interpreted as standing height.
#'   Note: at the moment, all LENGTHCM will be converted to HEIGHTCM. In the future, the algorithm will be updated to consider this difference.
#' @param agedays Numeric vector containing the age in days at each measurement.
#' @param sex Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F',
#'  'f' or 1 for females.
#' @param measurement Numeric vector containing the actual measurement data.  Weight must be in
#'   kilograms (kg), and linear measurements (height vs. length) in centimeters (cm).
#' @param orig.exclude Vector of exclusion assessment results from cleangrowth()
#' @param exclude_opt Number from 0 to 3 indicating which option to use to handle strings of carried-forwards:
#'    0. no change.
#'    1. when deciding to exclude values, if we have a string of carried forwards,
#'    drop the most deviant value, and all CFs in the same string, and move on as
#'    normal.
#'    2. when deciding to exclude values, if the most deviant in a
#'    string of carried forwards is flagged, check all the CFs in that
#'    string from 1:N. Exclude all after the first that is flagged for
#'    exclusion when comparing to the Include before and after. Do not
#'    remove things designated as include.
#'    3. when deciding to exclude values, if the most deviant in a
#'    string of carried forwards is flagged, check all the CFs in that
#'    string from 1:N. Exclude all after the first that is flagged for
#'    exclusion when comparing to the Include before and after. Make sure
#'    remove things designated as include.
#' @param sd.recenter Data frame or table with median SD-scores per day of life
#' @param ewma.exp Exponent to use for weighting measurements in the exponentially weighted moving
#'  average calculations. Defaults to -1.5. This exponent should be negative in order to weight growth
#'  measurements closer to the measurement being evaluated more strongly. Exponents that are further from
#'  zero (e.g. -3) will increase the relative influence of measurements close in time to the measurement
#'  being evaluated compared to using the default exponent.
#' @param ref.data.path Path to reference data. If not supplied, the year 2000
#' Centers for Disease Control (CDC) reference data will be used.
#' @param quietly Determines if function messages are to be displayed and if log files (parallel only)
#' are to be generated. Defaults to TRUE.
#' @param minfactor Sweep variable for computing mindiff.next.ht in 15f, default 0.5
#' @param maxfactor Sweep variable for computing maxdiff.next.ht in 15f, default 2
#' @param banddiff Sweep variable for computing mindiff.next.ht in 15f, default 3
#' @param banddiff_plus Sweep variable for computing maxdiff.next.ht in 15, default 5.5
#' @param min_ht.exp_under Sweep variable for computing ht.exp in 15f, default 2
#' @param min_ht.exp_over Sweep variable for computing ht.exp in 15f, default 0
#' @param max_ht.exp_under Sweep variable for computing ht.exp in 15f, default 0.33
#' @param max_ht.exp_over Sweep variable for computing ht.exp in 15f, default 1.5
#'
#' @return Re-evaluated exclusion assessments based on height velocity.
#'
#' @export
#' @rawNamespace import(plyr, except = c(failwith, id, summarize, count, desc, mutate, arrange, rename, is.discrete, summarise, summarize))
#' @rawNamespace import(dplyr, except = c(last, first, summarize, src, between))
#' @import data.table
#' @examples
#' \donttest{
#' # Run on a small subset of given data
#' df <- as.data.frame(syngrowth)
#' df <- df[df$subjid %in% unique(df[, "subjid"])[1:2], ]
#' clean_df <- cbind(df,
#'                   "gcr_result" = cleangrowth(df$subjid,
#'                                              df$param,
#'                                              df$agedays,
#'                                              df$sex,
#'                                              df$measurement))
#'
#' # Adjust carry forward values in cleaned data
#' adj_clean <- adjustcarryforward(subjid = clean_df$subjid,
#'                                 param = clean_df$param,
#'                                 agedays = clean_df$agedays,
#'                                 sex = clean_df$sex,
#'                                 measurement = clean_df$measurement,
#'                                 orig.exclude = clean_df$gcr_result)
#' }
adjustcarryforward <- function(subjid,
                               param,
                               agedays,
                               sex,
                               measurement,
                               orig.exclude,
                               exclude_opt = 0,
                               sd.recenter = NA,
                               ewma.exp = -1.5,
                               ref.data.path = "",
                               quietly = TRUE,
                               minfactor = 0.5,
                               maxfactor = 2,
                               banddiff = 3,
                               banddiff_plus = 5.5,
                               min_ht.exp_under = 2,
                               min_ht.exp_over = 0,
                               max_ht.exp_under = 0.33,
                               max_ht.exp_over = 1.5) {

  # Avoid "undefined global functions/variables" warnings
  v <- sd.orig <- sd.median <- tbc.sd <- agedays.next <- mid.agedays <- min.ht.vel <- NULL
  delta.agedays.next <- ht.exp <- max.ht.vel <- mindiff.next.ht <- temp.exclude <- NULL
  ecf_tmp <- ewma.all <- ewma.before <- ewma.after <- abs.tbc.sd <- whoinc.1.ht <- NULL
  whoinc.2.ht <- whoinc.3.ht <- whoinc.4.ht <- whoinc.6.ht <- max.whoinc.1.ht <- NULL
  max.whoinc.2.ht <- max.whoinc.3.ht <- max.whoinc.4.ht <- max.whoinc.6.ht <- NULL
  whoinc.age.ht <- whoinc.age.ht <- who.mindiff.next.ht <- who.maxdiff.next.ht <- NULL
  dewma.after <- maxdiff.next.ht <- v.prev <- v.next <- delta.prev.ht <- NULL
  mindiff.prev.ht <- delta.next.ht <- maxdiff.prev.ht <- pair <- dewma.after.prev <- NULL
  pair.prev <- dewma.before.next <- pair.next <- bef.g.aftm1 <- aft.g.befp1 <- NULL
  abs.tbc.sd.prev <- abs.tbc.sd.next <- exclude <- n <- dewma.before <- NULL
  tanner.months <- whoagegrp_ht <- whoagegrp.ht <- z.orig <- index <- temp.diff <- NULL

  # check option is valid
  if (!exclude_opt %in% 0:3){
    stop("Invalid exclude_opt. Enter a number from 0 to 3.")
  }

  # organize data into a dataframe along with a line "index" so the original data order can be recovered
  data.all <- data.table(
    line = seq_along(measurement),
    subjid = as.factor(subjid),
    param,
    agedays = as.integer(agedays),
    v = ifelse(measurement == 0, NaN, measurement),
    sex = as.integer(ifelse(
      sex %in% c(0, 'm', 'M'), 0, ifelse(sex %in% c(1, 'f', 'F'), 1, NA)
    )),
    orig.exclude = as.factor(orig.exclude)
  )

  ### ADJUSTCF EDIT
  data.all <- data.all[, n := 1:.N]
  ### ENDEDIT

  data.orig <- data.all

  setkey(data.all, subjid)

  subjid.unique <- unique(data.all$subjid)

  #### ADJUSTCF EDIT ####
  # for this purpose, want to subset dataset down to just "Exclude-Carried-Forward" and "Include" - assume all other measurements are invalid
  # want to remove any carried forward values whose non-carried forward is also excluded
  # data.all <- data.all %>%
  #   mutate(orig.exclude.lag = lag(orig.exclude, n = 1)) %>%
  #   filter(!(
  #     orig.exclude == "Exclude-Carried-Forward" &
  #       grepl("exclude", orig.exclude.lag, ignore.case = TRUE)
  #   )) %>%
  #   filter(orig.exclude %in% c("Exclude-Carried-Forward", "Include")) %>%
  #   select(-orig.exclude.lag)

  # NEW EDIT --
  # remove all the weight measurements
  data.all <- data.all %>%
    filter(.data$param %in% c("HEIGHTCM", "LENGTHCM"))

  # filter to only subjects with possible carried forwards - n is here to merge back
  # if they have all includes, filter them out
  data.all <- data.all %>%
    filter(.data$subjid %in% data.all$subjid[data.all$orig.exclude == "Exclude-Carried-Forward"]) %>%
    as.data.table()

  # here's what we want to filter out -- anything that's not carried forward/include
  # we're also going to include strings of carried forward
  # but we also need to make sure they're not coming from an excluded value

  # start of string to remove: everything that isn't include/excl-cf
  st <- which(!data.all$orig.exclude %in% c("Exclude-Carried-Forward", "Include"))
  # end of string: include or the end of a subject
  subj_end <- length(data.all$subjid)-match(unique(data.all$subjid),rev(data.all$subjid))+1
  end <- c(which(data.all$orig.exclude == "Include"),subj_end)
  end <- unique(sort(end))

  # remove anything between start and ends (including start, not including end)
  to_rem <- unlist(
    lapply(st, function(x){
      to_rem <- c(x:(end[end >= x][1]))
      if (to_rem[length(to_rem)] %in% subj_end){
        # if it's the last value, we want to get rid of that end
        return(to_rem)
      } else {
        # if it's an include, we want to keep it (don't remove)
        return(to_rem[-length(to_rem)])
      }
    })
  )
  to_rem <- unique(to_rem)

  if (length(to_rem) > 0){
    data.all <- data.all[-to_rem,]
  }

  # filter to only subjects with possible carried forwards again
  data.all <- data.all %>%
    filter(.data$subjid %in% data.all$subjid[data.all$orig.exclude == "Exclude-Carried-Forward"]) %>%
    as.data.table()

  ### END EDIT ####

  # load tanner height velocity data. sex variable is defined such that 0=male and 1=female
  # recode column names to match syntactic style ("." rather than "_" in variable names)
  tanner_ht_vel_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/tanner_ht_vel.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "tanner_ht_vel.csv.gz", sep = "")
  )

  tanner.ht.vel <- fread(tanner_ht_vel_path)

  setnames(tanner.ht.vel,
           colnames(tanner.ht.vel),
           gsub('_', '.', colnames(tanner.ht.vel)))
  setkey(tanner.ht.vel, sex, tanner.months)
  # keep track of column names in the tanner data
  tanner.fields <- colnames(tanner.ht.vel)
  tanner.fields <- tanner.fields[!tanner.fields %in% c('sex', 'tanner.months')]

  who_max_ht_vel_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/who_ht_maxvel_3sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "who_ht_maxvel_3sd.csv.gz", sep =
            "")
  )

  who_ht_vel_3sd_path <- ifelse(
    ref.data.path == "",
    system.file("extdata/who_ht_vel_3sd.csv.gz", package = "growthcleanr"),
    paste(ref.data.path, "who_ht_vel_3sd.csv.gz", sep =
            "")
  )
  who.max.ht.vel <- fread(who_max_ht_vel_path)
  who.ht.vel <- fread(who_ht_vel_3sd_path)
  setkey(who.max.ht.vel, sex, whoagegrp_ht)
  setkey(who.ht.vel, sex, whoagegrp_ht)
  who.ht.vel <- as.data.table(dplyr::full_join(who.ht.vel, who.max.ht.vel, by =
                                                c('sex', 'whoagegrp_ht')))

  setnames(who.ht.vel, colnames(who.ht.vel), gsub('_', '.', colnames(who.ht.vel)))
  setkey(who.ht.vel, sex, whoagegrp.ht)
  # keep track of column names in the who growth velocity data
  who.fields <- colnames(who.ht.vel)
  who.fields <- who.fields[!who.fields %in% c('sex', 'whoagegrp.ht')]

  # 1.  General principles
  # a.	All steps are done separately for each parameter unless otherwise noted
  # b.	All steps are done sorted by subject, parameter, and age (in days) for nonexcluded and nonmissing values only unless otherwise noted. This is very important.
  #     Sorting needs to be redone with each step to account for excluded and transformed  values.
  # c.	The next value refers to the value with the next highest age for the same parameter and the same subject, and the previous value refers to the value with the
  #     next lowest age for the same parameter and the same subject.
  # d.	You will need to set up a method for keeping track of whether a value is missing or excluded (and in what step). I use variables called exc_* that are =0
  #     if a value is to be included, =1 if missing, and =2 or higher if it is to be excluded, with each number indicating a different step. I also set up
  #     parameter-specific subjid_* variables that are = to subjid for included values and are blank if the value is missing or should be excluded. These subjid_*
  #     variables need to be updated with each step.
  # e.  All steps assume that data are sorted by subjid_*, parameter, and age (in days) for nonexcluded and nonmissing values only
  #     unless otherwise noted. Sorting needs to be redone after any transformations or exclusions to account for excluded and
  #     transformed values.
  # f.  The next value refers to the nonexcluded nonmissing value with the next highest age for the same parameter and the same
  #     subject, and the previous value refers to the nonexcluded nonmissing value with the next lowest age for the same parameter
  #     and the same subject.
  # g.  exc_* should only be replaced with a  higher value if exc_*==0 at the time of replacement, unless otherwise specified.


  # NOTE: in the R code below exclusion is documented as a series of factor levels, where all levels occuring before 'Exclude' in the sequence are considered
  # to be valid measurements.  We use the built in sorting of the data.table object and subsets rather than re-sorting at each step
  # to ensure that only valid measurements are used at the beginning of each step.
  # Also, unlike the Stata code, the measurement parameter (weight vs. height) is recorded as a factor in the data frame, rather than as a variable name

  # 2.  Data set-up
  # a.	I always code sex as 0=Male, 1=Female, so I recoded the variable sex that way and left a variable sexorigcode the way the data was sent to me (1=Female 2=Male)
  # b.	Remove rows that are duplicates for subjid, param, and measurement from further analysis
  #     NOTE: this step is not needed -- handled automatically by "temporary duplicate" step.
  # c.  I generated separate variables for weight (wt) and height (ht), as well as exc_* and subjid_* variables. Set exc_*=0 if value is not missing
  #     and exc_*=1 if value is missing. In all future steps, exc_* should only be changed if it is 0. This helps to keep track of which step excluded a value.
  #     I also kept the measurement variable there and untouched because sometimes wt and ht got transformed to something else.
  # d.	I made tables based on CDC growth curve parameters that include data for each day that I will send separately. The LMS parameters for each day are
  #     cubically interpolated from the values by month available on the CDC website. Create wt and ht z-scores for each value of each parameter (Z: WtZ and HtZ).
  # e.	There are variables in the table labelled cdc_*_csd_pos and cdc_*_csd_neg. For each age and sex, these correspond to ½ of the absolute value of the
  #     median and the value with a z-score of +2 (csd_pos) and -2 (csd_neg). These can be created to generate a score similar to the z-score but with an
  #     important difference. The z-scores created using the LMS method account for skewness in the distribution of the parameters (particularly weight), which
  #     can lead to small changes in z-score with large changes in weight in subjects with very high weight, and relatively large changes in z-score for smaller
  #     changes in weight in subjects with low weights.  The score we will create can be called an SD-score (SDorig: WtSDorig and HtSDorig that is calculated by
  #     dividing the difference between the value and the median by the SD score (use csd_pos if the value is above the median, csd_neg if the value is below the
  #     median). These SD-scores, rather than z-scores, now form the basis for the algorithm.

  # recategorize linear parameters as 'HEIGHTCM'
  # NOTE: this will be changed in future to consider this difference
  data.all[param == 'LENGTHCM', param := 'HEIGHTCM']

  # calculate z scores
  if (!quietly)
    cat(sprintf("[%s] Calculating z-scores...\n", Sys.time()))
  measurement.to.z <- read_anthro(ref.data.path, cdc.only = TRUE)
  data.all[, z.orig := measurement.to.z(param, agedays, sex, v)]

  # calculate "standard deviation" scores
  if (!quietly)
    cat(sprintf("[%s] Calculating SD-scores...\n", Sys.time()))
  data.all[, sd.orig := measurement.to.z(param, agedays, sex, v, TRUE)]

  # sort by subjid, param, agedays
  setkey(data.all, subjid, param, agedays)

  # add a new convenience index for bookkeeping
  data.all[, index := 1:.N]

  # enumerate the different exclusion levels
  exclude.levels <- c(
    'Missing',
    'No Change',
    'Include',
    'Exclude-Min-Height-Change',
    'Exclude-Max-Height-Change'
  )

  # Mark missing values for exclusion
  data.all[, exclude := factor(with(data.all, ifelse(
    is.na(v) |
      agedays < 0, 'Missing', 'No Change'
  )),
  levels = exclude.levels,
  ordered = TRUE)] # why is this ordered??

  # define field names needed by helper functions
  ewma.fields <- c('ewma.all', 'ewma.before', 'ewma.after')

  # 3.  SD-score recentering: Because the basis of the method is comparing SD-scores over time, we need to account for the fact that
  #     the mean SD-score for the population changes with age.
  # a.	Determine the median cdc*sd for each parameter by year of age (with sexes combined): median*sd.
  # b.	The median*sd should be considered to apply to midyear-age, defined as the age in days with the same value as the integer
  #     portion of (365.25*year + 365.25/2).
  # c.	Linearly interpolate median*sd for each parameter between each midyear-age, naming the interpolated values rc*sd.
  # d.	For ages below the first midyear-age, let rc*sd equal the median*sd for the earliest year.
  #     For ages above the last midyear_age, let rc*sd equal the median*sd for the last year.
  # e.	Subtract rcsd_* from SDorig to create the recentered SD-score.  This recentered SD-score, labeled tbc*sd
  #     (stands for "to be cleaned") will be used for most of the rest of the analyses.
  # f.	In future steps I will sometimes refer to measprev and measnext which refer to the previous or next wt or ht measurement
  #     for which exc_*==0 for the subject and parameter, when the data are sorted by subject, parameter, and agedays. SDprev and SDnext refer to the tbc*sd of the previous or next measurement.

  if (!quietly)
    cat(sprintf("[%s] Re-centering data...\n", Sys.time()))

  # see function definition below for explanation of the re-centering process
  # returns a data table indexed by param, sex, agedays
  if (!is.data.table(sd.recenter)) {
    # ADJUSTCF EDIT - be explicit about levels to keep
    keep.levels <- c(
      "Include",
      "Unit-Error-High",
      "Unit-Error-Low",
      "Unit-Error-Possible",
      "Swapped-Measurements"
    )
    sd.recenter <- data.all[orig.exclude %in% keep.levels, sd_median(param, sex, agedays, sd.orig)]
    # END EDIT
  }
  # add sd.recenter to data, and recenter
  setkey(data.all, param, sex, agedays)
  data.all <- sd.recenter[data.all]
  setkey(data.all, subjid, param, agedays)
  data.all[, tbc.sd := sd.orig - sd.median]

  # safety check: treat observations where tbc.sd cannot be calculated as missing
  data.all[is.na(tbc.sd), exclude := 'Missing']



  ######### END DATA PROCESING #########

  ######### START FLAGGING #########

  # 15.  Exclude heights based on absolute differences in measurement. The key to this step is that once we identify pairs of measurements with implausible
  #      amounts of absolute difference between them, we have to determine which of the two values in the pair is less likely to be representative and should
  #      be excluded. For subjects/parameters with 3 or more measurements, this is done by looking at the dewma_* of each of the 2 values in a pair using a ewma that
  #      excludes the other value in the pair. For subjects/parameters with 2 or more measurements, this is done by looking at the absolute value of the tbc*sd.
  #      The Tanner height velocity reference is used for measurements taken at >2yo, WHO will be used for <2yo. For a few pairs of measurements either could be used;
  #      WHO will be used if difference between ages is < 9 months.
  if (!quietly)
    cat(sprintf(
      "[%s] Exclude heights based on growth velocity...\n",
      Sys.time()
    ))

  # ADJUSTCF EDIT:
  # there are several different method for handling carried forward values
  # 0. no change
  # 1. when deciding to exclude values, if we have a string of carried forwards,
  #    drop the most deviant value, and all CFs in the same string, and move on as
  #    normal
  # 2. when deciding to exclude values, if the most deviant in a
  # string of carried forwards is flagged, check all the CFs in that
  # string from 1:N. exclude all after the first that is flagged for
  # exclusion when comparing to the Include before and after. do not
  # remove things designated as include.
  # 3. when deciding to exclude values, if the most deviant in a
  # string of carried forwards is flagged, check all the CFs in that
  # string from 1:N. exclude all after the first that is flagged for
  # exclusion when comparing to the Include before and after. make sure
  # remove things designated as include.


  # for (opt in all_opts){
  data.all[param == 'HEIGHTCM', exclude := (function(subj.df) {
    # assign some book keeping variables
    #subj.df[, `:=`(subjid = subjid, param='HEIGHTCM',index=1:.N)]
    subj.df[, index := 1:.N]

    # keep the original subject dataframe, since we will always use the original
    # includes to compare to
    subj.df_orig <- copy(subj.df)
    num.height.excluded = 0
    while (TRUE) {
      # use a closure to discard all the extra fields added to df with each iteration
      # "include" protects subsets (used in option 3)
      subj.df[!grepl("Exclude", exclude) & !grepl("Include", exclude),
              exclude := (function (df) {
                # put together the dataframe with original includes for comparisons
                # get all "includes"
                subj.df_orig_incl <- copy(subj.df_orig[orig.exclude == "Include",])
                # now get all the remaining and add them on
                df <- rbind(subj.df_orig_incl,
                            df[orig.exclude != "Include",])
                # sort them by original index
                df <- df[order(index),]

                # do steps 15a - 15q (functionalized for ease)
                eval_df <- calc_temp_exclusion_15(
                  copy(df),
                  ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
                  ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
                  min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over)

                # r.  If there is only one potential exclusion identified in step 15j for a subject and parameter,
                #     replace exc_ht=15 for that value if it met criteria i, ii, v, or vi  and exc_ht=16 if it met criteria iii, iv, vii, or viii
                # NOTE: these exclusions are assigned in the code above as 'Exclude-Min-Height-Change' and 'Exclude-Max-Height-Change'

                # remove includes from consideration of removal
                eval_df[orig.exclude == "Include", temp.diff := NaN]
                eval_df[orig.exclude == "Include", temp.exclude := NA]

                # count the amount of error codes we get
                all_exclude = !is.na(eval_df$temp.exclude)
                num.exclude = sum(all_exclude)

                # if there's only one, all of the groups agree
                # s.  If there is more than one potential exclusion identified in step 14h for a subject and parameter, determine which value has the largest temp_diff and
                #     replace exc_ht=15 for that value if it met criteria i, ii, v, or vi and exc_ht=16 for that value if it met criteria iii,  iv, vii, or viii.

                if (exclude_opt == 0){
                  # option 0: normal
                  if (num.exclude == 1){
                    eval_df[all_exclude, exclude := temp.exclude]
                  } else if (num.exclude > 1) {
                    # first order by decreasing temp.diff (where rep=TRUE)
                    worst.row = order(all_exclude, eval_df$temp.diff, decreasing = TRUE)[1]
                    eval_df[worst.row, exclude := temp.exclude]
                  }
                } else if (exclude_opt == 1){
                  # option 1: if there's a carried forward after, we want to exclude that too, until the next include
                  if (num.exclude == 1){
                    wh_exclude <- which(all_exclude)

                    # if it's in a string of carried forwards
                    if (
                      eval_df[wh_exclude, "orig.exclude"] == "Exclude-Carried-Forward" &&
                      wh_exclude != nrow(eval_df) &&
                      eval_df[wh_exclude+1, "orig.exclude"] == "Exclude-Carried-Forward"
                    ){
                      # find the next include OR the row end
                      next_incl <- which(
                        eval_df$orig.exclude[(wh_exclude+1):nrow(eval_df)] == "Include"
                      )[1]+wh_exclude
                      if (is.na(next_incl)){
                        next_incl <- nrow(eval_df)+1
                      }


                      # mark all the CFs after for exclusion
                      all_exclude[wh_exclude:(next_incl-1)] <- TRUE
                      # also copy all the temp.excludes
                      eval_df$temp.exclude[(wh_exclude+1):(next_incl-1)] <-
                        eval_df$temp.exclude[wh_exclude]
                    }

                    # exclude all the carried forwards before the next include
                    eval_df[all_exclude, exclude := temp.exclude]
                  } else if (num.exclude > 1) {
                    # first order by decreasing temp.diff (where rep=TRUE)
                    worst.row = order(all_exclude, eval_df$temp.diff, decreasing = TRUE)[1]

                    # option 1: if there's a carried forward after, we want to exclude that too, until the next include
                    wh_exclude <- worst.row
                    if (
                      eval_df[wh_exclude, "orig.exclude"] == "Exclude-Carried-Forward" &&
                      wh_exclude != nrow(eval_df) &&
                      eval_df[wh_exclude+1, "orig.exclude"] == "Exclude-Carried-Forward"
                    ){
                      # find the next include OR the row end
                      next_incl <- which(
                        eval_df$orig.exclude[(wh_exclude+1):nrow(eval_df)] == "Include"
                      )[1]+wh_exclude
                      if (is.na(next_incl)){
                        next_incl <- nrow(eval_df)+1
                      }

                      # mark all the CFs after for exclusion
                      worst.row <- c(worst.row:(next_incl-1))
                      # also copy all the temp.excludes
                      eval_df$temp.exclude[(wh_exclude+1):(next_incl-1)] <-
                        eval_df$temp.exclude[wh_exclude]
                    }

                    eval_df[worst.row, exclude := temp.exclude]
                  }
                } else if (exclude_opt == 2){
                  # 2. when deciding to exclude values, if the most deviant in a
                  # string of carried forwards is flagged, check all the CFs in that
                  # string from 1:N. exclude all after the first that is flagged for
                  # exclusion when comparing to the Include before and after. do not
                  # remove things designated as include.

                  # after you've evaluated them, do you reevaluate the ones you've kept?
                  if (num.exclude == 1){
                    wh_exclude <- which(all_exclude)

                    if (
                      eval_df[wh_exclude, "orig.exclude"] == "Exclude-Carried-Forward" &&
                      ((wh_exclude != nrow(eval_df) &&
                        eval_df[wh_exclude+1, "orig.exclude"] == "Exclude-Carried-Forward")
                       ||
                       (wh_exclude != 1 &&
                        eval_df[wh_exclude-1, "orig.exclude"] == "Exclude-Carried-Forward")
                      )
                    ){
                      # check all the carried forwards in a string
                      res <- check_cf_string(
                        copy(eval_df), wh_exclude,
                        ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
                        ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
                        min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over
                      )

                      verdict <- res$verdict
                      cf_ind <- res$cf_ind
                      next_incl <- res$next_incl
                      first_incl <- res$first_incl

                      if (verdict == "Exclude"){
                        # mark all the CFs after for exclusion
                        all_exclude[(cf_ind-1):(next_incl-1)] <- TRUE
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(cf_ind-1):(next_incl-1)] <-
                          eval_df$temp.exclude[(cf_ind-1)]
                      } else {
                        # if the verdict is include, we need to mark them all for
                        # "exclusion" to remove -- otherwise we end up in a loop
                        all_exclude[(first_incl+1):(next_incl-1)] <- TRUE
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(first_incl+1):(next_incl-1)] <-
                          "Include"
                      }
                    }

                    # exclude all the carried forwards before the next include
                    eval_df[all_exclude, exclude := temp.exclude]
                  } else if (num.exclude > 1) {
                    # first order by decreasing temp.diff (where rep=TRUE)
                    worst.row = order(all_exclude, eval_df$temp.diff, decreasing = TRUE)[1]

                    wh_exclude <- worst.row
                    if (
                      eval_df[wh_exclude, "orig.exclude"] == "Exclude-Carried-Forward" &&
                      ((wh_exclude != nrow(eval_df) &&
                        eval_df[wh_exclude+1, "orig.exclude"] == "Exclude-Carried-Forward")
                       ||
                       (wh_exclude != 1 &&
                        eval_df[wh_exclude-1, "orig.exclude"] == "Exclude-Carried-Forward")
                      )
                    ){
                      # check all the carried forwards in a string
                      res <- check_cf_string(
                        copy(eval_df), wh_exclude,
                        ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
                        ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
                        min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over
                      )

                      verdict <- res$verdict
                      cf_ind <- res$cf_ind
                      next_incl <- res$next_incl
                      first_incl <- res$first_incl

                      if (verdict == "Exclude"){
                        # mark all the CFs after for exclusion
                        worst.row <- c((cf_ind-1):(next_incl-1))
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(cf_ind-1):(next_incl-1)] <-
                          eval_df$temp.exclude[(cf_ind-1)]
                      } else {
                        # if the verdict is include, we need to mark them all for
                        # "exclusion" to remove -- otherwise we end up in a loop
                        worst.row <- c((first_incl+1):(next_incl-1))
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(first_incl+1):(next_incl-1)] <-
                          "Include"
                      }
                    }

                    eval_df[worst.row, exclude := temp.exclude]
                  }
                } else if (exclude_opt == 3){
                  # 3. when deciding to exclude values, if the most deviant in a
                  # string of carried forwards is flagged, check all the CFs in that
                  # string from 1:N. exclude all after the first that is flagged for
                  # exclusion when comparing to the Include before and after. make sure
                  # remove things designated as include.

                  # after you've evaluated them, do you reevaluate the ones you've kept?
                  if (num.exclude == 1){
                    wh_exclude <- which(all_exclude)
                    if (
                      eval_df[wh_exclude, "orig.exclude"] == "Exclude-Carried-Forward" &&
                      ((wh_exclude != nrow(eval_df) &&
                        eval_df[wh_exclude+1, "orig.exclude"] == "Exclude-Carried-Forward")
                       ||
                       (wh_exclude != 1 &&
                        eval_df[wh_exclude-1, "orig.exclude"] == "Exclude-Carried-Forward")
                      )
                    ){
                      # check all the carried forwards in a string
                      res <- check_cf_string(
                        copy(eval_df), wh_exclude,
                        ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
                        ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
                        min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over
                      )

                      verdict <- res$verdict
                      cf_ind <- res$cf_ind
                      next_incl <- res$next_incl
                      first_incl <- res$first_incl

                      if (verdict == "Exclude"){
                        # mark all the CFs after for exclusion
                        all_exclude[(cf_ind-1):(next_incl-1)] <- TRUE
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(cf_ind-1):(next_incl-1)] <-
                          eval_df$temp.exclude[(cf_ind-1)]

                        # we also want to keep all the implicit excludes
                        # mark all the CFs after for "exclusion"
                        all_exclude[(first_incl+1):(cf_ind)] <- TRUE
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(first_incl+1):(cf_ind)] <-
                          "Include"
                      } else {
                        # if the verdict is include, we need to mark them all for
                        # "exclusion" to remove -- otherwise we end up in a loop
                        all_exclude[(first_incl+1):(next_incl-1)] <- TRUE
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(first_incl+1):(next_incl-1)] <-
                          "Include"
                      }
                    }

                    # exclude all the carried forwards before the next include
                    eval_df[all_exclude, exclude := temp.exclude]
                  } else if (num.exclude > 1) {
                    # first order by decreasing temp.diff (where rep=TRUE)
                    worst.row = order(all_exclude, eval_df$temp.diff, decreasing = TRUE)[1]

                    wh_exclude <- worst.row
                    if (
                      eval_df[wh_exclude, "orig.exclude"] == "Exclude-Carried-Forward" &&
                      ((wh_exclude != nrow(eval_df) &&
                        eval_df[wh_exclude+1, "orig.exclude"] == "Exclude-Carried-Forward")
                       ||
                       (wh_exclude != 1 &&
                        eval_df[wh_exclude-1, "orig.exclude"] == "Exclude-Carried-Forward")
                      )
                    ){
                      # check all the carried forwards in a string
                      res <- check_cf_string(
                        copy(eval_df), wh_exclude,
                        ewma.fields, tanner.ht.vel, who.ht.vel, exclude.levels,
                        ewma.exp, minfactor, maxfactor, banddiff, banddiff_plus,
                        min_ht.exp_under, min_ht.exp_over, max_ht.exp_under, max_ht.exp_over
                      )

                      verdict <- res$verdict
                      cf_ind <- res$cf_ind
                      next_incl <- res$next_incl
                      first_incl <- res$first_incl

                      if (verdict == "Exclude"){
                        # mark all the CFs after for exclusion
                        worst.row <- c((cf_ind-1):(next_incl-1))
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(cf_ind-1):(next_incl-1)] <-
                          eval_df$temp.exclude[(cf_ind-1)]

                        # we also want to keep all the implicit excludes
                        # mark all the CFs after for "exclusion"
                        worst.row <- c((first_incl+1):(cf_ind), worst.row)
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(first_incl+1):(cf_ind)] <-
                          "Include"
                      } else {
                        # if the verdict is include, we need to mark them all for
                        # "exclusion" to remove -- otherwise we end up in a loop
                        worst.row <- c((first_incl+1):(next_incl-1))
                        # also copy all the temp.excludes
                        eval_df$temp.exclude[(first_incl+1):(next_incl-1)] <-
                          "Include"
                      }
                    }

                    eval_df[worst.row, exclude := temp.exclude]
                  }
                }

                return(eval_df$exclude)

              })(copy(.SD))]

      # t.  If there was at least one subject who had a potential exclusion identified in step 15q, repeat steps 15b-15q. If there were no subjects with potential
      #     exclusions identified in step 15q, move on to step 16.
      newly.excluded = sum(
        subj.df$exclude %in% c(
          'Exclude-Min-Height-Change',
          'Exclude-Max-Height-Change'
        )
      )

      if (newly.excluded > num.height.excluded) {
        num.height.excluded = newly.excluded
      } else {
        break
      }
    }
    setkey(subj.df, index)
    return(subj.df$exclude)
  })(copy(.SD)), by = .(subjid), .SDcols = c('sex', 'agedays', 'v', 'tbc.sd', 'exclude', 'orig.exclude')]

  # process what came from step 15

  # if you're outside of the bands OR if you are not a carry forward then you have no change
  data.all[
    !(
      data.all[, exclude] %in% c(
        'Exclude-Min-Height-Change',
        'Exclude-Max-Height-Change'
      )
    ) &
      (data.all$orig.exclude == "Exclude-Carried-Forward"), exclude := "Include"]
  # everything with exclude should not be changed
  data.all[
    data.all[, exclude] %in% c('Exclude-Min-Height-Change','Exclude-Max-Height-Change'),
    exclude := "No Change"]
  # all original includes should not be changed
  data.all[
    data.all$orig.exclude != "Exclude-Carried-Forward",
    exclude := "No Change"
  ]

  # }

  # formulating results and options
  acf_df <- data.frame(n = data.all$n)
  acf_df <- cbind(acf_df,
                  data.all[, exclude])
  colnames(acf_df)[-1] <- "adjustcarryforward"

  # return results
  return(rbind(
    acf_df,
    data.frame(
      filter(data.orig, !.data$n %in% data.all$n) %>% mutate(adjustcarryforward = "Not Considered")  %>% select("adjustcarryforward", "n")
    )
  ))
}

# testadjustcf.R functions: to test adjust carryforward

# Function to make a grid vector for future sweep
# inputs:
# - low: number to start vector
# - high: number to end vector
# - grid.length: length of vector
# - searchtype: type of vector to make ("random","line-grid","full-grid")
# - is_included:
# - fg_val: value to use for non-used param in full-grid search
# outputs
# - numeric vector with specified qualities
#' @keywords internal
#' @noRd
#' @importFrom stats runif
make_grid_vect <- function(low,
                           high,
                           grid.length,
                           searchtype,
                           is_included = TRUE,
                           fg_val = (low + high) / 2) {
  if (searchtype == "random") {
    # random parameter sweep
    mid <- (low + high) / 2
    # return half random below, midpoint, half random above
    gv <- c(
      runif(floor(grid.length / 2), low, mid),
      mid,
      runif(floor(grid.length / 2), mid, high)
    )

    return(gv)
  } else if (searchtype == "line-grid") {
    # line-grid parameter sweep
    return(seq(low, high, length = grid.length))
  } else if (searchtype == "full-grid") {
    if (is_included) {
      # full grid vectors are put together outside the list
      return(seq(low, high, length = grid.length))
    } else {
      if (is.na(fg_val)) {
        return(c((low + high) / 2))
      } else {
        return(c(fg_val))
      }
    }
  } else {
    stop("Invalid search type")
  }
}

# Function to execute sweep of parameter values for adjustcarryforward
# inputs:
# - grid_df: data frame of parameters to sweep
# outputs:
# - combined input file and result of each run
#' @keywords internal
#' @noRd
exec_sweep <- function(grid_df,
                       exclude_opt,
                       subjid,
                       param,
                       agedays,
                       sex,
                       measurement,
                       orig.exclude,
                       dm_filt,
                       quietly = FALSE) {
  for (index in 1:nrow(grid_df)) {
    if (!quietly) {
      cat(
        sprintf(
          "[%s] Calling adjustcarryforward(), run %s of %s\n",
          Sys.time(),
          index,
          nrow(grid_df)
        )
      )
    }
    out <-
      adjustcarryforward(
        subjid,
        param,
        agedays,
        sex,
        measurement,
        orig.exclude,
        exclude_opt = exclude_opt,
        quietly = quietly,
        minfactor = grid_df$minfactor[index],
        maxfactor = grid_df$maxfactor[index],
        banddiff = grid_df$banddiff[index],
        banddiff_plus = grid_df$banddiff_plus[index],
        min_ht.exp_under = grid_df$min_ht.exp_under[index],
        min_ht.exp_over = grid_df$min_ht.exp_over[index],
        max_ht.exp_under = grid_df$max_ht.exp_under[index],
        max_ht.exp_over = grid_df$max_ht.exp_over[index]
      )

    setnames(out, "adjustcarryforward", sprintf("run-%s", index))
    export <-
      merge(as.data.frame(dm_filt),
            out,
            by = "n",
            all = TRUE
      )

    # Combine into one wide set for simplicity
    if (index == 1) {
      combo <- export
    } else {
      combo <- merge(combo, out, by = "n", all = TRUE) # %>% select(-"n")
    }
  }

  return(combo)
}

#' Function to test adjust carried forward
#'
#' The goal of this script is to consider the height values that growthcleanr
#' excludes as “carried forward” for potential re-inclusion by using a reverse
#' absolute height velocity check based on step 15 of the Daymont et al.
#' algorithm
#'
#' @param infile Input data frame/data table, cleaned by cleangrowth(), with columns as described in main README.md
#' @param seed Numeric random seed, used only when performing random search
#' @param searchtype Type of search to perform: random (default), line-grid,
#' full-grid
#' @param grid.length Number of steps in grid to search
#' @param writeout Write output to file? Default FALSE.
#' @param outfile "Output file name, default 'test_adjustcarrforward_DATE_TIME', where DATE is the current system date and time"
#' @param quietly Verbose progress info
#' @param param "none", or data frame to specify which parameters to run full search on, and values to use if not, used only when performing full-grid search
#' @param debug Produce extra data files for debugging
#' @param maxrecs Limit to specified # subjects, default 0 (no limit)
#' @param exclude_opt Type of exclusion method for carried forward strings, 0 to 3. See adjustcarryforward documentation for more information
#' @param add_answers TRUE or FALSE, indicating whether or not to add answers (definely include/exclude) for the given dataset. Defaults to TRUE
#'
#' @return A list containing:
#'     testacf_res: data frame with adjustcarryforward results for each run,
#'     params: a data frame containing parameter values for each run.
#'     debug_filtered_data: data frame with original data, returned if debug TRUE
#'
#' @export
#' @import data.table
# NOTE: no examples, since this is a temporary function
testacf <- function(
  infile,
  seed = 7,
  searchtype = "random",
  grid.length = 9,
  writeout = FALSE,
  outfile = paste0(
    "test_adjustcarryforward_",
    format(Sys.time(), "%m-%d-%Y_%H-%M-%S")
  ),
  quietly = FALSE,
  param = "none",
  debug = FALSE,
  maxrecs = 0,
  exclude_opt = 0,
  add_answers = TRUE
){

  # Parse arguments for ease ----

  if (searchtype == "random") {
    # if the line search isn't odd, add one so that it includes the midpoint
    grid.length <-
      if (grid.length %% 2 == 0) {
        grid.length + 1
      } else {
        grid.length
      }
  }

  # Prepare the data ----

  # Read in data
  dm <- as.data.table(infile)
  setkey(dm, subjid, param, agedays)

  # limit dataset for debugging/performance reasons if specified
  if (maxrecs > 0) {
    if (!quietly) {
      cat(sprintf(
        "[%s] Taking max %s recs for debugging\n",
        Sys.time(),
        maxrecs
      ))
    }
    dm_filt <- dm[subjid %in% unique(dm[, subjid])[1:maxrecs], ]
  } else {
    dm_filt <- dm
  }

  dm_filt$n <- as.numeric(rownames(dm_filt))

  subjid <- with(dm_filt, subjid)
  param <- with(dm_filt, param)
  sex <- with(dm_filt, sex)
  agedays <- with(dm_filt, agedays)
  measurement <- with(dm_filt, measurement)
  orig.exclude <- with(dm_filt, gcr_result)

  # Find out answers ----

  if (add_answers){
    if (!quietly) {
      cat(
        sprintf(
          "[%s] Finding answers for cleaned dataset\n",
          Sys.time()
        )
      )
    }

    acf_answer_df <- acf_answers(subjid,
                                 param,
                                 agedays,
                                 sex,
                                 measurement,
                                 orig.exclude,
                                 quietly = quietly)

    dm_filt <- merge(as.data.frame(dm_filt),
                     acf_answer_df,
                     by = "n",
                     all = TRUE
    )
  }

  # Execute parameter sweep ----

  if (!quietly) {
    cat(
      sprintf(
        "[%s] Running parameter sweep of adjustcarryforward on cleaned dataset\n",
        Sys.time()
      )
    )
  }

  # Specify seed, if using random
  if (searchtype == "random") {
    set.seed(seed)
  }

  # Have a data frame of ranges
  param_n <- c(
    "minfactor",
    "maxfactor",
    "banddiff",
    "banddiff_plus",
    "min_ht.exp_under",
    "min_ht.exp_over",
    "max_ht.exp_under",
    "max_ht.exp_over"
  )

  p_range <- data.frame(
    "param" = param_n,
    low = c(0, 0, 0, 0, 0, -1, 0, 0),
    high = c(1, 4, 6, 11, 4, 1, 0.66, 3)
  )
  rownames(p_range) <- p_range$param

  # Define the params to test
  if (searchtype == "full-grid") {
    # specify parameter choices -- default to choosing all
    if (param != "none" & is.data.frame(param)) {
      param_df <- param
      colnames(param_df) <- c("param", "include", "value")
    } else {
      param_df <- data.frame(
        "param" = param_n,
        "include" = TRUE,
        "value" = NA
      )
    }

    if (searchtype == "full-grid") {
      cat(
        paste0(
          "[",
          Sys.time(),
          "] Note: due to the search type specified (full-grid), there will be ",
          grid.length^sum(param_df$include),
          " runs\n"
        )
      )
    }

    # now pass them in one by one
    grid_list <- list()
    for (i in 1:nrow(param_df)) {
      grid_list[[param_df$param[i]]] <-
        make_grid_vect(
          p_range[param_df$param[i], "low"],
          p_range[param_df$param[i], "high"],
          grid.length,
          searchtype,
          as.logical(param_df$include[i]),
          as.numeric(param_df$value[i])
        )
    }

    # make the full grid based on the list
    grid_df <- expand.grid(grid_list)
  } else {
    # for random and line-grid options
    grid_df <-
      as.data.frame(matrix(NA, nrow = grid.length, ncol = nrow(p_range)))
    colnames(grid_df) <- p_range$param
    for (i in 1:ncol(grid_df)) {
      grid_df[, p_range$param[i]] <-
        make_grid_vect(
          p_range[p_range$param[i], "low"],
          p_range[p_range$param[i], "high"],
          grid.length,
          searchtype
        )
    }
  }

  # Execute
  combo <- exec_sweep(grid_df,
                      exclude_opt,
                      subjid,
                      param,
                      agedays,
                      sex,
                      measurement,
                      orig.exclude,
                      dm_filt,
                      quietly)

  # Write out results ----

  if (writeout){
    # Combined adjusted set
    fwrite(combo %>% select(-"n"), paste0(outfile, ".csv"), row.names = FALSE)

    # Record the sweep parameters for review
    grid <- cbind("run" = 1:nrow(grid_df), grid_df)
    fwrite(grid, paste0(outfile, "_parameters.csv"), row.names = FALSE)

    # Any additional debugging output
    if (debug) {
      if (!quietly) {
        cat(sprintf("[%s] Writing debugging output files\n", Sys.time()))
      }
      # Filtered data
      fwrite(dm_filt, paste0(outfile, "_debug_filtered_data.csv"), row.names = FALSE)
    }
  }

  # Write out results ----

  out_list <- list(
    # Combined adjusted set
    "testacf_res" = combo %>% select(-"n"),
    "params" = cbind("run" = 1:nrow(grid_df), grid_df)
  )

  # any additional debugging output
  if (debug){
    out_list$debug_filtered_data <- dm_filt
  }

  return(
    out_list
  )
}

# Supporting pediatric growthcleanr functions
# Supporting functions for pediatric piece of algorithm

#' Helper function for cleanbatch to identify subset of observations that are either "included" or a "temporary extraneous"
#'
#' @keywords internal
#' @noRd
valid <- function(df,
                  include.temporary.extraneous = FALSE,
                  include.extraneous = FALSE,
                  include.carryforward = FALSE) {
  exclude <- if (is.data.frame(df))
    df$exclude
  else
    df
  return(
    exclude < 'Exclude'
    |
      include.temporary.extraneous &
      exclude == 'Exclude-Temporary-Extraneous-Same-Day'
    | include.extraneous & exclude == 'Exclude-Extraneous-Same-Day'
    |
      include.carryforward &
      exclude == 'Exclude-Carried-Forward'
  )
}

#' Helper function to treat NA values as FALSE
#'
#' @keywords internal
#' @noRd
na_as_false <- function(v) {
  v[is.na(v)] <- FALSE
  v
}

#' Function for temporary extraneous (step 5):
#' 5.  Temporary extraneous: I use extraneous to refer to more than more than one recorded value for a parameter on the same day,
#'     and we need to select which one to include in our analysis. The overall strategy will be to select a measurement using a simple
#'     strategy that will be used temporarily, and select permanently in a later step after we have a somewhat cleaner dataset that
#'     can help us identify the best extraneous.
#' a.  For subjects/parameters with extraneous: Determine median_tbc*sd for both parameters: the median tbc*sd for each subject
#'     and parameter including only non-extraneous values with exc_*==0. The median of the same parameter as the extraneous will
#'     be referred to as median_tbc*sd, the median of the other parameter will be referred to as median_tbcOsd.
#' b.	For each subject/parameter with extraneous and at least one value for the subject/parameter on a day with no extraneous,
#'     select the value closest to the median_tbc*sd for temporary inclusion, and assign all other extraneous exc_*=2.
#'  i.	For each subject/parameter with extraneous and no values for the subject/parameter on a day with no extraneous, select the
#'     value closest to the median_tbcOsd for temporary inclusion, and assign all other extraneous exc_*=2.
#'     If median_tbcOsd is missing because there are no values for the other parameter, randomly choose one extraneous value for
#'     each subject/parameter/age to keep as exc_*=0 and replace exc_*=2 for all other extraneous for that subject/parameter/age.
#'
#' @keywords internal
#' @noRd
temporary_extraneous <- function(df) {
  # avoid "no visible binding" warnings
  agedays <- delta.median.sd <- extraneous <- extraneous.this.day <- NULL
  index <- median.sd <- param <- subjid <- tbc.sd <- NULL

  # add subjid and param if needed (may be missing depending on where this is called from)
  if (is.null(df$subjid))
    df[, subjid := NA]
  if (is.null(df$param))
    df[, param := NA]
  # only operate on valid rows (but include rows that may have previously been flagged as a "temporary extraneous")
  valid.rows <- valid(df, include.temporary.extraneous = TRUE)
  # make a small copy of df with only the fields we need for efficiency
  df <- df[j = .(tbc.sd), keyby = .(subjid, param, agedays, index)]
  # initialize some useful fields
  df[, `:=`(
    median.sd = as.double(NA),
    delta.median.sd = as.double(NA),
    extraneous.this.day = FALSE,
    extraneous = FALSE
  )]
  # determine on which days there are extraneous measurements (more than 1 valid measurement on that age day)
  df[valid.rows, extraneous.this.day := (.N > 1), by = .(subjid, param, agedays)]
  # calculate median of measurements on days where there is no extraneous
  df[valid.rows &
       !extraneous.this.day, median.sd := median(tbc.sd), by = .(subjid, param)]
  # distribute median to other valid or potential extraneous rows for same parameter for that subject
  df[valid.rows, median.sd := sort(median.sd)[1], by = .(subjid, param)]
  # take care of subject/parameters with more than one day with a valid observation
  # determine the absolute difference between the measurements sd score and the median for that parameter for each child
  df[valid.rows &
       extraneous.this.day, delta.median.sd := abs(tbc.sd - median.sd), by = .(subjid, param)]
  # identify subjects that have duplicates/extraneous on all days of observation for that parameter (i.e. delta.median.sd undefined)
  subj.all.dups <- df[valid.rows &
                        extraneous.this.day &
                        is.na(delta.median.sd), unique(subjid)]
  df[valid.rows &
       subjid %in% subj.all.dups, delta.median.sd := (function(subj.df) {
         # iterate over parameters where delta.median.sd is not yet defined
         # pass 1: take median from other parameter(s)
         for (p in subj.df[is.na(delta.median.sd) &
                           extraneous.this.day, unique(param)]) {
           median.other.sd <- subj.df[param != p &
                                        !extraneous.this.day, median(tbc.sd)]
           subj.df[param == p, delta.median.sd := abs(tbc.sd - median.other.sd)]
         }
         return(subj.df$delta.median.sd)
       })(copy(.SD)), by = .(subjid)]
  # Final pass: take median as zero (i.e. if no measurements from a different parameter)
  # NOTE -- this is not exactly the same as taking a random parameter
  df[valid.rows &
       extraneous.this.day &
       is.na(delta.median.sd), delta.median.sd := abs(tbc.sd)]
  # flag any extraneous value on the same day that is not the minimum distance from the median sd score
  # NOTE: in the case of exact ducplicates, "which.min" will pick the first
  df[valid.rows &
       extraneous.this.day, extraneous := seq_along(delta.median.sd) != which.min(delta.median.sd), by =
       .(subjid, param, agedays)]
  # return the extraneous valid rows (i.e. the ones that should be temporarily excluded)
  return(df$extraneous & valid.rows)
}

#' Function to identify switches (step 6):
#'  7.  Identify switches (weight and height each recorded as the other parameter)
#' Utility function to find the value associated with the opposite parameter
#' For example, if param.1='WEIGHTKG' and param.2='HEIGHTCM', then a vector of height values observed on the same day as the weight values is returned and vice versa.
#'
#' @keywords internal
#' @noRd
swap_parameters <- function(param.1 = 'WEIGHTKG',
                            param.2 = 'HEIGHTCM',
                            field.name = 'tbc.sd',
                            df) {
  # avoid "no visible binding" warnings
  agedays <- agedays.other <- param <- param.other <- subjid <- subjid.other <- swap <- NULL

  valid.rows <- valid(df)
  # copy swap field to a new value for convenience in the code below
  df$swap <- df[, field.name, with = FALSE]
  # construct an indexed table with valid parameters for speed in swapping
  valid.other <- df[valid.rows, list(
    subjid.other = subjid,
    param.other = ifelse(
      param == param.1,
      param.2,
      ifelse(param == param.2, param.1, NA)
    ),
    agedays.other = agedays,
    swap
  )]
  setkey(valid.other, subjid.other, param.other, agedays.other)
  # clear swap field (should retain original type)
  df[, swap := NA]
  # use indexed table for speed -- assign all values in one statement
  df[valid.rows, swap := valid.other[list(subjid, param, agedays), swap]]
  return(df$swap)
}

#' Function to convert vector of ages, in days, into matrix
#'
#' @keywords internal
#' @noRd
as_matrix_delta <- function(agedays) {
  n <- length(agedays)
  delta <- abs(matrix(rep(agedays, n), n, byrow = TRUE) - agedays)

  return(delta)
}

#' BMI Anthro
#'
#' Part of default CDC-derived tables
#'
#' Contains BMI data for calculating BMI
#'
#' @name bmianthro
#'
#' @section bmianthro.txt.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' CDC Growth Percentile Table
#'
#' Part of default CDC-derived tables
#'
#' Contains percentiles for various ages, gender, and weights, pre-calculated by CDC
#'
#' @name growth_cdc_ext
#'
#' @section growthfile_cdc_ext.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' CDC Growth Percentile Table for Infants
#'
#' Part of default CDC-derived tables
#'
#' Contains percentiles for various ages, gender, and weights, pre-calculated by CDC for infants algorithm
#'
#' @name growth_cdc_ext_infants
#'
#' @section growthfile_cdc_ext_infants.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' WHO Growth Percentile Table
#'
#' Part of default WHO-derived tables
#'
#' Contains percentiles for various ages, gender, and weights, pre-calculated by WHO
#'
#' @name growth_who_ext
#'
#' @section growthfile_who.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' Length to Age Table
#'
#' Part of default CDC-derived tables
#'
#' Contains percentiles for various ages, gender, and weights, pre-calculated by CDC
#'
#' @name lenanthro
#'
#' @section lenanthro.txt.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' Infants reference medians
#'
#' Contains reference median values for default recentering in the infants
#' algorithm
#'
#' @name rc-reference-medians
#'
#' @section rcfile-2023-08-15_format.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' NHANES reference medians
#'
#' Contains reference median values for default recentering, derived from NHANES
#' years 2009-2018
#'
#' @name nhanes-reference-medians
#'
#' @section nhanes-reference-medians.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' Tanner Growth Velocity Table
#'
#' Part of default CDC-derived tables
#'
#' Contains velocities for growth pre-calculated by CDC
#'
#' @name tanner_ht_vel
#'
#' @section tanner_ht_vel.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' Tanner Growth Velocity Table for Infants
#'
#' Part of default CDC-derived tables
#'
#' Contains velocities for growth pre-calculated by CDC, used for the infants
#' algorithm
#'
#' @name tanner_ht_vel_rev
#'
#' @section tanner_ht_vel_rev.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' Tanner Growth Velocity Table with (2\eqn{\sigma})
#'
#' Part of default CDC-derived tables
#'
#' Contains velocities for growth pre-calculated by CDC, including those 2
#' standard deviations away.
#'
#' @name tanner_ht_vel_with_2sd
#'
#' @section tanner_ht_vel_with_2sd.csv.gz:
#'
#' Used in function `acf_answers()`
#'
#'
NULL

#' Weight Anthro Table
#'
#' Part of default CDC-derived tables
#'
#' Contains median and standard deviation for weight by age and gender
#'
#' @name weianthro
#'
#' @section weianthro.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' WHO Maximum Height Velocity for (3\eqn{\sigma})
#'
#' Part of default WHO-derived tables
#'
#' Contains three standard deviations for the World Health Organization values of
#' maximum height velocities.
#'
#' @name who_ht_maxvel
#'
#' @section who_ht_maxvel_3sd.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' WHO Maximum Height Velocity for (2\eqn{\sigma})
#'
#' Part of default WHO-derived tables
#'
#' Contains two standard deviations for the World Health Organization values of
#' maximum height velocities.
#'
#' @name who_ht_maxvel_2sd
#'
#' @section who_ht_maxvel_2sd.csv.gz:
#'
#' Used in function `acf_answers()`
#'
#'
NULL

#' WHO Height Velocity for (3\eqn{\sigma})
#'
#' Part of default WHO-derived tables
#'
#' Contains three standard deviations for the World Health Organization values of
#' height velocities.
#'
#' @name who_ht_vel_3sd
#'
#' @section who_ht_vel_3sd.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' WHO Height Velocity for (2\eqn{\sigma})
#'
#' Part of default WHO-derived tables
#'
#' Contains two standard deviations for the World Health Organization values of
#' height velocities.
#'
#' @name who_ht_vel_2sd
#'
#' @section who_ht_vel_2sd.csv.gz:
#'
#' Used in function `acf_answers()`
#'
#'
NULL

#' WHO Maximum Head Circumference Velocity for (3\eqn{\sigma})
#'
#' Part of default WHO-derived tables
#'
#' Contains three standard deviations for the World Health Organization values of
#' maximum head circumference velocities.
#'
#' @name who_hc_maxvel
#'
#' @section who_hc_maxvel_3sd_infants.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' WHO Head Circumference Velocity for (3\eqn{\sigma})
#'
#' Part of default WHO-derived tables
#'
#' Contains three standard deviations for the World Health Organization values of
#' head circumference velocities.
#'
#' @name who_hc_vel_3sd
#'
#' @section who_hc_vel_3sd_infants.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' CDC SAS BMI Input
#'
#' Contains input data for CDC SAS macro for calculating BMI values.
#'
#' @name test_syngrowth_wide
#'
#' @section test_syngrowth_wide.csv.gz:
#'
#' Used to test function `ext_bmiz()`
#'
#'
NULL

#' CDC SAS BMI Output
#'
#' Contains results of CDC SAS macro for calculating BMI values.
#'
#' @name test_syngrowth_sas_output_compare
#'
#' @section test_syngrowth_sas_output_compare.csv.gz:
#'
#' Used to test function `ext_bmiz()`
#'
#'
NULL

#' CDC BMI reference data
#'
#' Used for extended BMIz computation
#'
#' @name CDCref_d
#'
#' @section CDCref_d.csv.gz:
#'
#' Used for extended BMI computation
#'
#'
NULL

#' Fenton-Based Corrected Age Reference (Weight to GA Lookup)
#'
#' Approximated reference values for premature infant corrected-age estimation, derived from
#' data extracted from printed Fenton (2025) growth charts. Provides a
#' reverse lookup from integer weight (grams, rounded to 10) to estimated
#' gestational age in days, by sex.
#'
#' @name ref_for_corraga
#'
#' @section ref_for_corraga.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

#' Fenton-Based Corrected Z-Score Reference
#'
#' Approximated reference values for premature infant corrected z-score calculation, derived
#' from data extracted from printed Fenton (2025) growth charts. Allows determination of approximate
#' z-scores by gestational age and sex.
#'
#' @name ref_for_corrz
#'
#' @section ref_for_corrz.csv.gz:
#'
#' Used in function `cleangrowth()`
#'
#'
NULL

# Main pediatric growthcleanr function -- cleanbatch
# internal supporting functions for pediatrics can be found in: pediatric_support.R

#' Function to clean data (optionally in batches):
#' 4.  Dataset split (optional)
#' a.	In Stata, many of the rest of the steps require you to repeat steps as many times as there are observations for the subject with the highest number of
#'     observations. Therefore, in order to speed things up I separated the dataset at this point into three groups: short (subjects with <=10 total
#'     observations), medium (subjects with >10 and <=30 observations) and long (subjects with >30 observations). I performed all the rest of the steps
#'     separately for each of the 3 datasets and then re-combined them at the end. Note that all observations for a subject stay together. This step is
#'     definitely not necessary if it will not speed things up in whatever software program you're using.
#'
#' NOTE: to use mutliple processes in R we will process patients in batches using the plyr ddply function
#' set up function to process one patient.  Function to parallelize batches is below.
#'
#' @keywords internal
#' @import data.table
#' @importFrom stats median
#' @noRd
cleanbatch <- function(data.df,
                       log.path,
                       quietly,
                       parallel,
                       measurement.to.z,
                       ewma.fields,
                       ewma.exp,
                       recover.unit.error,
                       include.carryforward,
                       sd.extreme,
                       z.extreme,
                       exclude.levels,
                       tanner.ht.vel,
                       who.ht.vel,
                       lt3.exclude.mode,
                       error.load.threshold,
                       error.load.mincount) {
  # avoid "no visible binding" warnings
  abs.2ndlast.sd <- abs.tbc.sd <- abs.tbc.sd.next <- abs.tbc.sd.prev <- abssum2 <- NULL
  aft.g.befp1 <- agedays <- agedays.other <- bef.g.aftm1 <- delta <- NULL
  delta.agedays.next <- delta.agedays.prev <- delta.next.ht <- delta.next.sd <- NULL
  delta.prev.ht <- delta.prev.sd <- dewma.after <- dewma.after.prev <- NULL
  dewma.all <- dewma.before <- dewma.before.next <- dnext.sd <- dnext.sd.minus <- NULL
  dnext.sd.plus <- dprev.sd <- dprev.sd.minus <- dprev.sd.plus <- dup <- NULL
  dup.ratio <- ewma.after <- ewma.all <- ewma.before <- exclude <- exclude.count <- NULL
  extraneous <- extraneous.this.day <- first.of.three.or.more <- ht.exp <- NULL
  include.count <- index <- last.of.three.or.more <- line <- max.ht.vel <- NULL
  max.whoinc.1.ht <- max.whoinc.2.ht <- max.whoinc.3.ht <- max.whoinc.4.ht <- NULL
  max.whoinc.6.ht <- maxdiff.next.ht <- maxdiff.prev.ht <- median.other.sd <- NULL
  min.ht.vel <- mindiff.next.ht <- mindiff.prev.ht <- pair <- pair.next <- NULL
  pair.prev <- param <- param.other <- prev.v <- sd.median <- sex <- subjid <- NULL
  swap.flag.1 <- swap.flag.2 <- tanner.months <- tbc.other.sd <- tbc.sd <- NULL
  tbc.sd.d <- tbc.sd.max <- tbc.sd.min <- tbc.sd.minus <- tbc.sd.next <- tbc.sd.plus <- NULL
  tbc.sd.prev <- tbc.sd.sw <- tbc.sd.t <- temp.diff <- temp.exclude <- v <- NULL
  v.d <- v.minus <- v.next <- v.orig <- v.plus <- v.prev <- v.sw <- v.t <- NULL
  valid.interior.measurement <- who.maxdiff.next.ht <- who.mindiff.next.ht <- NULL
  whoagegrp.ht <- whoinc.1.ht <- whoinc.2.ht <- whoinc.3.ht <- whoinc.4.ht <- NULL
  whoinc.6.ht <- whoinc.age.ht <- z.orig <- NULL

  data.df <- data.table(data.df, key = c('subjid', 'param', 'agedays', 'index'))

  if (parallel & !is.na(log.path)) {
    sink(
      sprintf(
        "%s/cleangrowth-%s-batch-%03d.log",
        log.path,
        Sys.Date(),
        data.df$batch[1]
      )
    )
  }

  if (!quietly)
    cat(sprintf(
      "[%s] Processing Batch #%s...\n",
      Sys.time(),
      data.df$batch[1]
    ))

  # save a copy of all original measurement values before any transformation
  data.df[, v.orig := v]

  if (!quietly)
    cat(sprintf(
      "[%s] Preliminarily identify potential extraneous...\n",
      Sys.time()
    ))
  data.df$exclude[temporary_extraneous(data.df)] <- 'Exclude-Temporary-Extraneous-Same-Day'

  # capture a list of subjects with possible extraneous for efficiency later
  subj.dup <- data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', unique(subjid)]

  # 7a.  For each day on which a subject had both a weight and a height recorded, calculate tbc*sd_sw: SD scores as if the weight had been recorded as the height
  #      and the height had been recorded as the weight, recentered using rcsd_*.  I intentionally did not allow values that were the first or last for a subject/parameter to be replaced as a switch.
  # NOTE: this additional constraint related to first and last values is implemented below via the code
  # swap.flag.1 := if(.N>2) c(FALSE, rep(TRUE, .N-2), FALSE) else FALSE, by=.(subjid,param)

  if (!quietly)
    cat(sprintf(
      "[%s] Identify potentially swapped measurements...\n",
      Sys.time()
    ))

  data.df[, v.sw := swap_parameters(field.name = 'v', df = data.df)]
  # calculate "standard deviation" score for the "swapped" parameter and recenter
  data.df[, tbc.sd.sw := measurement.to.z(param, agedays, sex, v.sw, TRUE) - sd.median]

  # 7b.  Perform a EWMA calculation.
  #   i.	In addition to the standard all/bef/aft dewma_* variables calculate all/bef/aft dewma_*_sw variables by subtracting EWMASD from tbc*sd_sw
  data.df[, (ewma.fields) := as.double(NaN)]
  data.df[valid(exclude), (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE), by =
            .(subjid, param)]


  # 7c.  Label pairs of height/weight measurements recorded on the same day as a switch if ALL of the following criteria are met for BOTH parameters:
  #   i.	For the weight value:
  #       exc_wt==0 & tbcwtsd>4 & |tbcwtsd_sw|<3 & dewma_wt>3 & dewma_wt_bef>2 & dewma_wt_aft>2
  #         & |dewma_wt_sw|<0.3 & |dewma_wt_sw_bef|<0.5 & |dewma_wt_sw_aft|<0.5
  #   ii.	For the height value:
  #       exc_ht==0 & tbchtsd<-7 & |tbchtsd_sw|<3 & dewma_ht<-6 & dewma_ht_bef<-5 & dewma_ht_aft<-5 & |dewma_ht_sw|<0.3 & |dewma_ht_sw_bef|<0.5 & |dewma_ht_sw_aft|<0.5
  #

  # initialize swap.flag.1
  data.df[, `:=`(valid.interior.measurement = FALSE,
                 swap.flag.1 = FALSE)]

  # flag interior measurements
  data.df[valid(data.df), valid.interior.measurement := if (.N > 2)
    c(FALSE, rep(TRUE, .N - 2), FALSE)
    else
      FALSE, by = .(subjid, param)]
  data.df[valid.interior.measurement &
            (
              param == 'WEIGHTKG' &
                tbc.sd > 4 &
                abs(tbc.sd.sw) < 3 &
                (tbc.sd - ewma.all) > 3 &
                (tbc.sd - ewma.before) > 2 &
                (tbc.sd - ewma.after) > 2
              &
                abs(tbc.sd.sw - ewma.all) < 0.3 &
                abs(tbc.sd.sw - ewma.before) < 0.5 &
                abs(tbc.sd.sw - ewma.after) < 0.5
              |
                param == 'HEIGHTCM' &
                tbc.sd < -7 &
                abs(tbc.sd.sw) < 3 &
                (tbc.sd - ewma.all) < -6 &
                (tbc.sd - ewma.before) < -5 &
                (tbc.sd - ewma.after) < -5
              &
                abs(tbc.sd.sw - ewma.all) < 0.3 &
                abs(tbc.sd.sw - ewma.before) < 0.5 &
                abs(tbc.sd.sw - ewma.after) < 0.5
            ), swap.flag.1 := TRUE]


  # 7d.  For pairs of measurements that meet criteria for a switch, do the following:
  #   i.	Replace wt with the value that was originally recorded as the ht, and replace ht with the value that was originally recorded as the wt
  #   ii.	Replace tbc*sd with the values for tbc*sd_sw
  data.df$swap.flag.2 <- swap_parameters(field.name = 'swap.flag.1', df = data.df)
  data.df[swap.flag.1 &
            swap.flag.2, `:=`(v = v.sw,
                              tbc.sd = tbc.sd.sw,
                              exclude = 'Swapped-Measurements')]

  # look for possible unit errors
  # calculate additional z-scores if we will be attempting to correct unit errors
  if (recover.unit.error) {
    # 8.  Identify unit errors (weight/height recorded in wrong units)
    # a.  Generate variables transformed by conversion factors for kg/lbs and inches/cm:
    #    i.	  wt_d_2=wt/2.204622
    #    ii.	wt_t_2=wt*2.204622
    #    iii.	ht_d_2=ht/2.54
    #    iv.	ht_t_2=ht*2.54
    if (!quietly)
      cat(sprintf("[%s] Identify and recover unit errors...\n", Sys.time()))
    valid.rows <- valid(data.df)
    data.df[, `:=`(
      v.d = v / ifelse(param == 'WEIGHTKG', 2.204622, 2.54),
      v.t = v * ifelse(param == 'WEIGHTKG', 2.204622, 2.54)
    )]

    # 8b.  Calculate SD scores for each transformed variable and recenter these (tbc*sd_d_2 and tbc_*sd_t_2)
    data.df$tbc.sd.d <- with(data.df,
                             measurement.to.z(param, agedays, sex, v.d, TRUE) - sd.median)
    data.df$tbc.sd.t <- with(data.df,
                             measurement.to.z(param, agedays, sex, v.t, TRUE) - sd.median)

    # 8c.  Perform a EWMA calculation
    #    i.	  In addition to the standard all/bef/aft dewma_* variables calculate all/bef/aft variables for  dewma_*_d_2 and dewma_*_t_2  by subtracting EWMASD from
    #         tbc*sd_d_2 and tbc_*_sd_t_2
    # NOTE: the additional dewma fields are calculated in the tests below
    data.df[, (ewma.fields) := as.double(NaN)]
    data.df[valid(exclude), (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE), by =
              .(subjid, param)]

    # 8d.  Also calculate d_prevsd_*=tbc*sd-tbc*sdprev and d_nextsd_*=tbc*sdi-tbc*sdnext. d_prevsd_* will be missing for the first value and d_nextsd_* should be
    #     missing for the last value. I intentionally did not allow values that were the first or last for a subject/parameter to be replaced as unit errors.
    # structure c(NA, field.name[-.N]) == get.prev
    # structure c(field.name[-1], NA) == get.next
    data.df[, `:=`(delta.prev.sd = as.double(NaN),
                   delta.next.sd = as.double(NaN))]
    data.df[valid.rows, `:=`(
      delta.prev.sd = tbc.sd - c(NA, tbc.sd[-.N]),
      delta.next.sd = tbc.sd - c(tbc.sd[-1], NA)
    ), by = .(subjid, param)]

    # 8e.	Identify a value as a unit error if one of the following sets of criteria are met:
    #    i.  	For wt_d_2: dewma_wt>3 & dewma_wt_bef>2 & dewma_wt_aft>2 & tbcwtsd>3 & d_nextsd_wt>2 & d_nextsd_wt is not missing & d_prevsd_wt>2 & d_prevsd_wt is not missing
    #        & abs(dewma_wt_d_2)<0.3 & abs(dewma_wt_d_2_bef)<0.5 & abs(dewma_wt_d_2_aft)<0.5 & abs(tbcwtsd_d_2)<3 & exc_wt==0
    #    ii.	For wt_t_2: dewma_wt<-3 & dewma_wt_bef<-2 & dewma_wt_aft<-2 & tbcwtsd<-3 & d_nextsd_wt<-2 & d_nextsd_wt is not missing & d_prevsd_wt<-2 & d_prevsd_wt is not missing
    #        & abs(dewma_wt_t_2)<0.3 & abs(dewma_wt_t_2_bef)<0.5 & abs(dewma_wt_t_2_aft)<0.5 & abs(tbcwtsd_t_2)<3 & exc_wt==0
    #    iii.	For ht_d_2: dewma_ht>5 & dewma_ht_bef>4 & dewma_ht_aft>4 & tbchtsd>7 & d_nextsd_ht>4 & d_nextsd_ht is not missing & d_prevsd_ht>4 & d_prevsd_ht is not missing
    #        & abs(dewma_ht_d_2)<0.3 & abs(dewma_ht_d_2_bef)<0.5 & abs(dewma_ht_d_2_aft)<0.5 & abs(tbchtsd_d_2)<3 & exc_ht==0
    #    iv.	For ht_t_2: dewma_ht<-5 & dewma_ht_bef<-4 & dewma_ht_aft<-4 & tbchtsd<-7 & d_nextsd_ht<-4 & d_nextsd_ht is not missing & d_prevsd_ht<-4 & d_prevsd_ht is not missing
    #        & abs(dewma_ht_t_2)<0.3 & abs(dewma_ht_t_2_bef)<0.5 & abs(dewma_ht_t_2_aft)<0.5 & abs(tbchtsd_t_2)<3 & exc_ht==0
    # f.  For values that are identified as unit errors by one of the sets of criteria above, replace wt or ht with the corresponding transformed value and
    #     replace tbc*sd with the recentered sd-score for the transformed value

    # process unit error low first (8.e.i and iii from above), re-factor slightly for efficiency
    data.df[valid(data.df) &
              abs(tbc.sd.d - ewma.all) < 0.3 &
              abs(tbc.sd.d - ewma.before) < 0.5 &
              abs(tbc.sd.d - ewma.after) < 0.5 & abs(tbc.sd.d) < 3
            &
              (
                param == 'WEIGHTKG' &
                  (tbc.sd - ewma.all) > 3 &
                  (tbc.sd - ewma.before) > 2 &
                  (tbc.sd - ewma.after) > 2 & tbc.sd > 3
                &
                  !is.na(delta.next.sd) &
                  delta.next.sd > 2 &
                  !is.na(delta.prev.sd) & delta.prev.sd > 2
                |
                  param == 'HEIGHTCM' &
                  (tbc.sd - ewma.all) > 5 &
                  (tbc.sd - ewma.before) > 4 &
                  (tbc.sd - ewma.after) > 4 & tbc.sd > 7
                &
                  !is.na(delta.next.sd) &
                  delta.next.sd > 4 &
                  !is.na(delta.prev.sd) & delta.prev.sd > 4
              ),
            `:=`(v = v.d,
                 tbc.sd = tbc.sd.d,
                 exclude = 'Unit-Error-High')]

    # process unit error high second (8.e.ii and iv from above), re-factor slightly for efficiency
    data.df[valid(data.df) &
              abs(tbc.sd.t - ewma.all) < 0.3 &
              abs(tbc.sd.t - ewma.before) < 0.5 &
              abs(tbc.sd.t - ewma.after) < 0.5 & abs(tbc.sd.t) < 3
            &
              (
                param == 'WEIGHTKG' &
                  (tbc.sd - ewma.all) < -3 &
                  (tbc.sd - ewma.before) < -2 &
                  (tbc.sd - ewma.after) < -2 & tbc.sd < -3
                &
                  !is.na(delta.next.sd) &
                  delta.next.sd < -2 &
                  !is.na(delta.prev.sd) & delta.prev.sd < -2
                |
                  param == 'HEIGHTCM' &
                  (tbc.sd - ewma.all) < -5 &
                  (tbc.sd - ewma.before) < -4 &
                  (tbc.sd - ewma.after) < -4 & tbc.sd < -7
                &
                  !is.na(delta.next.sd) &
                  delta.next.sd < -4 &
                  !is.na(delta.prev.sd) & delta.prev.sd < -4
              ),
            `:=`(v = v.t,
                 tbc.sd = tbc.sd.t,
                 exclude = 'Unit-Error-Low')]
  }

  # 9.  Exclude values that are carried forward. For the purposes of this analysis, any value that is identical to the preceding value for the same parameter
  #     and subject is considered carried forward. The chances of having identical measurements, even at an age/interval when little or no growth would be expected,
  #     is fairly small, and when this is the case the carried forward measurements provide little new information.
  # a.	Calculate d_prev_wt=wt-wtprev and d_prev_ht=ht-htprev. Use original measurements rather than transformed
  #     measurements (unit errors and switches).
  # b.	Unlike most steps, do this step for all extraneous values (exc_*==2) in addition to included values (exc_*==0), comparing all values for one day to all
  #     values from the prior day – if there are any values with a d_prev*==0, the value on the latter day should be excluded.
  # c.	Replace exc_*=3 for all values with d_prev*==0 & (exc_*==0 OR exc_*==2)
  if (!include.carryforward) {
    if (!quietly)
      cat(sprintf(
        "[%s] Exclude measurements carried forward...\n",
        Sys.time()
      ))
    # for efficiency, bring get.prev and get.next inline here (working on valid rows within a single parameter for a single subject)
    # structure c(NA, field.name[-.N]) == get.prev
    data.df[, prev.v := as.double(NaN)]
    data.df[valid(data.df), prev.v := c(NA, v.orig[-.N]), by = .(subjid, param)]

    # optimize "carry forward" for children without extraneous.
    data.df[!(subjid %in% subj.dup) &
              v.orig == prev.v, exclude := 'Exclude-Carried-Forward']

    # need to handle children with extraneous measurements on same day separately
    data.df[subjid %in% subj.dup &
              valid(data.df, include.temporary.extraneous = TRUE), exclude := (function(df) {
                setkey(df, agedays)
                ages = unique(agedays)
                # no point in looking for measurements carried forward if all measurements are from a single day of life
                if (length(ages) > 1) {
                  # iterate over each age
                  for (i in 2:length(ages)) {
                    # find the set of measurements from the previous age in days
                    all.prev.v = df[agedays == ages[i - 1], v.orig]
                    # if a measurement for the current age is found in the set of measurements from the previous age, then mark it as carried forward
                    df[agedays == ages[i] &
                         v.orig %in% all.prev.v, exclude := 'Exclude-Carried-Forward']
                  }
                }
                return(df$exclude)
              })(copy(.SD)), .SDcols = c('agedays', 'exclude', 'v.orig'), by = .(subjid, param)]
  }

  # 9d.  Replace exc_*=0 if exc_*==2 & redo step 5 (temporary extraneous)
  data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
  data.df[temporary_extraneous(data.df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']

  # 10.  Exclude extreme errors with SD cutoffs. For this, a cutoff of |SD|>25 is used. Because of differences in SD and z score, there are some very extreme values
  #      with a |z|>25 that are implausible with an |SD|<25, so both are used to exclude extreme errors. This works better than using a lower value for the limit
  #      for |SD|.
  # a.	Generally we only evaluate measurements where exc_*==0, but for this step we also need to evaluate measurements with exc_*==2
  # b.	Replace exc_*=4 if |tbc*sd|>25 & (exc_*==0 OR exc_*==2)
  # c.	Replace exc_*=4 if |*z|>25 & (exc_*==0 OR exc_*==2) & the value is not switched or transformed.
  if (!quietly)
    cat(sprintf(
      "[%s] Exclude extreme measurements based on SD...\n",
      Sys.time()
    ))
  data.df[na_as_false(
    valid(data.df, include.temporary.extraneous = TRUE) & abs(tbc.sd) > sd.extreme
    |
      exclude %in% c('Include', 'Exclude-Temporary-Extraneous-Same-Day') &
      abs(z.orig) > z.extreme
  ),
  exclude := 'Exclude-SD-Cutoff']

  # 10d. Redo temporary extraneous as in step 5.
  data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
  data.df[temporary_extraneous(data.df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']

  # 11.  Exclude extreme errors with EWMA
  # a.	Erroneous measurements can distort the EWMA for measurements around them. Therefore, if the EWMA method identifies more than one value for a subject and
  #     parameter that meets criteria for exclusion, we will only exclude the value that deviates the most from expected in any given step. Then we will repeat the
  #     entire process until no measurements are identified that meet criteria for exclusion.
  # b.	Perform a EWMA calculation
  #   i.	Only use values where exc_*==0 to determine the EWMAs. However, calculate dewma_* variables for values where exc_*==0 or exc_*==2
  if (!quietly)
    cat(sprintf(
      "[%s] Exclude extreme measurements based on EWMA...\n",
      Sys.time()
    ))
  data.df <- data.df[, exclude := (function(df) {
    num.ewma.excluded <- 0
    # optimization: determine whether this subject has any extraneous
    has.extraneous <- subjid %in% subj.dup
    while (TRUE) {
      df[, (ewma.fields) := as.double(NaN)]
      df[valid(exclude), (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE)]

      # note: at this point, only one ewma exists per param on a given day for a subject, so sort(ewma.all)[1] will returns the non-missing ewma.all
      # restrict to children with possible extraneous for efficiency
      if (has.extraneous) {
        df[, `:=`(
          ewma.all = sort(ewma.all)[1],
          ewma.before = sort(ewma.before)[1],
          ewma.after = sort(ewma.after)[1]
        ), by = .(agedays)]
      }
      df[, `:=`(
        dewma.all = tbc.sd - ewma.all,
        dewma.before = tbc.sd - ewma.before,
        dewma.after = tbc.sd - ewma.after
      )]

      # 11c.  Identify all values that meet all of the following criteria as potential exclusions:
      #   i.	There are 3 or more measurements for that subject and parameter
      #   ii.	(dewma_*>3.5 & dewma_*_bef>3 & dewma_*_aft>3 & tbc*sd>3.5) OR (dewma_*<-3.5 & dewma_*_bef<-3 & d & dewma_*_aft<-3 & tbc*sd<-3.5)
      #   iii.	exc_*==0

      num.valid <- sum(valid(df))
      rep <- na_as_false(with(
        df,
        num.valid >= 3 & valid(df)
        &
          (
            dewma.all > 3.5 & dewma.before > 3 & dewma.after > 3 & tbc.sd > 3.5
            |
              dewma.all < -3.5 &
              dewma.before < -3 & dewma.after < -3 & tbc.sd < -3.5
          )
      ))
      num.exclude <- sum(rep)
      # 11d.  If there is only one potential exclusion identified in step 11c for a subject and parameter, replace exc_*=5 for that value
      if (num.exclude == 1)
        df[rep, exclude := 'Exclude-EWMA-Extreme']
      # 11e.  If there is more than one potential exclusion identified in step 11c for a subject and parameter, calculate abssum_*=|tbc*sd+dewma_*| for each exclusion and
      #     replace exc_*=5 for the value with the highest abssum_*
      if (num.exclude > 1) {
        # first order by decreasing abssum
        worst.row <- with(df, order(rep, abs(tbc.sd + (
          tbc.sd - ewma.all
        )), decreasing = TRUE))[1]
        df[worst.row, exclude := 'Exclude-EWMA-Extreme']
      }

      # 11f.  For subjects/parameters with only 2 values, calculate abstbc*sd=|tbc*sd|
      # g.  Replace exc_*=6 for values that meet all of the following criteria
      #   i.	There are 2 measurements for that subject and parameter
      #   ii.	(dewma_*>3.5 & tbc*sd>3.5) OR (dewma_*<-3.5 & tbc*sd<-3.5)
      #   iii.	If there are 2 measurements for a subject/parameter that meet criteria ii, only replace exc_*=6 for the value with the larger abstbc*sd
      rep <- na_as_false(with(
        df,
        num.valid == 2 &
          (
            tbc.sd - ewma.all > 3.5 &
              tbc.sd > 3.5 |
              tbc.sd - ewma.all < -3.5 & tbc.sd < -3.5
          )
      ))
      num.exclude <- sum(rep)
      if (num.exclude == 1)
        df[rep, exclude := 'Exclude-EWMA-Extreme-Pair']
      if (num.exclude > 1) {
        # first order by decreasing abssum
        worst.row <- with(df, order(rep, abs(tbc.sd), decreasing = TRUE))[1]
        df[worst.row, exclude := 'Exclude-EWMA-Extreme-Pair']
      }

      # 11h.  Recalculate temporary extraneous as in step 5
      # optimize: only perform these steps if this subject is known to have extraneous measurements
      if (has.extraneous) {
        df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
        df[temporary_extraneous(df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']
      }

      # 11i.  If there was at least one subject who had a potential exclusion identified in step 11c, repeat steps 11b-11g. If there were no subjects with potential
      #     exclusions identified in step 11c, move on to step 12.
      newly.excluded <- sum(df$exclude %in% c('Exclude-EWMA-Extreme', 'Exclude-EWMA-Extreme-Pair'))
      if (newly.excluded > num.ewma.excluded) {
        num.ewma.excluded <- newly.excluded
      } else {
        break
      }
    }
    return(df$exclude)
  })(copy(.SD)), by = .(subjid, param), .SDcols = c('index', 'sex', 'agedays', 'tbc.sd', 'exclude')]



  # 12. Redo extraneous using EWMA. This will be the final time extraneous are done.
  #     For some extraneous it is very difficult to tell which one is likely representative. If the extraneous are very similar to each
  #     other, we will select one. If it is very difficult to tell which one is correct and the extraneous are not very similar, we will
  #     exclude all extraneous for that subject/parameter on that day
  # a.	Replace exc_*=0 for all temporarily excluded extraneous (exc_*==0)
  if (!quietly)
    cat(sprintf("[%s] Exclude extraneous based on EWMA...\n", Sys.time()))
  data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']

  # 12b. Select which extraneous to include in EWMA calculations using the same criteria as in step 5. However, do not include values
  #      in these medians that were excluded in steps 9-11 (exc_*=3, 4, 5 or 6)
  #  i.  Determine median_tbc*sd and median_tbcOsd as in step 5.
  # ii.	 For each subject/parameter with extraneous and at least one non-extraneous value, select the value closest to the
  #      median_tbc*sd for inclusion in EWMA calculations.

  # This is functionally the same as re-doing the "temporary extraneous" step before doing the EWMA
  temp.dups <- temporary_extraneous(data.df)
  data.df[temp.dups, exclude := 'Exclude-Temporary-Extraneous-Same-Day']

  # prepare a list of valid rows and initialize variables for convenience
  valid.rows <- valid(data.df)
  data.df[, `:=`(
    ewma.all = as.double(NaN),
    abssum2 = as.double(NaN),
    median.other.sd = as.double(NaN),
    extraneous = FALSE
  )]

  # 12c.	Calculate a EWMA step for all subjects/parameters with duplicates/extraneous and at
  #       least one non-extraneous value with the following modifications
  #   i.	For calculating the EWMA, include only the extraneous selected in 12c
  #  ii.	Calculate dewma_* for all values of extraneous
  # iii.	You do not need to calculate EWMAbef or EWMAaft for this step

  # determine proportion of days with extraneous/duplication for each parameter ahead of time for efficiency
  dup.ratio.df <- data.df[subjid %in% subj.dup &
                            (valid.rows |
                               temp.dups), list(dup = (.N > 1)), by = .(subjid, param, agedays)][j = list(dup.ratio =
                                                                                                            mean(dup)), keyby = .(subjid, param)]
  # identify subject/parameters where there is duplication but at least one day with no extraneous for that parameter
  subj.param.not.all.dups <- dup.ratio.df[dup.ratio < 1.0, list(subjid, param)]
  # identify subject/parameters where there is duplication for all days for that parameter
  subj.param.all.dups <- dup.ratio.df[dup.ratio == 1, list(subjid, param)]
  subj.all.dups <- subj.param.all.dups[, unique(subjid)]
  # perform ewma for subjects with extraneous
  data.df[subjid %in% subj.dup &
            valid.rows, ewma.all := ewma(agedays, tbc.sd, ewma.exp, ewma.adjacent =
                                           FALSE), by = .(subjid, param)]
  # note: at this point, only one ewma.all exists per param on a given day for a subject, so sort(ewma.all)[1] will returns the non-missing ewma.all
  data.df[subjid %in% subj.dup, ewma.all := sort(ewma.all)[1], by = .(subjid, param, agedays)]

  # iv.  Calculate abssum2_*=|2*dewma_*|+|tbc*sd| (note that this is different from how we calculated abssum_* in step 11).
  # NOTE: only children with more than one ageday with valid measurements will have a valid ewma from above
  data.df[, abssum2 := 2 * abs(tbc.sd - ewma.all) + abs(tbc.sd)]

  # 12d.  For each subject/parameter/age with extraneous and at least one non-extraneous value:
  #   i.  Replace exc_*=7 for all values except the value that has the smallest abssum2_*.
  data.df[data.table(subj.param.not.all.dups), extraneous := seq_along(abssum2) != which.min(abssum2), by =
            .(subjid, param, agedays)]
  data.df[temp.dups, exclude := 'Include']
  data.df[(valid.rows |
             temp.dups) &
            extraneous, exclude := 'Exclude-Extraneous-Same-Day']
  #  ii.  Determine dup_tot_* (# of days with extraneous for that subject/parameter) and nodup_tot_* (# of days with nonexlcuded
  #       non-extraneous for that subject/parameter).
  # iii.  If dup_tot_*/(dup_tot_*+nodup_tot_*) is greater than 1/2, replace exc_*=7 for all extraneous for that subject/parameter
  #       for each age where the  largest measurement minus the smallest measurement for that subject/parameter/age is larger than
  #       the maximum difference (ht 3cm; wt 0-9.999 kg 0.25kg; wt 10-29.9999 kg 0.5 kg; wt 30kg and higher 1 kg).

  data.df[data.table(dup.ratio.df[dup.ratio > 1 / 2, list(subjid, param)]), exclude := (function(df) {
    df[, `:=`(tbc.sd.min = as.double(NaN),
              tbc.sd.max = as.double(NaN))]
    df[valid(
      exclude,
      include.extraneous = TRUE,
      include.temporary.extraneous = TRUE
    ), `:=`(tbc.sd.min = min(tbc.sd),
            tbc.sd.max = max(tbc.sd))]
    df[tbc.sd.max - tbc.sd.min > ifelse(param == 'HEIGHTCM',
                                        3,
                                        ifelse(
                                          param == 'WEIGHTKG',
                                          ifelse(tbc.sd.min < 10, 0.25, ifelse(tbc.sd.min < 30, 0.5, 1)),
                                          NA
                                        )),
       exclude := 'Exclude-Extraneous-Same-Day']
    return(df$exclude)
  })(copy(.SD)), .SDcols = c('exclude', 'tbc.sd'), by = .(subjid, param, agedays)]

  # 12e.	For each subject/parameter/age with extraneous and no nonextraneous values:
  #   i.	Replace exc_*=7 for all values except the value with the smallest |tbc*sd-median_tbcOsd|. If median_tbcOsd is missing because there are no values
  #       for the other parameter, randomly choose one extraneous value for each subject/parameter/age to keep as exc_*=0 and replace exc_*=7 for all other
  #       extraneous for that subject/parameter/age.
  #  ii.  If the largest measurement minus the smallest measurement for that subject/parameter/age is larger than the maximum difference
  #       (ht 3cm; wt 0-9.999 kg 0.25kg; wt 10-29.9999 kg 0.5 kg; wt 30kg and higher 1 kg)., replace exc_*=7 for all extraneous for that
  #       subject/parameter/age.

  # calculate median for other parameter (restrict to subjects with all duplication for at least one parameter)
  data.df[subjid %in% subj.all.dups, exclude := (function(subj.df) {
    # flag days that have extraneous / potentially valid parameters
    subj.df[, `:=`(
      extraneous.this.day = FALSE,
      extraneous = FALSE,
      tbc.sd.min = as.double(NaN),
      tbc.sd.max = as.double(NaN)
    )]
    valid.rows = valid(
      subj.df,
      include.extraneous = TRUE,
      include.temporary.extraneous = TRUE
    )
    subj.df[valid.rows, extraneous.this.day := (.N > 1), by = .(param, agedays)]
    for (p in subj.df[j = unique(param)]) {
      median.sd <- subj.df[param != p &
                             !extraneous.this.day, median(tbc.sd)]
      subj.df[param == p, median.other.sd := median.sd]
    }
    # safety check -- assign median.other.sd==0 to ensure "which.min" functions correctly below
    subj.df[is.na(median.other.sd), median.other.sd := 0]
    # identify rows as extraneous where |tbc*sd-median_tbcOsd| is not at the minimum value
    subj.df[extraneous.this.day == TRUE, extraneous := (seq_along(median.other.sd) != which.min(abs(tbc.sd - median.other.sd))), by =
              .(param, agedays)]
    subj.df[extraneous.this.day &
              !extraneous, exclude := 'Include']
    subj.df[extraneous.this.day &
              extraneous, exclude := 'Exclude-Extraneous-Same-Day']
    subj.df[extraneous.this.day == TRUE, `:=`(tbc.sd.min = min(tbc.sd),
                                              tbc.sd.max = max(tbc.sd)), by = .(param, agedays)]
    subj.df[tbc.sd.max - tbc.sd.min > ifelse(param == 'HEIGHTCM',
                                             3,
                                             ifelse(
                                               param == 'WEIGHTKG',
                                               ifelse(tbc.sd.min < 10, 0.25, ifelse(tbc.sd.min < 30, 0.5, 1)),
                                               NA
                                             )),
            exclude := 'Exclude-Extraneous-Same-Day']

    # identify kids who had an SD or EWMA extreme excluded that was a extraneous and re-label as "Exclude-Extraneous-Same-Day"
    subj.df[, extraneous.this.day := FALSE]
    # consider any non-missing measurement when determining presence of extraneous
    subj.df[exclude != 'Missing', extraneous.this.day := (.N > 1), by =
              .(param, agedays)]
    subj.df[extraneous.this.day &
              exclude %in% c('Exclude-SD-Cutoff',
                             'Exclude-EWMA-Extreme',
                             'Exclude-EWMA-Extreme-Pair'), exclude := 'Exclude-Extraneous-Same-Day']

    return(subj.df$exclude)
  })(copy(.SD)), .SDcols = c('param', 'agedays', 'exclude', 'tbc.sd'), by =
    .(subjid)]

  # 12f.  For any values that were excluded with exc_*=4, 5, or 6 that are also extraneous, replace exc_*=7.
  data.df[subjid %in% subj.dup, exclude := (function(subj.df) {
    if (.N > 1) {
      subj.df[exclude %in% c('Exclude-SD-Cutoff',
                             'Exclude-EWMA-Extreme',
                             'Exclude-EWMA-Extreme-Pair'), exclude := 'Exclude-Extraneous-Same-Day']
    }
    return(subj.df$exclude)
  })(copy(.SD)), .SDcols = c('exclude'), by = .(subjid, param, agedays)]

  # 13.  Calculate plus/minus measurements with allowable errors and corresponding recentered SD scores
  # a.	In order to help determine if a deviation from EWMA could be explained by an allowable degree of error, calculate the following:
  #    i.	wt_plus=wt+0.05*wt
  #    ii.	wt_minus=wt-0.05*wt
  #    iii.	ht_plus=ht+1
  #    iv.	ht_minus=ht-1
  # b.	Foreach of the above calculated and then recenter the SD score for the new value, generating tbc*sd_plus and tbc*sd_minus
  data.df[, delta := ifelse(param == 'WEIGHTKG', .05 * v, 1)]
  data.df[, `:=`(v.minus = v - delta, v.plus = v + delta)]
  data.df[, `:=`(
    tbc.sd.minus = measurement.to.z(param, agedays, sex, v.minus, TRUE),
    tbc.sd.plus = measurement.to.z(param, agedays, sex, v.plus, TRUE)
  )]

  # 14.  Exclude moderate errors based on EWMA
  # a.	This step is similar to step 11, with repeated exclusions of 1 value at a time, but with different criteria than step 11. There are several criteria used
  #     as checks to make sure that values that have a large dewma_* are truly not likely to be representative.
  # b.	Perform a EWMA calculation
  #   i.	In addition to standard dewma_* variables, calculate dewma_*_plus and dewma_*_minus using the tbc*sd scores generated in step 13
  if (!quietly)
    cat(sprintf("[%s] Exclude moderate errors based on EWMA...\n", Sys.time()))
  data.df[, exclude := (function(subj.df) {
    num.ewma.excluded <- 0
    while (TRUE) {
      valid.rows <- valid(subj.df)
      # initialize fields
      subj.df[, `:=`(
        ewma.all = as.double(NaN),
        ewma.before = as.double(NaN),
        ewma.after = as.double(NaN),
        tbc.sd.prev = as.double(NaN),
        tbc.sd.next = as.double(NaN),
        delta.agedays.prev = as.integer(NaN),
        delta.agedays.next = as.integer(NaN),
        abs.2ndlast.sd = as.double(NaN),
        tbc.other.sd = as.double(NaN)
      )]
      subj.df[valid.rows, (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE), by =
                param]
      subj.df[, `:=`(
        dewma.all = tbc.sd - ewma.all,
        dewma.before = tbc.sd - ewma.before,
        dewma.after = tbc.sd - ewma.after
      )]

      # 14c.	Calculate d_prevsd=tbc*sd-tbc*sdprev; d_prevsd_minus=tbc*sd_minus-tbc*sdprev; d_prevsd_plus=tbc*sd_plus-tbc*sdprev
      #     and d_nextsd=tbc*sd-tbc*sdnext; d_nextsd_minus=tbc*sd_minus-tbc*sdnext; d_nextsd_plus=tbc*sd_plus-tbc*sdnext

      # for efficiency, bring get.prev and get.next inline here (working on valid rows within a single parameter for a single subject)
      # structure c(NA, field.name[-.N]) == get.prev
      # structure c(field.name[-1], NA) == get.next
      subj.df[valid.rows, `:=`(tbc.sd.prev = c(NA, tbc.sd[-.N]),
                               tbc.sd.next = c(tbc.sd[-1], NA)), by = param]
      subj.df[, `:=`(
        dprev.sd = tbc.sd - tbc.sd.prev,
        dprev.sd.minus = tbc.sd.minus - tbc.sd.prev,
        dprev.sd.plus = tbc.sd.plus - tbc.sd.prev,
        dnext.sd = tbc.sd - tbc.sd.next,
        dnext.sd.minus = tbc.sd.minus - tbc.sd.next,
        dnext.sd.plus = tbc.sd.plus - tbc.sd.next
      )]

      # 14d.	Calculate d_agedays_prev=agedays-agedaysprev and d_agedays_next=agedaysnext-agedays
      subj.df[valid.rows, `:=`(
        delta.agedays.prev = as.integer(agedays - c(NA, agedays[-.N])),
        delta.agedays.next = as.integer(agedays - c(agedays[-1], NA))
      ), by = param]

      # 14e.	Generate abs_2ndlast_sd=|tbc*sd| for the second-to-last measurement for a subject/parameter
      # assign this value (if defined) to all agedays for that subject
      subj.df[valid.rows, abs.2ndlast.sd := ifelse(.N >= 2, abs(tbc.sd[.N -
                                                                         1]), as.double(NaN)), by = param]

      # 14f.	Calculate tbcOsd which is the tbc*sd for the OTHER parameter for the same subject and ageday  with exc_*==0 (this may be missing).
      # NOTE: move this simplified version of swap_parameters here for efficiency
      valid.other <- subj.df[valid.rows, list(
        param.other = ifelse(param == 'WEIGHTKG', 'HEIGHTCM', 'WEIGHTKG'),
        agedays.other = agedays,
        tbc.sd
      )]
      setkey(valid.other, param.other, agedays.other)
      subj.df[valid.rows, tbc.other.sd := valid.other[list(param, agedays), tbc.sd]]

      subj.df[, exclude := (function(df) {
        # 14g.  Calculate median_tbcOsd which is the median_tbc*sd for the OTHER parameter for the same subject with exc_*==0 (this may be missing)
        # also assign param variable for use in subsequent steps
        median.tbc.other.sd <- median(df$tbc.other.sd, na.rm = TRUE)
        num.valid <- sum(valid(df$exclude))

        # 14h.	Identify values for possible exclusion if they meet one of the following sets of criteria. Generate a temporary exclusion variable temp_exc_* equal
        #     to the number indicated to keep track of which set of criteria were met
        df$temp.exclude <-
          factor(NA, levels = exclude.levels, ordered = TRUE)

        # 14h.i.	Replace temp_exc_*=8 if the value is one of 3 or more measurements for a subject/parameter AND the value is neither the first nor the last measurement
        #       AND one of the following sets of criteria are met
        #     1.	dewma_*>1 & dewma_*_bef>1 & dewma_*_aft>1 & d_nextsd_*>1 & d_prevsd_*>1 & d_prevsd_plus_*>1 & d_prevsd_minus_*>1 & d_nextsd_plus_*>1 & d_nextsd_minus_*>1
        #     2.	dewma_*<-1 & dewma_*_bef<-1 & dewma_*_aft<-1 & d_nextsd_*<-1 & d_prevsd_*<-1 & d_prevsd_plus_*<-1 & d_prevsd_minus_*<-1 & d_nextsd_plus_*<-1 & d_nextsd_minus_*<-1

        # NOTE: NA as treated as false in context of data.table, which takes care of the NA values that result when
        # evaluating the boolean expression for a first or last value of a parameter for a child,
        # which ensures the flag is only set for measurements that have both a preceding and subsequent measurement
        df[dewma.all > 1 &
             dewma.before > 1 &
             dewma.after > 1 &
             dprev.sd > 1 &
             dprev.sd.plus > 1 &
             dprev.sd.minus > 1 &
             dnext.sd > 1 &
             dnext.sd.plus > 1 & dnext.sd.minus > 1
           |
             dewma.all < -1 &
             dewma.before < -1 &
             dewma.after < -1 &
             dprev.sd < -1 &
             dprev.sd.plus < -1 &
             dprev.sd.minus < -1 &
             dnext.sd < -1 &
             dnext.sd.plus < -1 & dnext.sd.minus < -1,
           temp.exclude := 'Exclude-EWMA-8']

        # 14h.ii.	Replace temp_exc_*=9 (mark for potential exclusion) if the value is the first of 3 or more measurements for a subject/parameter
        #      AND d_agedays_next< 365.25 AND one of the following sets of criteria are met
        #
        #     1.	dewma_*>2 & dewma_*_aft>1 & d_nextsd_*>1 & d_nextsd_plus_*>1 & d_nextsd_minus_*>1 & agedays>=30
        #     2.	dewma_*<-2 & dewma_*_aft<-1 & d_nextsd_*<-1 & d_nextsd_plus_*<-1 & d_nextsd_minus_*<-1 & agedays>=30
        #     3.  dewma_*>2.5 & dewma_*_aft>1 & d_nextsd_*>1 & d_nextsd_plus_*>1 & d_nextsd_minus_*>1 & agedays<30
        #     4.  dewma_*<-4 & dewma_*_aft<-1 & d_nextsd_*<-1 & d_nextsd_plus_*<-1 & d_nextsd_minus_*<-1 & agedays<30

        # take advantage of other variables we have calculated to infer that a row is the first of three or more valid measurements for a paramete
        df$first.of.three.or.more <- FALSE
        df$last.of.three.or.more <- FALSE
        df[is.na(delta.agedays.prev) &
             num.valid >= 3, first.of.three.or.more := TRUE]
        df[is.na(delta.agedays.next) &
             num.valid >= 3, last.of.three.or.more := TRUE]

        df[first.of.three.or.more & delta.agedays.next < 365.25
           &
             (
               dewma.all > 2 &
                 dewma.after > 1 &
                 dnext.sd > 1 &
                 dnext.sd.plus > 1 & dnext.sd.minus > 1 & agedays >= 30
               |
                 dewma.all < -2 &
                 dewma.after < -1 &
                 dnext.sd < -1 &
                 dnext.sd.plus < -1 & dnext.sd.minus < -1 & agedays >= 30
               |
                 dewma.all > 2.5 &
                 dewma.after > 1 &
                 dnext.sd > 1 &
                 dnext.sd.plus > 1 & dnext.sd.minus > 1 & agedays < 30
               |
                 dewma.all < -4 &
                 dewma.after < -1 &
                 dnext.sd < -1 &
                 dnext.sd.plus < -1 & dnext.sd.minus < -1 & agedays < 30
             ),
           temp.exclude := 'Exclude-EWMA-9']

        # 14h.iii.	Replace temp_exc_*=10 if the value is the first of 3 or more measurements for a subject/parameter AND d_agedays_next>365.25
        #         AND one of the following sets of criteria are met
        #     1.	dewma_*>3 & dewma_*_aft>1 & d_nextsd_*>1 & d_nextsd_plus_*>1 & d_nextsd_minus_*>1
        #     2.	dewma_*<-3 & dewma_*_aft<-1 & d_nextsd_*<-1 & d_nextsd_plus_*<-1 & d_nextsd_minus_*<-1

        # NOTE: changed age comparison to >= to ensure complete coverage of possible conditions in the event case fractional days are present in the dataset
        df[first.of.three.or.more & delta.agedays.next >= 365.25
           &
             (
               dewma.all > 3 &
                 dewma.after > 1 &
                 dnext.sd > 1 &
                 dnext.sd.plus > 1 & dnext.sd.minus > 1
               |
                 dewma.all < -3 &
                 dewma.after < -1 &
                 dnext.sd < -1 &
                 dnext.sd.plus < -1 & dnext.sd.minus < -1
             ),
           temp.exclude := 'Exclude-EWMA-10']

        # 14h.iv.	Replace temp_exc_*=11 if the value is the last of 3 or more measurements for a subject/parameter AND d_agedays_prev<730.5 AND abs_2ndlast_sd <2
        #       AND one of the following sets of criteria are met
        #     1.	dewma_*>2 & dewma_*_bef>1 & d_prevsd_*>1 & d_prevsd_plus_*>1 & d_prevsd_minus_*>1
        #     2.	dewma_*<-2 & dewma_*_bef<-1 & d_prevsd_*<-1 & d_prevsd_plus_*<-1 & d_prevsd_minus_*<-1
        df[last.of.three.or.more &
             delta.agedays.prev < 730.5 & abs.2ndlast.sd < 2
           &
             (
               dewma.all > 2 &
                 dewma.before > 1 &
                 dprev.sd > 1 &
                 dprev.sd.plus > 1 & dprev.sd.minus > 1
               |
                 dewma.all < -2 &
                 dewma.before < -1 &
                 dprev.sd < -1 &
                 dprev.sd.plus < -1 & dprev.sd.minus < -1
             ),
           temp.exclude := 'Exclude-EWMA-11']

        # 14h.v.	Replace temp_exc_*=12 if the value is the last of 3 or more measurements for a subject/parameter AND d_agedays_prev<730.5 AND abs_2ndlast_sd >=2
        #       AND one of the following sets of criteria are met
        #     1.	dewma_*>abs_2ndlast_sd & dewma_*_bef>1 & d_prevsd_*>1 & d_prevsd_plus_*>1 & d_prevsd_minus_*>1
        #     2.	dewma_*<(-1*abs_2ndlast_sd) & dewma_*_bef<-1 & d_prevsd_*<-1 & d_prevsd_plus_*<-1 & d_prevsd_minus_*<-1
        df[last.of.three.or.more &
             delta.agedays.prev < 730.5 & abs.2ndlast.sd >= 2
           &
             (
               dewma.all > abs.2ndlast.sd &
                 dewma.before > 1 &
                 dprev.sd > 1 &
                 dprev.sd.plus > 1 & dprev.sd.minus > 1
               |
                 dewma.all < -abs.2ndlast.sd &
                 dewma.before < -1 &
                 dprev.sd < -1 &
                 dprev.sd.plus < -1 & dprev.sd.minus < -1
             ),
           temp.exclude := 'Exclude-EWMA-12']

        # 14h.vi.	Replace temp_exc_*=13 if the value is the last of 3 or more measurements for a subject/parameter AND d_agedays_prev>730.5 AND  abs_2ndlast_sd <2
        #       AND one of the following sets of criteria are met
        #     1.	dewma_*>3 & dewma_*_bef>1 & d_prevsd_*>1 & d_prevsd_plus_*>1 & d_prevsd_minus_*>1 & ((tbc*sd-tbcOsd)>4
        #         OR ((tbc*sd-median_tbcOsd)>4 & tbcOsd is missing) OR median_tbcOsd is missing)
        #     2.	dewma_*<-3 & dewma_*_bef<-1 & d_prevsd_*<-1 & d_prevsd_plus_*<-1 & d_prevsd_minus_*<-1 & ((tbc*sd-tbcOsd)<-4
        #         OR ((tbc*sd-median_tbcOsd)<-4 & tbcOsd is missing) OR median_tbcOsd is missing)

        # NOTE: changed age comparison to >= to ensure complete coverage of possible conditions in the event case fractional days are present in the dataset
        df[last.of.three.or.more &
             delta.agedays.prev >= 730.5 & abs.2ndlast.sd < 2
           &
             (
               dewma.all > 3 &
                 dewma.before > 1 &
                 dprev.sd > 1 &
                 dprev.sd.plus > 1 & dprev.sd.minus > 1
               &
                 (
                   tbc.sd - tbc.other.sd > 4 |
                     is.na(tbc.other.sd) &
                     tbc.sd - median.tbc.other.sd > 4 |
                     is.na(median.tbc.other.sd)
                 )
               |
                 dewma.all < -3 &
                 dewma.before < -1 &
                 dprev.sd < -1 &
                 dprev.sd.plus < -1 & dprev.sd.minus < -1
               &
                 (
                   tbc.sd - tbc.other.sd < -4 |
                     is.na(tbc.other.sd) &
                     tbc.sd - median.tbc.other.sd < -4 |
                     is.na(median.tbc.other.sd)
                 )
             ),
           temp.exclude := 'Exclude-EWMA-13']

        # 14h.vii.	Replace temp_exc_*=14 if the value is the last of 3 or more measurements for a subject/parameter AND d_agedays_prev>730.5 AND  abs_2ndlast_sd>=2
        #         AND one of the following sets of criteria are met
        #     1.	dewma_*>(1+abs_2ndlast_sd) & dewma_*_bef>1 & d_prevsd_*>1 & d_prevsd_plus_*>1 & d_prevsd_minus_*>1 & ((tbc*sd-tbcOsd)>4
        #         OR ((tbc*sd-median_tbcOsd)>4 & tbcOsd is missing) OR median_tbcOsd is missing)
        #     2.	dewma_*<(-1-abs_2ndlast_sd) & dewma_*_bef<-1 & d_prevsd_*<-1 & d_prevsd_plus_*<-1 & d_prevsd_minus_*<-1 & ((tbc*sd-tbcOsd)<-4
        #         OR ((tbc*sd-median_tbcOsd)<-4 & tbcOsd is missing) OR median_tbcOsd is missing)
        df[last.of.three.or.more &
             delta.agedays.prev >= 730.5 & abs.2ndlast.sd >= 2
           &
             (
               dewma.all > 1 + abs.2ndlast.sd &
                 dewma.before > 1 &
                 dprev.sd > 1 &
                 dprev.sd.plus > 1 & dprev.sd.minus > 1
               &
                 (
                   tbc.sd - tbc.other.sd > 4 |
                     is.na(tbc.other.sd) &
                     tbc.sd - median.tbc.other.sd > 4 |
                     is.na(median.tbc.other.sd)
                 )
               |
                 dewma.all < -1 - abs.2ndlast.sd &
                 dewma.before < -1 &
                 dprev.sd < -1 &
                 dprev.sd.plus < -1 & dprev.sd.minus < -1
               &
                 (
                   tbc.sd - tbc.other.sd < -4 |
                     is.na(tbc.other.sd) &
                     tbc.sd - median.tbc.other.sd < -4 |
                     is.na(median.tbc.other.sd)
                 )
             ),
           temp.exclude := 'Exclude-EWMA-14']

        # 14i.	If there is only one potential exclusion identified in step 14h for a subject and parameter, replace exc_*=temp_exc_* for that value
        rep <- !is.na(df$temp.exclude)
        num.exclude <- sum(rep)
        if (num.exclude == 1)
          df[rep, exclude := temp.exclude]

        # 14j.	If there is more than one potential exclusion identified in step 14h for a subject and parameter, calculate abssum_*=|tbc*sd+dewma_*| for each exclusion
        #     and replace exc_*=temp_exc_* for the value with the highest abssum_*
        if (num.exclude > 1) {
          # first order by decreasing abssum  (where rep=TRUE)
          worst.row <- with(df, order(rep, abs(tbc.sd + dewma.all), decreasing =
                                        TRUE))[1]
          df[worst.row, exclude := temp.exclude]
        }

        return(df$exclude)
      })(copy(.SD)), by = param]

      # k.	If there was at least one subject who had a potential exclusion identified in step 14h, repeat steps 14b-14j. If there were no subjects with potential
      #     exclusions identified in step 14h, move on to step 15.
      newly.excluded <- sum(subj.df$exclude >= 'Exclude-EWMA-8' &
                              subj.df$exclude <= 'Exclude-EWMA-14')
      if (newly.excluded > num.ewma.excluded) {
        num.ewma.excluded <- newly.excluded
      } else {
        break
      }
    }

    return(subj.df$exclude)
  })(copy(.SD)), by = subjid, .SDcols = c(
    'index',
    'sex',
    'param',
    'agedays',
    'v',
    'tbc.sd',
    'v.minus',
    'v.plus',
    'tbc.sd.minus',
    'tbc.sd.plus',
    'exclude'
  )]

  # 15.  Exclude heights based on absolute differences in measurement. The key to this step is that once we identify pairs of measurements with implausible
  #      amounts of absolute difference between them, we have to determine which of the two values in the pair is less likely to be representative and should
  #      be excluded. For subjects/parameters with 3 or more measurements, this is done by looking at the dewma_* of each of the 2 values in a pair using a ewma that
  #      excludes the other value in the pair. For subjects/parameters with 2 or more measurements, this is done by looking at the absolute value of the tbc*sd.
  #      The Tanner height velocity reference is used for measurements taken at >2yo, WHO will be used for <2yo. For a few pairs of measurements either could be used;
  #      WHO will be used if difference between ages is < 9 months.
  if (!quietly)
    cat(sprintf(
      "[%s] Exclude heights based on growth velocity...\n",
      Sys.time()
    ))
  data.df[param == 'HEIGHTCM', exclude := (function(subj.df) {
    # assign some book keeping variables
    #subj.df[, `:=`(subjid = subjid, param='HEIGHTCM',index=1:.N)]
    subj.df[, index := 1:.N]

    num.height.excluded <- 0
    while (TRUE) {
      # use a closure to discard all the extra fields added to df with each iteration
      subj.df[valid(exclude), exclude := (function (df) {
        # with each iteration we are operating only on valid rows to allow simplification of code below

        # initialize fields
        df[, (ewma.fields) := as.double(NaN)]
        df[, `:=`(
          v.prev = as.double(NaN),
          v.next = as.double(NaN),
          dewma.after.prev = as.double(NaN),
          dewma.before.next = as.double(NaN),
          abs.tbc.sd.prev = as.double(NaN),
          abs.tbc.sd.next = as.double(NaN),
          agedays.next = as.integer(NaN),
          abs.2ndlast.sd = as.double(NaN),
          mindiff.prev.ht = as.double(NaN),
          mindiff.next.ht = as.double(NaN),
          maxdiff.prev.ht = as.double(NaN),
          maxdiff.next.ht = as.double(NaN),
          pair.prev = FALSE,
          pair.next = FALSE
        )]

        # ewma fields are needed later -- calculate now for efficiency
        df[, (ewma.fields) := ewma(agedays, tbc.sd, ewma.exp, TRUE)]

        # calculate some usefule values (e.g. dewma values and tbc.sd) for use in later steps
        df[, `:=`(
          dewma.all = tbc.sd - ewma.all,
          dewma.before = tbc.sd - ewma.before,
          dewma.after = tbc.sd - ewma.after,
          abs.tbc.sd = abs(tbc.sd)
        )]

        # 15a.  As with steps 11 and 14, only one value will be excluded per round, and the step will be repeated until there are no more values to exclude
        # b.  For each height, calculate the d_age=agedays of next value-agedays of current value
        # NOTE: obtain next measurement, ewma.before and abs.tbc.sd as well since they are needed later

        # for efficiency, bring get.next inline here (working on valid rows within a single parameter for a single subject)
        # structure c(field.name[-1], NA) == get.next
        df[, `:=`(
          agedays.next = c(agedays[-1], NA),
          v.next = c(v[-1], NA),
          dewma.before.next = c(dewma.before[-1], NA),
          abs.tbc.sd.next = c(abs.tbc.sd[-1], NA)
        )]
        df$delta.agedays.next <- with(df, agedays.next - agedays)

        # 15c.	For each height, calculate mid_agedays=0.5*(agedays of next value + agedays of current value)
        df$mid.agedays <- 0.5 * (df$agedays.next + df$agedays)

        # 15d.	Generate variable tanner_months= 6+12*(round(mid_agedays/365.25))
        # only calculate for rows that relate to height (may speed up subsequent processing)
        df$tanner.months <-
          with(df, 6 + 12 * (round(mid.agedays / 365.25)))

        # 15e.	Merge with dataset tanner_ht_vel using sex and tanner_months – this will give you min_ht_vel and max_ht_vel
        setkey(df, sex, tanner.months)
        df <- tanner.ht.vel[df]

        # 15f.	Calculate the following:
        #   i.	mindiff_ht=0.5*min_ht_vel*(d_agedays/365.25)^2-3 if d_agedays<365.25
        #   ii.	replace mindiff_ht=0.5*min_ht_vel-3 if d_agedays>365.25
        df[, ht.exp := ifelse(delta.agedays.next < 365.25, 2, 0)]
        df[, `:=`(maxdiff.next.ht = as.double(NA),
                  mindiff.next.ht = as.double(NaN))]
        df[, mindiff.next.ht := 0.5 * min.ht.vel * (delta.agedays.next /
                                                      365.25) ^ ht.exp - 3]

        # 15f.iii.	maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^1.5+5.5 if d_agedays>365.25
        #   iv.	replace maxdiff_ht=2*max_ht_vel*(d_agedays/365.25)^0.33+5.5 if d_agedays<365.25
        df[, ht.exp := ifelse(delta.agedays.next < 365.25, 0.33, 1.5)]
        df[, maxdiff.next.ht := 2 * max.ht.vel * (delta.agedays.next /
                                                    365.25) ^ ht.exp + 5.5]

        # 15g.	Generate variable whoagegrp_ht=agedays/30.4375 rounded to the nearest integer
        df[, whoagegrp.ht := round(agedays / 30.4375)]

        # 15h.	Generate variable whoinc_age_ht based on values of d_agedays_ht according the the following table
        #   d_agedays_ht	whoinc_age_ht
        #   20-45	        1
        #   46-75	        2
        #   76-106	      3
        #   107-152	      4
        #   153-198	      6
        #   All others	  missing
        df[, whoinc.age.ht := ifelse(delta.agedays.next < 20 ,
                                     NA,
                                     ifelse(
                                       delta.agedays.next <= 45,
                                       1,
                                       ifelse(
                                         delta.agedays.next <= 75,
                                         2,
                                         ifelse(
                                           delta.agedays.next <= 106,
                                           3,
                                           ifelse(
                                             delta.agedays.next <= 152,
                                             4,
                                             ifelse(delta.agedays.next <= 198, 6, NA)
                                           )
                                         )
                                       )
                                     ))]

        # i.	Merge using sex and whoagegrp_ht using who_ht_vel_3sd and who_ht_maxvel_3sd; this will give you varaibles whoinc_i_ht and maxwhoinc_i_ht
        #     for various intervals where i is 1,2, 3,4, 6 and corresponds to whoinc_age_ht.
        setkey(df, sex, whoagegrp.ht)
        df <- who.ht.vel[df]

        # restore original sort order (ensures valid.rows variable applies to correct rows)
        setkey(df, index)

        # 15j.	Generate variable who_mindiff_ht=whoinc_i_ht according to the value if whoinc_age_ht; make who_mindiff_ht missing if whoinc_i_ht or whoinc_age_ht is missing.
        df[, who.mindiff.next.ht := ifelse(
          delta.agedays.next < 20 ,
          NA,
          ifelse(
            delta.agedays.next <= 45,
            whoinc.1.ht,
            ifelse(
              delta.agedays.next <= 75,
              whoinc.2.ht,
              ifelse(
                delta.agedays.next <= 106,
                whoinc.3.ht,
                ifelse(
                  delta.agedays.next <= 152,
                  whoinc.4.ht,
                  ifelse(delta.agedays.next <= 198, whoinc.6.ht, NA)
                )
              )
            )
          )
        )]

        # 15k.	Generate variable who_maxdiff_ht=max_whoinc_i_ht according to the value if whoinc_age_ht; make who_maxdiff_ht missing if max_whoinc_i_ht or
        #     whoinc_age_ht is missing.
        df[, who.maxdiff.next.ht := ifelse(
          delta.agedays.next < 20 ,
          NA,
          ifelse(
            delta.agedays.next <= 45,
            max.whoinc.1.ht,
            ifelse(
              delta.agedays.next <= 75,
              max.whoinc.2.ht,
              ifelse(
                delta.agedays.next <= 106,
                max.whoinc.3.ht,
                ifelse(
                  delta.agedays.next <= 152,
                  max.whoinc.4.ht,
                  ifelse(delta.agedays.next <= 198, max.whoinc.6.ht, NA)
                )
              )
            )
          )
        )]

        # 15l.	Scale allowed value based on d_agedays_ht:
        #   1.	replace who_mindiff_`p'=who_mindiff_`p'*d_agedays_`p'/(whoinc_age_`p'*30.4375) if d_agedays_`p'<(whoinc_age_`p'*30.4375)
        #   2.	replace who_maxdiff_`p'=who_maxdiff_`p'*d_agedays_`p'/(whoinc_age_`p'*30.4375) if d_agedays_`p'>(whoinc_age_`p'*30.4375)
        df[delta.agedays.next < whoinc.age.ht * 30.4375,
           `:=`(
             who.mindiff.next.ht = who.mindiff.next.ht * delta.agedays.next / (whoinc.age.ht *
                                                                                 30.4375),
             who.maxdiff.next.ht = who.maxdiff.next.ht * delta.agedays.next /
               (whoinc.age.ht * 30.4375)
           )]

        # 15m.	Replace mindiff_ht/maxdiff_ht with adjusted WHO value if Tanner value is missing or if both are present and age difference is < 9 months:
        #   1.	replace mindiff_`p'=0.5*who_mindiff_`p'-3 if who_mindiff_`p' is not missing & d_agedays_`p'<(9*30.4375)
        #   2.	replace maxdiff_`p'=2*who_maxdiff_`p'+3 if who_maxdiff_`p' is not missing & d_agedays_`p'<(9*30.4375)
        #   3.	replace mindiff_`p'=0.5*who_mindiff_`p'-3 if mindiff_`p' is missing & who_mindiff_`p' is not missing
        #   4.	replace maxdiff_`p'=2*who_maxdiff_`p'+3 if maxdiff_`p is missing & who_maxdiff_`p' is not missing

        # refactored logic slightly for efficiency
        df[!is.na(who.mindiff.next.ht) &
             (delta.agedays.next < 9 * 30.4375 |
                is.na(mindiff.next.ht)),
           `:=`(
             mindiff.next.ht = 0.5 * who.mindiff.next.ht - 3,
             maxdiff.next.ht = 2.0 * who.maxdiff.next.ht + 3
           )]

        # 15m.5.  replace mindiff_`p'=-3 if mindiff_`p' is missing
        df[is.na(mindiff.next.ht), mindiff.next.ht := -3]

        # 15n.	Determine the min/maxdiffs for the previous age: mindiff_prev_ht, maxdiff_prev_ht
        # NOTE: obtain previous height, ewma.after value and abs.tbc.sd as well since they are needed in next steps

        # for efficiency, bring get.prev inline here (working on valid rows within a single parameter for a single subject)
        # structure c(NA, tbc.sd[-.N]) == get.prev
        df[, `:=`(
          v.prev = c(NA, v[-.N]),
          dewma.after.prev = c(NA, dewma.after[-.N]),
          abs.tbc.sd.prev = c(NA, abs.tbc.sd[-.N]),
          mindiff.prev.ht = c(NA, mindiff.next.ht[-.N]),
          maxdiff.prev.ht = c(NA, maxdiff.next.ht[-.N])
        )]

        # 15o.	Determine d_prev_ht=ht-htprev (set to missing for the first value for a subject) and d_next_ht=htnext-ht (set to missing for the last value for a subject)
        df[, `:=`(delta.prev.ht = v - v.prev,
                  delta.next.ht = v.next - v)]

        # 15p.  Perform a EWMA calculation with the following modifications:
        #  i.	  Generate a variable pair=1 if (d_prev_ht<mindiff_prev_ht OR d_ht<mindiff_ht OR d_prev_ht>maxdiff_prev_ht  OR d_ht>maxdiff_ht) AND exc_ht==0
        df[, pair := na_as_false(
          delta.prev.ht < mindiff.prev.ht |
            delta.next.ht < mindiff.next.ht |
            delta.prev.ht > maxdiff.prev.ht |
            delta.next.ht > maxdiff.next.ht
        )]

        # for efficiency, bring get.prev and get.next inline here (working on valid rows within a single parameter for a single subject)
        # structure c(NA, field.name[-.N]) == get.prev
        # structure c(field.name[-1], NA) == get.next
        df[, `:=`(pair.prev = c(FALSE, pair[-.N]),
                  pair.next = c(pair[-1], FALSE))]

        #  ii.	Generate bef_g_aftm1=1 if |Δewma_htbef| for the value of interest is greater than |Δewma_htaft| for the previous value
        #       AND the value of interest is not the first height value for that subject AND pair==1 AND pair for the previous value==1

        #  iii.	Generate aft_g_befp1=1 if |Δewma_htaft| for the value of interest is greater than |Δewma_htbef| for the next value
        #       AND the value of interest is not the last height value for that subject AND pair==1 AND pair for the next value==1
        # NOTE: pair.next will be NA last height, which will result in a FALSE value below
        df[, `:=`(
          bef.g.aftm1 = na_as_false(
            abs(dewma.before) > abs(dewma.after.prev)  & pair & pair.prev
          ),
          aft.g.befp1 = na_as_false(
            abs(dewma.after)  > abs(dewma.before.next) & pair & pair.next
          )
        )]

        #  iv.	Determine tbchtsd for each value as well as the one before prev_tbchtsd and after next_tbchtsd it
        # NOTE: done previously for efficiency

        # 15p.v.  Determine the total number of ht values for each subject (tot_ht)
        # NOTE: all rows are valid due to constraint in subj.df[...] statement
        num.valid <- .N

        # 15q.	Identify a value for possible exclusion if one of the following sets of criteria are met. For values identified by each set of criteria determine
        #       the value of temp_diff using the formula given
        #   i.	d_prev_ht<mindiff_prev_ht & bef_g_aftm1_ht==1 & exc_ht==0 & mindiff_prev_ht is not missing
        #     a.  (temp_diff=|dewma_ht_bef|)
        df[, temp.diff := as.double(NaN)]
        df$temp.exclude <-
          factor(NA, levels = exclude.levels, ordered = TRUE)
        df[delta.prev.ht < mindiff.prev.ht & bef.g.aftm1,
           `:=`(temp.diff = abs(dewma.before),
                temp.exclude = 'Exclude-Min-Height-Change')]

        #   ii.	d_ht<mindiff_ht & aft_g_befp1_ht==1 & exc_ht==0 & mindiff_ht is not missing
        #     a.	(temp_diff=|dewma_ht_aft|)
        df[delta.next.ht < mindiff.next.ht & aft.g.befp1,
           `:=`(temp.diff = abs(dewma.after),
                temp.exclude = 'Exclude-Min-Height-Change')]

        #   iii.	d_prev_ht>maxdiff_prev_ht & bef_g_aftm1_ht==1 & exc_ht==0 & mindiff_prev_ht is not missing
        #     a.  (temp_diff=|dewma_ht_bef|)
        df[delta.prev.ht > maxdiff.prev.ht & bef.g.aftm1,
           `:=`(temp.diff = abs(dewma.before),
                temp.exclude = 'Exclude-Max-Height-Change')]

        #   iv.	d_ht>maxdiff_ht & aft_g_befp1_ht==1 & exc_ht==0 & mindiff_ht is not missing
        #     a.  (temp_diff=|dewma_ht_aft|)
        df[delta.next.ht > maxdiff.next.ht & aft.g.befp1,
           `:=`(temp.diff = abs(dewma.after),
                temp.exclude = 'Exclude-Max-Height-Change')]

        #   v.	d_prev_ht<mindiff_prev_ht & tot_ht==2 & |tbchtsd|>|prev_tbchtsd|
        #     a. for v-viii temp_diff is kept as missing
        #   vi. d_ht<mindiff_ht & tot_ht==2 & |tbchtsd|>|next_tbchtsd|
        df[delta.prev.ht < mindiff.prev.ht &
             num.valid == 2 & abs.tbc.sd > abs.tbc.sd.prev
           |
             delta.next.ht < mindiff.next.ht &
             num.valid == 2 & abs.tbc.sd > abs.tbc.sd.next,
           temp.exclude := 'Exclude-Min-Height-Change']

        #   vii.	d_prev_ht>maxdiff_prev_ht & tot_ht==2 & |tbchtsd|>|prev_tbchtsd|
        #   viii. d_ht>maxdiff_ht & tot_ht==2 & |tbchtsd|>|next_tbchtsd|
        df[delta.prev.ht > maxdiff.prev.ht &
             num.valid == 2 & abs.tbc.sd > abs.tbc.sd.prev
           |
             delta.next.ht > maxdiff.next.ht &
             num.valid == 2 & abs.tbc.sd > abs.tbc.sd.next,
           temp.exclude := 'Exclude-Max-Height-Change']

        # r.  If there is only one potential exclusion identified in step 15j for a subject and parameter,
        #     replace exc_ht=15 for that value if it met criteria i, ii, v, or vi  and exc_ht=16 if it met criteria iii, iv, vii, or viii
        # NOTE: these exclusions are assigned in the code above as 'Exclude-Min-Height-Change' and 'Exclude-Max-Height-Change'
        rep <- !is.na(df$temp.exclude)
        num.exclude <- sum(rep)
        if (num.exclude == 1)
          df[rep, exclude := temp.exclude]

        # s.  If there is more than one potential exclusion identified in step 14h for a subject and parameter, determine which value has the largest temp_diff and
        #     replace exc_ht=15 for that value if it met criteria i, ii, v, or vi and exc_ht=16 for that value if it met criteria iii,  iv, vii, or viii.

        if (num.exclude > 1) {
          # first order by decreasing temp.diff (where rep=TRUE)
          worst.row <- order(rep, df$temp.diff, decreasing = TRUE)[1]
          df[worst.row, exclude := temp.exclude]
        }

        return(df$exclude)
      })(copy(.SD))]

      # t.  If there was at least one subject who had a potential exclusion identified in step 15q, repeat steps 15b-15q. If there were no subjects with potential
      #     exclusions identified in step 15q, move on to step 16.
      newly.excluded <- sum(
        subj.df$exclude %in% c(
          'Exclude-Min-Height-Change',
          'Exclude-Max-Height-Change'
        )
      )
      if (newly.excluded > num.height.excluded) {
        num.height.excluded <- newly.excluded
      } else {
        break
      }
    }

    setkey(subj.df, index)
    return(subj.df$exclude)
  })(copy(.SD)), by = .(subjid), .SDcols = c('sex', 'agedays', 'v', 'tbc.sd', 'exclude')]

  # 16.  Exclude measurements for subjects/parameters with only 1 or 2 measurements with exc_*=0. This step uses a variety of criteria,
  #      including the tbc*sd of the other parameter.
  if (!quietly)
    cat(sprintf(
      "[%s] Exclude single measurements and pairs...\n",
      Sys.time()
    ))
  data.df$tbc.other.sd <- swap_parameters(df = data.df)
  data.df[, exclude := (function(df) {
    valid.rows <- which(valid(df))
    # calculate median SD for other parameter (used in steps below)
    median.tbc.other.sd <- median(df$tbc.other.sd)
    # a.  Identify subjects/parameters with only 2 values with exc_*==0, and determine the following:
    #   1.  d_tbc*sd_other=the absolute value difference in tbc*sd
    #   2.	d_agedays_other=the difference in agedays between the 2 values
    #   3.	tbcOsd=the tbc*sd for the other parameter on the same ageday
    # NOTE: this is done as the first step below for efficiency
    #   4.	median_tbcOsd= the median tbc*sd for the other parameter
    #   5.	abs_d_*_O is the absolute value of the difference between tbc*sd and tbcOsd if tbcOsd is not missing;
    #       it is the absolute value of the difference between tbc*sd and median_tbcOsd if tbcOsd is missing but median_tbcOsd is not missing;
    #       and is missing if tbcOsd and median_tbcOsd are both missing
    if (length(valid.rows) == 2) {
      delta.tbc.sd.other <- abs(df$tbc.sd[valid.rows[1]] - df$tbc.sd[valid.rows[2]])
      delta.agedays.other <- abs(df$agedays[valid.rows[1]] - df$agedays[valid.rows[2]])
      abs.delta.other <- with(df, ifelse(
        !is.na(tbc.other.sd),
        abs(tbc.sd - tbc.other.sd),
        ifelse(
          !is.na(median.tbc.other.sd),
          abs(tbc.sd - median.tbc.other.sd),
          NA
        )
      ))

      # b.	For subjects/parameters with 2 values with exc_*==0, replace exc_*=17 if one of the following sets of criteria are met:
      #   1.	If |d_agedays_other|>365.25 and |d_tbc*sd_other|>3; replace exc_*=17 for the value of the pair that has the largest abs_d_*_O. If abs_d_*O is missing, replace exc_*=17 for the value of the pair with the higher |tbc*sd|
      #   2.	If |d_agedays_other|<365.25 and |d_tbc*sd_other|>2; replace exc_*=18 for the value of the pair that has the largest abs_d_*_O. If abs_d_*O is missing, replace exc_*=18 for the value of the pair with the higher |tbc*sd|
      #
      # MODE default
      #   1.  uses current behavior outlined in b, in else statement
      #
      # MODE flag.both
      #   1.  similar to b, except in cases where only HTs values exist or only WTs values exist, both are dropped instead of only 1 (first if)
      #
      #
      # check whether each value in double has other parameters (make sure they don't first)
      if (lt3.exclude.mode == "flag.both" &&
          (is.na(df$tbc.other.sd[valid.rows[1]])) &&
          (is.na(df$tbc.other.sd[valid.rows[2]])) &&
          is.na(abs.delta.other)) {
        # if they don't then do test
        # check default thresholds, if they fail then exclude both regardless of tbc.sd
        if (na_as_false(delta.agedays.other >= 365.25 &
                        delta.tbc.sd.other > 3)) {
          df$exclude[valid.rows[1]] = 'Exclude-Pair-Delta-19'
          df$exclude[valid.rows[2]] = 'Exclude-Pair-Delta-19'
        } else if (na_as_false(delta.agedays.other < 365.25 &
                               delta.tbc.sd.other > 2)) {
          df$exclude[valid.rows[1]] = 'Exclude-Pair-Delta-19'
          df$exclude[valid.rows[2]] = 'Exclude-Pair-Delta-19'
        }

      } else {
        # default method if other mode is not used or if other parameters exist
        worst.row <- order(valid(df),
                           abs(df$tbc.sd),
                           abs(abs.delta.other),
                           decreasing = TRUE)[1]
        if (na_as_false(delta.agedays.other >= 365.25 &
                        delta.tbc.sd.other > 3)) {
          df$exclude[worst.row] <- 'Exclude-Pair-Delta-17'
        } else if (na_as_false(delta.agedays.other < 365.25 &
                               delta.tbc.sd.other > 2)) {
          df$exclude[worst.row] <- 'Exclude-Pair-Delta-18'
        }
      }
    }
    # c.	Identify subjects/parameters with exactly 1 value for which exc_*=0. This will include subjects/parameters for which a value was excluded in step 16b.
    #     Determine tbcOsd and median_tbcOsd as described in step 16a3 and 16a4 above.
    # d.	For subjects/parameters with 1 value for which exc_*=0,  replace exc_*=19 if one of the following sets of criteria are met:
    #   1.	|tbc*sd|>3 & |tbc*sd-tbcOsd|>5 & tbcOsd is not missing
    #   2.	|tbc*sd|>3 & |tbc*sd-median_tbcOsd|>5 & tbcOsd is missing & median_tbcOsd is not missing
    #   3.	|tbc*sd|>5 & tbcOsd is missing & median_tbcOsd is missing
    valid.rows <- valid(df)
    # NOTE: valid.rows is now a boolean.  was a row number in code above
    if (sum(valid.rows) == 1) {
      #cat("param = ", df[valid.rows, param])
      df[valid.rows &
           (
             abs(tbc.sd) > 3 &
               abs(
                 tbc.sd - ifelse(
                   !is.na(tbc.other.sd),
                   tbc.other.sd,
                   median.tbc.other.sd
                 )
               ) > 5 |
               abs(tbc.sd) > 5 &
               is.na(tbc.other.sd) & is.na(median.tbc.other.sd)
           ),
         exclude := 'Exclude-Single-Outlier']
    }

    # ensure factor levels didn't accidentally get mangled
    return(factor(df$exclude, levels = exclude.levels, ordered = TRUE))

  })(copy(.SD)), by = .(subjid, param), .SDcols = c('agedays', 'tbc.sd', 'tbc.other.sd', 'exclude', 'param')] #added param here for debugging, i think it's dropped otherwise

  # 17.  Exclude measurements based on error load for the subject
  # a.	For each subject/parameter determine the following:
  #   1.	tot_exc_*=the total number of values for which exc_* is equal to 4, 5, 6, or 8-19
  #   2.	tot_inc_*=the total number of values for which exc_*=0
  # b.	For subjects/parameters where tot_exc_*>0.5 x tot_inc_*; replace all values where exc_* had been 0 to exc_*=20
  # c.	For subjects/parameters where tot_exc_*>tot_inc_*, replace all values for the OTHER parameter where exc_* had been 0 to exc_*=21
  #
  # CD e-mail 2/10/15: It looks like there are two problems. One is carried forward measurements. The other is that I forgot to include an
  # important part of the rule in the English - tot_exc_* has to be >=2. I'm forwarding an updated English version.
  # Also, because your exclusions are handled a little differently I wanted to specify that unit errors and swaps are not included
  # in tot_exc_* and are included in tot_inc_*, whereas carried forwards and extraneous are not included in either count.
  # NOTE: updated to include optional argument "include.carryforward=TRUE" in the valid() function, and added the
  # constraints "exclude.count.this.parameter >= 2" and "exclude.count.other.parameter >= 2" in the code below

  if (!quietly)
    cat(
      sprintf(
        "[%s] Exclude all measurements if maximum threshold of errors is exceeded...\n",
        Sys.time()
      )
    )
  # NOTE: restrict analysis to non-missing rows
  data.df[exclude != 'Missing', exclude := (function(subj.df) {
    inc.exc <- subj.df[, j = list(exclude.count = sum(
      !valid(
        exclude,
        include.extraneous = TRUE,
        include.carryforward = TRUE
      )
    ),
    include.count = sum(valid(exclude))), keyby = param]
    for (p in unique(subj.df$param)) {
      exclude.count.this.parameter <- inc.exc[param == p, exclude.count]
      exclude.count.other.parameter <- sum(inc.exc[param != p, exclude.count])
      if (exclude.count.this.parameter > error.load.threshold * inc.exc[param ==
                                                                        p, include.count] &
          exclude.count.this.parameter >= error.load.mincount) {
        subj.df[param == p &
                  valid(exclude), exclude := 'Exclude-Too-Many-Errors']
      } else if (exclude.count.other.parameter > sum(inc.exc[param != p, include.count]) &
                 exclude.count.other.parameter >= error.load.mincount) {
        subj.df[param == p &
                  valid(exclude), exclude := 'Exclude-Too-Many-Errors-Other-Parameter']
      }
    }
    # ensure factor levels didn't accidentally get mangled
    return(factor(subj.df$exclude, levels = exclude.levels, ordered =
                    TRUE))
  })(copy(.SD)), by = subjid, .SDcols = c('param', 'exclude')]

  if (!quietly)
    cat(sprintf("[%s] Completed Batch #%s...\n", Sys.time(), data.df$batch[1]))
  if (!quietly & parallel)
    sink()
  return(data.df[j = .(line, exclude, tbc.sd, tbc.other.sd, param)]) #debugging
}

#' Split input data into multiple files
#'
#' \code{splitinput} Splits input based on keepcol specified, yielding csv files each with at least the mininum
#' number of rows that are written and saved separately (except for the last split file written, which may be
#' smaller). Allows splitting input data while ensuring all records for each individual subject will stay together
#' in one file. Pads split filenames with zeros out to five digits for consistency, assuming < 100,000 file count
#' result.
#'
#' @param df data frame to split
#' @param fname new name for each of the split files to start with
#' @param fdir directory to put each of the split files (use "." for  working directory). Must be changed from default (NA), which will trigger error.
#' @param min_nrow minimum number of rows for each split file (default 10000)
#' @param keepcol the column name (default "subjid") to use to keep records with the same values together in the same single split file
#'
#' @return the count number referring to the last split file written
#'
#' @export
#' @importFrom utils tail
#' @examples
#' \donttest{
#' # Run on given data
#' df <- as.data.frame(syngrowth)
#'
#' # Run with all defaults (specifying directory)
#' splitinput(df, fdir = tempdir())
#'
#' # Specifying the name, directory and minimum row size
#' splitinput(df, fname = "syngrowth", fdir = tempdir(), min_nrow = 5000)
#'
#' # Specifying a different subject ID column
#' colnames(df)[colnames(df) == "subjid"] <- "sub_id"
#' splitinput(df, fdir = tempdir(), keepcol = "sub_id")
#' }
splitinput <-
  function(df,
           fname = deparse(substitute(df)),
           fdir = NA,
           min_nrow = 10000,
           keepcol = 'subjid') {
    # first, check if the given directory exists
    if (!is.na(fdir) & fdir != "." & is.character(fdir) & !dir.exists(fdir)){
      stop("invalid directory")
    }

    fname_counter <- 0
    row_count <- 0
    split_df <- data.frame()

    # split the data frame by the grouping that user specifies
    split_sample <- split(df, df[[keepcol]])

    # grab the individual grouping names
    split_sample_names <- names(split_sample)

    for (name in split_sample_names) {
      # append the rows from name, store new total row_count for current split file
      split_df <- rbind(split_df, split_sample[[name]])
      current_nrow <- nrow(split_df)

      # check if updated row count will exceed min row count,
      # if min nrow is exceeded, then write.csv the current split file and clear the split dataframe starter (start from 0)
      if (current_nrow > min_nrow) {
        fname_counter_str <- sprintf("%05d", fname_counter) #pad 0s
        write.csv(
          split_df,
          file = file.path(fdir, paste(fname, fname_counter_str, "csv", sep = ".")),
          row.names = FALSE
        )

        split_df <- data.frame() #reset split_df
        fname_counter <- fname_counter + 1
      } else if (name == tail(split_sample_names, 1)) {
        #for last part, just write
        fname_counter_str <- sprintf("%05d", fname_counter) #pad 0s
        write.csv(
          split_df,
          file = file.path(fdir, paste(fname, fname_counter_str, "csv", sep = ".")),
          row.names = FALSE,
          na = ""
        )
      }
    }

    return(fname_counter)
  }


#' Recode binary sex variable for compatibility
#'
#' \code{recode_sex} recodes a binary sex variable for a given source column in a data frame or data table.
#' Useful in transforming output from growthcleanr::cleangrowth() into a format suitable for growthcleanr::ext_bmiz().
#'
#' @param input_data a data frame or data table to be transformed. Expects a source column containing a binary sex variable.
#' @param sourcecol name of sex descriptor column. Defaults to "sex"
#' @param sourcem variable indicating "male" sex in input data. Defaults to "0"
#' @param sourcef variable indicating "female" sex in input data. Defaults to "1"
#' @param targetcol desired name of recoded sex descriptor column. Defaults to "sex_recoded"
#' @param targetm desired name of recoded sex variable indicating "male" sex in output data. Defaults to 1
#' @param targetf desired name of recoded sex variable indicating "female" sex in output data. Defaults to 2
#'
#' @return Returns a data table with recoded sex variables.
#'
#' @export
#' @examples
#' # Run on given data
#' df <- as.data.frame(syngrowth)
#'
#' # Run with all defaults
#' df_r <- recode_sex(df)
#'
#' # Specify different targets
#' df_rt <- recode_sex(df, targetcol = "sexr", targetm = "Male", targetf = "Female")
#'
#' # Specify different inputs
#' df_ri <- recode_sex(df_rt, sourcecol = "sexr", sourcem = "Male", sourcef = "Female")
recode_sex <- function(input_data,
                       sourcecol = "sex",
                       sourcem = "0",
                       sourcef = "1",
                       targetcol = "sex_recoded",
                       targetm = 1L,
                       targetf = 2L) {
  # cast to DT for faster processing
  input_table <- data.table(input_data)
  #replace targetcol variables with targetm where sourcecol = sourcem
  input_table[input_table[[sourcecol]] == sourcem, targetcol] <-
    targetm
  #replace targetcol variables with targetf where sourcecol = sourcef
  input_table[input_table[[sourcecol]] == sourcef, targetcol] <-
    targetf

  #return table
  return(input_table)
}


#' Transform data in growthcleanr format into wide structure for BMI calculation
#'
#' \code{longwide} transforms data from long to wide format. Ideal for
#' transforming output from growthcleanr::cleangrowth() into a format suitable
#' for growthcleanr::ext_bmiz().
#'
#' @param long_df A data frame to be transformed. Expects columns: id, subjid,
#' sex, agedays, param, measurement, and gcr_result.
#' @param id name of observation ID column
#' @param subjid name of subject ID column
#' @param sex name of sex descriptor column
#' @param agedays name of age (in days) descriptor column
#' @param param name of parameter column to identify each type of measurement
#' @param measurement name of measurement column containing the actual
#' measurement data
#' @param gcr_result name of column of results from growthcleanr::cleangrowth()
#' @param include_all Determines whether the function keeps all exclusion codes.
#' If TRUE, all exclusion types are kept and the inclusion_types argument is
#' ignored. Defaults to FALSE.
#' @param inclusion_types Vector indicating which exclusion codes from the
#' cleaning algorithm should be included in the data, given that include_all is
#' FALSE. For all options, see growthcleanr::cleangrowth(). Defaults to
#' c("Include").
#' @param extra_cols Vector of additional columns to include in the output. If
#' a column C1 differs on agedays matched height and weight values, then include
#' separate ht_C1 and wt_C1 columns as well as a match_C1 column that gives
#' booleans indicating where ht_C1 and wt_C1 are the same. If the agedays
#' matched height and weight columns are identical, then only include a single
#' version of C1. Defaults to empty vector (not keeping any additional columns).
#' @param keep_unmatched_data boolean indicating whether to keep height/weight
#' observations that do not have a matching weight/height on that day
#'
#' @return Returns a data frame transformed from long to wide. Includes only
#' values flagged with indicated inclusion types. Potentially includes
#' additional columns if arguments are passed to extra_cols. For each subject,
#' heights without corresponding weights for a given age (and vice versa) will
#' be dropped unless keep_unmatched_data is set to TRUE.
#'
#' @export
#' @importFrom rlang .data
#' @rawNamespace import(tidyr, except = extract)
#' @rawNamespace import(dplyr, except = c(last, first, summarize, src, between))
#' @examples
#' \donttest{
#' # Run on a small subset of given data
#' df <- as.data.frame(syngrowth)
#' df <- df[df$subjid %in% unique(df[, "subjid"])[1:2], ]
#' df <- cbind(df,
#'             "gcr_result" = cleangrowth(df$subjid,
#'                                        df$param,
#'                                        df$agedays,
#'                                        df$sex,
#'                                        df$measurement))
#' # Convert to wide format
#' wide_df <- longwide(df)
#'
#' # Include all inclusion types
#' wide_df <- longwide(df, include_all = TRUE)
#'
#' # Specify all inclusion codes
#' wide_df <- longwide(df, inclusion_types = c("Include", "Exclude-Carried-Forward"))
#' }
longwide <-
  function(long_df,
           id = "id",
           subjid = "subjid",
           sex = "sex",
           agedays = "agedays",
           param = "param",
           measurement = "measurement",
           gcr_result = "gcr_result",
           include_all = FALSE,
           inclusion_types = c("Include"),
           extra_cols = NULL,
           keep_unmatched_data = FALSE) {

    # check that all needed columns are present, if not, throw an error.
    if(all(c(id, subjid, sex, agedays, param, measurement, gcr_result,
             extra_cols) %in% colnames(long_df))) {
      long_df %>%
        select("id", "subjid", "sex", "agedays", "param", "measurement", all_of(gcr_result),
               all_of(extra_cols)) -> obs_df
    } else {
      # catch error if any variables were not found
      stop("not all specified columns were present")
    }

    # extract values flagged with indicated inclusion types:
    if (include_all == TRUE) {
      obs_df <- obs_df
    } else if (include_all == FALSE) {
      obs_df <- obs_df[obs_df[,gcr_result] %in% inclusion_types,]
    } else{
      stop(paste0("include_all is not a logical of length 1. It is a ",
                  typeof(include_all), " of length ", length(include_all)))
    }

    # calculate age in years
    obs_df$agey <- round(obs_df$agedays / 365.25, 4)

    # calculate age in months
    obs_df$agem <- round((obs_df$agey * 12), 4)

    # recode sex to expected ext_bmiz() format
    obs_df <- recode_sex(
      input_data = obs_df,
      sourcecol = "sex",
      sourcem = "0",
      sourcef = "1",
      targetcol = "sex_recoded",
      targetm = 1L,
      targetf = 2L
    )

    obs_df %>%
      mutate(sex = .data$sex_recoded) %>%
      mutate(param = as.character(.data$param)) %>%
      select("subjid", "id", "agey", "agem", "agedays", "sex", "param", "measurement",
             all_of(extra_cols)) -> clean_df


    # check for unique weight and height ids
    if (any(duplicated(clean_df$id))) {
      stop("duplicate IDs in long_df")
    }

    # separate heights and weights using unique ids
    clean_df %>%
      pivot_wider(names_from = "param", values_from = "measurement") -> param_separated

    # extract heights and weights attached to ids
    param_separated %>%
      filter(!is.na(.data$HEIGHTCM)) %>%
      filter(is.na(.data$WEIGHTKG)) %>% # why this, shouldn't the !is.na height take care of it
      mutate(ht_id = .data$id) %>%
      select(-"id") %>%
      select(-"WEIGHTKG") -> height

    param_separated %>%
      filter(is.na(.data$HEIGHTCM)) %>%
      filter(!is.na(.data$WEIGHTKG)) %>%
      mutate(wt_id = .data$id) %>%
      select(-"id") %>%
      select(-"HEIGHTCM") -> weight

    # join based on subjid, age, and sex, potentially keep unmatched values too
    wide_df <- merge(height,
                     weight,
                     by = c("subjid", "agey", "agem", "agedays", "sex"),
                     all = keep_unmatched_data) %>%
      mutate(wt = .data$WEIGHTKG, ht = .data$HEIGHTCM)


    for(col in extra_cols) {
      ht_vals <- wide_df[,paste0(col, ".x")]
      wt_vals <- wide_df[,paste0(col, ".y")]
      match_vals <- ht_vals == wt_vals
      match_vals[is.na(ht_vals) & is.na(wt_vals)] <- TRUE

      # If the columns match exactly, then just put a single column with that
      # name in and drop the other two.
      if(sum(match_vals, na.rm=TRUE)==length(match_vals)) {
        wide_df[,col] <- ht_vals
        wide_df[,c(paste0(col, ".x"), paste0(col, ".y"))] <- list(NULL)
      } else {
        wide_df[,paste0("match_", col)] <- match_vals
        colnames(wide_df)[colnames(wide_df)==paste0(col, ".x")] <- paste0("ht_",
                                                                          col)
        colnames(wide_df)[colnames(wide_df)==paste0(col, ".y")] <- paste0("wt_",
                                                                          col)
      }
    }

    # Drop WEIGHTKG and HEIGHTCM columns since they are in ht and wt
    wide_df <- wide_df %>% select(-c("HEIGHTCM", "WEIGHTKG"))

    return(wide_df)
}

#' Compute BMI using standard formula
#'
#' \code{simple_bmi} Computes BMI using standard formula. Assumes input compatible with
#' output from longwide().
#'
#' @param wide_df A data frame or data table containing heights and weights in
#' wide format, e.g., after transformation with longwide()
#' @param wtcol name of observation height value column, default 'wt'
#' @param htcol name of subject weight value column, default 'ht'
#'
#' @return Returns a data table with the added column "bmi"
#'
#' @export
#' @import data.table
#' @rawNamespace import(dplyr, except = c(last, first, summarize, src, between))
#' @examples
#' \donttest{
#' # Simple usage
#' # Run on a small subset of given data
#' df <- as.data.frame(syngrowth)
#' df <- df[df$subjid %in% unique(df[, "subjid"])[1:2], ]
#' df <- cbind(df,
#'             "gcr_result" = cleangrowth(df$subjid,
#'                                        df$param,
#'                                        df$agedays,
#'                                        df$sex,
#'                                        df$measurement))
#' # Convert to wide format
#' wide_df <- longwide(df)
#' wide_df_with_bmi <- simple_bmi(wide_df)
#'
#' # Specifying different column names; note that quotes are used
#' colnames(wide_df)[colnames(wide_df) %in% c("wt", "ht")] <-
#'   c("weight", "height")
#' wide_df_with_bmi <- simple_bmi(wide_df, wtcol = "weight", htcol = "height")
#' }
simple_bmi <- function(wide_df, wtcol = "wt", htcol = "ht") {
  # Verify the specified columns are present
  if (!all(c(wtcol, htcol) %in% colnames(wide_df))) {
    stop("Specified column names are not all present")
  }

  # coerce to data table
  if (!is.data.table(wide_df)){
    wide_df <- as.data.table(wide_df)
  }

  # add bmi column
  wide_df[, "bmi" := get(wtcol) / ((get(htcol) * 0.01) ^ 2)]
  return(wide_df)
}

testthat::skip_on_cran()

test_that("splitinput splits files correctly with default values", {
  create_df <- function(num_ids, num_obs) {
    df <- data.frame(matrix(NA, nrow = num_ids * num_obs, ncol = 5))
    colnames(df)[1] <- "subjid"
    df$subjid <- rep(paste0("subj", 1:num_ids), each = num_obs)

    return(df)
  }

  # create input dataframe for default value test
  num_ids <- 1000 * 6
  num_obs <- 7
  df <- create_df(num_ids, num_obs)
  # where to put output
  fcount <- splitinput(df,
                       fname = "df",
                       fdir = tempdir())

  # check that it yielded the correct number of files, with the correct name
  f_log <- grepl("df.*.csv", list.files(tempdir()))
  expect_equal(sum(f_log),
               ceiling(num_ids * num_obs / 10000))

  # check file contents
  sp_list <- lapply(list.files(tempdir())[f_log],
                    function(x) {
                      read.csv(paste0(tempdir(), "/", x))
                    })

  # check that each, except the last, is above the default limit
  # also check that all columns are accounted for
  for (x in sp_list[1:(length(sp_list) - 1)]) {
    expect_gte(nrow(x), 10000)
    expect_equal(ncol(x), 5)
  }

  # check that subjects are not split, between files
  all_subj <- lapply(sp_list, function(x) {
    unique(x$subjid)
  })
  expect_equal(length(Reduce(intersect, all_subj)), 0)

  # remove created csvs
  file.remove(list.files(tempdir(), full.names = T)[f_log])
})

test_that("splitinput splits files correctly with custom values", {
  create_df <- function(num_ids, num_obs) {
    df <- data.frame(matrix(NA, nrow = num_ids * num_obs, ncol = 5))
    colnames(df)[1] <- "subjid"
    df$subjid <- rep(paste0("subj", 1:num_ids), each = num_obs)

    return(df)
  }

  remove_files <- function(f_log) {
    file.remove(list.files(tempdir(), full.names = T)[f_log])
  }

  # create input dataframe
  num_ids <- 1
  num_obs <- 20
  df <- create_df(num_ids, num_obs)
  # run splitinput with new name, less than the default observations
  fcount <- splitinput(df,
                       fname = "onesub",
                       fdir = tempdir())

  # check that it yielded the correct number of files, with the correct name
  f_log <- grepl("onesub.*.csv", list.files(tempdir()))
  expect_equal(sum(f_log),
               ceiling(num_ids * num_obs / 10000))

  # check file contents
  sp_list <- lapply(list.files(tempdir())[f_log],
                    function(x) {
                      read.csv(paste0(tempdir(), "/", x))
                    })

  # check that the only file has all the observations
  expect_equal(nrow(sp_list[[1]]), num_ids * num_obs)
  expect_equal(ncol(sp_list[[1]]), 5)

  # remove created csvs
  remove_files(f_log)

  # try reducing the amount of minimum rows
  fcount <- splitinput(df,
                       min_nrow = 2,
                       fname = "lessrows",
                       fdir = tempdir())

  # check that it did not split the file
  f_log <- grepl("lessrows.*.csv", list.files(tempdir()))
  expect_equal(sum(f_log), 1)

  remove_files(f_log)

  # check that it splits correctly for a given minimum amount of rows
  df <- create_df(2, 10)
  df$X2 <- c(1:(2 * 10)) # creating fake observations
  df <- df[sample(1:nrow(df), nrow(df)), ] # reorder
  fcount <- splitinput(df,
                       min_nrow = 5,
                       fname = "multless",
                       fdir = tempdir())

  # check that it yielded the correct number of files, with the correct name
  f_log <- grepl("multless.*.csv", list.files(tempdir()))
  expect_equal(sum(f_log), 2)

  # check file contents
  sp_list <- lapply(list.files(tempdir())[f_log],
                    function(x) {
                      read.csv(paste0(tempdir(), "/", x))
                    })

  # check that each file is above the limit
  # also check that all columns are accounted for
  for (x in sp_list[1:(length(sp_list))]) {
    expect_gte(nrow(x), 5)
    expect_equal(ncol(x), 5)
  }

  # check that subjects are not split, between files
  all_subj <- lapply(sp_list, function(x) {
    unique(x$subjid)
  })
  expect_equal(length(Reduce(intersect, all_subj)), 0)

  # remove created csvs
  remove_files(f_log)
})

test_that("splitinput throws errors when expected", {
  # run splitinput with several wrong directory names
  expect_error(splitinput(data.frame(), fdir = "hello"))
  expect_error(splitinput(data.frame(), fdir = T))
  expect_error(splitinput(data.frame(), fdir = data.frame()))
})

test_that("recode_sex works as expected with defaults", {
  # create a dataframe according to recode_sex defaults
  num_obs <- 20
  set.seed(7) # for replicability
  df <- data.frame(matrix(NA, nrow = num_obs, ncol = 1))
  colnames(df) <- "sex"
  df$sex <- sample(c("0", "1"), num_obs, replace = T)

  # run recode sex with all defaults
  r_df <- recode_sex(df)

  # check that column names are correct
  expect(all(colnames(r_df) %in% c("sex", "sex_recoded")),
         "column names incorrect")

  # all observations are accounted for
  expect_equal(df$sex, r_df$sex)

  # check that sex was recoded according to specifications
  expect_equal(r_df$sex_recoded[r_df$sex == "0"], rep(1, sum(df$sex == "0")))
  expect_equal(r_df$sex_recoded[r_df$sex == "1"], rep(2, sum(df$sex == "1")))
})

test_that("recode_sex works as expected with custom inputs", {
  # create a dataframe with customization
  num_obs <- 31
  set.seed(7) # for replicability
  df <- data.frame(matrix(NA, nrow = num_obs, ncol = 1))
  colnames(df) <- "sex_type"
  df$sex_type <- sample(c("M", "F"), num_obs, replace = T)

  # run recode sex with all defaults
  r_df <- recode_sex(
    df,
    sourcecol = "sex_type",
    sourcem = "M",
    sourcef = "F",
    targetcol = "sex_r",
    targetm = "m",
    targetf = "f"
  )

  # check that column names are correct
  expect(all(colnames(r_df) %in% c("sex_type", "sex_r")),
         "column names incorrect")

  # all observations are accounted for
  expect_equal(df$sex_type, r_df$sex_type)

  # check that sex was recoded according to specifications
  expect_equal(r_df$sex_r[r_df$sex_type == "M"],
               rep("m", sum(df$sex_type == "M")))
  expect_equal(r_df$sex_r[r_df$sex_type == "F"],
               rep("f", sum(df$sex_type == "F")))
})

test_that("longwide works as expected with default values", {
  # use synthetic data, running cleaning on a subset
  data("syngrowth")
  sub_syn <-
    syngrowth[syngrowth$subjid %in% unique(syngrowth$subjid)[1:100], ]
  sub_syn <- cbind(
    sub_syn,
    "gcr_result" = cleangrowth(
      subjid = sub_syn$subjid,
      param = sub_syn$param,
      agedays = sub_syn$agedays,
      sex = sub_syn$sex,
      measurement = sub_syn$measurement
    )
  )

  # run longwide on changed data
  wide_syn <- longwide(sub_syn)

  # check that it has the correct amount of columns
  expect_equal(ncol(wide_syn), 9)

  # check that all subjects are accounted for
  ss <- unique(sub_syn$subjid)
  ws <- unique(wide_syn$subjid)

  expect_equal(sort(ss), sort(ws))

  # check that all subjects' measurements with at least two occurrences appear;
  # works since cleangrowth() function will only identify at most one height and
  # one weight to include for given ageday
  all_obs <- sapply(unique(sub_syn$subjid), function(i) {
    sub_group <- sub_syn[sub_syn$gcr_result == "Include",]
    sum(table(sub_group$agedays[sub_group$subjid == i]) >= 2)
  })

  # it should be that the number of occurrences for each subject in "all_obs" is
  # the same as in "wide_syn", so check that but make sure sorted by subjid
  ws_subj_counts <- table(wide_syn$subjid)
  ws_subj_counts <- setNames(as.vector(ws_subj_counts), names(ws_subj_counts))

  expect_equal(all_obs[sort(names(all_obs))],
               ws_subj_counts[sort(names(ws_subj_counts))])

  # get all the observation ID's in the wide data
  obs_ids <- c(wide_syn$wt_id, wide_syn$ht_id)

  # here specified inclusion type is just "include"
  # check that it includes specified inclusion types
  expect(
    all(sub_syn$gcr_result[sub_syn$id %in% obs_ids] == "Include"),
    "longwide() includes inclusion values that were not specified"
  )

  # check that all sexes have been correctly recoded by picking out all the
  # unique subject ID and sex pairs from each data set, sorting them by ID, and
  # looking if the sex code in the new set is 1 plus the old
  orig_sex <- sub_syn[!duplicated(sub_syn$subjid), c("subjid", "sex")]
  orig_sex <- orig_sex[order(orig_sex$subjid),]

  aft_sex <- wide_syn[!duplicated(wide_syn$subjid), c("subjid", "sex")]
  aft_sex <- aft_sex[order(aft_sex$subjid),]

  expect_equal(orig_sex$sex+1, aft_sex$sex)

  # spot check that data is correct
  set.seed(7)
  # check height ids
  ht_sub <-
    sub_syn[sub_syn$param == "HEIGHTCM" & sub_syn$id %in% obs_ids, ]
  for (x in ht_sub$id[sample(1:nrow(ht_sub), 20)]) {
    w_idx <- wide_syn$ht_id == x
    ht_idx <- ht_sub$id == x

    # check ages
    expect_equal(wide_syn$agey[w_idx], round(ht_sub$agedays[ht_idx] / 365.25), 4)
    expect_equal(wide_syn$agem[w_idx],
                 round(round(ht_sub$agedays[ht_idx] / 365.25), 4) * 12, 4)
    expect_equal(wide_syn$agedays[w_idx], ht_sub$agedays[ht_idx])

    # check height
    expect_equal(wide_syn$ht[w_idx], ht_sub$measurement[ht_idx])

  }

  # check weight ids
  wt_sub <-
    sub_syn[sub_syn$param == "WEIGHTKG" & sub_syn$id %in% obs_ids, ]
  for (x in wt_sub$id[sample(1:nrow(wt_sub), 20)]) {
    w_idx <- wide_syn$wt_id == x
    wt_idx <- wt_sub$id == x

    # check ages
    expect_equal(wide_syn$agey[w_idx], round(wt_sub$agedays[wt_idx] / 365.25), 4)
    expect_equal(wide_syn$agem[w_idx],
                 round(round(wt_sub$agedays[wt_idx] / 365.25), 4) * 12, 4)
    expect_equal(wide_syn$agedays[w_idx], wt_sub$agedays[wt_idx])

    # check weight
    expect_equal(wide_syn$wt[w_idx], wt_sub$measurement[wt_idx])
  }
})

test_that("longwide works as expected with extra columns", {
  # use synthetic data, running cleaning on a subset
  data("syngrowth")
  sub_syn <-
    syngrowth[syngrowth$subjid %in% unique(syngrowth$subjid)[1:100], ]
  sub_syn <- cbind(
    sub_syn,
    "gcr_result" = cleangrowth(
      subjid = sub_syn$subjid,
      param = sub_syn$param,
      agedays = sub_syn$agedays,
      sex = sub_syn$sex,
      measurement = sub_syn$measurement
    )
  )

  # add extra columns to sub_syn; one where all values match, and one where they
  # don't
  set.seed(7)
  sub_syn$r1 <- sample(c("A", "B", "C", "D"), size = nrow(sub_syn),
                       replace = TRUE)
  sub_syn$r2 <- "E"

  # run longwide on changed data
  wide_syn <- longwide(sub_syn, extra_cols = c("r1", "r2"))

  # check that it has the correct amount of columns
  expect_equal(ncol(wide_syn), ncol(sub_syn)+4)

  # check that additional columns are correctly named
  expect_equal(c("ht_r1", "wt_r1", "match_r1", "r2") %in%
                    colnames(wide_syn), rep(TRUE, 4))


  # run through same tests as default values

  # check that all subjects are accounted for
  ss <- unique(sub_syn$subjid)
  ws <- unique(wide_syn$subjid)

  expect_equal(sort(ss), sort(ws))

  # check that all subjects' measurements with at least two occurrences appear;
  # works since cleangrowth() function will only identify at most one height and
  # one weight to include for given ageday
  all_obs <- sapply(unique(sub_syn$subjid), function(i) {
    sub_group <- sub_syn[sub_syn$gcr_result == "Include",]
    sum(table(sub_group$agedays[sub_group$subjid == i]) >= 2)
  })

  # it should be that the number of occurrences for each subject in "all_obs" is
  # the same as in "wide_syn", so check that but make sure sorted by subjid
  ws_subj_counts <- table(wide_syn$subjid)
  ws_subj_counts <- setNames(as.vector(ws_subj_counts), names(ws_subj_counts))

  expect_equal(all_obs[sort(names(all_obs))],
               ws_subj_counts[sort(names(ws_subj_counts))])

  # get all the observation ID's in the wide data
  obs_ids <- c(wide_syn$wt_id, wide_syn$ht_id)

  # here specified inclusion type is just "include"
  # check that it includes specified inclusion types
  expect(
    all(sub_syn$gcr_result[sub_syn$id %in% obs_ids] == "Include"),
    "longwide() includes inclusion values that were not specified"
  )

  # check that all sexes have been correctly recoded
  orig_sex <- sub_syn[!duplicated(sub_syn$subjid), c("subjid", "sex")]
  orig_sex <- orig_sex[order(orig_sex$subjid),]

  aft_sex <- wide_syn[!duplicated(wide_syn$subjid), c("subjid", "sex")]
  aft_sex <- aft_sex[order(aft_sex$subjid),]

  expect_equal(orig_sex$sex+1, aft_sex$sex)

  # spot check that data is correct; add in checks for additional columns
  set.seed(7)
  # check height ids
  ht_sub <-
    sub_syn[sub_syn$param == "HEIGHTCM" & sub_syn$id %in% obs_ids, ]
  for (x in ht_sub$id[sample(1:nrow(ht_sub), 20)]) {
    w_idx <- wide_syn$ht_id == x
    ht_idx <- ht_sub$id == x

    # check ages
    expect_equal(wide_syn$agey[w_idx], round(ht_sub$agedays[ht_idx] / 365.25), 4)
    expect_equal(wide_syn$agem[w_idx],
                 round(round(ht_sub$agedays[ht_idx] / 365.25), 4) * 12, 4)
    expect_equal(wide_syn$agedays[w_idx], ht_sub$agedays[ht_idx])

    # check height
    expect_equal(wide_syn$ht[w_idx], ht_sub$measurement[ht_idx])

    # check extra columns
    expect_equal(wide_syn$ht_r1[w_idx], ht_sub$r1[ht_idx])
    expect_equal(wide_syn$r2[w_idx], ht_sub$r2[ht_idx])
  }

  # check weight ids
  wt_sub <-
    sub_syn[sub_syn$param == "WEIGHTKG" & sub_syn$id %in% obs_ids, ]
  for (x in wt_sub$id[sample(1:nrow(wt_sub), 20)]) {
    w_idx <- wide_syn$wt_id == x
    wt_idx <- wt_sub$id == x

    # check ages
    expect_equal(wide_syn$agey[w_idx], round(wt_sub$agedays[wt_idx] / 365.25), 4)
    expect_equal(wide_syn$agem[w_idx],
                 round(round(wt_sub$agedays[wt_idx] / 365.25), 4) * 12, 4)
    expect_equal(wide_syn$agedays[w_idx], wt_sub$agedays[wt_idx])

    # check weight
    expect_equal(wide_syn$wt[w_idx], wt_sub$measurement[wt_idx])

    # check extra columns
    expect_equal(wide_syn$wt_r1[w_idx], wt_sub$r1[wt_idx])
    expect_equal(wide_syn$r2[w_idx], wt_sub$r2[wt_idx])
  }
})

test_that("longwide works as expected when not dropping unmatched values", {
  # use synthetic data, running cleaning on a subset
  data("syngrowth")
  sub_syn <-
    syngrowth[syngrowth$subjid %in% unique(syngrowth$subjid)[1:100], ]
  sub_syn <- cbind(
    sub_syn,
    "gcr_result" = cleangrowth(
      subjid = sub_syn$subjid,
      param = sub_syn$param,
      agedays = sub_syn$agedays,
      sex = sub_syn$sex,
      measurement = sub_syn$measurement
    )
  )

  # to improve coverage add extra columns to sub_syn; one where all values
  # match, and one where they don't
  sub_syn$r1 <- sample(c("A", "B", "C", "D"), size = nrow(sub_syn),
                       replace = TRUE)
  sub_syn$r2 <- "E"

  # run longwide on changed data
  wide_syn <- longwide(sub_syn, extra_cols = c("r1", "r2"),
                       keep_unmatched_data = TRUE)

  # check for correct number of columns and correct naming; here r2 will not
  # be all matches since there are unmatched heights/weights
  expect_equal(ncol(wide_syn), ncol(sub_syn)+6)

  expect_equal(c("ht_r1", "wt_r1", "match_r1", "ht_r2", "wt_r2", "match_r2")
               %in% colnames(wide_syn), rep(TRUE, 6))

  # check that all measurements are in the data set by making sure all
  # observation ID's appear
  obs_ids <- c(wide_syn$wt_id, wide_syn$ht_id)

  # there will be NA's for the unmatched ID's, drop these to compare the lists
  obs_ids <- obs_ids[!is.na(obs_ids)]

  # pull the ID's from the long form data set and make sure it is the same list
  # as in the wide form (need to sort first)
  ss_ids <- sub_syn$id[sub_syn$gcr_result=="Include"]

  expect_equal(sort(obs_ids), sort(ss_ids))
})

test_that("longwide works as expected with other exclusion codes", {
  # just checking the custom-ness, so use a smaller subset for speed
  # use synthetic data, running cleaning on a subset
  data("syngrowth")
  sub_syn <-
    syngrowth[syngrowth$subjid %in% unique(syngrowth$subjid)[1:20], ]
  sub_syn <- cbind(
    sub_syn,
    "cv" = cleangrowth(
      subjid = sub_syn$subjid,
      param = sub_syn$param,
      agedays = sub_syn$agedays,
      sex = sub_syn$sex,
      measurement = sub_syn$measurement
    )
  )

  # run longwide on changed data with all exclusion types included
  wide_syn <- longwide(sub_syn,
                       gcr_result = "cv",
                       include_all = TRUE)

  # check that it has the correct amount of columns
  expect_equal(ncol(wide_syn), 9)

  # check that all subjects are accounted for
  expect(all(unique(sub_syn$subjid) %in% unique(wide_syn$subjid)),
         "not all subjects appear in wide format")

  # quick test that a few particular height/weight ID's with an exclusion codes
  # besides "Include" appear in "wide_syn"
  expect(3 %in% wide_syn$ht_id, "missing height with ID 3 that should be
         included")
  expect(16 %in% wide_syn$wt_id, "missing height with ID 16 that should be
         included")
  expect(104 %in% wide_syn$ht_id, "missing height with ID 104 that should be
         included")
  expect(109 %in% wide_syn$wt_id, "missing weight with ID 109 that should be
         included")

  # run longwide on changed data with some exclusion types included
  inc_types <- c("Include",
                 "Exclude-Carried-Forward",
                 "Exclude-Extraneous-Same-Day")
  wide_syn <- longwide(sub_syn,
                       gcr_result = "cv",
                       inclusion_types = inc_types)

  # check that it has the correct amount of columns
  expect_equal(ncol(wide_syn), 9)

  # check that it includes specified inclusion types
  obs_ids <- c(wide_syn$wt_id, wide_syn$ht_id)
  expect(
    all(sub_syn$gcr_result[sub_syn$id %in% obs_ids] %in% inc_types),
    "longwide() includes inclusion values that were not specified"
  )
})

test_that("longwide throws errors correctly", {
  # use synthetic data, running cleaning on a very small subset
  data("syngrowth")
  sub_syn <-
    syngrowth[syngrowth$subjid %in% unique(syngrowth$subjid)[1:5], ]
  sub_syn <- cbind(
    sub_syn,
    "gcr_result" = cleangrowth(
      subjid = sub_syn$subjid,
      param = sub_syn$param,
      agedays = sub_syn$agedays,
      sex = sub_syn$sex,
      measurement = sub_syn$measurement
    )
  )

  # test with deleting a necessary column
  expect_error(longwide(sub_syn[, -2]))

  # test include_all not being correct
  expect_error(longwide(sub_syn, include_all = "hello"))

  # test duplicated ids
  sub_syn$id <- 1
  expect_error(longwide(sub_syn))
})

test_that("simple_bmi works as expected", {
  data("syngrowth")

  # Similar strategy as for longwide, create subset for speed
  sub_syn <-
    syngrowth[syngrowth$subjid %in% unique(syngrowth$subjid)[101:200],]
  sub_syn <- cbind(
    sub_syn,
    "cv" = cleangrowth(
      subjid = sub_syn$subjid,
      param = sub_syn$param,
      agedays = sub_syn$agedays,
      sex = sub_syn$sex,
      measurement = sub_syn$measurement
    )
  )

  wide_syn <-
    longwide(sub_syn, gcr_result = "cv", include_all = TRUE)
  bmi_syn <- simple_bmi(wide_syn)
  expect_equal(TRUE, "wt" %in% names(bmi_syn))
  expect_equal(bmi_syn$bmi,
               bmi_syn$wt / ((bmi_syn$ht * .01) ^ 2))

  # Verify that invalid column names throw an error
  expect_error(simple_bmi(wide_syn, ht = "invalid_column"))
  expect_error(simple_bmi(wide_syn, wt = "invalid_wt_col", ht = "invalid_ht_col"))
})

# Main growthcleanr algorithm and other exported functions
# Main growthcleanr algorithm (cleangrowth): adult and pediatric are in *_clean.R
#(main bulk of algorithm) and *_support.R (supporting functions)

#' Clean growth measurements
#'
#' @param subjid Vector of unique identifiers for each subject in the database.
#' @param param Vector identifying each measurement, may be 'WEIGHTKG', 'WEIGHTLBS', 'HEIGHTCM', 'HEIGHTIN', 'LENGTHCM', or 'HEADCM'.
#'   'HEIGHTCM'/'HEIGHTIN' vs. 'LENGTHCM' only affects z-score calculations between ages 24 to 35 months (730 to 1095 days).
#'   All linear measurements below 731 days of life (age 0-23 months) are interpreted as supine length, and
#'   all linear measurements above 1095 days of life (age 36+ months) are interpreted as standing height.
#'   Note: at the moment, all LENGTHCM will be converted to HEIGHTCM. In the future, the algorithm will be updated to consider this difference.
#'   Additionally, imperial 'HEIGHTIN' and 'WEIGHTLBS' measurements are converted to
#'   metric during algorithm calculations.
#' @param agedays Numeric vector containing the age in days at each measurement.
#' @param sex Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F', 'f' or 1 for females.
#' @param measurement Numeric vector containing the actual measurement data.  Weight must be in
#'   kilograms (kg), and linear measurements (height vs. length) in centimeters (cm).
#' @param recover.unit.error Indicates whether the cleaning algorithm should
#' attempt to identify unit errors (I.e. inches vs. cm, lbs vs. kg). If unit
#' errors are identified, the value will be corrected and retained within the
#' cleaning algorithm as a valid measurement.  Defaults to FALSE.
#' @param sd.extreme Measurements more than sd.extreme standard deviations from
#' the mean (either above or below) will be flagged as invalid. Defaults to 25.
#' @param z.extreme Measurements with an absolute z-score greater than
#' z.extreme will be flagged as invalid. Defaults to 25.
#' @param lt3.exclude.mode Determines type of exclusion procedure to use for 1 or 2 measurements of one type without
#' matching same ageday measurements for the other parameter. Options include "default" (standard growthcleanr approach),
#' and "flag.both" (in case of two measurements of one type without matching values for the other parameter, flag both
#' for exclusion if beyond threshold)
#' @param height.tolerance.cm maximum decrease in height tolerated for sequential measurements
#' @param error.load.mincount minimum count of exclusions on parameter before
#' considering excluding all measurements. Defaults to 2.
#' @param error.load.threshold threshold of percentage of excluded measurement count to included measurement
#' count that must be exceeded before excluding all measurements of either parameter. Defaults to 0.5.
#' @param sd.recenter specifies how to recenter medians. May be a data frame or
#' table w/median SD-scores per day of life by gender and parameter, or "NHANES"
#' or "derive" as a character vector.
#' \itemize{
#'   \item If `sd.recenter` is specified as a data set, use the data set
#'   \item If `sd.recenter` is specified as "`nhanes`", use NHANES reference medians
#'   \item If `sd.recenter` is specified as "`derive`", derive from input
#'   \item If `sd.recenter` is not specified or `NA`:
#'     \itemize{
#'       \item If the input set has at least 5,000 observations, derive medians from input
#'       \item If the input set has fewer than 5,000 observations, use NHANES
#'     }
#' }
#'
#' If specifying a data set, columns must include param, sex, agedays, and sd.median
#' (referred to elsewhere as "modified Z-score"), and those medians will be used
#' for recentering. A summary of how the NHANES reference medians were derived is
#' available in README.md. Defaults to NA.
#' @param sdmedian.filename Name of file to save sd.median data calculated on the input dataset to as CSV.
#' Defaults to "", for which this data will not be saved. Use for extracting medians for parallel processing
#' scenarios other than the built-in parallel option.
#' @param sdrecentered.filename Name of file to save re-centered data to as CSV. Defaults to "", for which this
#' data will not be saved. Useful for post-processing and debugging.
#' @param include.carryforward Determines whether Carry-Forward values are kept in the output. Defaults to False.
#' @param ewma.exp Exponent to use for weighting measurements in the
#' exponentially weighted moving average calculations. Defaults to -1.5.
#' This exponent should be negative in order to weight growth measurements
#' closer to the measurement being evaluated more strongly. Exponents that are
#' further from zero (e.g. -3) will increase the relative influence of
#' measurements close in time to the measurement being evaluated compared to
#' using the default exponent.
#' @param ref.data.path Path to reference data. If not supplied, the year 2000
#' Centers for Disease Control (CDC) reference data will be used.
#' @param log.path Path to log file output when running in parallel (non-quiet mode). Default is NA. A new
#' directory will be created if necessary. Set to NA to disable log files.
#' @param parallel Determines if function runs in parallel.  Defaults to FALSE.
#' @param num.batches Specify the number of batches to run in parallel. Only
#' applies if parallel is set to TRUE. Defaults to the number of workers
#' returned by the getDoParWorkers function in the foreach package.
#' @param quietly Determines if function messages are to be displayed and if log files (parallel only) are to be generated.
#' Defaults to TRUE
#' @param adult_cutpoint Number between 18 and 20, describing ages when the
#'  pediatric algorithm should not be applied (< adult_cutpoint), and the adult
#'   algorithm should apply (>= adult_cutpoint). Numbers outside this range will be
#'   changed to the closest number within the range. Defaults to 20.
#' @param weight_cap Positive number, describing a weight cap in kg (rounded to the
#' nearest .1, +/- .1) within the adult dataset. If there is no weight cap, set
#'  to Inf. Defaults to Inf.
#' @param adult_columns_filename Name of file to save original adult data, with additional output columns to
#' as CSV. Defaults to "", for which this data will not be saved. Useful
#' for post-analysis. For more information on this output, please see README.
#' @param prelim_infants TRUE/FALSE. Run the in-development release of the infants algorithm (expands pediatric algorithm to improve performance for children 0 – 2 years). Not recommended for use in research. For more information regarding the logic of the algorithm, see the vignette 'Preliminary Infants Algorithm.' Defaults to FALSE.
#'
#' @return Vector of exclusion codes for each of the input measurements.
#'
#'    Possible values for each code are:
#'
#' * 'Include', 'Unit-Error-High', 'Unit-Error-Low', 'Swapped-Measurements', 'Missing',
#' *  'Exclude-Carried-Forward', 'Exclude-SD-Cutoff', 'Exclude-EWMA-Extreme', 'Exclude-EWMA-Extreme-Pair',
#' *  'Exclude-Extraneous-Same-Day',
#' *  'Exclude-EWMA-8', 'Exclude-EWMA-9', 'Exclude-EWMA-10', 'Exclude-EWMA-11', 'Exclude-EWMA-12', 'Exclude-EWMA-13', 'Exclude-EWMA-14',
#' *  'Exclude-Min-Height-Change', 'Exclude-Max-Height-Change',
#' *  'Exclude-Pair-Delta-17', 'Exclude-Pair-Delta-18', 'Exclude-Pair-Delta-19',
#' *  'Exclude-Single-Outlier', 'Exclude-Too-Many-Errors', 'Exclude-Too-Many-Errors-Other-Parameter'
#' @md
#'
#' @export
#' @import data.table
#' @rawNamespace import(plyr, except = c(failwith, id, summarize, count, desc, mutate, arrange, rename, is.discrete, summarise, summarize))
#' @import foreach
#' @import doParallel
#' @import parallel
#' @importFrom utils write.csv
#' @rawNamespace import(R.utils, except = c(extract))
#' @examples
#' \donttest{
#' # Run calculation using a small subset of given data
#' df_stats <- as.data.frame(syngrowth)
#' df_stats <- df_stats[df_stats$subjid %in% unique(df_stats[, "subjid"])[1:5], ]
#'
#' clean_stats <-cleangrowth(subjid = df_stats$subjid,
#'                          param = df_stats$param,
#'                          agedays = df_stats$agedays,
#'                          sex = df_stats$sex,
#'                          measurement = df_stats$measurement)
#'
#' # Once processed you can filter data based on result value
#' df_stats <- cbind(df_stats, "clean_result" = clean_stats)
#' clean_df_stats <- df_stats[df_stats$clean_result == "Include",]
#'
#' # Parallel processing: run using 2 cores and batches
#' clean_stats <- cleangrowth(subjid = df_stats$subjid,
#'                            param = df_stats$param,
#'                            agedays = df_stats$agedays,
#'                            sex = df_stats$sex,
#'                            measurement = df_stats$measurement,
#'                            parallel = TRUE,
#'                            num.batches = 2)
#' }
cleangrowth <- function(subjid,
                        param,
                        agedays,
                        sex,
                        measurement,
                        recover.unit.error = FALSE,
                        sd.extreme = 25,
                        z.extreme = 25,
                        lt3.exclude.mode = "default",
                        height.tolerance.cm = 2.5,
                        error.load.mincount = 2,
                        error.load.threshold = 0.5,
                        sd.recenter = NA,
                        sdmedian.filename = "",
                        sdrecentered.filename = "",
                        include.carryforward = FALSE,
                        ewma.exp = -1.5,
                        ref.data.path = "",
                        log.path = NA,
                        parallel = FALSE,
                        num.batches = NA,
                        quietly = TRUE,
                        adult_cutpoint = 20,
                        weight_cap = Inf,
                        adult_columns_filename = "",
                        prelim_infants = FALSE) {
  # avoid "no visible binding" warnings
  N <- age_years <- batch <- exclude <- index <- line <- NULL
  newbatch <- sd.median <- sd.orig <- tanner.months <- tbc.sd <- NULL
  v <- v_adult <- whoagegrp.ht <- whoagegrp_ht <- z.orig <- NULL
  z.orig_cdc <- z.orig_who <- sd.orig_cdc <- sd.orig_who <- NULL
  id <- result <- NULL

  sd.orig_uncorr <- agemonths <- intwt <- fengadays <- pmagedays <- cagedays <-
    unmod_zscore <- fen_wt_m <- fen_wt_l <- fen_wt_s <- cwho_cv <- ccdc_cv <-
    sd.c_cdc <- sd.c_who <- sd.c <- sd.corr <- seq_win <- sd.corr_abssumdiff <-
    sd.orig_abssumdiff <- orig_colnames <- ctbc.sd <- sum_sde <- no_sde <-
    sum_val <- no_dup_val <- no_outliers <- no_bigdiff <- nottoofar <- nnte <-
    nnte_full <- NULL

  # preprocessing ----

  # organize data into a dataframe along with a line "index" so the original data order can be recovered
  data.all.ages <- data.table(
    line = seq_along(measurement),
    subjid = as.factor(subjid),
    param,
    agedays = as.integer(agedays),
    v = ifelse(measurement == 0, NaN, measurement),
    v_adult = measurement,
    sex = as.integer(ifelse(
      sex %in% c(0, 'm', 'M'), 0, ifelse(sex %in% c(1, 'f', 'F'), 1, NA)
    ))
  )

  # quality checks
  if (!is.numeric(adult_cutpoint)){
    stop("adult_cutpoint not numeric. Please enter a number between 18 and 20.")
  }
  if (!is.numeric(weight_cap) | weight_cap < 0){
    stop("weight_cap not numeric. Please enter a positive number.")
  }
  if (any(!param %in% c("LENGTHCM", "HEIGHTCM", "WEIGHTKG", "HEIGHIN",
                        "WEIGHTLBS", "HEADCM"))){
    cat(sprintf("[%s] Parameters included that do not match 'param' specifications. Marking as missing...\n", Sys.time()))
    data.all.ages <-
      data.all.ages[
        !param %in% c("LENGTHCM", "HEIGHTCM", "WEIGHTKG", "HEIGHIN",
                      "WEIGHTLBS", "HEADCM"), v := NA]
    data.all.ages <-
      data.all.ages[
        !param %in% c("LENGTHCM", "HEIGHTCM", "WEIGHTKG", "HEIGHIN",
                      "WEIGHTLBS", "HEADCM"), v_adult := NA]
  }

  # rate limit cutpoint -- min 18, max 20
  cutpoint_update <-
    if (adult_cutpoint < 18){
      18
    } else if (adult_cutpoint > 20){
      20
    } else {
      adult_cutpoint
    }

  # split by cutpoint
  # for ease, data.all will refer to pediatric data; data.adult will refer to
  # adult data -- copy to make sure they're separate
  data.all <- copy(data.all.ages[agedays < adult_cutpoint*365.25,])
  data.adult <- copy(data.all.ages[agedays >= adult_cutpoint*365.25,])

  # TODO: ADD PARALLEL FOR ADULTS

  # constants for pediatric
  # enumerate the different exclusion levels
  if (prelim_infants){
    # different for infants
    exclude.levels.peds <- c(
      'Include',
      'Unit-Error-High',
      'Unit-Error-Low',
      'Unit-Error-Possible',
      'Swapped-Measurements',
      'Exclude',
      'Missing',
      'Not cleaned',
      'Exclude-Temporary-Extraneous-Same-Day',
      'Exclude-Carried-Forward',
      # added CF exclusions
      "Exclude-1-CF-deltaZ-<0.05",
      "Exclude-1-CF-deltaZ-<0.1-wholehalfimp",
      "Exclude-Teen-2-plus-CF-deltaZ-<0.05",
      "Exclude-Teen-2-plus-CF-deltaZ-<0.1-wholehalfimp",
      'Exclude-EWMA-Extreme',
      'Exclude-EWMA-Extreme-Pair',
      'Exclude-SDE-Identical',
      'Exclude-SDE-All-Exclude',
      'Exclude-SDE-All-Extreme',
      'Exclude-SDE-EWMA',
      'Exclude-SDE-One-Day',
      "Exclude-EWMA2-middle",
      "Exclude-EWMA2-birth-WT",
      "Exclude-EWMA2-birth-WT-ext",
      "Exclude-EWMA2-first",
      "Exclude-EWMA2-first-ext",
      "Exclude-EWMA2-last",
      "Exclude-EWMA2-last-high",
      "Exclude-EWMA2-last-ext",
      "Exclude-EWMA2-last-ext-high",
      "Exclude-EWMA2-birth-HT-HC",
      "Exclude-EWMA2-birth-HT-HC-ext",
      "Exclude-Min-diff",
      "Exclude-Max-diff",
      "Exclude-2-meas->1-year",
      "Exclude-2-meas-<1-year",
      "Exclude-1-meas",
      "Exclude-Error-load",

      # old

      'Exclude-Extraneous-Same-Day',
      'Exclude-SD-Cutoff',
      'Exclude-EWMA-8',
      'Exclude-EWMA-9',
      'Exclude-EWMA-10',
      'Exclude-EWMA-11',
      'Exclude-EWMA-12',
      'Exclude-EWMA-13',
      'Exclude-EWMA-14',
      'Exclude-Min-Height-Change',
      'Exclude-Max-Height-Change',
      'Exclude-Pair-Delta-17',
      'Exclude-Pair-Delta-18',
      'Exclude-Pair-Delta-19',
      'Exclude-Single-Outlier',
      'Exclude-Too-Many-Errors',
      'Exclude-Too-Many-Errors-Other-Parameter',

      #new
      "Exclude-Absolute-BIV",
      "Exclude-Standardized-BIV",
      "Exclude-Evil-Twins",
      "Exclude-EWMA1-Extreme"
    )
  } else {
    exclude.levels.peds <- c(
      'Include',
      'Unit-Error-High',
      'Unit-Error-Low',
      'Unit-Error-Possible',
      'Swapped-Measurements',
      'Exclude',
      'Missing',
      'Exclude-Temporary-Extraneous-Same-Day',
      'Exclude-Carried-Forward',
      'Exclude-SD-Cutoff',
      'Exclude-EWMA-Extreme',
      'Exclude-EWMA-Extreme-Pair',
      'Exclude-Extraneous-Same-Day',
      'Exclude-EWMA-8',
      'Exclude-EWMA-9',
      'Exclude-EWMA-10',
      'Exclude-EWMA-11',
      'Exclude-EWMA-12',
      'Exclude-EWMA-13',
      'Exclude-EWMA-14',
      'Exclude-Min-Height-Change',
      'Exclude-Max-Height-Change',
      'Exclude-Pair-Delta-17',
      'Exclude-Pair-Delta-18',
      'Exclude-Pair-Delta-19',
      'Exclude-Single-Outlier',
      'Exclude-Too-Many-Errors',
      'Exclude-Too-Many-Errors-Other-Parameter'
    )
  }

  exclude.levels.adult <- c(
    "Include",
    "Exclude-Adult-BIV",
    "Exclude-Adult-Hundreds",
    "Exclude-Adult-Hundreds-RV",
    "Exclude-Adult-Unit-Errors",
    "Exclude-Adult-Unit-Errors-RV",
    "Exclude-Adult-Transpositions",
    "Exclude-Adult-Transpositions-RV",
    "Exclude-Adult-Weight-Cap-Identical",
    "Exclude-Adult-Weight-Cap",
    "Exclude-Adult-Swapped-Measurements",
    "Exclude-Adult-Identical-Same-Day",
    "Exclude-Adult-Extraneous-Same-Day",
    "Exclude-Adult-Distinct-Pairs",
    "Exclude-Adult-Distinct-3-Or-More",
    "Exclude-Adult-EWMA-Extreme",
    "Exclude-Adult-EWMA-Extreme-RV",
    "Exclude-Adult-Distinct-Ordered-Pairs",
    "Exclude-Adult-EWMA-Moderate",
    "Exclude-Adult-Possibly-Impacted-By-Weight-Cap",
    "Exclude-Adult-Distinct-Single",
    "Exclude-Adult-Too-Many-Errors"
  )

  exclude.levels <- base::union(exclude.levels.peds, exclude.levels.adult)

  # if there's no pediatric data, no need to go through this rigamarole
  if (nrow(data.all) > 0){

    # pediatric: height velocity calculations and preprocessing ----

    # for pediatric data, convert in and lbs to cm and kg (adult is done within algo)
    data.all[param == "HEIGHTIN", v := v*2.54]
    data.all[param == "HEIGHTIN", param := "HEIGHTCM"]
    data.all[param == "WEIGHTLBS", v := v/2.2046226]
    data.all[param == "WEIGHTLBS", param := "WEIGHTKG"]

    # if parallel processing is desired, load additional modules
    if (parallel) {
      if (is.na(num.batches)) {
        num.batches <- getDoParWorkers()
      }
      if (prelim_infants){
        # variables needed for parallel workers
        var_for_par <- c("temporary_extraneous", "valid", "swap_parameters",
                         "na_as_false", "ewma", "read_anthro", "as_matrix_delta",
                         "sd_median",

                         "temporary_extraneous_infants",
                         "get_dop", "calc_oob_evil_twins",
                         "calc_and_recenter_z_scores")
      } else {
        var_for_par <- c("temporary_extraneous", "valid", "swap_parameters",
                         "na_as_false", "ewma", "read_anthro", "as_matrix_delta",
                         "sd_median")
      }

      cl <- makeCluster(num.batches)
      clusterExport(cl = cl, varlist = var_for_par, envir = environment())
      registerDoParallel(cl)
    } else {
      if (is.na(num.batches))
        num.batches <- 1
    }

    setkey(data.all, subjid)
    subjid.unique <- data.all[j = unique(subjid)]
    batches.all <- data.table(
      subjid = subjid.unique,
      batch = sample(num.batches, length(subjid.unique), replace = TRUE),
      key = 'subjid'
    )
    data.all <- batches.all[data.all]

    if (!quietly){
      cat(sprintf("[%s] Begin processing pediatric data...\n", Sys.time()))
    }

    # load tanner height velocity data. sex variable is defined such that 0=male and 1=female
    # recode column names to match syntactic style ("." rather than "_" in variable names)
    tanner_ht_vel_path <- ifelse(
      ref.data.path == "",
      system.file(file.path("extdata", "tanner_ht_vel.csv.gz"), package = "growthcleanr"),
      file.path(ref.data.path, "tanner_ht_vel.csv.gz")
    )

    tanner.ht.vel <- fread(tanner_ht_vel_path)

    setnames(tanner.ht.vel,
             colnames(tanner.ht.vel),
             gsub('_', '.', colnames(tanner.ht.vel)))
    setkey(tanner.ht.vel, sex, tanner.months)
    # keep track of column names in the tanner data
    tanner.fields <- colnames(tanner.ht.vel)
    tanner.fields <- tanner.fields[!tanner.fields %in% c('sex', 'tanner.months')]

    if (!prelim_infants){
      who_max_ht_vel_path <- ifelse(
        ref.data.path == "",
        system.file(file.path("extdata", "who_ht_maxvel_3sd.csv.gz"), package = "growthcleanr"),
        file.path(ref.data.path, "who_ht_maxvel_3sd.csv.gz")
      )

      who_ht_vel_3sd_path <- ifelse(
        ref.data.path == "",
        system.file(file.path("extdata", "who_ht_vel_3sd.csv.gz"), package = "growthcleanr"),
        file.path(ref.data.path, "who_ht_vel_3sd.csv.gz")
      )
    } else {
      who_max_ht_vel_path <- ifelse(
        ref.data.path == "",
        system.file(file.path("extdata", "who_hc_maxvel_3sd_infants.csv.gz"), package = "growthcleanr"),
        file.path(ref.data.path, "who_hc_maxvel_3sd_infants.csv.gz")
      )

      who_ht_vel_3sd_path <- ifelse(
        ref.data.path == "",
        system.file(file.path("extdata", "who_hc_vel_3sd_infants.csv.gz"), package = "growthcleanr"),
        file.path(ref.data.path, "who_hc_vel_3sd_infants.csv.gz")
      )
    }
    who.max.ht.vel <- fread(who_max_ht_vel_path)
    who.ht.vel <- fread(who_ht_vel_3sd_path)
    setkey(who.max.ht.vel, sex, whoagegrp_ht)
    setkey(who.ht.vel, sex, whoagegrp_ht)
    who.ht.vel <- as.data.table(dplyr::full_join(who.ht.vel, who.max.ht.vel, by =
                                                   c('sex', 'whoagegrp_ht')))

    setnames(who.ht.vel, colnames(who.ht.vel), gsub('_', '.', colnames(who.ht.vel)))
    setkey(who.ht.vel, sex, whoagegrp.ht)
    # keep track of column names in the who growth velocity data
    who.fields <- colnames(who.ht.vel)
    who.fields <- who.fields[!who.fields %in% c('sex', 'whoagegrp.ht')]

    # 1.  General principles
    # a.	All steps are done separately for each parameter unless otherwise noted
    # b.	All steps are done sorted by subject, parameter, and age (in days) for nonexcluded and nonmissing values only unless otherwise noted. This is very important.
    #     Sorting needs to be redone with each step to account for excluded and transformed  values.
    # c.	The next value refers to the value with the next highest age for the same parameter and the same subject, and the previous value refers to the value with the
    #     next lowest age for the same parameter and the same subject.
    # d.	You will need to set up a method for keeping track of whether a value is missing or excluded (and in what step). I use variables called exc_* that are =0
    #     if a value is to be included, =1 if missing, and =2 or higher if it is to be excluded, with each number indicating a different step. I also set up
    #     parameter-specific subjid_* variables that are = to subjid for included values and are blank if the value is missing or should be excluded. These subjid_*
    #     variables need to be updated with each step.
    # e.  All steps assume that data are sorted by subjid_*, parameter, and age (in days) for nonexcluded and nonmissing values only
    #     unless otherwise noted. Sorting needs to be redone after any transformations or exclusions to account for excluded and
    #     transformed values.
    # f.  The next value refers to the nonexcluded nonmissing value with the next highest age for the same parameter and the same
    #     subject, and the previous value refers to the nonexcluded nonmissing value with the next lowest age for the same parameter
    #     and the same subject.
    # g.  exc_* should only be replaced with a  higher value if exc_*==0 at the time of replacement, unless otherwise specified.


    # NOTE: in the R code below exclusion is documented as a series of factor levels, where all levels occuring before 'Exclude' in the sequence are considered
    # to be valid measurements.  We use the built in sorting of the data.table object and subsets rather than re-sorting at each step
    # to ensure that only valid measurements are used at the beginning of each step.
    # Also, unlike the Stata code, the measurement parameter (weight vs. height) is recorded as a factor in the data frame, rather than as a variable name

    # 2.  Data set-up
    # a.	I always code sex as 0=Male, 1=Female, so I recoded the variable sex that way and left a variable sexorigcode the way the data was sent to me (1=Female 2=Male)
    # b.	Remove rows that are extraneous for subjid, param, and measurement from further analysis
    #     NOTE: this step is not needed -- handled automatically by "temporary extraneous" step.
    # c.  I generated separate variables for weight (wt) and height (ht), as well as exc_* and subjid_* variables. Set exc_*=0 if value is not missing
    #     and exc_*=1 if value is missing. In all future steps, exc_* should only be changed if it is 0. This helps to keep track of which step excluded a value.
    #     I also kept the measurement variable there and untouched because sometimes wt and ht got transformed to something else.
    # d.	I made tables based on CDC growth curve parameters that include data for each day that I will send separately. The LMS parameters for each day are
    #     cubically interpolated from the values by month available on the CDC website. Create wt and ht z-scores for each value of each parameter (Z: WtZ and HtZ).
    # e.	There are variables in the table labelled cdc_*_csd_pos and cdc_*_csd_neg. For each age and sex, these correspond to ½ of the absolute value of the
    #     median and the value with a z-score of +2 (csd_pos) and -2 (csd_neg). These can be created to generate a score similar to the z-score but with an
    #     important difference. The z-scores created using the LMS method account for skewness in the distribution of the parameters (particularly weight), which
    #     can lead to small changes in z-score with large changes in weight in subjects with very high weight, and relatively large changes in z-score for smaller
    #     changes in weight in subjects with low weights.  The score we will create can be called an SD-score (SDorig: WtSDorig and HtSDorig that is calculated by
    #     dividing the difference between the value and the median by the SD score (use csd_pos if the value is above the median, csd_neg if the value is below the
    #     median). These SD-scores, rather than z-scores, now form the basis for the algorithm.

    # recategorize linear parameters as 'HEIGHTCM'
    # NOTE: this will be changed in future to consider this difference
    data.all[param == 'LENGTHCM', param := 'HEIGHTCM']

    # calculate z/sd scores
    if(prelim_infants){
      if (!quietly)
        cat(sprintf("[%s] Calculating z-scores...\n", Sys.time()))
      # removing z calculations, as they are not used
      # for infants, use z and who
      measurement.to.z <- read_anthro(ref.data.path, cdc.only = TRUE,
                                      prelim_infants = TRUE)
      measurement.to.z_who <- read_anthro(ref.data.path, cdc.only = FALSE,
                                          prelim_infants = TRUE)

      # calculate "standard deviation" scores
      if (!quietly)
        cat(sprintf("[%s] Calculating SD-scores...\n", Sys.time()))
      data.all[, sd.orig_cdc := measurement.to.z(param, agedays, sex, v, TRUE)]
      data.all[, sd.orig_who := measurement.to.z_who(param, agedays, sex, v, TRUE)]

      # smooth z-scores/SD scores between ages 1 - 3yo using weighted scores
      # older uses cdc, younger uses who
      data.all$ageyears <- data.all$agedays/365.25

      who_weight <- 4 - (data.all$ageyears)
      cdc_weight <- (data.all$ageyears) - 2

      smooth_val <- data.all$ageyears >= 2 &
        data.all$ageyears <= 4 &
        data.all$param != "HEADCM"
      data.all[smooth_val,
               sd.orig := (data.all$sd.orig_cdc[smooth_val]*cdc_weight[smooth_val] +
                             data.all$sd.orig_who[smooth_val]*who_weight[smooth_val])/2]

      # otherwise use WHO and CDC for older and younger, respectively
      who_val <- data.all$param == "HEADCM" |
        data.all$ageyears < 2

      data.all[(who_val & !smooth_val) | (smooth_val & is.na(data.all$sd.orig_cdc)),
               sd.orig := data.all$sd.orig_who[(who_val & !smooth_val)  | (smooth_val & is.na(data.all$sd.orig_cdc))]]

      cdc_val <- data.all$param != "HEADCM" &
        data.all$ageyears > 4

      data.all[(cdc_val & !smooth_val) | (smooth_val & is.na(data.all$sd.orig_who)),
               sd.orig := data.all$sd.orig_cdc[(cdc_val & !smooth_val) | (smooth_val & is.na(data.all$sd.orig_who))]]

      # NOTE: SD SCORES IN CODE ARE Z IN INFANT DOCS -- USE sd.orig ONLY

      # keep the original, uncorrected, unrecentered zscores
      data.all[,sd.orig_uncorr := sd.orig]

      # NOTE: MAY WANT TO SUBSET HERE

      # 2b: corrected z scores ----

      # keep the original column names -- we're adding a ton of columns that we
      # want to filter out after correction
      orig_colnames <- copy(colnames(data.all))

      # start by reading in approximated fenton reference values
      ref_for_corraga <- fread(
        system.file(file.path("extdata", "ref_for_corraga.csv.gz"),
                    package = "growthcleanr"))
      ref_for_corrz <- fread(
        system.file(file.path("extdata", "ref_for_corrz.csv.gz"),
                    package = "growthcleanr"))

      # add age in months
      data.all[, agemonths := agedays/30.4375]

      potcorr <- data.all$param == "WEIGHTKG" &
        data.all$sd.orig < -2 &
        data.all$agemonths < 10

      # integer weight is in grams, rounded to the nearest 10
      data.all[potcorr, intwt := round(v*100)*10]
      # replace to facilitate merging with approximated Fenton reference values
      data.all[intwt >= 250 & intwt <=560, intwt := 570]

      # merge with approximated Fenton reference values
      data.all <- merge(
        data.all, ref_for_corraga, by = c("sex", "intwt"),
        all.x = TRUE)

      data.all[fengadays < 259, pmagedays := agedays + fengadays]
      data.all[fengadays < 259, cagedays := pmagedays - 280]
      # replace fengadays with pmagedays to facilitate merging
      data.all[, fengadays := pmagedays]

      # merge with approximated Fenton reference values
      data.all <- merge(
        data.all, ref_for_corrz, by = c("sex", "fengadays"),
        all.x = TRUE)

      # add unmodified zscore using weight in unrounded grams
      data.all[, unmod_zscore :=
                 ((v*1000/fen_wt_m)^fen_wt_l - 1)/(fen_wt_l * fen_wt_s)]

      # read in who/cdc data
      growthfile_who <- fread(
        system.file(file.path("extdata", "growthfile_who.csv.gz"),
                    package = "growthcleanr"))
      growthfile_cdc <- fread(
        system.file(file.path("extdata", "growthfile_cdc_ext_infants.csv.gz"),
                    package = "growthcleanr"))

      # merge in the who/cdc data
      data.all <- merge(data.all, growthfile_who,
                        by.x = c("sex", "cagedays"),
                        by.y = c("sex", "agedays"),
                        all.x = TRUE)
      data.all <- merge(data.all, growthfile_cdc,
                        by.x = c("sex", "cagedays"),
                        by.y = c("sex", "agedays"),
                        all.x = TRUE)

      # adjust WHO and CDC heights based on age
      data.all[, cwho_cv := v]
      data.all[, ccdc_cv := v]
      data.all[param == "HEIGHTCM" & agedays > 730 & cagedays <= 730,
               cwho_cv := cwho_cv + 0.8]
      data.all[param == "HEIGHTCM" & agedays > 730 & cagedays <= 730,
               ccdc_cv := ccdc_cv + 0.7]

      # create the corrected z scores
      # use read anthro, but pass in different arguments
      # pass in the corrected height
      data.all[, sd.c_cdc :=
                 measurement.to.z(param, cagedays, sex, ccdc_cv, TRUE)]
      data.all[, sd.c_who :=
                 measurement.to.z_who(param, cagedays, sex, cwho_cv, TRUE)]

      # smooth using weights as in original z score creation
      who_weight <- 4 - (data.all$agedays/365.25)
      cdc_weight <- (data.all$agedays/365.25) - 2

      smooth_val <- data.all$agedays/365.25 >= 2 &
        data.all$agedays/365.25 <= 4 &
        data.all$param != "HEADCM"
      data.all[smooth_val,
               sd.c := (sd.c_cdc[smooth_val]*cdc_weight[smooth_val] +
                             sd.c_who[smooth_val]*who_weight[smooth_val])/2]

      # otherwise use WHO and CDC for older and younger, respectively
      who_val <- data.all$param == "HEADCM" |
        data.all$agedays/365.25 < 2
      data.all[who_val | (smooth_val & is.na(data.all$sd.c_cdc)),
               sd.c := data.all$sd.c_who[who_val  | (smooth_val & is.na(data.all$sd.c_cdc))]]

      cdc_val <- data.all$param != "HEADCM" |
        data.all$agedays/365.25 > 4
      data.all[cdc_val | (smooth_val & is.na(data.all$sd.c_who)),
               sd.c := data.all$sd.c_cdc[cdc_val | (smooth_val & is.na(data.all$sd.c_who))]]

      # smooth corrected and uncorrected z scores
      uncorrweight <-  4 - (data.all$agedays/365.25)
      corrweight <- (data.all$agedays/365.25) - 2
      smooth_val <- data.all$agedays/365.25 >= 2 &
        data.all$agedays/365.25 <= 4
      data.all[smooth_val,
               sd.corr := (sd.c[smooth_val]*corrweight[smooth_val] +
                             sd.orig[smooth_val]*uncorrweight[smooth_val])/2]
      # for < 2 & potential correction, use corrected score
      data.all[agedays/365.25 <= 2 & potcorr, sd.corr := sd.c]
      # for > 4, use original z score
      data.all[agedays/365.25 >= 4, sd.corr := sd.orig]
      # for not potential corrections, use the original z score
      data.all[!potcorr, sd.corr := sd.orig]
      # if the who/corrected score is not available for any reason, use the
      # original
      data.all[is.na(sd.corr), sd.corr := sd.orig]

      # check for consistent growth
      examine_only <- data.all$param == "WEIGHTKG" &
        data.all$subjid %in% data.all$subjid[potcorr]
      tmp <- copy(data.all[examine_only,])
      tmp <- tmp[order(subjid, agedays),]
      tmp[, seq_win := sequence(.N), by = subjid]
      # we're only looking at the first 4 values, and they need to be < 2 years
      tmp <- tmp[seq_win <= 4 & (agedays/365.25) < 2,]
      # don't look at subjects where there is only 1 score and there is no
      # value for either
      tmp <- tmp[!(is.na(sd.corr & sd.orig)),]
      tmp <- tmp[subjid %in% names(table(subjid) > 1),]
      # create differences, absolute sum them
      tmp[, sd.corr_abssumdiff := abs(sum(sd.corr[1] - sd.corr)), by = subjid]
      tmp[, sd.orig_abssumdiff := abs(sum(sd.orig[1] - sd.orig)), by = subjid]
      # find subjects where corrected value needs to be replaced
      sub_replace <- unique(tmp[sd.corr_abssumdiff > sd.orig_abssumdiff,
                                subjid])

      # replace accordingly in the main dataframe
      data.all[subjid %in% sub_replace, sd.corr := sd.orig]

      orig_colnames <- c(orig_colnames, "sd.corr")

      # remove many added columns
      data.all <- data.all[, colnames(data.all) %in% orig_colnames,
                           with = FALSE]
    } else {
      # calculate z scores
      if (!quietly)
        cat(sprintf("[%s] Calculating z-scores...\n", Sys.time()))
      measurement.to.z <- read_anthro(ref.data.path, cdc.only = TRUE)
      data.all[, z.orig := measurement.to.z(param, agedays, sex, v)]

      # calculate "standard deviation" scores
      if (!quietly)
        cat(sprintf("[%s] Calculating SD-scores...\n", Sys.time()))
      data.all[, sd.orig := measurement.to.z(param, agedays, sex, v, TRUE)]
    }

    # sort by subjid, param, agedays
    setkey(data.all, subjid, param, agedays)

    # add a new convenience index for bookkeeping
    data.all[, index := 1:.N]

    # Mark missing values for exclusion
    data.all[, exclude := factor(with(data.all, ifelse(
      is.na(v) |
        agedays < 0, 'Missing', 'Include'
    )),
    levels = exclude.levels,
    ordered = TRUE)]
    # also mark certain measurements to not consider
    data.all[param == "HEADCM" & agedays > (3*365.25), exclude := "Not cleaned"]

    # define field names needed by helper functions
    ewma.fields <- c('ewma.all', 'ewma.before', 'ewma.after')

    # 3.  SD-score recentering: Because the basis of the method is comparing SD-scores over time, we need to account for the fact that
    #     the mean SD-score for the population changes with age.
    # a.	Determine the median cdc*sd for each parameter by year of age (with sexes combined): median*sd.
    # b.	The median*sd should be considered to apply to midyear-age, defined as the age in days with the same value as the integer
    #     portion of (365.25*year + 365.25/2).
    # c.	Linearly interpolate median*sd for each parameter between each midyear-age, naming the interpolated values rc*sd.
    # d.	For ages below the first midyear-age, let rc*sd equal the median*sd for the earliest year.
    #     For ages above the last midyear_age, let rc*sd equal the median*sd for the last year.
    # e.	Subtract rcsd_* from SDorig to create the recentered SD-score.  This recentered SD-score, labeled tbc*sd
    #     (stands for "to be cleaned") will be used for most of the rest of the analyses.
    # f.	In future steps I will sometimes refer to measprev and measnext which refer to the previous or next wt or ht measurement
    #     for which exc_*==0 for the subject and parameter, when the data are sorted by subject, parameter, and agedays. SDprev and SDnext refer to the tbc*sd of the previous or next measurement.

    if (!quietly)
      cat(sprintf("[%s] Re-centering data...\n", Sys.time()))

    # see function definition below for explanation of the re-centering process
    # returns a data table indexed by param, sex, agedays. can use NHANES reference
    # data, derive from input, or use user-supplied data.
    if (!is.data.table(sd.recenter)) {
      # INFANTS CHANGES:
      # use recentering file derived from work, independent of sex
      if (prelim_infants){
        infants_reference_medians_path <- ifelse(
          ref.data.path == "",
          system.file(file.path("extdata",
                                "rcfile-2023-08-15_format.csv.gz"),
                      package = "growthcleanr"),
          file.path(ref.data.path, "rcfile-2023-08-15_format.csv.gz")
        )
        sd.recenter <- fread(infants_reference_medians_path)
        if (!quietly)
          cat(
            sprintf(
              "[%s] Using infants reference medians...\n",
              Sys.time()
            )
          )
      } else if ((is.character(sd.recenter) &
                  tolower(sd.recenter) == "nhanes") |
          (!(is.character(sd.recenter) &
             tolower(sd.recenter) == "derive") & (data.all[, .N] < 5000))) {

        # Use NHANES medians if the string "nhanes" is specified instead of a data.table
        # or if sd.recenter is not specified as "derive" and N < 5000.

        nhanes_reference_medians_path <- ifelse(
          ref.data.path == "",
          system.file(file.path("extdata", "nhanes-reference-medians.csv.gz"), package = "growthcleanr"),
          file.path(ref.data.path, "nhanes-reference-medians.csv.gz")
        )
        sd.recenter <- fread(nhanes_reference_medians_path)
        if (!quietly)
          cat(
            sprintf(
              "[%s] Using NHANES reference medians...\n",
              Sys.time()
            )
          )
      } else {
        # Derive medians from input data
        sd.recenter <- data.all[exclude < 'Exclude', sd_median(param, sex, agedays, sd.orig)]
        if (!quietly)
          cat(
            sprintf(
              "[%s] Using re-centering medians derived from input...\n",
              Sys.time()
            )
          )
        if (sdmedian.filename != "") {
          write.csv(sd.recenter, sdmedian.filename, row.names = FALSE)
          if (!quietly)
            cat(
              sprintf(
                "[%s] Wrote re-centering medians to %s...\n",
                Sys.time(),
                sdmedian.filename
              )
            )
        }
      }
    } else {
      # Use specified data
      if (!quietly)
        cat(
          sprintf(
            "[%s] Using specified re-centering medians...\n",
            Sys.time()
          )
        )
    }

    # ensure recentering medians are sorted correctly
    setkey(sd.recenter, param, sex, agedays)

    # add sd.recenter to data, and recenter
    setkey(data.all, param, sex, agedays)
    data.all <- sd.recenter[data.all]

    setkey(data.all, subjid, param, agedays)
    data.all[, tbc.sd := sd.orig - sd.median]
    if (prelim_infants){
      # separate out corrected and noncorrected values
      data.all[, ctbc.sd := sd.corr - sd.median]
    }

    if (sdrecentered.filename != "") {
      write.csv(data.all, sdrecentered.filename, row.names = FALSE)
      if (!quietly)
        cat(
          sprintf(
            "[%s] Wrote re-centered data to %s...\n",
            Sys.time(),
            sdrecentered.filename
          )
        )
    }

    # notification: ensure awareness of small subsets in data
    if (!quietly) {
      year.counts <- data.all[, .N, floor(agedays / 365.25)]
      if (year.counts[N < 100, .N] > 0) {
        cat(
          sprintf(
            "[%s] Note: input data has at least one age-year with < 100 subjects...\n",
            Sys.time()
          )
        )
      }
    }

    # safety check: treat observations where tbc.sd cannot be calculated as missing
    data.all[is.na(tbc.sd), exclude := 'Missing']

    if (prelim_infants){
      # 4: identify subset that don't need to be cleaned (nnte) ----

      # identify those meeting all subjects meeting these criteria as "no need
      # to ewma"
      # does the subject have sdes

      # keep the original column names -- we're adding a ton of columns that we
      # want to filter out after correction
      orig_colnames <- copy(colnames(data.all))
      # no SDEs
      data.all[, sum_sde := .N, by = c("subjid", "param", "agedays")]
      data.all[, no_sde := sum_sde == 1]
      # does the subject have identical values
      data.all[, sum_val := .N, by = c("subjid", "param", "v")]
      data.all[, no_dup_val := sum_val == 1]
      # all tbc.sd are within [-3,3] -- 0 is false
      data.all[, no_outliers := sum((tbc.sd > -3 & tbc.sd < 3) |
                                      is.na(tbc.sd)) == (.N),
               by = c("subjid", "param")]
      data.all[, no_outliers := no_outliers == 1]
      # all max - min tbd.sc < 2.5
      data.all[, no_bigdiff :=
                 rep((abs(max(tbc.sd, na.rm = TRUE) - min(tbc.sd, na.rm = TRUE)) < 2.5),
                     .N),
               by = c("subjid", "param")]
      # the previous value can't be too far from the current value
      data.all[, seq_win := sequence(.N), by = c("subjid", "param")]
      data.all[, nottoofar :=
                 (abs(tbc.sd - dplyr::lag(tbc.sd)) < 1 | seq_win == 1) &
                 (abs(tbc.sd - dplyr::lead(tbc.sd)) < 1 | seq_win == .N),
               by = c("subjid", "param")]
      data.all[is.na(nottoofar),  nottoofar :=  TRUE]

      # cumulative: no need to ewma -- needs to work for all within a subject &
      # parameter
      data.all[, nnte := no_sde & no_dup_val & no_outliers & no_bigdiff & nottoofar]
      # NNTE can be calculated by parameter -- but it's occasionally easier for
      # calculations to require all parameters to be nnte
      data.all[, nnte_full := sum(nnte) == .N, by = c("subjid", "param")]
      data.all[, nnte := sum(nnte) == .N, by = c("subjid")]


      # remove many added columns -- except for nnte
      orig_colnames <- c(orig_colnames, "nnte", "nnte_full")
      data.all <- data.all[, colnames(data.all) %in% orig_colnames,
                           with = FALSE]
    }
    # pediatric: cleanbatch (most of steps) ----

    # NOTE: the rest of cleangrowth's steps are done through cleanbatch().

    # optionally process batches in parallel
    if (!quietly)
      cat(sprintf(
        "[%s] Cleaning growth data in %d batch(es)...\n",
        Sys.time(),
        num.batches
      ))
    if (num.batches == 1) {
      if (!prelim_infants){
        ret.df <- cleanbatch(data.all,
                             log.path = log.path,
                             quietly = quietly,
                             parallel = parallel,
                             measurement.to.z = measurement.to.z,
                             ewma.fields = ewma.fields,
                             ewma.exp = ewma.exp,
                             recover.unit.error = recover.unit.error,
                             include.carryforward = include.carryforward,
                             sd.extreme = sd.extreme,
                             z.extreme = z.extreme,
                             exclude.levels = exclude.levels,
                             tanner.ht.vel = tanner.ht.vel,
                             who.ht.vel = who.ht.vel,
                             lt3.exclude.mode = lt3.exclude.mode,
                             error.load.threshold = error.load.threshold,
                             error.load.mincount = error.load.mincount)
      } else {
        ret.df <- cleanbatch_infants(
          data.all,
          log.path = log.path,
          quietly = quietly,
          parallel = parallel,
          measurement.to.z = measurement.to.z,
          ewma.fields = ewma.fields,
          recover.unit.error = recover.unit.error,
          include.carryforward = include.carryforward,
          sd.extreme = sd.extreme,
          z.extreme = z.extreme,
          exclude.levels = exclude.levels,
          tanner.ht.vel = tanner.ht.vel,
          who.ht.vel = who.ht.vel,
          lt3.exclude.mode = lt3.exclude.mode,
          error.load.threshold = error.load.threshold,
          error.load.mincount = error.load.mincount,
          ref.data.path = ref.data.path)
      }
    } else {
      # create log directory if necessary
      if (!is.na(log.path)) {
        cat(sprintf("[%s] Writing batch logs to '%s'...\n", Sys.time(), log.path))
        ifelse(!dir.exists(log.path), dir.create(log.path, recursive = TRUE), FALSE)
      }

      if (!prelim_infants){
        ret.df <- ddply(
          data.all,
          .(batch),
          cleanbatch,
          .parallel = parallel,
          .paropts = list(.packages = "data.table"),
          log.path = log.path,
          quietly = quietly,
          parallel = parallel,
          measurement.to.z = measurement.to.z,
          ewma.fields = ewma.fields,
          ewma.exp = ewma.exp,
          recover.unit.error = recover.unit.error,
          include.carryforward = include.carryforward,
          sd.extreme = sd.extreme,
          z.extreme = z.extreme,
          exclude.levels = exclude.levels,
          tanner.ht.vel = tanner.ht.vel,
          who.ht.vel = who.ht.vel,
          lt3.exclude.mode = lt3.exclude.mode,
          error.load.threshold = error.load.threshold,
          error.load.mincount = error.load.mincount
        )
      } else {
        ret.df <- ddply(
          data.all,
          .(batch),
          cleanbatch_infants,
          .parallel = parallel,
          .paropts = list(.packages = "data.table"),
          log.path = log.path,
          quietly = quietly,
          parallel = parallel,
          measurement.to.z = measurement.to.z,
          ewma.fields = ewma.fields,
          recover.unit.error = recover.unit.error,
          include.carryforward = include.carryforward,
          sd.extreme = sd.extreme,
          z.extreme = z.extreme,
          exclude.levels = exclude.levels,
          tanner.ht.vel = tanner.ht.vel,
          who.ht.vel = who.ht.vel,
          lt3.exclude.mode = lt3.exclude.mode,
          error.load.threshold = error.load.threshold,
          error.load.mincount = error.load.mincount,
          ref.data.path = ref.data.path
        )
      }
      stopCluster(cl)
    }


    if (!quietly)
      cat(sprintf("[%s] Done with pediatric data!\n", Sys.time()))
  } else {
    ret.df <- data.table()

    if (!quietly)
      cat(sprintf("[%s] No pediatric data. Moving to adult data...\n", Sys.time()))
  }

  # adult: send to cleanadult to do most of the work ----

  # no need to do this if there's no data
  if (nrow(data.adult) > 0){
    if (!quietly){
      cat(sprintf("[%s] Begin processing adult data...\n", Sys.time()))
    }

    # if parallel processing is desired, load additional modules
    if (parallel) {
      if (is.na(num.batches)) {
        num.batches <- getDoParWorkers()
      }
      # variables needed for parallel workers
      var_for_par <- c(
        "cleanadult", "check_between", "round_pt", "get_float_rem",
        "as.matrix.delta_dn", "ewma_dn", "remove_biv", "remove_biv_high",
        "remove_biv_low", "identify_rv", "temp_sde", "redo_identify_rv",
        "rem_hundreds", "rem_unit_errors", "get_num_places", "switch_tens_ones",
        "rem_transpositions", "ht_allow", "ht_change_groups",
        "ht_3d_growth_compare", "remove_ewma_wt", "remove_mod_ewma_wt"
      )

      cl <- makeCluster(num.batches)
      clusterExport(cl = cl, varlist = var_for_par, envir = environment())
      registerDoParallel(cl)
    } else {
      if (is.na(num.batches))
        num.batches <- 1
    }

    # this is where we do most of the adult work
    # is the randomness necessary here?
    subjid.unique <- data.adult[j = unique(subjid)]
    batches.adult <- data.table(
      subjid = subjid.unique,
      newbatch = sample(num.batches, length(subjid.unique), replace = TRUE)
    )
    data.adult <- merge(data.adult, batches.adult, by = "subjid")

    # add age in years
    data.adult[, age_years := agedays/365.25]
    # rename for ease of use
    data.adult[, measurement := v_adult]
    data.adult[, id := line]

    if (num.batches == 1) {
      # do the cleaning
      res <- cleanadult(data.adult, weight_cap = weight_cap)
    } else {
      res <- ddply(
        data.adult,
        .(newbatch),
        cleanadult,
        .parallel = parallel,
        .paropts = list(.packages = "data.table"),
        weight_cap = weight_cap
      )

      res <- as.data.table(res)
    }

    # replace result with missing if measurement or agedays are missing
    res[is.na(measurement) | agedays < 0, result := "Missing"]

    if (parallel){
      stopCluster(cl)
    }

    if (adult_columns_filename != "") {
      write.csv(res, adult_columns_filename, row.names = FALSE, na = "")
      if (!quietly){
        cat(
          sprintf(
            "[%s] Wrote adult data with additional columns to to %s...\n",
            Sys.time(),
            adult_columns_filename
          )
        )
      }
    }

    if (!quietly)
      cat(sprintf("[%s] Done with adult data!\n", Sys.time()))
  } else {
    res <- data.table()

    if (!quietly){
      cat(sprintf("[%s] No adult data. Moving to postprocessing...\n", Sys.time()))
    }
  }

  if (any(nrow(data.all) > 0, nrow(data.adult) > 0)) {
    # join with pediatric data
    full_out <- data.table(
      line = c(ret.df$line, res$line),
      exclude = c(as.character(ret.df$exclude), res$result),
      mean_sde = c(rep(NA, nrow(ret.df)), res$mean_sde)
    )
    full_out[, exclude := factor(exclude, levels = exclude.levels)]
    full_out <- full_out[order(line),]
    # remove column added for keeping track
    full_out[, line := NULL]

    return(full_out$exclude)
  } else {
    return(c())
  }

}

#' Function to calculate z-scores and csd-scores based on anthro tables.
#'
#' @param path Path to supplied reference anthro data. Defaults to package anthro tables.
#' @param cdc.only Whether or not only CDC data should be used. Defaults to false.
#' @param prelim_infants TRUE/FALSE. Run the in-development release of the infants algorithm (expands pediatric algorithm to improve performance for children 0 – 2 years). Not recommended for use in research. For more information regarding the logic of the algorithm, see the vignette 'Preliminary Infants Algorithm.' Defaults to FALSE.
#'
#' @return Function for calculating BMI based on measurement, age in days, sex, and measurement value.
#' @export
#' @import data.table
#' @importFrom utils read.csv read.table
#' @examples
#' \donttest{
#' # Return calculating function with all defaults
#' afunc <- read_anthro()
#'
#' # Return calculating function while specifying a path and using only CDC data
#' afunc <- read_anthro(path = system.file("extdata", package = "growthcleanr"),
#'                      cdc.only = TRUE)
#' }
read_anthro <- function(path = "", cdc.only = FALSE, prelim_infants = FALSE) {
  # avoid "no visible bindings" warning
  src <- param <- sex <- age <- ret <- m <- NULL
  csdneg <- csdpos <- s <- NULL

  # set correct path based on input reference table path (if any)
  if (!prelim_infants){
    weianthro_path <- ifelse(
      path == "",
      system.file(file.path("extdata", "weianthro.txt.gz"), package = "growthcleanr"),
      file.path(path, "weianthro.txt.gz")
    )
    lenanthro_path <- ifelse(
      path == "",
      system.file(file.path("extdata", "lenanthro.txt.gz"), package = "growthcleanr"),
      file.path(path, "lenanthro.txt.gz")
    )
    bmianthro_path <- ifelse(
      path == "",
      system.file(file.path("extdata", "bmianthro.txt.gz"), package = "growthcleanr"),
      file.path(path, "bmianthro.txt.gz")
    )
    growth_cdc_ext_path <- ifelse(
      path == "",
      system.file(file.path("extdata", "growthfile_cdc_ext.csv.gz"), package = "growthcleanr"),
      file.path(path, "growthfile_cdc_ext.csv.gz")
    )
  } else {
    weianthro_path <- lenanthro_path <- bmianthro_path <-
      ifelse(
        path == "",
        system.file(file.path("extdata", "growthfile_who.csv.gz"), package = "growthcleanr"),
        file.path(path, "growthfile_who.csv.gz")
      )
    growth_cdc_ext_path <- ifelse(
      path == "",
      system.file(file.path("extdata", "growthfile_cdc_ext_infants.csv.gz"), package = "growthcleanr"),
      file.path(path, "growthfile_cdc_ext_infants.csv.gz")
    )
  }
  growth_cdc_ext <- read.csv(gzfile(growth_cdc_ext_path))

  l <- if (!prelim_infants){
    list(
      with(
        read.table(gzfile(weianthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'WEIGHTKG',
          sex = sex - 1,
          age,
          l,
          m,
          s,
          csdpos = as.double(NA),
          csdneg = as.double(NA)
        )
      ),
      with(
        read.table(gzfile(lenanthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'HEIGHTCM',
          sex = sex - 1,
          age,
          l,
          m,
          s,
          csdpos = as.double(NA),
          csdneg = as.double(NA)
        )
      ),
      with(
        read.table(gzfile(bmianthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'BMI',
          sex = sex - 1,
          age,
          l,
          m,
          s,
          csdpos = as.double(NA),
          csdneg = as.double(NA)
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'WEIGHTKG',
          sex,
          age = agedays,
          l = cdc_wt_l,
          m = cdc_wt_m,
          s = cdc_wt_s,
          csdpos = cdc_wt_csd_pos,
          csdneg = cdc_wt_csd_neg
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'HEIGHTCM',
          sex,
          age = agedays,
          l = cdc_ht_l,
          m = cdc_ht_m,
          s = cdc_ht_s,
          csdpos = cdc_ht_csd_pos,
          csdneg = cdc_ht_csd_neg
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'BMI',
          sex,
          age = agedays,
          l = cdc_bmi_l,
          m = cdc_bmi_m,
          s = cdc_bmi_s,
          csdpos = cdc_bmi_csd_pos,
          csdneg = cdc_bmi_csd_neg
        )
      )
    )
  } else {
    list(
      with(
        read.csv(gzfile(weianthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'WEIGHTKG',
          sex,
          age = agedays,
          l = who_wt_l,
          m = who_wt_m,
          s = who_wt_s,
          csdpos = who_wt_csd_pos,
          csdneg =  who_wt_csd_neg
        )
      ),
      with(
        read.csv(gzfile(lenanthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'HEIGHTCM',
          sex,
          age = agedays,
          l = who_ht_l,
          m = who_ht_m,
          s = who_ht_s,
          csdpos = who_ht_csd_pos,
          csdneg =  who_ht_csd_neg
        )
      ),
      with(
        read.csv(gzfile(lenanthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'HEADCM',
          sex,
          age = agedays,
          l = who_hc_l,
          m = who_hc_m,
          s = who_hc_s,
          csdpos = who_hc_csd_pos,
          csdneg =  who_hc_csd_neg
        )
      ),
      with(
        read.csv(gzfile(bmianthro_path), header = TRUE),
        data.frame(
          src = 'WHO',
          param = 'BMI',
          sex,
          age = agedays,
          l = who_bmi_l,
          m = who_bmi_m,
          s = who_bmi_s,
          csdpos = who_bmi_csd_pos,
          csdneg =  who_bmi_csd_neg
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'WEIGHTKG',
          sex,
          age = agedays,
          l = cdc_wt_l,
          m = cdc_wt_m,
          s = cdc_wt_s,
          csdpos = cdc_wt_csd_pos,
          csdneg = cdc_wt_csd_neg
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'HEIGHTCM',
          sex,
          age = agedays,
          l = cdc_ht_l,
          m = cdc_ht_m,
          s = cdc_ht_s,
          csdpos = cdc_ht_csd_pos,
          csdneg = cdc_ht_csd_neg
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'HEADCM',
          sex,
          age = agedays,
          l = cdc_hc_l,
          m = cdc_hc_m,
          s = cdc_hc_s,
          csdpos = cdc_hc_csd_pos,
          csdneg = cdc_hc_csd_neg
        )
      ),
      with(
        growth_cdc_ext,
        data.frame(
          src = 'CDC',
          param = 'BMI',
          sex,
          age = agedays,
          l = cdc_bmi_l,
          m = cdc_bmi_m,
          s = cdc_bmi_s,
          csdpos = cdc_bmi_csd_pos,
          csdneg = cdc_bmi_csd_neg
        )
      )
    )
  }

  anthro <- rbindlist(l)


  setkey(anthro, src, param, sex, age)

  return(function(param, agedays, sex, measurement, csd = FALSE) {
    # For now, we will only use CDC growth reference data, note that the cubically interpolated file
    # we are using has linear measurments derived from length data for children < 731 days, and height thereafter
    src <- ifelse(agedays < 3*365.25 & !cdc.only, 'WHO', 'CDC')

    # keep column sequence the same fo efficient join
    dt <- data.table(src, param, sex, agedays, measurement)
    dt <- anthro[dt]

    dt[, ret := as.double(NA)]
    if (csd) {
      dt[measurement < m, ret := (measurement - m) / csdneg]
      dt[measurement >= m, ret := (measurement - m) / csdpos]
    } else {
      dt[l == 0, ret := log(measurement / m) / s]
      dt[l != 0, ret := (((measurement / m) ^ l) - 1) / (l * s)]
    }

    return(dt$ret)
  })
}

#' Exponentially Weighted Moving Average (EWMA)
#'
#' \code{ewma} calculates the exponentially weighted moving average (EWMA) for a set of numeric observations over time.
#'
#' @param agedays Vector of age in days for each z score (potentially transformed to adjust weighting).
#'
#' @param z Input vector of numeric z-score data.
#'
#' @param ewma.exp Exponent to use for weighting.
#'
#' @param ewma.adjacent Specify whether EWMA values excluding adjacent measurements should be calculated.  Defaults to TRUE.
#'
#' @return Data frame with 3 variables:
#' * The first variable (ewma.all) contains the EWMA at observation time
#'   excluding only the actual observation for that time point.
#' * The second variable (ewma.before) contains the EWMA for each observation excluding both the actual observation
#'   and the immediate prior observation.
#' * The third variable (ewma.after) contains the EWMA for each observation excluding both the actual observation
#'   and the subsequent observation.
#'@md
#'
#' @export
#' @examples
#' # Run on 1 subject, 1 type of parameter
#' df_stats <- as.data.frame(syngrowth)
#' df_stats <- df_stats[df_stats$subjid == df_stats$subjid[1] &
#'                        df_stats$param == "HEIGHTCM", ]
#'
#' # Get the uncentered z-scores
#' measurement_to_z <- read_anthro(cdc.only = TRUE)
#' sd <- measurement_to_z(df_stats$param,
#'                        df_stats$agedays,
#'                        df_stats$sex,
#'                        df_stats$measurement,
#'                        TRUE)
#'
#' # Calculate exponentially weighted moving average
#' e_df <- ewma(df_stats$agedays, sd, ewma.exp = -1.5)
ewma <- function(agedays, z, ewma.exp, ewma.adjacent = TRUE) {
  # 6.  EWMA calculation description: Most of the next steps will involve calculating the exponentially weighted moving average for each subject and parameter. I will
  #     describe how to calculate EWMASDs, and will describe how it needs to be varied in subsequent steps.
  # a.	The overall goal of the EWMASD calculation is to identify the difference between the SD-score and what we might predict that DS-score should be, in order to
  #     determine whether it should be excluded.
  # b.	Only nonmissing SD-scores for a parameter that are not designated for exclusion are included in the following calculations.
  # c.	For each SD-score SDi and associated agedaysi calculate the following for every other z-score (SDj…SDn) and associated agedays (agedaysj…agedaysn)  for the
  #     same subject and parameter
  #   i.	ΔAgej=agedaysj-agedaysi
  #   ii.	EWMAZ=SDi=[Σj→n(SDj*((5+ΔAgej)^-1.5))]/[ Σj→n((5+ΔAgej)^-1.5)]
  #   iii.	For most EWMASD calculations, there are 3 EWMASDs that need to be calculated. I will note if not all of these need to be done for a given step.
  #     1.	EWMASDall calculated as above
  #     2.	EWMAZbef calculated excluding the SD-score just before the SD-score of interest (sorted by agedays). For the first observation for a parameter for a
  #         subject, this should be identical to EWMASDall rather than missing.
  #     3.	EWMAZaft calculated excluding the z-score just after the SD-score of interest (sorted by agedays). For the lastobservation for a parameter for a subject,
  #         this should be identical to EWMASDall rather than missing.
  #   iv.	For each of the three EWMASDs, calculate the dewma_*=SD-EWMASD
  # d.	EWMASDs and ΔEWMASDs will change if a value is excluded or manipulated using one of the methods below, therefore EWMASDs and ΔEWMASDs be recalculated for each
  #     step where they are needed.
  # e.	For these calculations, use variables that allow for precise storage of numbers (in Stata this is called 'double') because otherwise rounding errors can cause
  #     problems in a few circumstances

  n <- length(agedays)
  # initialize response variables
  ewma.all <- ewma.before <- ewma.after <- vector('numeric', 0)
  if (n > 0) {
    # organize into data frame and sort into order of increasing age,
    # but retain original sort order information in index
    if (!all(agedays == cummax(agedays)))
      warning("EWMA ordering is not sorted; double check") #add in a check to make sure the inputs are already sorted (they should be)
    index <- order(agedays)


    # calculate matrix of differences in age, and add 5 to each delta per Daymont algorithm
    delta <- as_matrix_delta(agedays)
    delta <- ifelse(delta == 0, 0, (delta + 5) ^ ewma.exp)

    # calculate EWMAs, and return in order of original data
    ewma.all[index] <- delta %*% z / apply(delta, 1, sum)

    if (ewma.adjacent) {
      if (n > 2) {
        delta2 <- delta
        delta2[col(delta2) == row(delta2) - 1] <- 0
        ewma.before[index] <- delta2 %*% z / apply(delta2, 1, sum)
        delta3 <- delta
        delta3[col(delta3) == row(delta3) + 1] <- 0
        ewma.after[index] <- delta3 %*% z / apply(delta3, 1, sum)
      } else {
        ewma.before <- ewma.after <- ewma.all
      }
    }
  }
  # return all 3 EWMAs as a data frame
  return(if (ewma.adjacent)
    data.frame(ewma.all, ewma.before, ewma.after)
    else
      data.frame(ewma.all))
}

#' Calculate median SD score by age for each parameter.
#'
#' @param param Vector identifying each measurement, may be 'WEIGHTKG', or 'HEIGHTCM'.
#' @param agedays Numeric vector containing the age in days at each measurement.
#' @param sex Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F', 'f' or 1 for females.
#' @param sd.orig Vector of previously calculated standard deviation (SD) scores for each measurement before re-centering.
#'
#' @return Table of data with median SD-scores per day of life by gender and parameter.
#'
#' @export
#' @import data.table
#' @examples
#' # Run on 1 subject
#' df_stats <- as.data.frame(syngrowth)
#' df_stats <- df_stats[df_stats$subjid == df_stats$subjid[1], ]
#'
#' # Get the original standard deviations
#' measurement_to_z <- read_anthro(cdc.only = TRUE)
#' sd.orig <- measurement_to_z(df_stats$param,
#'                        df_stats$agedays,
#'                        df_stats$sex,
#'                        df_stats$measurement,
#'                        TRUE)
#'
#' # Calculate median standard deviations
#' sd.m <- sd_median(df_stats$param,
#'                   df_stats$sex,
#'                   df_stats$agedays,
#'                   sd.orig)
sd_median <- function(param, sex, agedays, sd.orig) {
  # 3.  SD-score recentering: Because the basis of the method is comparing SD-scores over time, we need to account for the fact that
  #     the mean SD-score for the population changes with age.
  # a.  Determine the median cdc*sd for each parameter by year of age (with sexes combined): median*sd.
  # b.	The median*sd should be considered to apply to midyear-age, defined as the age in days with the same value as the integer
  #     portion of (365.25*year + 365.25/2).
  # c.	Linearly interpolate median*sd for each parameter between each midyear-age, naming the interpolated values rc*sd.
  # d.	For ages below the first midyear-age, let rc*sd equal the median*sd for the earliest year.
  #     For ages above the last midyear_age, let rc*sd equal the median*sd for the last year.
  # e.	Subtract rcsd_* from SDorig to create the recentered SD-score.  This recentered SD-score, labeled tbc*sd
  #     (stands for "to be cleaned") will be used for most of the rest of the analyses.
  # f.	In future steps I will sometimes refer to measprev and measnext which refer to the previous or next wt or ht measurement
  #     for which exc_*==0 for the subject and parameter, when the data are sorted by subject, parameter, and agedays. SDprev and SDnext refer to the tbc*sd of the previous or next measurement.

  # avoid "no visible binding" warnings
  ageyears <- sd.median <- NULL

  dt <- data.table(param, sex, agedays, ageyears = floor(agedays / 365.25), sd.orig)
  setkey(dt, param, sex, agedays)
  # determine ages (in days) we need to cover from min to max age in years
  agedays.to.cover <- dt[, (floor(min(ageyears) * 365.25):floor(max(ageyears +
                                                                     1) * 365.25))]
  # group all measurements above age 19 together (copied from CD stata code e-mailed 4/3/15)
  dt[ageyears > 19, ageyears := 19]
  dt.median <- dt[!is.na(sd.orig), list(sex = c(0, 1), sd.median = rep(median(sd.orig), 2)), by =
                   .(param, ageyears)]
  dt.median <- dt.median[, {
    x_pts <- floor((ageyears + 0.5) * 365.25)
    y_pts <- sd.median
    list(
      agedays = agedays.to.cover,
      sd.median = if (length(x_pts) < 2) {
        # fewer than 2 age-years of data (e.g. adult_cutpoint near the
        # minimum age in the dataset): fall back to constant recentering
        rep(if (length(y_pts) > 0) y_pts[1] else NA_real_,
            length(agedays.to.cover))
      } else {
        approx(x_pts, y_pts, xout = agedays.to.cover, rule = 2)$y
      }
    )
  }, by = .(param, sex)]
  setkey(dt.median, param, sex, agedays)
  return(dt.median)
}

testthat::skip_on_cran()

test_that("growthcleanr works as expected on pediatric synthetic data", {

  # Run cleangrowth() on syngrowth data
  data <- as.data.table(syngrowth)

  # syngrowth hasn't changed in length
  expect_equal(77721, data[, .N])
  setkey(data, subjid, param, agedays)

  # subset to pediatric data
  data_peds <- copy(data[agedays < 20 * 365.25, ])

  # Create small samples; one for NHANES recentering, one for derive. Note that
  # we are not auto-detecting large sizes because running cleangrowth is too
  # long for CRAN test suite.
  #
  # Note that we're creating distinct data tables to avoid accidentally
  # reusing the same by reference.
  d100_nhanes <- as.data.table(data_peds)[subjid %in% unique(data[, subjid])[1:100], ]
  expect_equal(832, d100_nhanes[, .N])

  # And for overriding NHANES/derive option
  d100_derive <- as.data.table(data_peds)[subjid %in% unique(data[, subjid])[1:100], ]
  expect_equal(832, d100_derive[, .N])

  # Clean samples: specify sd.recenter should use NHANES
  cd100_nhanes <-
    d100_nhanes[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement,
      sd.recenter = "NHANES"
    )]

  # Specifying "derive" from data instead of NHANES
  cd100_derived <-
    d100_derive[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement,
      sd.recenter = "derive"
    )]


  # Spot check individual results
  gcr_result <- function (dt, rowid) {
    return(as.character(dt[id == rowid]$gcr_result))
  }

  # Results for these records should not change w/sample size (NHANES vs. derived)
  expect_equal("Exclude-EWMA-8", gcr_result(cd100_nhanes, 35119))
  expect_equal("Exclude-EWMA-8", gcr_result(cd100_derived, 35119))

  expect_equal("Exclude-Min-Height-Change", gcr_result(cd100_nhanes, 38718))
  expect_equal("Exclude-Min-Height-Change", gcr_result(cd100_derived, 38718))

  expect_equal("Include", gcr_result(cd100_nhanes, 23766))
  expect_equal("Include", gcr_result(cd100_derived, 23766))

  # Results for these records can change w/NHANES vs. derived
  expect_equal("Exclude-Extraneous-Same-Day", gcr_result(cd100_nhanes, 25))
  expect_equal("Include", gcr_result(cd100_derived, 25))

  expect_equal("Exclude-Carried-Forward", gcr_result(cd100_nhanes, 40094))
  expect_equal("Exclude-Carried-Forward", gcr_result(cd100_derived, 40094))

  expect_equal("Include", gcr_result(cd100_nhanes, 62606))
  expect_equal("Exclude-Extraneous-Same-Day", gcr_result(cd100_derived, 62606))

  # Check counts of exclusions by category
  catcount <- function (df, category) {
    return(as.numeric(df %>% filter(gcr_result == category) %>% select(n)))
  }

  d100_exclusions <-
    cd100_nhanes %>% group_by(gcr_result) %>% tally(sort = TRUE)
  expect_equal(563, catcount(d100_exclusions, "Include"))
  expect_equal(113, catcount(d100_exclusions, "Exclude-Carried-Forward"))
  expect_equal(3, catcount(d100_exclusions, "Exclude-EWMA-8"))

  d100_derived_exclusions <-
    cd100_derived %>% group_by(gcr_result) %>% tally(sort = TRUE)
  expect_equal(563, catcount(d100_derived_exclusions, "Include"))
  expect_equal(113, catcount(d100_derived_exclusions, "Exclude-Carried-Forward"))
  expect_equal(3, catcount(d100_derived_exclusions, "Exclude-EWMA-8"))

  # also run a test to make sure missing data returns as "missing"
  d5_miss <- as.data.table(data_peds)[subjid %in% unique(data[, subjid])[1:5], ]

  # add missing data randomly for 5 values
  set.seed(10)
  d5_miss$measurement[sample(1:nrow(d5_miss), 5)] <- NA

  d5_miss <-
    d5_miss[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement
    )]

  expect_equal(sum(d5_miss$gcr_result == "Missing"), 5)

})

test_that("growthcleanr works as expected on adult synthetic data", {

  # Run cleangrowth() on syngrowth data
  data <- as.data.table(syngrowth)

  # syngrowth hasn't changed in length
  expect_equal(77721, data[, .N])
  setkey(data, subjid, param, agedays)

  # subset to adult data
  data_adult <- copy(data[agedays >= 18 * 365.25, ])

  # Create small sample
  d100 <- as.data.table(data_adult)[subjid %in% unique(data_adult[, subjid])[1:100], ]
  expect_equal(2023, d100[, .N])

  # Clean sample
  cd100 <-
    d100[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement
    )]

  # Clean again with lower cutpoint
  cd100cp <-
    copy(d100)[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement,
      adult_cutpoint = 18
    )]


  # Spot check individual results
  gcr_result <- function (dt, rowid) {
    return(as.character(dt[id == rowid]$gcr_result))
  }

  # These results should not change with cutpoint
  expect_equal("Include", gcr_result(cd100, 18695))
  expect_equal("Include", gcr_result(cd100cp, 18695))

  expect_equal("Exclude-Adult-Identical-Same-Day", gcr_result(cd100, 167))
  expect_equal("Exclude-Adult-Identical-Same-Day", gcr_result(cd100cp, 167))

  expect_equal("Exclude-Adult-BIV", gcr_result(cd100, 22009))
  expect_equal("Exclude-Adult-BIV", gcr_result(cd100cp, 22009))

  # Results for these records should change due to younger cutpoint
  expect_equal("Exclude-Extraneous-Same-Day", gcr_result(cd100, 69740))
  expect_equal("Exclude-Adult-Extraneous-Same-Day", gcr_result(cd100cp, 69740))

  expect_equal("Exclude-Carried-Forward", gcr_result(cd100, 55171))
  expect_equal("Include", gcr_result(cd100cp, 55171))

  expect_equal("Exclude-Extraneous-Same-Day", gcr_result(cd100, 25259))
  expect_equal("Exclude-Adult-Distinct-3-Or-More", gcr_result(cd100cp, 25259))

  # Check counts of exclusions by category
  catcount <- function (df, category) {
    return(as.numeric(df %>% filter(gcr_result == category) %>% select(n)))
  }

  d100_exclusions <-
    cd100 %>% group_by(gcr_result) %>% tally(sort = TRUE)
  expect_equal(1570, catcount(d100_exclusions, "Include"))
  expect_equal(347, catcount(d100_exclusions, "Exclude-Adult-Extraneous-Same-Day"))
  expect_equal(6, catcount(d100_exclusions, "Exclude-Adult-Distinct-3-Or-More"))
  expect_equal(13, catcount(d100_exclusions, "Exclude-Carried-Forward"))
  expect_equal(13, catcount(d100_exclusions, "Exclude-Adult-BIV"))

  d100cp_exclusions <-
    cd100cp %>% group_by(gcr_result) %>% tally(sort = TRUE)
  expect_equal(1580, catcount(d100cp_exclusions, "Include"))
  expect_equal(368, catcount(d100cp_exclusions, "Exclude-Adult-Extraneous-Same-Day"))
  expect_equal(7, catcount(d100cp_exclusions, "Exclude-Adult-Distinct-3-Or-More"))
  expect_true(is.na(catcount(d100cp_exclusions, "Exclude-Carried-Forward")))
  expect_equal(14, catcount(d100cp_exclusions, "Exclude-Adult-BIV"))

  # also run a test to make sure missing data returns as "missing"
  d5_miss <- as.data.table(data_adult)[subjid %in% unique(data[, subjid])[1:5], ]

  # add missing data randomly for 5 values
  set.seed(10)
  d5_miss$measurement[sample(1:nrow(d5_miss), 5)] <- NA

  d5_miss <-
    d5_miss[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement
    )]

  expect_equal(sum(d5_miss$gcr_result == "Missing"), 5)

})

test_that("growthcleanr works without either adult or pediatric data", {
  # creating small only adult and only pediatric data
  # using default cutpoint -- 20
  only_peds <- syngrowth[syngrowth$agedays < 20*365.25,][1:20,]
  only_adult <- syngrowth[syngrowth$agedays >= 20*365.25,][1:20,]
  nobody <- syngrowth[syngrowth$agedays > 120*365.25,]

  # testing cleangrowth works without adult data
  peds_res <- cleangrowth(
    only_peds$subjid,
    only_peds$param,
    only_peds$agedays,
    only_peds$sex,
    only_peds$measurement,
    quietly = TRUE
  )

  expect_equal(length(peds_res), nrow(only_peds))

  # testing cleangrowth works without pediatric data
  adult_res <- cleangrowth(
    only_adult$subjid,
    only_adult$param,
    only_adult$agedays,
    only_adult$sex,
    only_adult$measurement,
    quietly = TRUE
  )

  expect_equal(length(adult_res), nrow(only_adult))

  # testing cleangrowth works with no data
  no_res <- cleangrowth(
    nobody$subjid,
    nobody$param,
    nobody$agedays,
    nobody$sex,
    nobody$measurement,
    quietly = TRUE
  )

  expect_equal(length(no_res), nrow(nobody))

})

test_that("growthcleanr runs preliminary infants algorithm", {

  # Run cleangrowth() on syngrowth data
  data <- as.data.table(syngrowth)

  # syngrowth hasn't changed in length
  expect_equal(77721, data[, .N])
  setkey(data, subjid, param, agedays)

  # subset to pediatric data
  data_peds <- copy(data[agedays < 20 * 365.25, ])

  # Create very small sample; just to test no errors in running
  d5_nhanes <- as.data.table(data_peds)[subjid %in% unique(data[, subjid])[1:5], ]
  expect_equal(42, d5_nhanes[, .N])

  # Clean samples with infants algorithm
  cd5_nhanes <-
    d5_nhanes[, gcr_result := cleangrowth(
      subjid,
      param,
      agedays,
      sex,
      measurement,
      prelim_infants = TRUE
    )]

  expect_equal(length(cd5_nhanes$gcr_result), 42)

  # test that there are a correct number of levels
  expect_equal(length(levels(cd5_nhanes$gcr_result)), 80)
})

# Main adult growthcleanr function
# internal supporting functions for adults can be found in: adult_support.R

# TODO: ADD A DATAFRAME FOR STEPS

#' function to clean height and weight data for adults
#' inputs:
#' df: data.table with 7 columns:
#'   id: row id, must be unique
#'   subjid: subject id
#'   sex: sex of subject
#'   age_years: age, in years
#'   param: HEIGHTCM or WEIGHTKG
#'   measurement: height or weight measurement
#' outputs:
#'   df, with additional columns:
#'     result, which specifies whether the height measurement should be included,
#'       or is implausible (designated with an error code).
#'     mean_sde, mean of similar same day extraneous values
#' @keywords internal
#' @importFrom stats setNames
#' @noRd
cleanadult <- function(df, weight_cap = Inf){

  # avoid "no visible binding" warnings
  age_years <- embed <- i.all_exc_weight_cap <- i.gain_groups <- i.keep <- NULL
  i.loss_groups <- i.mean_ht <- i.mean_sde <- i.orig_sde <- mean_sde <- NULL
  id <- median <- orig_sde <- param <- result <- NULL

  # method specific constants ----
  # this includes specified cutoffs, etc.

  # BIVs
  biv_df <- data.frame(
    "low" = c(50, 20),
    "high" = c(244, 500)
  )
  rownames(biv_df) <- c("height", "weight")

  # begin implementation ----

  # parallel coerces to a dataframe
  if (!is.data.table(df)){
    df <- as.data.table(df)
  }

  # preallocate final designation
  df[, id := as.character(id)]
  df[, result := "Include"]
  df[, orig_sde := as.numeric(NA)]
  df[, mean_sde := as.numeric(NA)]
  df[, mean_ht := as.numeric(NA)]
  df[, loss_groups := as.numeric(NA)]
  df[, gain_groups := as.numeric(NA)]
  df[, all_exc_weight_cap := FALSE]
  # rownames(df) <- df$id # NO ROWNAMES IN DATA.TABLE
  # go through each subject
  for (i in unique(df$subjid)){
    slog <- df$subjid == i

    # start with height (steps 1 - 7) ----

    h_df <- copy(df[param %in% c("HEIGHTCM", "HEIGHTIN") & slog,])

    # order by age
    h_df <- h_df[order(age_years, id),]

    h_subj_keep <- rep("Include", nrow(h_df))
    h_subj_mean_sde <- rep(NA, nrow(h_df))
    names(h_subj_keep) <- names(h_subj_mean_sde) <-
      h_df$id

    h_subj_df <- copy(h_df)

    # if there are no valid heights, skip
    if (nrow(h_df) > 0){
      # add metric (m) and imperial (im) measurements
      # also keep original measurements (will overwrite for ease in functions)
      h_subj_df$meas_m <- h_subj_df$meas_im <- h_subj_df$meas_orig <-
        h_subj_df$measurement
      h_subj_df$meas_m[h_subj_df$param == "HEIGHTIN"] <-
        h_subj_df$measurement[h_subj_df$param == "HEIGHTIN"]*2.54
      h_subj_df$meas_im[h_subj_df$param == "HEIGHTCM"] <-
        h_subj_df$measurement[h_subj_df$param == "HEIGHTCM"]/2.54
      # convert age years to days -- already in there
      h_subj_df$age_days <- h_subj_df$agedays
    }

    # 1h, H BIV ----
    # 1h. remove biologically impossible height records
    step <- "Exclude-Adult-BIV"

    if (nrow(h_subj_df) > 0){
      # overwrite measurement with metric (bivs are in metric)
      h_subj_df$measurement <- h_subj_df$meas_m

      criteria <- remove_biv(h_subj_df, "height", biv_df)
      h_subj_keep[criteria] <- step

      h_subj_df <- h_subj_df[!criteria,]
    }

    # 3h, H temp extraneous ----
    # 3h. Calculate temporary extraneous for exclusion in intermediate steps
    step <- "temp extraneous" # no exclusions here

    if (nrow(h_subj_df) > 0){
      # overwrite measurement with metric (step done in metric)
      h_subj_df$measurement <- h_subj_df$meas_m

      # adds the "extraneous" column, designating whether or not the row is
      # temporarily extraneous (not to be considered in the future)
      # also adds age days and "diff" (unused outside of this function)
      h_subj_df <- temp_sde(h_subj_df)
    }

    # 5h, H hundreds ----
    # 5h. when height goes down by 100 cm -- is it valid?
    step <- "Exclude-Adult-Hundreds"

    inc_df <- if (nrow(h_subj_df) > 0){
      # we only want to consider subjects without temp extraneous
      copy(h_subj_df[!h_subj_df$extraneous,])
    } else {
      h_subj_df
    }

    # only do this if there are at least two values
    if (nrow(inc_df) > 1){
      # calculate ewma (using metric)
      ewma_res <- ewma_dn(inc_df$age_days, inc_df$meas_m)
      # delta ewma
      dewma <- (inc_df$meas_m- ewma_res)
      colnames(dewma) <- paste0("d",colnames(ewma_res))

      criteria <- rem_hundreds(inc_df, dewma, "meas_m", 100, "height")

      # update and remove
      h_subj_keep[as.character(inc_df$id)][criteria] <- step

      # don't get rid of extraneous just yet
      h_subj_df <- h_subj_df[h_subj_df$id %in% inc_df$id[!criteria] |
                               h_subj_df$extraneous,]

      # reevaluate temp same day
      h_subj_df <- temp_sde(h_subj_df)
    }

    # 6h, H unit errors ----
    # 6h. checking whether or not height should be a different type of value
    step <- "Exclude-Adult-Unit-Errors"

    inc_df <- if (nrow(h_subj_df) > 0){
      # we only want to consider subjects without temp extraneous
      copy(h_subj_df[!h_subj_df$extraneous,])
    } else {
      h_subj_df
    }

    # only do this if there are at least two values
    if (nrow(inc_df) > 1){

      criteria <- rem_unit_errors(inc_df, ptype = "height")

      # update and remove
      h_subj_keep[as.character(inc_df$id)][criteria] <- step

      # don't get rid of extraneous just yet
      h_subj_df <- h_subj_df[h_subj_df$id %in% inc_df$id[!criteria] |
                               h_subj_df$extraneous,]

      # reevaluate temp same day
      h_subj_df <- temp_sde(h_subj_df)
    }

    # 7h, H transpositions ----
    # 7h. checking whether or not 10s and 1s digit should be switched
    step <- "Exclude-Adult-Transpositions"

    inc_df <- if (nrow(h_subj_df) > 0){
      # we only want to consider subjects without temp extraneous
      copy(h_subj_df[!h_subj_df$extraneous,])
    } else {
      h_subj_df
    }

    # only do this if there are at least two unique values
    if (nrow(inc_df) > 1 & length(unique(inc_df$meas_m)) > 2){

      criteria <- rem_transpositions(inc_df, ptype = "height")

      # update and remove
      h_subj_keep[as.character(inc_df$id)][criteria] <- step

      # don't get rid of extraneous just yet
      h_subj_df <- h_subj_df[h_subj_df$id %in% inc_df$id[!criteria] |
                               h_subj_df$extraneous,]

      # reevaluate temp same day
      h_subj_df <- temp_sde(h_subj_df)

    }

    # then do weight (steps 1 - 7) ----

    w_df <- copy(df[df$param == "WEIGHTKG" & slog,])

    # order by age
    w_df <- w_df[order(age_years, id),]

    w_subj_keep <- rep("Include", nrow(w_df))
    w_subj_mean_sde <- rep(NA, nrow(w_df))
    names(w_subj_keep) <- names(w_subj_mean_sde) <-
      w_df$id

    w_subj_df <- copy(w_df)

    if (nrow(w_df) > 0){
      # add metric (m) and imperial (im) measurements
      # also keep original measurements (will overwrite for ease in functions)
      w_subj_df$meas_m <- w_subj_df$meas_im <- w_subj_df$meas_orig <-
        w_subj_df$measurement
      w_subj_df$meas_m[w_subj_df$param == "WEIGHTLBS"] <-
        w_subj_df$measurement[w_subj_df$param == "WEIGHTLBS"]/2.2046226
      w_subj_df$meas_im[w_subj_df$param == "WEIGHTKG"] <-
        w_subj_df$measurement[w_subj_df$param == "WEIGHTKG"]*2.2046226

      # convert age years to days -- already exists
      w_subj_df$age_days <- w_subj_df$agedays
    }

    # 1w, W BIV ----
    # 1w. remove biologically impossible weight records

    # if there are no valid heights, skip
    if (nrow(w_subj_df) > 0){
      step <- "Exclude-Adult-BIV"

      # overwrite measurement with metric (bivs are in metric)
      w_subj_df$measurement <- w_subj_df$meas_m

      criteria <- remove_biv(w_subj_df, "weight", biv_df)
      w_subj_keep[criteria] <- step

      w_subj_df <- w_subj_df[!criteria,]
    }

    # 2w, W repeated values ----
    # 2w. Identify repeated values (RV) -- values that are the same over different days
    step <- "repeated values" # no exclusions here

    w_subj_df <- identify_rv(w_subj_df)

    # 3w, W temp extraneous ----
    # 3w. Calculate temporary extraneous for exclusion in intermediate steps
    step <- "temp extraneous" # no exclusions here

    if (nrow(w_subj_df) > 0){
      # overwrite measurement with metric (step done in metric)
      w_subj_df$measurement <- w_subj_df$meas_m

      # adds the "extraneous" column, designating whether or not the row is
      # temporarily extraneous (not to be considered in the future)
      # also adds age days and "diff" (unused outside of this function)
      w_subj_df <- temp_sde(w_subj_df, ptype = "weight")

      # redo RVs just if any first RVs became extraneous
      w_subj_df <- redo_identify_rv(w_subj_df)
    }

    # 4w, W weight cap ----
    # 4w. excluding specified "weight caps" if a user specifies
    # step specified in function

    inc_df <- if (nrow(w_subj_df) > 0){
      # we only want to consider subjects without temp extraneous
      copy(w_subj_df[!w_subj_df$extraneous,])
    } else {
      w_subj_df
    }

    # only do this if there are at least two values, and if there's a weight cap
    # to evaluate
    if (nrow(inc_df) > 1 & weight_cap < Inf){
      # weight cap is evaluated with +/ .1 (for precision)
      wc_low <- round(weight_cap, 1) - .1
      wc_high <- round(weight_cap, 1) + .1

      # if all are weight cap, implausible
      is_wc <- check_between(inc_df$meas_m, wc_low, wc_high)
      criteria <- if (all(is_wc)){
        rep(TRUE, nrow(inc_df))
      } else {
        rep(FALSE, nrow(inc_df))
      }

      # if we have somewhere between 1 and < all weight caps, we can evaluate
      if (any(is_wc) & !all(is_wc) & nrow(inc_df) > 2){
        # calculate ewma (using metric)
        ewma_res <- ewma_dn(inc_df$age_days, inc_df$meas_m)
        # delta ewma
        dewma <- (inc_df$meas_m- ewma_res)
        colnames(dewma) <- paste0("d",colnames(ewma_res))

        # figure out weight difference between points
        wt_next <- c(NA, diff(inc_df$meas_m))
        wt_prev <- c(diff(inc_df$meas_m), NA)

        exc_wc <-
          is_wc &
          ((dewma$dewma.all > 50 &
              dewma$dewma.before > .9*dewma$dewma.all &
              dewma$dewma.all > .9*dewma$dewma.all) |
             (dewma$dewma.all < -50 &
                dewma$dewma.before < -.9*dewma$dewma.all &
                dewma$dewma.all < -.9*dewma$dewma.all)) &
          ((wt_prev > 50 & wt_next > 50) |
             (wt_prev < -50 & wt_next < -50))

        criteria <- exc_wc
        criteria[is.na(criteria)] <- FALSE
      }

      # implausible ids from the step
      impl_ids <- as.character(inc_df$id)[criteria]

      step <- if (all(is_wc)){
        "Exclude-Adult-Weight-Cap-Identical"
      } else {
        "Exclude-Adult-Weight-Cap"
      }

      # update and remove
      w_subj_keep[c(impl_ids)] <- step

      # don't get rid of extraneous just yet
      w_subj_df <- w_subj_df[!w_subj_df$id %in% impl_ids,]

      # reevaluate first RV
      w_subj_df <- identify_rv(w_subj_df)
      # reevaluate temp same day
      w_subj_df <- temp_sde(w_subj_df, ptype = "weight")
      # redo RVs just if any first RVs became extraneous
      w_subj_df <- redo_identify_rv(w_subj_df)
    }

    # 5w, W hundreds ----
    # 5w. when weight goes up/down by 100/200 kg/100-300 lbs -- is it valid?
    step <- "Exclude-Adult-Hundreds"

    inc_df <- if (nrow(w_subj_df) > 0){
      # we only want to consider subjects without temp extraneous and rvs
      copy(w_subj_df[!w_subj_df$extraneous & !w_subj_df$is_rv,])
    } else {
      w_subj_df
    }

    # only do this if there are at least two values
    if (nrow(inc_df) > 1){
      # calculate ewma (using metric)
      ewma_res <- ewma_dn(inc_df$age_days, inc_df$meas_m)
      # delta ewma
      dewma <- (inc_df$meas_m- ewma_res)
      colnames(dewma) <- paste0("d",colnames(ewma_res))

      criteria <- rep(FALSE, nrow(inc_df))
      for (mtype in c("m", "im")){
        # only test certain 100s for metric v imperial
        test_hundreds <- if(mtype == "m"){
          c(100, 200)
        } else {
          c(100, 200, 300)
        }

        # go through and test each combination of hundreds
        for (th in test_hundreds){
          criteria <- criteria |
            rem_hundreds(inc_df, dewma, paste0("meas_", mtype), th, "weight")
        }
      }

      # implausible ids from the step
      impl_ids <- as.character(inc_df$id)[criteria]
      # if it's a repeated value, we want to get rid of it as well
      rv_impl_ids <- as.character(
        w_subj_df$id[w_subj_df$meas_m %in% inc_df$meas_m[criteria &
                                                           inc_df$is_first_rv] &
                       w_subj_df$is_rv]
      )

      # update and remove
      w_subj_keep[impl_ids] <- step
      w_subj_keep[rv_impl_ids] <- paste0(step, "-RV")

      # don't get rid of extraneous just yet
      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids, rv_impl_ids),]

      # reevaluate temp same day
      w_subj_df <- temp_sde(w_subj_df, ptype = "weight")
      # redo RVs just if any first RVs became extraneous
      w_subj_df <- redo_identify_rv(w_subj_df)
    }

    # 6w, W unit errors ----
    # 6w. if a record recorded as metric should be imperial for interior values
    step <- "Exclude-Adult-Unit-Errors"

    inc_df <- if (nrow(w_subj_df) > 0){
      # we only want to consider subjects without temp extraneous and rvs
      copy(w_subj_df[!w_subj_df$extraneous & !w_subj_df$is_rv,])
    } else {
      w_subj_df
    }

    # only do this if there are at least two values
    if (nrow(inc_df) > 1){

      criteria <- rem_unit_errors(inc_df, ptype = "weight")

      # implausible ids from the step
      impl_ids <- as.character(inc_df$id)[criteria]
      # if it's a repeated value, we want to get rid of it as well
      rv_impl_ids <- as.character(
        w_subj_df$id[w_subj_df$meas_m %in% inc_df$meas_m[criteria &
                                                           inc_df$is_first_rv] &
                       w_subj_df$is_rv]
      )

      # update and remove
      w_subj_keep[impl_ids] <- step
      w_subj_keep[rv_impl_ids] <- paste0(step, "-RV")

      # don't get rid of extraneous just yet
      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids, rv_impl_ids),]

      # reevaluate temp same day
      w_subj_df <- temp_sde(w_subj_df, ptype = "weight")
      # redo RVs just if any first RVs became extraneous
      w_subj_df <- redo_identify_rv(w_subj_df)
    }

    # 7w, W transpositions ----
    # 7w. if a record should have swapped the 10s and 1s digits
    step <- "Exclude-Adult-Transpositions"

    inc_df <- if (nrow(w_subj_df) > 0){
      # we only want to consider subjects without temp extraneous and rvs
      copy(w_subj_df[!w_subj_df$extraneous & !w_subj_df$is_rv,])
    } else {
      w_subj_df
    }

    # only do this if there are at least two values
    if (nrow(inc_df) > 1 & length(inc_df$meas_m) > 2){

      criteria <- rem_transpositions(inc_df, ptype = "weight")

      # implausible ids from the step
      impl_ids <- as.character(inc_df$id)[criteria]
      # if it's a repeated value, we want to get rid of it as well
      rv_impl_ids <- as.character(
        w_subj_df$id[w_subj_df$meas_m %in% inc_df$meas_m[criteria &
                                                           inc_df$is_first_rv] &
                       w_subj_df$is_rv]
      )

      # update and remove
      w_subj_keep[impl_ids] <- step
      w_subj_keep[rv_impl_ids] <- paste0(step, "-RV")

      # don't get rid of extraneous just yet
      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids, rv_impl_ids),]

      # reevaluate temp same day
      w_subj_df <- temp_sde(w_subj_df, ptype = "weight")
      # redo RVs just if any first RVs became extraneous
      w_subj_df <- redo_identify_rv(w_subj_df)
    }

    # do step 8: swaps (both height and weight) ----
    # 8. checking whether heights and weights should have been swapped
    step <- "Exclude-Adult-Swapped-Measurements"

    if (nrow(h_subj_df) > 0 & nrow(w_subj_df) > 0){
      # we only want to consider subjects without temp extraneous
      h_inc_df <- copy(h_subj_df[!h_subj_df$extraneous,])
      w_inc_df <- copy(w_subj_df[!w_subj_df$extraneous,])

      # calculate ewma
      h_ewma_res <- ewma_dn(h_inc_df$age_days, h_inc_df$meas_m)
      w_ewma_res <- ewma_dn(w_inc_df$age_days, w_inc_df$meas_m)
      # add ewma to dataframes
      h_inc_df <- cbind(h_inc_df, h_ewma_res)
      w_inc_df <- cbind(w_inc_df, w_ewma_res)

      # possible removal of height/weights by bmi
      # h = height, w = weight
      comb_df <- comb_df_orig <-
        merge(h_inc_df, w_inc_df, by = "age_days", all = TRUE,
              suffixes = c(".h", ".w"))
      # remove ones that don't match
      comb_df <- comb_df[!(is.na(comb_df$id.h) | (is.na(comb_df$id.w))),]

      # you need at least three values for this -- can't evaluate next/previous for
      # first and last values
      if (nrow(comb_df) >= 3){
        # identify safe swaps
        # check exact conversion multiples of weight kg to lbs
        # .1 lbs for higher precision, .5 lbs for values with lower precision)
        noswap <- get_float_rem(comb_df$meas_m.w, 2.2046226 * .1) < .01
        noswap <- noswap |
          (get_float_rem(comb_df$meas_m.w, 2.2046226 * .5) < .1 &
             get_float_rem(comb_df$meas_m.w, .1) == 0)
        # check exact conversion multiples of height cm to in
        # .25 in for higher precision (.01), 1 in for lower precision (.1)
        noswap <- noswap |
          get_float_rem(comb_df$meas_m.h, 2.54 * .25) < .01
        noswap <- noswap |
          (get_float_rem(comb_df$meas_m.h, 2.54) < .1 &
             get_float_rem(comb_df$meas_m.h, .1) == 0)

        opposite_map <- c("h" = "w", "w" = "h")

        for (typ in c("h", "w")){
          # create "swaps"
          comb_df[,paste0("swap_",typ)] <-
            comb_df[, paste0("meas_m.", opposite_map[typ]), with = FALSE]

          # calculate difference between values -- DOING ABS HERE
          comb_df[,paste0("swap_prev_",typ)] <-
            c(NA, abs(unlist(
              comb_df[, paste0("swap_",typ), with = FALSE][2:nrow(comb_df)] -
                comb_df[, paste0("meas_m.",typ), with = FALSE][1:(nrow(comb_df)-1)]
            )))
          comb_df[,paste0("swap_next_",typ)] <-
            c(abs(unlist(
              comb_df[, paste0("swap_",typ), with = FALSE][1:(nrow(comb_df)-1)] -
                comb_df[, paste0("meas_m.",typ), with = FALSE][2:nrow(comb_df)]
              )), NA)

          # create dewmas
          for (dtyp in c("all", "before", "after")){
            comb_df[,paste0("dewma.",dtyp,".", typ)] <-
              comb_df[, paste0("meas_m.",typ), with = FALSE] -
              comb_df[,paste0("ewma.",dtyp,".", typ), with = FALSE]

            # dewma for swaps
            comb_df[,paste0("dewma.",dtyp,".swap_", typ)] <-
              comb_df[, paste0("swap_",typ), with = FALSE] -
              comb_df[,paste0("ewma.",dtyp,".", typ), with = FALSE]
          }
        }

        # criteria check - 4 parts
        # check if dewma and one or more diff of recorded values is high
        wt_far <-
          (comb_df$dewma.all.w > 30 &
             comb_df$dewma.before.w > (.9*comb_df$dewma.all.w) &
             comb_df$dewma.after.w > (.9*comb_df$dewma.all.w)) |
          (comb_df$dewma.all.w < -30 &
             comb_df$dewma.before.w < (-.9*comb_df$dewma.all.w) &
             comb_df$dewma.after.w < (-.9*comb_df$dewma.all.w))
        ht_far <-
          (comb_df$dewma.all.h > 12.7 &
             comb_df$dewma.before.h > (.9*comb_df$dewma.all.h) &
             comb_df$dewma.after.h > (.9*comb_df$dewma.all.h)) |
          (comb_df$dewma.all.h < -12.7 &
             comb_df$dewma.before.h < (-.9*comb_df$dewma.all.h) &
             comb_df$dewma.after.h < (-.9*comb_df$dewma.all.h))

        # check if the swapped value is close
        swap_wt_close <-
          (abs(comb_df$dewma.all.swap_w) <= 2 |
             abs(comb_df$swap_prev_w) <= 2 |
             abs(comb_df$swap_next_w) <= 2)
        swap_ht_close <-
          (abs(comb_df$dewma.all.swap_h) <= 2.54 |
             abs(comb_df$swap_prev_h) <= 2.54 |
             abs(comb_df$swap_next_h) <= 2.54)

        exc_swap <- wt_far & ht_far & swap_wt_close & swap_ht_close
        # replace noswaps
        exc_swap[noswap] <- FALSE
        criteria <- exc_swap
        criteria[is.na(criteria)] <- FALSE

        # implausible ids from the step
        impl_ids <- as.character(comb_df$id.w)[criteria]
        # do not remove repeated values

        # update and remove -- weight
        w_subj_keep[impl_ids] <- step

        # don't get rid of extraneous just yet -- shouldn't be in
        w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids) |
                                 !w_subj_df$id %in% comb_df$id.w,]

        # update and remove -- height
        h_subj_keep[as.character(comb_df$id.h)][criteria] <- step

        # don't get rid of extraneous just yet
        h_subj_df <- h_subj_df[h_subj_df$id %in% comb_df$id.h[!criteria] |
                                 h_subj_df$extraneous |
                                 !h_subj_df$id %in% comb_df$id.h,]

        # reevaluate temp same day -- don't need to reevaluate if nothing has
        # changed
        if (any(criteria)){
          h_subj_df <- temp_sde(h_subj_df)
          w_subj_df <- temp_sde(w_subj_df, ptype = "weight")
          # redo RVs just if any first RVs became extraneous
          w_subj_df <- redo_identify_rv(w_subj_df)
        }
      }
    }

    # height steps (steps 9-11) ----
    # 9h, H same day extraneous ----
    # 9h. reevaluate same day extraneous for exclusion
    # step specified inside

    # don't do this if there aren't any non extraneous for the subject
    if (nrow(h_subj_df) > 0 & any(h_subj_df$extraneous)){
      step <- "Exclude-Adult-Identical-Same-Day"

      # identify duplicate days
      dup_days <- unique(h_subj_df$age_days[h_subj_df$extraneous])

      # first, exclude identical measurements on the same day
      ide_ids <- c() # where we keep the identical ones
      for (dd in dup_days){
        # count amount of unique values
        s_df <- copy(h_subj_df[h_subj_df$age_days == dd,])
        ide_tab <- table(s_df$meas_m)
        if (any(ide_tab > 1)){
          # for each identical, keep only the first one, by id
          # (ordered initially)
          ide_ids <- c(ide_ids, s_df$id[
            as.character(s_df$meas_m) %in% names(ide_tab[ide_tab > 1])
            ][duplicated(
            s_df$meas_m[
              as.character(s_df$meas_m) %in% names(ide_tab[ide_tab > 1])
            ]
          )])
        }
      }
      criteria <- h_subj_df$id %in% ide_ids

      # we're going to update h_subj_df before moving on to the rest of this
      # subject
      h_subj_keep[as.character(h_subj_df$id)][criteria] <- step

      h_subj_df <- h_subj_df[!criteria,]

      if (any(criteria)){
        # reevaluate temp same day
        h_subj_df <- temp_sde(h_subj_df)
        # identify duplicate days
        dup_days <- unique(h_subj_df$age_days[h_subj_df$extraneous])
      }

      step <- "Exclude-Adult-Extraneous-Same-Day"
      # now the rest!

      # check if heights on duplicate days are trivially the same, keep both,
      # use mean for all of those
      rem_ids <- c()
      for (dd in dup_days){
        s_df <- copy(h_subj_df[h_subj_df$age_days == dd,])
        sde_range <- max(s_df$meas_m) - min(s_df$meas_m)
        if (sde_range < 2.541){ # 1 inch +eps
          # update imperial
          h_subj_df$meas_im[h_subj_df$age_days == dd] <-
            mean(s_df$meas_m)/2.54

          h_subj_df$mean_sde[h_subj_df$age_days == dd] <-
            h_subj_df$meas_m[h_subj_df$age_days == dd] <-
            mean(s_df$meas_m)

          # remove all except the first by id
          rem_ids <- c(rem_ids, h_subj_df$id[h_subj_df$age_days == dd][
            duplicated(h_subj_df$meas_m[h_subj_df$age_days == dd])])
        }
      }
      criteria <- h_subj_df$id %in% rem_ids

      # we're going to update h_subj_df before moving on to the rest of this
      # subject
      h_subj_keep[as.character(h_subj_df$id)][criteria] <- step
      # also update mean sde
      h_subj_mean_sde[as.character(h_subj_df$id)] <- h_subj_df$mean_sde

      h_subj_df <- h_subj_df[!criteria,]

      if (any(criteria)){
        # reevaluate temp same day
        h_subj_df <- temp_sde(h_subj_df)
        # identify duplicate days
        dup_days <- unique(h_subj_df$age_days[h_subj_df$extraneous])
      }

      # next, calculate the duplicate ratio -- what proportion of days are
      # duplicated
      dup_ratio <- mean(!is.na(h_subj_df$diff))
      # also check whether or not any same-days are adjacent -- need 4 at least
      # rolling windows of day differences -- we are looking for 0,x,0
      if (nrow(h_subj_df) > 3){
        roll <- embed(diff(h_subj_df$age_days), 3)
        adjacent <- any(sapply(1:nrow(roll), function(x){
          all(c(roll[x,1] == 0, roll[x,2] != 0, roll[x,3] == 0))
        }))
      } else {
        adjacent <- FALSE
      }

      # if dup ratio is too high, or any adjacent same days, we exclude all
      # same day extraneous
      criteria <-
        if ((dup_ratio > .25) | adjacent){
          !is.na(h_subj_df$diff)
        } else {
          rep(FALSE, nrow(h_subj_df))
        }

      # if criteria didn't catch it, we now compare with medians
      if (!any(criteria) & any(h_subj_df$extraneous)){
        med <- median(h_subj_df$meas_m[
          !h_subj_df$age_days %in% dup_days
        ])
        h_subj_df$absdiff <- NA
        h_subj_df$absdiff[h_subj_df$age_days %in% dup_days] <-
          abs(med - h_subj_df$meas_m[h_subj_df$age_days %in% dup_days])

        # go through each duplicate day and figure out criteria
        # keep the first one that satisfies this criteria
        mm <- h_subj_df$meas_m # shorthand
        rem_ids <- c()
        for (dd in dup_days){
          # find prior and next index
          curr_ind <- which(h_subj_df$age_days == dd)[1]
          p_ind <- which(h_subj_df$age_days == dd)[1] - 1
          n_ind <- which(h_subj_df$age_days == dd)[
            sum(h_subj_df$age_days == dd)] + 1
          # modify for first and last
          if (p_ind < 0){
            p_ind <- 1
          }
          if (n_ind > nrow(h_subj_df)){
            n_ind <- nrow(h_subj_df)
          }

          # are before, after, median close enough? 1 inch
          compare <-
            abs(mm[curr_ind] - mm[p_ind]) < 2.541 &
            abs(mm[curr_ind] - mm[n_ind]) < 2.541 &
            h_subj_df$absdiff[h_subj_df$age_days == dd] < 2.541

          comp_id <-
            if (sum(compare) == 1){
              h_subj_df$id[which(compare)]
            } else {
              ""
            }
          # the rest are to be excluded
          rem_ids <- c(rem_ids,
                       h_subj_df$id[h_subj_df$age_days == dd &
                                      h_subj_df$id != comp_id]
          )
        }
        # update criteria
        criteria <- h_subj_df$id %in% rem_ids
      }

      h_subj_keep[as.character(h_subj_df$id)][criteria] <- step

      h_subj_df <- h_subj_df[!criteria,]
    }

    # 10hab, H distinct values ----
    # 10ha-b. evaluate based on numbers of distinct values
    # steps specified in specific caveat

    # get number of distinct values
    num_distinct <- length(unique(h_subj_df$meas_m))

    # preallocate criteria
    criteria <- rep(FALSE, nrow(h_subj_df))

    # preallocate values used for mean ht later
    # pairs
    pairhtloss <- pairhtgain <- FALSE
    # 3d gain/loss
    glist_loss <- glist_gain <- list()
    origexc <- g3_g2_check <- g3_g1_check <- g2_g1_check <- TRUE

    # go through each type of exclusion
    if (num_distinct == 2){
      # 10ha, H distinct pairs ----
      step <- "Exclude-Adult-Distinct-Pairs"

      # identify "height 1 and 2" and their corresponding ages
      ht_1 <- unique(h_subj_df$meas_m[order(h_subj_df$age_days)])[1]
      ht_2 <- unique(h_subj_df$meas_m[order(h_subj_df$age_days)])[2]
      ht_1_log <- h_subj_df$meas_m == ht_1
      ht_2_log <- h_subj_df$meas_m == ht_2
      ht_1_imp <- unique(h_subj_df$meas_im[ht_1_log])
      ht_2_imp <- unique(h_subj_df$meas_im[ht_2_log])
      ht_1_ageyears <- h_subj_df$age_years[ht_1_log]
      ht_2_ageyears <- h_subj_df$age_years[ht_2_log]

      # check if pairs outside two inch range
      # imperial will also be unique
      exc_2d <- round(abs(ht_1_imp - ht_2_imp), 2) > 2

      # only if outside the range
      if (exc_2d){
        # calculate first and second ageyears
        ageyears1 <- sort(ht_1_ageyears)[1]
        ageyears2 <- sort(ht_2_ageyears)[1]

        # calculate "height allow": allowed growth for ages <= 25
        if (ageyears1 <= 25){
          htcompare <- ifelse(ageyears2 > 25, 25, ageyears2)

          # using short circuiting
          hta <-
            if ((ageyears2 - ageyears1) < 2){
              ht_allow(15.5, ageyears1, htcompare)
            } else if ((ageyears2 - ageyears1) <= 3){
              ht_allow(13, ageyears1, htcompare)
            } else if ((ageyears2 - ageyears1) > 3){
              ht_allow(12, ageyears1, htcompare)
            }

          # potential reallow: height gain for youth gain in height
          pairhtgain <-
            (ht_2_imp - ht_1_imp) <= (hta + 2) &
            (ht_2_imp - ht_1_imp) > 0 &
            min(ht_2_ageyears) > max(ht_1_ageyears)
        } else {
          # this is not a check to apply if they're outside the age range
          pairhtgain <- FALSE
        }

        # potential reallow: falls are <= 3 (+2) in or <= 5 (+2) for ageyears > 50
        pairhtloss <-
          (ht_1 > ht_2) &
          ((ht_1_imp - ht_2_imp) <= (3+2) |
             ((ht_1_imp - ht_2_imp) <= (5+2) & ageyears2 >= 50)) &
          min(ht_2_ageyears) > max(ht_1_ageyears)

        # exclude if there are no reallow for loss and gain
        exc_pairs <- !(pairhtloss | pairhtgain)
        if (exc_pairs){
          criteria <- rep(TRUE, nrow(h_subj_df))
        }

        # generate keep ratios for height 1 and 2
        keepht1 <- sum(ht_1_log) >= (sum(ht_2_log)*(4/3))
        keepht2 <- sum(ht_2_log) >= (sum(ht_1_log)*(4/3))

        # reallow if the ratios are satisfied
        if (keepht1){
          criteria[ht_1_log] <- FALSE
        }
        if (keepht2){
          criteria[ht_2_log] <- FALSE
        }
      }

    } else if (num_distinct >= 3){
      # 10ha, H distinct 3 or more ----
      step <- "Exclude-Adult-Distinct-3-Or-More"

      h_subj_df <- h_subj_df[order(h_subj_df$age_years),]
      # create w2 (w/in 2) and o2  (outside 2) groups
      w2_groups <- lapply(unique(h_subj_df$meas_m), function(x){
        h_subj_df$meas_m[check_between(h_subj_df$meas_m, x, x+5.081)]
      })
      o2_groups <- lapply(unique(h_subj_df$meas_m), function(x){
        h_subj_df$meas_m[!check_between(h_subj_df$meas_m, x, x+5.081)]
      })
      # will coerce to character
      names(w2_groups) <- names(o2_groups) <- unique(h_subj_df$meas_m)

      # calculate keep ratio (if nothing's outside, it's infinite = good)
      ratio_w2o2 <- sapply(unique(h_subj_df$meas_m), function(x){
        length(w2_groups[[as.character(x)]])/length(o2_groups[[as.character(x)]])
      })
      # calculate which ones are ok (1.5 x more in w2 than o2)
      okratio <- ratio_w2o2 > 3/2

      # if there are multiple ok, we need some tiebreakers
      if (sum(okratio) > 1){
        # w2 groups for consideration
        consider_w2 <- unique(h_subj_df$meas_m)[okratio]

        # create the total and distinct score
        score <- sapply(consider_w2, function(x){
          length(w2_groups[[as.character(x)]]) +
            .5*length(unique(w2_groups[[as.character(x)]]))
        })

        best_scores <- which(score == max(score))
        if (length(best_scores) > 1){
          # if there are multiple best scores, we move to our next tiebreaker
          # calculating mean absolute distance between w2 and o2
          mean_abs_dist <- sapply(consider_w2[best_scores], function(x){
            mean(abs(mean(w2_groups[[as.character(x)]]) -
                       o2_groups[[as.character(x)]]))
          })
          # if NA, means that there is nothing in the o2 groups and it is
          # indeed the best
          mean_abs_dist[is.na(mean_abs_dist)] <- Inf

          # this will choose the first if there's a tiebreaker
          best_w2 <- consider_w2[best_scores][which.max(mean_abs_dist)]
        } else {
          best_w2 <- consider_w2[best_scores]
        }
      } else if (sum(okratio) == 1){
        best_w2 <- unique(h_subj_df$meas_m)[okratio]
      } else {
        best_w2 <- "none"
      }

      if (best_w2 != "none"){ # one of the w2s was selected -- o2s are out
        criteria[!h_subj_df$meas_m %in% w2_groups[[as.character(best_w2)]]] <- TRUE
      } else { # none -- all are false
        criteria <- rep(TRUE, nrow(h_subj_df))
      }

      # 3D loss ----

      # if anything was excluded, we want to possibly reinclude with gain/loss code
      if (any(criteria)){
        # create 3 groups:
        # g1: ht 1 and all within 2 inches
        # g2: first height outside g1 + 2 inches
        # g3: first height outside g2 + 2 inches
        # ^ if any heights out, we want to go with what we had originally

        gtotal_loss <- ht_change_groups(h_subj_df, 3, type = "loss")
        glist <- glist_loss <- gtotal_loss$meas # groups with measurements
        galist <- gtotal_loss$age # groups with ages, in years

        # if there are any outside these groups, we don't attempt to fix
        # we need all of them to be in some group
        # if (all(Reduce("|", glist))){
        if (length(glist) > 0 & length(glist) <= 3){
          # create a mean ht and a min age for each group
          mean_ht <- sapply(glist, function(g){
            suppressWarnings(mean(g))
          })
          min_age <- sapply(galist, function(ga){
            suppressWarnings(min(ga))
          })

          # check g2 v g1 -- true indicates use the original exclusions
          g2_g1_check <-
            !(if (!is.na(mean_ht[2])){
              (mean_ht[2] - mean_ht[1]) < 0 &
                ((min_age[2] < 50 &
                    (mean_ht[2] - mean_ht[1]) > ((-5 * 2.54) +.001)) |
                   (min_age[2] >= 50 &
                      (mean_ht[2] - mean_ht[1]) > ((-7 * 2.54) +.001)))
            } else {
              FALSE
            })

          # check g3 v g2 -- true indicates use the original exclusions
          g3_g2_check <-
            if (!is.na(mean_ht[2]) & !is.na(mean_ht[3])){
              (mean_ht[3] - mean_ht[2]) > 0 |
                (min_age[3] < 50 &
                   (mean_ht[3] - mean_ht[2]) < ((-5 * 2.54) +.001)) |
                (min_age[3] >= 50 &
                   (mean_ht[3] - mean_ht[2]) < ((-7 * 2.54) +.001))
            } else {
              FALSE
            }

          # check g3 v g1 -- true indicates use the original exclusions
          g3_g1_check <-
            if (!is.na(mean_ht[2]) & !is.na(mean_ht[3])){
              (min_age[3] < 50 &
                 (mean_ht[3] - mean_ht[1]) < ((-6 * 2.54) +.001)) |
                (min_age[2] < 50 & min_age[3] >= 50 &
                   (mean_ht[3] - mean_ht[1]) < ((-8 * 2.54) +.001)) |
                (min_age[2] >= 50 &
                   (mean_ht[3] - mean_ht[1]) < ((-9 * 2.54) +.001))
            } else {
              FALSE
            }

          # if all are false, reinclude all?
          if (all(!c(g3_g2_check, g3_g1_check, g2_g1_check))){
            criteria <- rep(FALSE, nrow(h_subj_df))
          }
        }

      }

      # 3D gain ----

      # account for gain -- only for subjects with < 25 years
      if (any(criteria) & min(h_subj_df$age_years) < 25){
        # create 6 groups:
        # g1: ht 1 and all within 2 inches
        # g2: first height outside g1 + 2 inches ....
        # ^ if any heights out, we want to go with what we had originally

        gtotal_gain <- ht_change_groups(h_subj_df, 6, type = "gain")
        glist <- glist_gain <- gtotal_gain$meas # groups with measurements
        galist <- gtotal_gain$age # groups with ages

        # if there are any outside these groups, we don't attempt to fix
        # we need all of them to be in some group
        if (length(glist) > 0 & length(glist) <= 6){
          # create a mean ht and a min age for each group
          mean_ht <- sapply(glist, function(g){
            suppressWarnings(mean(g))
          })
          min_age <- sapply(galist, function(ga){
            suppressWarnings(min(ga))
          })

          # preallocating on whether or not we want to go by original exclusion
          origexc <- FALSE
          # first, compare to the age before
          origexc <- origexc |
            ht_3d_growth_compare(mean_ht, min_age, glist, compare = "before")
          # then, compare to the first age group
          origexc <- origexc |
            ht_3d_growth_compare(mean_ht, min_age, glist, compare = "first")

          # if all are false, reinclude all?
          if (!origexc){
            criteria <- rep(FALSE, nrow(h_subj_df))
          }
        }
      }

    }

    # we're going to do some processing to get the growth groups and loss groups
    # for eventual output
    loss_groups <- gain_groups <- rep(NA, length(h_subj_keep))
    names(loss_groups) <- names(gain_groups) <- names(h_subj_keep)

    hold_groups <-
      if (length(glist_loss) > 0){
        unlist(lapply(1:length(glist_loss), function(x){
          setNames(rep(x, length(glist_loss[[x]])), names(glist_loss[[x]]))
        }))
      } else {
        loss_groups
      }
    loss_groups[names(hold_groups)] <- hold_groups

    hold_groups <-
      if (length(glist_gain) > 0){
        unlist(lapply(1:length(glist_gain), function(x){
          setNames(rep(x, length(glist_gain[[x]])), names(glist_gain[[x]]))
        }))
      } else {
        c()
      }
    gain_groups[names(hold_groups)] <- hold_groups

    # update and remove
    h_subj_keep[as.character(h_subj_df$id)][criteria] <- step

    h_subj_df <- h_subj_df[!criteria,]

    # 11h, H mean height ----
    # 11h. If there are any heights left, generate a stable height to be used
    # for analysis and output.

    # if there are any heights left
    meanht <- setNames(rep(NA, length(h_subj_keep)), names(h_subj_keep))
    if (nrow(h_subj_df) > 0){
      # we're relying on variables created above
      # if pairhtloss/gain are true, we're dealing with pairs
      hold_meanht <-
        if (pairhtloss | pairhtgain){
          h_subj_df$meas_orig
        } else if (length(glist_loss) > 0 &
                   all(!c(g3_g2_check, g3_g1_check, g2_g1_check))){
          # has loss -- everything was reincluded
          # we were in a 3d scenario -- we repeat mean for every one in the group
          unlist(lapply(glist_loss, function(x){
              rep(mean(h_subj_df$meas_orig[h_subj_df$id %in% names(x)]),
                  length(x))
          }))
        } else if (length(glist_gain) > 0 &
                   !origexc){
          # has gain -- everything was reincluded
          # we were in a 3d scenario -- we repeat mean for every one in the group
          unlist(lapply(glist_gain, function(x){
            rep(mean(h_subj_df$meas_orig[h_subj_df$id %in% names(x)]),
                length(x))
          }))
        } else {
          # it's just the mean of the remaining heights
          rep(mean(h_subj_df$meas_orig), nrow(h_subj_df))
        }
      # give mean height id names to correspond them back later
      names(hold_meanht) <- h_subj_df$id
      meanht[names(hold_meanht)] <- hold_meanht
    }

    # finish weight steps (steps 9-12) ----

    # 9w, W extreme EWMA ----
    # 9w. mark extreme values using EWMA method
    step <- "Exclude-Adult-EWMA-Extreme"

    if (nrow(w_subj_df) > 0){
      # first, remove ewma without temp extraneous and repeated values
      # we only want to consider subjects without temp extraneous
      inc_df_first <- copy(w_subj_df[!w_subj_df$extraneous & !w_subj_df$is_rv,])
      # only do this if there are at least two values
      criteria_first <-
        if (nrow(inc_df_first) > 1){
          remove_ewma_wt(inc_df_first)
        } else {
          rep(FALSE, nrow(inc_df_first))
        }

      # then, remove ewma just without temp extraneous
      inc_df_rv <- copy(w_subj_df[
        !w_subj_df$extraneous & !w_subj_df$id %in% inc_df_first$id[criteria_first],
      ])
      # only do this if there are at least two UNIQUE values
      criteria_rv <-
        if (length(unique(inc_df_rv)) > 1){
          remove_ewma_wt(inc_df_rv)
        } else {
          rep(FALSE, nrow(inc_df_rv))
        }

      # implausible ids from the step
      impl_ids <- c(
        as.character(inc_df_first$id)[criteria_first],
        as.character(inc_df_rv$id)[criteria_rv]
      )
      # if it's a repeated value, we want to get rid of it as well
      rv_impl_ids <- as.character(
        w_subj_df$id[w_subj_df$meas_m %in%
                       inc_df_first$meas_m[criteria_first &
                                             inc_df_first$is_first_rv] &
                       w_subj_df$is_rv],
        w_subj_df$id[w_subj_df$meas_m %in%
                       inc_df_rv$meas_m[criteria_rv &
                                          inc_df_rv$is_first_rv] &
                       w_subj_df$is_rv]
      )

      # update and remove
      w_subj_keep[impl_ids] <- step
      w_subj_keep[rv_impl_ids] <- paste0(step, "-RV")

      # don't get rid of extraneous just yet
      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids, rv_impl_ids),]

      # don't need to do this if we don't find anything to remove
      if (length(c(impl_ids, rv_impl_ids)) > 0){
        # reevaluate first rvs
        w_subj_df <- identify_rv(w_subj_df)

        # don't need to reevaluate temp same day -- next step we get rid of them
      }
    }

    # 10w, W same day extraneous ----
    # 10w. check and remove same day extraneous values

    # includes all RV

    # don't do this if there aren't any non extraneous for the subject
    if (nrow(w_subj_df) > 0 & any(w_subj_df$extraneous)){
      step <- "Exclude-Adult-Identical-Same-Day"

      # identify duplicate days
      dup_days <- unique(w_subj_df$age_days[w_subj_df$extraneous])

      # first, exclude identical measurements on the same day
      ide_ids <- c() # where we keep the identical ones
      for (dd in dup_days){
        # count amount of unique values
        s_df <- copy(w_subj_df[w_subj_df$age_days == dd,])
        ide_tab <- table(s_df$meas_m)
        if (any(ide_tab > 1)){
          # for each identical, keep only the first one, by id
          # (ordered initially)
          ide_ids <- c(ide_ids, s_df$id[
            as.character(s_df$meas_m) %in% names(ide_tab[ide_tab > 1])
          ][duplicated(
            s_df$meas_m[
              as.character(s_df$meas_m) %in% names(ide_tab[ide_tab > 1])
            ]
          )])
        }
      }
      criteria <- w_subj_df$id %in% ide_ids

      # we're going to update h_subj_df before moving on to the rest of this
      # subject
      w_subj_keep[as.character(w_subj_df$id)][criteria] <- step

      w_subj_df <- w_subj_df[!criteria,]

      if (any(criteria)){
        # reevaluate temp same day
        w_subj_df <- temp_sde(w_subj_df)
        # identify duplicate days
        dup_days <- unique(w_subj_df$age_days[w_subj_df$extraneous])
      }

      step <- "Exclude-Adult-Extraneous-Same-Day"
      # now the rest!

      # check if heights on duplicate days are trivially the same, keep both,
      # use mean for all of those
      rem_ids <- c()
      for (dd in dup_days){
        s_df <- copy(w_subj_df[w_subj_df$age_days == dd,])
        sde_range <- max(s_df$meas_m) - min(s_df$meas_m)
        if (sde_range < 1.001){ # 1 pound
          # update imperial
          w_subj_df$meas_im[w_subj_df$age_days == dd] <-
            mean(s_df$meas_m)*2.2046226

          w_subj_df$mean_sde[w_subj_df$age_days == dd] <-
            w_subj_df$meas_m[w_subj_df$age_days == dd] <-
            mean(s_df$meas_m)

          # remove all except the first by id
          rem_ids <- c(rem_ids, w_subj_df$id[w_subj_df$age_days == dd][
            duplicated(w_subj_df$meas_m[w_subj_df$age_days == dd])])
        }
      }
      criteria <- w_subj_df$id %in% rem_ids

      # we're going to update w_subj_df before moving on to the rest of this
      # subject
      w_subj_keep[as.character(w_subj_df$id)][criteria] <- step
      # also update mean sde
      w_subj_mean_sde[as.character(w_subj_df$id)] <- w_subj_df$mean_sde

      w_subj_df <- w_subj_df[!criteria,]

      if (any(criteria)){
        # reevaluate temp same day
        w_subj_df <- temp_sde(w_subj_df)
        # identify duplicate days
        dup_days <- unique(w_subj_df$age_days[w_subj_df$extraneous])
      }

      # next, calculate the duplicate ratio -- what proportion of days are
      # duplicated
      dup_ratio <- mean(!is.na(w_subj_df$diff))
      # also check whether or not any same-days are adjacent -- need 4 at least
      # rolling windows of day differences -- we are looking for 0,x,0
      if (nrow(w_subj_df) > 3){
        roll <- embed(diff(w_subj_df$age_days), 3)
        adjacent <- any(sapply(1:nrow(roll), function(x){
          all(c(roll[x,1] == 0, roll[x,2] != 0, roll[x,3] == 0))
        }))
      } else {
        adjacent <- FALSE
      }

      # if dup ratio is too high, or any adjacent same days, we exclude all
      # same day extraneous
      criteria <-
        if ((dup_ratio > .25) | adjacent){
          !is.na(w_subj_df$diff)
        } else {
          rep(FALSE, nrow(w_subj_df))
        }

      # if criteria didn't catch it, we now compare with medians
      if (!any(criteria) & any(w_subj_df$extraneous)){
        # calculate ewma
        # calculate ewma (using metric)
        ewma_res <- ewma_dn(w_subj_df$age_days, w_subj_df$meas_m,
                            ewma.adjacent = FALSE)
        # delta ewma
        dewma <- (w_subj_df$meas_m- ewma_res)
        colnames(dewma) <- paste0("d",colnames(ewma_res))

        # go through each duplicate day and figure out criteria
        # keep the first one that satisfies this criteria
        rem_ids <- c()
        for (dd in dup_days){
          compare <- abs(dewma$dewma.all[w_subj_df$age_days == dd]) <= 2.001
          less_5 <- abs(dewma$dewma.all[w_subj_df$age_days == dd]) < 5

          comp_id <-
            if (sum(compare & less_5) == 1){
              w_subj_df$id[which(compare)]
            } else {
              ""
            }
          # the rest are to be excluded
          rem_ids <- c(rem_ids,
                       w_subj_df$id[w_subj_df$age_days == dd &
                                      w_subj_df$id != comp_id]
          )
        }
        # update criteria
        criteria <- w_subj_df$id %in% rem_ids
      }

      w_subj_keep[as.character(w_subj_df$id)][criteria] <- step

      w_subj_df <- w_subj_df[!criteria,]

      # redo RVs just if any first RVs became extraneous
      w_subj_df <- identify_rv(w_subj_df)
    }

    # 11w, W distinct/moderate EWMA ----
    # 11w. Check remaining weight values depending on amount of values within
    # EWMA

    # check if there are two distinct, and if so, are they ordered?
    pair_distinct <- if (length(unique(w_subj_df$meas_m)) == 2){
      if (
        max(w_subj_df$age_years[w_subj_df$meas_m == unique(w_subj_df$meas_m)[1]]) <
        min(w_subj_df$age_years[w_subj_df$meas_m == unique(w_subj_df$meas_m)[2]])
      ){
        TRUE
      } else {
        FALSE
      }
    } else {
      FALSE
    }

    # for pairs: use all RV
    # for >=2, use first RV, remove, then use all RV, remove

    # first, identify implausible for distinct pairs
    if (pair_distinct){
      # 11wa, W distinct ordered pairs ----
      # 11wa. Check pairs (2 distinct ordered values), where all first values are
      # of ages less than second values
      step <- "Exclude-Adult-Distinct-Ordered-Pairs"

      # use all RV

      # get first and last weight -- already ordered by age
      # all first are before all last, so these will be the two distinct values
      wt_first <- w_subj_df$meas_m[1]
      wt_last <- w_subj_df$meas_m[nrow(w_subj_df)]
      # weight difference at beginning and end of series
      wt_diff <- wt_last - wt_first

      # get age difference, in years
      ageyears_diff <-
        min(w_subj_df$age_years[w_subj_df$meas_m == unique(w_subj_df$meas_m)[2]]) -
        max(w_subj_df$age_years[w_subj_df$meas_m == unique(w_subj_df$meas_m)[1]])

      # compute "weight allow" how much change is allowed over time
      wta <- 4 + 18*log(1 + (ageyears_diff*12))
      # cap at 60
      wta[wta > 60] <- 60

      # is the difference outside the allowed weight change?
      exc_pairs <- abs(wt_diff) > wta

      # another criteria: consider the weight ratio
      wt_perc <-
        if (wt_last/wt_first < 1){
          wt_last/wt_first
        } else {
          wt_first/wt_last
        }

      # set a limit for wts to be percent of other wts, focused on lower wts
      perc_limit <- .7
      if (all(w_subj_df$meas_m > 45)){
        perc_limit <- .4
      }

      # exclude if outside the limit (exc pairs will be a vector)
      exc_pairs <- exc_pairs | wt_perc < perc_limit

      criteria <- exc_pairs

      # do removal
      # implausible ids from the step
      impl_ids <- as.character(w_subj_df$id)[criteria]

      # update and remove
      w_subj_keep[c(impl_ids)] <- step

      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids),]

    } else if (length(unique(w_subj_df$meas_m)) >= 2){ # (but not ordered)
      # 11wb, W moderate EWMA ----
      # 11wb. Check all other types, using a more moderate EWMA cutoff and other
      # criteria
      step <- "Exclude-Adult-EWMA-Moderate"

      # DO TWO STEPS: FIRST RVS, REMOVE, THEN DO ALL RVS

      # first, do only first rvs
      inc_df <- copy(w_subj_df[!w_subj_df$is_rv,])

      criteria <- remove_mod_ewma_wt(inc_df)

      # implausible ids from the step
      impl_ids <- as.character(inc_df$id)[criteria]

      # update and remove
      w_subj_keep[impl_ids] <- step

      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids),]

      # then, do all RVs
      inc_df <- copy(w_subj_df)

      criteria <- remove_mod_ewma_wt(inc_df)

      # implausible ids from the step
      impl_ids <- as.character(inc_df$id)[criteria]

      # update and remove
      w_subj_keep[impl_ids] <- step

      w_subj_df <- w_subj_df[!w_subj_df$id %in% c(impl_ids),]
    }

    # no need to do rvs -- no need in the future

    # 12w, W weight cap influence ----
    # 12w. Check, if a weight cap is specified, if there are any remaining after
    # the preceding steps -- they may have some influence on remaining data.
    step <- "Exclude-Adult-Possibly-Impacted-By-Weight-Cap"

    if (weight_cap < Inf & nrow(w_subj_df) > 0){
      # weight cap is evaluated with +/ .1 (for precision)
      wc_low <- round(weight_cap, 1) - .1
      wc_high <- round(weight_cap, 1) + .1
      is_wc <- check_between(inc_df$meas_m, wc_low, wc_high)

      # if there are any weight caps left, we need to propagate out
      if (any(is_wc)){
        # if there are any left, we need to mark everything for exclusion
        criteria <- w_subj_keep == "Include"

        # remove based on above criteria
        rem_ids <- names(w_subj_keep)[criteria]

        # update and remove
        w_subj_keep[rem_ids] <- step

        # remove everything from weight, we don't use it anymore
        w_subj_df <- w_subj_df[c(),]
      }
    }

    # 13, distinct 1 ----
    # 13.  determine if single values in height/weight fall within BMI criteria
    step <- "Exclude-Adult-Distinct-Single"

    # only do this if there's 1 distinct in either height or weight
    if (length(unique(h_subj_df$meas_m)) == 1 |
        length(unique(w_subj_df$meas_m)) == 1){
      # match and create BMIs

      if (nrow(h_subj_df) > 0 & nrow(w_subj_df) > 0){
        # possible removal of height/weights by bmi
        # h = height, w = weight
        comb_df <- comb_df_orig <-
          merge(h_subj_df, w_subj_df, by = "age_days", all = TRUE,
                suffixes = c(".h", ".w"))
        # remove ones that don't match
        comb_df <- comb_df[!(is.na(comb_df$id.h) | (is.na(comb_df$id.w))),]
      } else {
        comb_df <- data.table()
      }

      # must be matches
      if (nrow(comb_df) > 0){
        # calculate bmi
        comb_df$bmi <- comb_df$meas_m.w/((comb_df$meas_m.h/100)^2)

        # TODO: DATAFRAME OF THESE 8 PARAMETERS
        # check if it's in a more common range, then have wider limits
        h_exc_btw <-
          check_between(comb_df$bmi, 16, 60) &
          (comb_df$meas_m.h < 88 | comb_df$meas_m.h > 231) &
          length(unique(h_subj_df$meas_m)) == 1
        w_exc_btw <-
          check_between(comb_df$bmi, 16, 60) &
          (comb_df$meas_m.w < 30 | comb_df$meas_m.w > 271) &
          length(unique(w_subj_df$meas_m)) == 1

        # if there's a bmi for the subject, but outside the limits, AND none
        # within the limits AND it's 1D, we exclude
        h_bmi_out <-
          all(!check_between(comb_df$bmi, 16, 60)) &
          !check_between(comb_df$meas_m.h, 139, 206) &
          length(unique(h_subj_df$meas_m)) == 1
        w_bmi_out <-
          all(!check_between(comb_df$bmi, 16, 60)) &
          !check_between(comb_df$meas_m.w, 40, 225) &
          length(unique(w_subj_df$meas_m)) == 1

        # if any are true for the above or below, remove all for that parameter

        # remove based on above criteria
        rem_ids_ht <- h_subj_df$id[any(h_exc_btw | h_bmi_out)]
        rem_ids_wt <- w_subj_df$id[any(w_exc_btw | w_bmi_out)]

        # update and remove
        h_subj_keep[rem_ids_ht] <- step
        w_subj_keep[rem_ids_wt] <- step

        h_subj_df <- h_subj_df[!h_subj_df$id %in% rem_ids_ht,]
        w_subj_df <- w_subj_df[!w_subj_df$id %in% rem_ids_wt,]
      }

      # combine again, then check if still 1D
      if (nrow(h_subj_df) > 0 & nrow(w_subj_df) > 0){
        # h = height, w = weight
        comb_df <- comb_df_orig <-
          merge(h_subj_df, w_subj_df, by = "age_days", all = TRUE,
                suffixes = c(".h", ".w"))
        # remove ones that don't match
        comb_df <- comb_df[!(is.na(comb_df$id.h) | (is.na(comb_df$id.w))),]
      } else {
        comb_df <- data.table()
      }
      # no bmis available -- no matches
      if (nrow(comb_df) == 0) {
        # no bmis available
        if (nrow(h_subj_df) > 0){
          exc_ht <-
            !check_between(h_subj_df$meas_m, 139, 206) &
            length(unique(h_subj_df$meas_m)) == 1

          # remove based on above criteria
          rem_ids_ht <- h_subj_df$id[exc_ht]

          # update and remove
          h_subj_keep[rem_ids_ht] <- step

          h_subj_df <- h_subj_df[!h_subj_df$id %in% rem_ids_ht,]
        }

        if (nrow(w_subj_df) > 0){
          exc_wt <-
            !check_between(w_subj_df$meas_m, 40, 225) &
            length(unique(w_subj_df$meas_m)) == 1

          # update and remove
          rem_ids_wt <- w_subj_df$id[exc_wt]

          w_subj_keep[rem_ids_wt] <- step

          w_subj_df <- w_subj_df[!w_subj_df$id %in% rem_ids_wt,]
        }
      }
    }

    # 14h, H error load ----
    # 14h. compute error load -- whether there are too many errors and all
    # should be excluded
    step <- "Exclude-Adult-Too-Many-Errors"

    # no need to do this if everything is already excluded
    if (nrow(h_subj_df) > 0){
      # compute errors -- without sde
      num_err <- sum(!grepl("Same-Day", h_subj_keep) & h_subj_keep != "Include")
      # compute include -- without sde
      num_inc <- sum(!grepl("Same-Day", h_subj_keep))

      # if there are greater than 40% errors for a subject, fill everything
      # with errors
      if (num_err/num_inc > .4){
        criteria <- h_subj_keep == "Include"

        # remove based on above criteria
        rem_ids <- names(h_subj_keep)[criteria]

        # update and remove
        h_subj_keep[rem_ids] <- step

        # no need to update h, since we're done
      }
    }

    # 14w, W error load ----
    # 14w. compute error load -- whether there are too many errors and all
    # should be excluded
    step <- "Exclude-Adult-Too-Many-Errors"

    # no need to do this if everything is already excluded
    if (nrow(w_subj_df) > 0){
      # compute errors -- without sde
      num_err <- sum(!grepl("Same-Day", w_subj_keep) & w_subj_keep != "Include")
      # compute include -- without sde
      num_inc <- sum(!grepl("Same-Day", w_subj_keep))

      # if there are greater than 40% errors for a subject, fill everything
      # with errors
      if (num_err/num_inc > .4){
        criteria <- w_subj_keep == "Include"

        # remove based on above criteria
        rem_ids <- names(w_subj_keep)[criteria]

        # update and remove
        w_subj_keep[rem_ids] <- step

        # no need to update w, since we're done
      }
    }

    # additional columns ----

    # first, let's deal with the weight cap
    # preallocate the all exc weight cap variable (for additional output)
    all_exc_weight_cap <- rep(FALSE, length(w_subj_keep))
    names(all_exc_weight_cap) <- names(w_subj_keep)
    all_exc_weight_cap[names(w_subj_keep)[grepl("Weight-Cap", w_subj_keep)]] <-
      TRUE

    # add to output ----

    # add results to full dataframe
    if (nrow(h_df) > 0){
      # create a data.table for easy joining
      h_out <- data.table(
        id = names(h_subj_keep),
        keep = h_subj_keep,
        orig_sde = h_df$measurement,
        mean_sde = h_subj_mean_sde,
        mean_ht = meanht,
        loss_groups = loss_groups,
        gain_groups = gain_groups
      )

      df[h_out, result := i.keep, on = .(id)]
      df[h_out, orig_sde := i.orig_sde, on = .(id)]
      df[h_out, mean_sde := i.mean_sde, on = .(id)]
      df[h_out, mean_ht := i.mean_ht, on = .(id)]
      df[h_out, loss_groups := i.loss_groups, on = .(id)]
      df[h_out, gain_groups := i.gain_groups, on = .(id)]
    }
    if (nrow(w_df) > 0){
      # create a data.table for easy joining
      w_out <- data.table(
        id = names(w_subj_keep),
        keep = w_subj_keep,
        orig_sde = w_df$measurement,
        mean_sde = w_subj_mean_sde,
        all_exc_weight_cap = all_exc_weight_cap
      )

      df[w_out, result := i.keep, on = .(id)]
      df[w_out, orig_sde := i.orig_sde, on = .(id)]
      df[w_out, mean_sde := i.mean_sde, on = .(id)]
      df[w_out, all_exc_weight_cap := i.all_exc_weight_cap, on = .(id)]
    }

    # complete ----

  }

  return(df)
}

# Main pediatric growthcleanr function -- infants cleanbatch
# internal supporting functions for pediatrics can be found in: infants_support.R
# note

#' Function to clean data (optionally in batches):
#' 4.  Dataset split (optional)
#' a.	In Stata, many of the rest of the steps require you to repeat steps as many times as there are observations for the subject with the highest number of
#'     observations. Therefore, in order to speed things up I separated the dataset at this point into three groups: short (subjects with <=10 total
#'     observations), medium (subjects with >10 and <=30 observations) and long (subjects with >30 observations). I performed all the rest of the steps
#'     separately for each of the 3 datasets and then re-combined them at the end. Note that all observations for a subject stay together. This step is
#'     definitely not necessary if it will not speed things up in whatever software program you're using.
#'
#' NOTE: to use mutliple processes in R we will process patients in batches using the plyr ddply function
#' set up function to process one patient.  Function to parallelize batches is below.
#'
#' @keywords internal
#' @import data.table
#' @importFrom stats median embed
#' @noRd
cleanbatch_infants <- function(data.df,
                               log.path,
                               quietly,
                               parallel,
                               measurement.to.z,
                               ewma.fields,
                               recover.unit.error,
                               include.carryforward,
                               sd.extreme,
                               z.extreme,
                               exclude.levels,
                               tanner.ht.vel,
                               who.ht.vel,
                               lt3.exclude.mode,
                               error.load.threshold,
                               error.load.mincount,
                               ref.data.path) {
  # avoid "no visible binding" warnings
  abs.2ndlast.sd <- abs.tbc.sd <- abs.tbc.sd.next <- abs.tbc.sd.prev <- abssum2 <- NULL
  aft.g.befp1 <- agedays <- agedays.other <- bef.g.aftm1 <- delta <- NULL
  delta.agedays.next <- delta.agedays.prev <- delta.next.ht <- delta.next.sd <- NULL
  delta.prev.ht <- delta.prev.sd <- dewma.after <- dewma.after.prev <- NULL
  dewma.all <- dewma.before <- dewma.before.next <- dnext.sd <- dnext.sd.minus <- NULL
  dnext.sd.plus <- dprev.sd <- dprev.sd.minus <- dprev.sd.plus <- dup <- NULL
  dup.ratio <- ewma.after <- ewma.all <- ewma.before <- exclude <- exclude.count <- NULL
  extraneous <- extraneous.this.day <- first.of.three.or.more <- ht.exp <- NULL
  include.count <- index <- last.of.three.or.more <- line <- max.ht.vel <- NULL
  max.whoinc.1.ht <- max.whoinc.2.ht <- max.whoinc.3.ht <- max.whoinc.4.ht <- NULL
  max.whoinc.6.ht <- maxdiff.next.ht <- maxdiff.prev.ht <- median.other.sd <- NULL
  min.ht.vel <- mindiff.next.ht <- mindiff.prev.ht <- pair <- pair.next <- NULL
  pair.prev <- param <- param.other <- prev.v <- sd.median <- sex <- subjid <- NULL
  swap.flag.1 <- swap.flag.2 <- tanner.months <- tbc.other.sd <- tbc.sd <- NULL
  tbc.sd.d <- tbc.sd.max <- tbc.sd.min <- tbc.sd.minus <- tbc.sd.next <- tbc.sd.plus <- NULL
  tbc.sd.prev <- tbc.sd.sw <- tbc.sd.t <- temp.diff <- temp.exclude <- v <- NULL
  v.d <- v.minus <- v.next <- v.orig <- v.plus <- v.prev <- v.sw <- v.t <- NULL
  valid.interior.measurement <- who.maxdiff.next.ht <- who.mindiff.next.ht <- NULL
  whoagegrp.ht <- whoinc.1.ht <- whoinc.2.ht <- whoinc.3.ht <- whoinc.4.ht <- NULL
  whoinc.6.ht <- whoinc.age.ht <- z.orig <- NULL

  oob <- sd_med <- med_diff <- max_diff <- sum_oob <- i.exclude <- NULL


  # avoid no visible warning errors
  sum_sde <- no_sde <- cf <- wholehalfimp <- seq_win <- cs <- absdiff <-
    sd.orig_uncorr <- seq_win <- absdiff <- wholehalfimp <- ageyears <- ctbc.sd <-
    ..col_replace <- c.ewma.all <- pot_excl <- c.dewma.all <- p_plus <-
    p_minus <- ctbc.sd <- c.ewma.all <- tbc_diff_next <- tbc_diff_prior <-
    tbc_diff_plus_next <- tbc.p_plus <- tbc_diff_plus_prior <-
    tbc_diff_minus_next <- tbc.p_minus <- tbc_diff_minus_prior <- addcrithigh <-
    addcritlow <- tbc_dop <- i.tbc.sd <- rowind <- abssum <- c.dewma.all <-
    whoagegrp_ht <- d_agedays <- mindiff <- maxdiff <- who_mindiff_ht <-
    who_maxdiff_ht <- mindiff_prev <- maxdiff_prev <- whoinc.age.hc <-
    who_maxdiff_hc <- who_mindiff_hc <- diff_prev <-
    diff_next <- aft.g.aftm1 <- val_excl <-
    absval <- comp_diff <- err_ratio <-
    NULL

  data.df <- data.table(data.df, key = c('subjid', 'param', 'agedays', 'index'))

  if (parallel & !is.na(log.path)) {
    sink(
      sprintf(
        "%s/cleangrowth-%s-batch-%03d.log",
        log.path,
        Sys.Date(),
        data.df$batch[1]
      )
    )
  }

  if (!quietly)
    cat(sprintf(
      "[%s] Processing Batch #%s...\n",
      Sys.time(),
      data.df$batch[1]
    ))

  # NOTE: in each step, we redo temp SDEs

  # 5: temporary SDEs ----

  # save a copy of all original measurement values before any transformation
  data.df[, v.orig := v]

  if (!quietly)
    cat(sprintf(
      "[%s] Preliminarily identify potential extraneous...\n",
      Sys.time()
    ))
  data.df$exclude[temporary_extraneous_infants(data.df)] <- 'Exclude-Temporary-Extraneous-Same-Day'

  # capture a list of subjects with possible extraneous for efficiency later
  subj.dup <- data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', unique(subjid)]

  # 6:  carried forwards ----

  if (!include.carryforward) {
    if (!quietly)
      cat(sprintf(
        "[%s] Exclude measurements carried forward...\n",
        Sys.time()
      ))

    # save original column names
    orig_colnames <- colnames(data.df)

    # we only running carried forwards on valid values, non NNTE values,
    # and non single values
    tmp <- table(paste0(data.df$subjid, "_", data.df$param))
    not_single <- paste0(data.df$subjid, "_", data.df$param) %in% names(tmp)[tmp > 1]
    valid_set <- valid(data.df, include.temporary.extraneous = TRUE) &
      !data.df$nnte_full &
      not_single
    data.sub <- data.df[valid_set,]

    # for ease, order by subject, parameter, and agedays
    data.sub <- data.sub[order(subjid, param, agedays)]

    # if no sdes, exclude carried forward values
    data.sub[, sum_sde := .N, by = c("subjid", "param", "agedays")]
    data.sub[, no_sde := !paste0(data.sub$subjid, "_", data.sub$param) %in%
               paste0(data.sub$subjid, "_", data.sub$param)[data.sub$sum_sde > 1]]
    data.sub[no_sde == TRUE, cf :=  (v.orig - shift(v.orig)) == 0, by = c("subjid", "param")]
    data.sub[is.na(cf), cf := FALSE]

    # for values with sdes, compare to all values from the prior day
    data.sub[no_sde == FALSE, cf := (function(df){
      ages <- unique(agedays)
      if (length(ages) > 1){
        for (i in 2:length(ages)) {
          # find the set of measurements from the previous age in days
          all.prev.v <- df[agedays == ages[i - 1], v.orig]
          # if a measurement for the current age is found in the set of measurements from the previous age, then mark it as carried forward
          df[agedays == ages[i] &
               v.orig %in% all.prev.v, cf := TRUE]
        }
      }
    })(copy(.SD)), .SDcols = c('agedays', 'cf', 'v.orig'), by = .(subjid, param)]

    # merge in the carried forwards
    cf_idx <- data.sub$index[data.sub$cf]
    data.df[index %in% cf_idx, exclude := "Exclude-Carried-Forward"]

    # redo temp sde
    data.df$exclude[temporary_extraneous_infants(data.df)] <- 'Exclude-Temporary-Extraneous-Same-Day'

    # find out if values are whole or half imperial
    data.df[param == "WEIGHTKG",
            wholehalfimp := (v.orig * 2.20462262)%%1 < 0.01]
    data.df[param == "HEIGHTCM",
            wholehalfimp := (v.orig * 1.27)%%1 < 0.01]
    data.df[param == "HEIGHTCM",
            wholehalfimp := (v.orig * 1.27 / 2)%%1 < 0.01]
    data.df[param == "HEADCM",
            wholehalfimp := (v.orig * 1.27)%%1 < 0.01]
    data.df[param == "HEADCM",
            wholehalfimp := (v.orig * 1.27 / 2)%%1 < 0.01]
    # find string carried forward length
    data.df[, seq_win := sequence(rle(as.character(exclude))$lengths),
            by = c("subjid", "param")]
    data.df[exclude != "Exclude-Carried-Forward", seq_win := NA]
    # relabel all the initial values as "0" -- reorder
    data.df <- data.df[order(subjid, param, v.orig, agedays)]
    data.df[which(seq_win == 1)-1, seq_win :=  0]
    # reorder correctly
    data.df <- data.df[order(subjid, param, agedays)]
    # also create labels
    data.df[, cs := rep(1:length(rle(as.character(exclude))$lengths),
                        times = rle(as.character(exclude))$lengths),
            by = c("subjid", "param")]
    # fix the first one -- have to do this outside of data table
    data.df[which(seq_win == 0), "cs"] <-
      data.df[which(seq_win == 0)+1, "cs"]
    data.df[is.na(seq_win), cs := NA]

    # find th diff between initial and cf
    data.df[!is.na(seq_win), absdiff :=
              abs(sd.orig_uncorr - sd.orig_uncorr[1]),
            by = c("subjid", "param", "cs")]

    # handle CFs by case
    data.df[!is.na(seq_win), exclude := (function(df){
      # only 1 cf in string
      if (max(seq_win) == 1){
        df[seq_win != 0 & absdiff < 0.05,
           exclude := "Exclude-1-CF-deltaZ-<0.05"]

        # use short circuiting to have the lower end covered (>= 0.05)
        df[seq_win != 0 & absdiff < .1 & wholehalfimp,
           exclude := "Exclude-1-CF-deltaZ-<0.1-wholehalfimp"]
      } else if (max(seq_win) > 1){
        df[seq_win != 0 & agedays/365.25 > 16 & sex == 1 & absdiff < 0.05,
           exclude := "Exclude-Teen-2-plus-CF-deltaZ-<0.05"]
        df[seq_win != 0 & agedays/365.25 > 17 & sex == 0 & absdiff < 0.05,
           exclude := "Exclude-Teen-2-plus-CF-deltaZ-<0.05"]

        df[seq_win != 0 & agedays/365.25 > 16 & sex == 1 & absdiff < .1 &
             wholehalfimp,
           exclude := "Exclude-Teen-2-plus-CF-deltaZ-<0.1-wholehalfimp"]
        df[seq_win != 0 & agedays/365.25 > 17 & sex == 0 & absdiff < .1 &
             wholehalfimp,
           exclude := "Exclude-Teen-2-plus-CF-deltaZ-<0.1-wholehalfimp"]
      }

      return(df$exclude)
    })(copy(.SD)),
    .SDcols = c('agedays', "seq_win", 'absdiff', "sex", "cs", "wholehalfimp",
                "exclude"),
    by = c("subjid", "param", "cs")]
  }

  # BIV ----

  # Wt limits from do-file Oct 10 2022, in dataset, highest plausible weight=29.5kg, 65 lbs
  # HC Limits based on analysis in do-file from Oct 11 2022: in dataset, lowest plausible 18.5, z=-13, highest plausible 63.5, z=11.2  (at birth, highest=42, z=6)


  if (!quietly)
    cat(sprintf(
      "[%s] Exclude BIVs...\n",
      Sys.time()
    ))

  # add age in years
  data.df[, ageyears := agedays/365.25]

  # calculate the valid step
  valid_set <- valid(data.df, include.temporary.extraneous = TRUE) &
    !data.df$nnte_full

  exc_nam <- "Exclude-Absolute-BIV"

  # identify absolute cutoffs
  # Min/max weight at birth based on published births
  data.df[valid_set & param == "WEIGHTKG" & v < 0.2 & agedays == 0,
          exclude := exc_nam]
  data.df[valid_set & param == "WEIGHTKG" & v < 1 & agedays != 0,
          exclude := exc_nam]
  # Min weight after birth based on rules about who can go home with alowance for significant weight loss -- this is for outpatient data only so very low weight babies would still be in NICU
  data.df[valid_set & param == "WEIGHTKG" & v > 10.5 &
            agedays == 0,
          exclude := exc_nam]
  # Max weight for <2y based on eval (see do-file from Oct 10 2022), highest plaus weight 29.5kg
  data.df[valid_set & param == "WEIGHTKG" & v < 1 &
            agedays == 0,
          exclude := exc_nam]
  # Max weight for <2y based on eval (see do-file from Oct 10 2022), highest plaus weight 29.5kg
  data.df[valid_set & param == "WEIGHTKG" & v > 35 &
            ageyears < 2,
          exclude := exc_nam]
  # Max weight for all based on published data
  data.df[valid_set & param == "WEIGHTKG" & v > 600,
          exclude := exc_nam]

  # Min/max HC based on analysis in do file from
  # Also, 18 is z=-6 for 22 0/7 in Fenton and 65 is z=6 for 40 0/7
  data.df[valid_set & param == "HEIGHTCM" & v < 18,
          exclude := exc_nam]
  data.df[valid_set & param == "HEIGHTCM" & v > 244,
          exclude := exc_nam]
  data.df[valid_set & param == "HEIGHTCM" & v > 65 &
            agedays == 0,
          exclude := exc_nam]

  # Min/max HC based on analysis in do file from Oct 11 2022
  # Also, 13 is z=-6 for 22 0/7 in Fenton and
  data.df[valid_set & param == "HEADCM" & v < 13,
          exclude := exc_nam]
  data.df[valid_set & param == "HEADCM" & v > 75,
          exclude := exc_nam]
  data.df[valid_set & param == "HEADCM" & v > 50 &
            agedays == 0,
          exclude := exc_nam]

  exc_nam <- "Exclude-Standardized-BIV"

  # identify z cutoff
  # ***Note, using unrecentered values***
  #  *For weight only do after birth
  data.df[valid_set & param == "WEIGHTKG" & sd.orig_uncorr < -25 &
            ageyears < 1,
          exclude := exc_nam]
  data.df[valid_set & param == "WEIGHTKG" & sd.orig_uncorr < -15 &
            ageyears >= 1,
          exclude := exc_nam]
  data.df[valid_set & param == "WEIGHTKG" & sd.orig_uncorr > 22,
          exclude := exc_nam]

  # *Max z-score for height based on analysis of CHOP data because 15/25 too loose for upper limits
  data.df[valid_set & param == "HEIGHTCM" & sd.orig_uncorr < -25 &
            ageyears < 1,
          exclude := exc_nam]
  data.df[valid_set & param == "HEIGHTCM" & sd.orig_uncorr < -15 &
            ageyears >= 1,
          exclude := exc_nam]
  data.df[valid_set & param == "HEIGHTCM" & sd.orig_uncorr > 8,
          exclude := exc_nam]

  # head circumference
  data.df[valid_set & param == "HEADCM" & sd.orig_uncorr < -15,
          exclude := exc_nam]
  data.df[valid_set & param == "HEADCM" & sd.orig_uncorr > 15,
          exclude := exc_nam]

  # then, do some explicit overwriting for the 0 case (otherwise will be
  # set as missing)
  data.df[v == 0, exclude := exc_nam]

  # 9d.  Replace exc_*=0 if exc_*==2 & redo step 5 (temporary extraneous)
  data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
  data.df[temporary_extraneous_infants(data.df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']

  # evil twins ----
  # Evil Twins: An important weakness in the original pediatric growthcleanr algorithm was that it often failed to identify two or more implausible measurements that occurred next to each other, even if they were extremely deviant from a child’s other measurements. This step is now added to identify these multiple extreme values, although it also identifies some single values.

  exc_nam <- "Exclude-Evil-Twins"

  # create the valid set
  # we only running carried forwards on valid values, non NNTE values,
  # and non single values, and non pair
  tmp <- table(paste0(data.df$subjid, "_", data.df$param))
  not_single_pairs <- paste0(data.df$subjid, "_", data.df$param) %in% names(tmp)[tmp > 2]
  valid_set <- valid(data.df, include.temporary.extraneous = FALSE) &
    !data.df$nnte_full & # does not use the "other"
    not_single_pairs

  # make sure it's ordered by subjid and parameter
  data.df <- data.df[order(subjid, param),]

  # 9A/B/C
  # first, find out of any re possible evil twins in the first place
  start_df <- calc_oob_evil_twins(data.df[valid_set,])

  if (any(start_df$oob)) {
    if (!quietly)
      cat(sprintf(
        "[%s] Exclude evil twins...\n",
        Sys.time()
      ))

    # start to evaluate and remove evil twins
    data.df[valid_set, exclude := (function(df) {
      # where we are updating results
      upd.df <- copy(df)
      upd.df <- calc_oob_evil_twins(upd.df)
      # count the amount of oobs for each subject/param and distribute it out
      upd.df[, `:=` (sum_oob = sum(oob, na.rm = TRUE)), by =.(subjid, param)]

      any_oob <- any(upd.df$sum_oob >= 2)
      # while there are multiple oob, we want to remove
      while (any_oob){

        # 9D
        # now calculate the maximum difference from the median tbc.sd
        upd.df[, `:=` (sd_med = median(tbc.sd, na.rm = TRUE)), by =.(subjid, param)]
        upd.df[, `:=` (med_diff = abs(tbc.sd - sd_med)), by =.(subjid, param)]
        upd.df[, `:=` (max_diff = med_diff  == max(med_diff)), by =.(subjid, param)]
        # for ones with no tbc.sd, mark as false
        upd.df[is.na(max_diff), max_diff := FALSE]

        upd.df[sum_oob > 0 & max_diff, exclude := exc_nam]

        df[upd.df[exclude == exc_nam,], exclude := i.exclude, on = .(line)]

        #9E
        # reupdate valid (to recalculate OOB -- others are not included)
        upd.df <- calc_oob_evil_twins(df[valid(df),])
        upd.df[, `:=` (sum_oob = sum(oob, na.rm = TRUE)), by =.(subjid, param)]

        any_oob <- any(upd.df$sum_oob >= 2)

      }

      return(df$exclude)
    })(copy(.SD))]

  }

  # 9F.  redo temp extraneous
  data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
  data.df[temporary_extraneous_infants(data.df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']

  # Extreme EWMA ----

  # 11.  Exclude extreme errors with EWMA
  # a.	Erroneous measurements can distort the EWMA for measurements around them. Therefore, if the EWMA method identifies more than one value for a subject and
  #     parameter that meets criteria for exclusion, we will only exclude the value that deviates the most from expected in any given step. Then we will repeat the
  #     entire process until no measurements are identified that meet criteria for exclusion.
  # b.	Perform a EWMA calculation
  #   i.	Only use values where exc_*==0 to determine the EWMAs. However, calculate dewma_* variables for values where exc_*==0 or exc_*==2
  if (!quietly)
    cat(sprintf(
      "[%s] Exclude extreme measurements based on EWMA...\n",
      Sys.time()
    ))
  data.df <- data.df[, exclude := (function(df) {
    num.ewma.excluded <- 0
    # optimization: determine whether this subject has any extraneous
    has.extraneous <- subjid %in% subj.dup

    # only enter if there's more than one valid value
    tmp <- table(paste0(df$subjid, "_", df$param))
    not_single_pairs <- paste0(df$subjid, "_", df$param) %in% names(tmp)[tmp > 2]
    valid_set <- valid(df, include.temporary.extraneous = TRUE) &
      !df$nnte_full & # does not use the "other"
      not_single_pairs

    if (sum(valid_set) > 1){
      while (TRUE) {
        tmp <- table(paste0(df$subjid, "_", df$param))
        not_single_pairs <- paste0(df$subjid, "_", df$param) %in% names(tmp)[tmp > 2]
        valid_set <- valid(df, include.temporary.extraneous = TRUE) &
          !df$nnte_full & # does not use the "other"
          not_single_pairs


        df[, (ewma.fields) := as.double(NaN)]

        # first, calculate which exponent we want to put through (pass a different
        # on for each exp)
        # subset df to only valid rows
        df_sub <- df[valid_set,]
        tmp <- data.frame(
          "before" = abs(df_sub$agedays - c(NA, df_sub$agedays[1:(nrow(df_sub)-1)])),
          "after" = abs(df_sub$agedays - c(df_sub$agedays[2:(nrow(df_sub))], NA))
        )
        maxdiff <- sapply(1:nrow(tmp), function(x){max(tmp[x,], na.rm = TRUE)})
        exp_vals <- rep(-1.5, nrow(tmp))
        exp_vals[maxdiff > 365.25] <- -2.5
        exp_vals[maxdiff > 730.5] <- -3.5
        df[valid_set, exp_vals := exp_vals]

        # calculate ewma -- need to do it in a special way to not include
        # temp extaneous

        # first calculate ewma for rows without temp extraneous
        no_tmp_extr <- valid(df, include.temporary.extraneous = FALSE)

        df[valid_set & no_tmp_extr, (ewma.fields) := ewma(agedays, tbc.sd, exp_vals, TRUE)]
        df[valid_set & no_tmp_extr, paste0("c.",ewma.fields) := ewma(agedays, ctbc.sd, exp_vals, TRUE)]

        # now go through the temp extraneous and do the ewmas just for those
        # rows
        if (sum(df$exclude == "Exclude-Temporary-Extraneous-Same-Day") > 0){
          # go through each temp extraneous
          for (j in
               which(df$exclude == "Exclude-Temporary-Extraneous-Same-Day")){
            # subset to only the temp extraneous and valid values
            df_sub <- df[valid_set & (no_tmp_extr | (1:nrow(df) == j)),]

            # run ewma on those values
            df_sub[, (ewma.fields) := ewma(agedays, tbc.sd, exp_vals, TRUE)]
            df_sub[, paste0("c.",ewma.fields) := ewma(agedays, ctbc.sd, exp_vals, TRUE)]

            # merge the value back in
            col_replace <- c(ewma.fields, paste0("c.",ewma.fields))
            df_val <-
              df_sub[df_sub$exclude == "Exclude-Temporary-Extraneous-Same-Day",
                     ..col_replace]
            df[j, (col_replace) := df_val]
          }

        }

        # include the corrected
        df[, `:=`(
          dewma.all = tbc.sd - ewma.all,
          dewma.before = tbc.sd - ewma.before,
          dewma.after = tbc.sd - ewma.after,

          c.dewma.all = ctbc.sd - c.ewma.all
        )]


        # calculate potential exclusions
        df[valid_set, pot_excl :=
             (dewma.all > 3.5 & dewma.before > 3 & dewma.after > 3 & tbc.sd > 3.5 &
                ((!is.na(ctbc.sd) & c.dewma.all > 3.5) | is.na(ctbc.sd))
             ) |
             # for both
             (
               dewma.all < -3.5 &
                 dewma.before < -3 & dewma.after < -3 & tbc.sd < -3.5 &
                 ((!is.na(ctbc.sd) & c.dewma.all > 3.5) | is.na(ctbc.sd))
             )
        ]
        df[valid_set,][c(1, nrow(df)), pot_excl := FALSE]
        df[!valid_set, pot_excl := FALSE] # to compensate for later


        # 11c.  Identify all values that meet all of the following criteria as potential exclusions:
        #   i.	There are 3 or more measurements for that subject and parameter
        #   ii.	(dewma_*>3.5 & dewma_*_bef>3 & dewma_*_aft>3 & tbc*sd>3.5) OR (dewma_*<-3.5 & dewma_*_bef<-3 & d & dewma_*_aft<-3 & tbc*sd<-3.5)
        #   iii.	exc_*==0

        num.exclude <- sum(df$pot_excl)
        # 11d.  If there is only one potential exclusion identified in step 11c for a subject and parameter, replace exc_*=5 for that value
        if (num.exclude == 1)
          df[pot_excl, exclude := 'Exclude-EWMA1-Extreme']
        # 11e.  If there is more than one potential exclusion identified in step 11c for a subject and parameter, calculate abssum_*=|tbc*sd+dewma_*| for each exclusion and
        #     replace exc_*=5 for the value with the highest abssum_*
        if (num.exclude > 1) {
          # first order by decreasing abssum
          worst.row <- with(df, order(pot_excl, abs(tbc.sd + dewma.all),
                                      decreasing = TRUE))[1]
          df[worst.row, exclude := 'Exclude-EWMA1-Extreme']
        }

        # 11h.  Recalculate temporary extraneous as in step 5
        # optimize: only perform these steps if this subject is known to have extraneous measurements
        if (has.extraneous) {
          df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
          df[temporary_extraneous_infants(df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']
        }

        # 11i.  If there was at least one subject who had a potential exclusion identified in step 11c, repeat steps 11b-11g. If there were no subjects with potential
        #     exclusions identified in step 11c, move on to step 12.
        newly.excluded <- sum(df$exclude %in% c('Exclude-EWMA1-Extreme'))
        if (newly.excluded > num.ewma.excluded) {
          num.ewma.excluded <- newly.excluded
        } else {
          break
        }
      }
    }
    return(df$exclude)
  })(copy(.SD)), by = .(subjid, param), .SDcols = c('index', 'sex', 'agedays', 'tbc.sd', 'ctbc.sd', 'nnte_full', 'exclude')]

  # 9d.  Replace exc_*=0 if exc_*==2 & redo step 5 (temporary extraneous)
  data.df[exclude == 'Exclude-Temporary-Extraneous-Same-Day', exclude := 'Include']
  data.df[temporary_extraneous_infants(data.df), exclude := 'Exclude-Temporary-Extraneous-Same-Day']

  # 13: SDEs ----

  if (!quietly)
    cat(sprintf(
      "[%s] Exclude same day extraneous...\n",
      Sys.time()
    ))

  # prepare a list of valid rows and initialize variables for convenience
  valid.rows <- valid(data.df, include.temporary.extraneous = TRUE) &
    !data.df$nnte_full  # does not use the "other"

  data.df <- data.df[valid.rows, exclude := (function(subj_df) {
    # keep the original indices and exclude to overwrite
    orig_idx <- copy(subj_df$index)
    exclude_all <- copy(subj_df$exclude)
    if (nrow(subj_df) > 0 & any(subj_df$exclude == "Exclude-Temporary-Extraneous-Same-Day")){
      # for convenience
      subj_df[, extraneous := exclude == "Exclude-Temporary-Extraneous-Same-Day"]

      step <- "Exclude-SDE-Identical"

      # identify duplicate days
      dup_days <- unique(subj_df$agedays[subj_df$extraneous])

      # first, exclude identical measurements on the same day
      ide_ids <- c() # where we keep the identical ones
      for (dd in dup_days){
        # count amount of unique values
        s_df <- copy(subj_df[subj_df$agedays == dd,])
        ide_tab <- table(s_df$v)
        if (any(ide_tab > 1)){
          # for each identical, keep only the first one, by id
          # (ordered initially)
          ide_ids <- c(ide_ids, s_df$index[
            as.character(s_df$v) %in% names(ide_tab[ide_tab > 1])
          ][duplicated(
            s_df$v[
              as.character(s_df$v) %in% names(ide_tab[ide_tab > 1])
            ]
          )])
        }
      }
      criteria <- subj_df$index %in% ide_ids

      # we're going to update excludes before moving on to the rest of this
      # subject
      exclude_all[criteria] <- step

      subj_df <- subj_df[!criteria,]

      if (any(criteria)){
        # reevaluate temp same day
        subj_df[, extraneous := temporary_extraneous_infants(subj_df)]

        # identify duplicate days
        dup_days <- unique(subj_df$agedays[subj_df$extraneous])
      }

      # 13B: identify similar groups
      similar_ids <- c()
      for (dd in dup_days){
        # first, calculate if all SDEs on a single day are similar
        s_df <- copy(subj_df[subj_df$agedays == dd,])

        similar <- if (subj_df$param[1] == "WEIGHTKG"){
          max(s_df$v)/min(s_df$v) <= 1.03 & max(s_df$v) - min(s_df$v) <= 2.5
        } else if (subj_df$param[1] == "HEIGHTCM"){
          (max(s_df$v) - min(s_df$v) <  2.541 & min(s_df$v) < 127) |
            (max(s_df$v) - min(s_df$v) <  5.081 & min(s_df$v) >= 127)
        } else { # head circumference
          max(s_df$v) - min(s_df$v) <  1.271
        }

        if (similar){
          similar_ids <- c(similar_ids, s_df$index)
        }
      }

      # 13C: exclude non similar groups

      step <- "Exclude-SDE-All-Exclude"
      # now the rest!

      # next, calculate the duplicate ratio -- what proportion of days are
      # duplicated
      dup_ratio <-
        length(unique(subj_df$agedays[duplicated(subj_df$agedays)]))/
        length(unique(subj_df$agedays))
      # also check whether or not any same-days are adjacent -- need 4 at least
      # rolling windows of day differences -- we are looking for 0,x,0
      if (nrow(subj_df) > 3){
        roll <- embed(diff(subj_df$agedays), 3)
        adjacent <- sapply(1:nrow(roll), function(x){
          all(c(roll[x,1] == 0, roll[x,2] != 0, roll[x,3] == 0))
        })

        # get age days & rows that had adjacent
        idx_roll <- c(embed(1:nrow(subj_df),4)[adjacent,, drop = FALSE])
        idx_adj <- which(subj_df$agedays %in% subj_df$agedays[idx_roll])
      } else {
        adjacent <- FALSE
        idx_adj <- idx_roll <- c()
      }

      # if dup ratio is too high, we exclude all
      # if adjacent, we exclude only those relevant sdes
      # same day extraneous -- for strict (default)
      criteria <-
          # looser criteria:
          # 1.	Exclude for ratio if ratio >1/3
          # 2.	Exclude for adjacent if 2 are adjacent AND ratio >1/4
          # 3.	Exclude for adjacent if 2 are adjacent AND those 2 are the first 2 or last 2 measurements for the subject/parameter
          #   note: if the first or last is sde and it's adjacent, by default
          #   that means it's the first or last two that are adjacent
          # 4.	Exclude for adjacent if 3 or more are adjacent
          if (dup_ratio > (1/3)){
            subj_df$extraneous
          } else if (
            (sum(adjacent) == 1 & dup_ratio > (1/4)) |
            (sum(adjacent) == 1 & c(1) %in% idx_adj) |
            (sum(adjacent) == 1 &
             c(nrow(subj_df)) %in% idx_adj) |
            (sum(adjacent) >= 2)
          ){
            # adjacent sum == 1 means 2 are adjacent
            # only exclude sdes that are adjacent
            subj_df$agedays %in% subj_df$agedays[idx_roll]
          } else {
            rep(FALSE, nrow(subj_df))
          }
      # re-include similar groups
      criteria[subj_df$index %in% similar_ids] <- FALSE

      # we're going to update excludes before moving on to the rest of this
      # subject
      exclude_all[orig_idx %in% subj_df$index[criteria]] <- step

      subj_df <- subj_df[!criteria,]

      if (any(criteria)){
        # reevaluate temp same day
        subj_df[, extraneous := temporary_extraneous_infants(subj_df)]

        # identify duplicate days
        dup_days <- unique(subj_df$agedays[subj_df$extraneous])
      }

      step_extreme <- "Exclude-SDE-All-Extreme"
      step <- "Exclude-SDE-EWMA"
      # next, check for trivial differences for SDEs on the same day
      # this only works if there are non-sde days

      if (any(subj_df$extraneous)){
        # 13D: calculating ewmas

        # check for SDEs by EWMA -- alternate calculate excluding other SDEs
        all_sdes <- duplicated(subj_df$agedays) |
          duplicated(subj_df$agedays, fromLast = TRUE)

        # first, calculate which exponent we want to put through (pass a different
        # on for each exp)
        tmp <- data.frame(
          "before" = abs(subj_df$agedays - c(NA, subj_df$agedays[1:(nrow(subj_df)-1)])),
          "after" = abs(subj_df$agedays - c(subj_df$agedays[2:(nrow(subj_df))], NA))
        )
        maxdiff <- sapply(1:nrow(tmp), function(x){max(tmp[x,], na.rm = TRUE)})
        exp_vals <- rep(-1.5, nrow(tmp))
        exp_vals[maxdiff > 365.25] <- -2.5
        exp_vals[maxdiff > 730.5] <- -3.5
        subj_df[, exp_vals := exp_vals]

        # calculate the time difference for all values, as well as exponential
        delta <- as_matrix_delta(subj_df$agedays)
        delta <- ifelse(delta == 0, 0, (delta) ^ subj_df$exp_vals)

        rem_ids_extreme <- c()
        rem_ids <- c()
        for (dd in dup_days){
          # first, calculate if all SDEs on a single day are similar
          s_df <- copy(subj_df[subj_df$agedays == dd,])

          similar <- if (subj_df$param[1] == "WEIGHTKG"){
            max(s_df$v)/min(s_df$v) <= 1.03 & max(s_df$v) - min(s_df$v) <= 2.5
          } else if (subj_df$param[1] == "HEIGHTCM"){
            (max(s_df$v) - min(s_df$v) <  2.541 & min(s_df$v) < 127) |
              (max(s_df$v) - min(s_df$v) <  5.081 & min(s_df$v) >= 127)
          } else { # head circumference
            max(s_df$v) - min(s_df$v) <  1.271
          }

          if (!similar){
            # 13E

            # calculate dewma for each SDE on this day
            dewma <- sapply(1:nrow(s_df), function(x){
              ind <- subj_df$index == s_df[[x, "index"]]

              ewma_res <- sum(subj_df$tbc.sd[!all_sdes]*delta[ind,!all_sdes])/
                sum(delta[ind, !all_sdes])

              # delta ewma
              return(s_df$tbc.sd[x] - ewma_res)
            })
            absdewma <- abs(dewma)

            if (min(absdewma) > 1){
              rem_ids_extreme <- c(rem_ids_extreme, s_df$index)
            } else {
              de_day <- which.min(abs(dewma))

              # keep the value with the lowest EWMA -- do not keep rest
              rem_ids <- c(rem_ids, s_df$index[-de_day])
            }
          }
        }
        # re-include similar groups
        criteria_extreme <- subj_df$index %in% rem_ids_extreme
        criteria <- subj_df$index %in% rem_ids
        # we're going to update excludes before moving on to the rest of this
        # subject
        exclude_all[orig_idx %in% subj_df$index[criteria_extreme]] <- step_extreme
        exclude_all[orig_idx %in% subj_df$index[criteria]] <- step

        subj_df <- subj_df[!(criteria | criteria_extreme),]

        if (any(c(criteria, criteria_extreme))){
          # reevaluate temp same day
          subj_df[, extraneous := temporary_extraneous_infants(subj_df)]

          # identify duplicate days
          dup_days <- unique(subj_df$agedays[subj_df$extraneous])
        }
      }

      # all remaining are similar
      step_extreme <- "Exclude-SDE-All-Extreme"
      step <- "Exclude-SDE-One-Day"
      if (any(subj_df$extraneous)){
        # check for SDEs
        all_sdes <- duplicated(subj_df$agedays) |
          duplicated(subj_df$agedays, fromLast = TRUE)

        rem_ids <- c()
        rem_ids_extreme <- c()
        for (dd in dup_days){
          # first, calculate if all SDEs on a single day are similar
          s_df <- copy(subj_df[subj_df$agedays == dd,])

          # find the DOP (designated other parameter)
          dop <-
            data.df[subjid == s_df$subjid[1] & param == get_dop(s_df$param[1]) &
                      agedays == s_df$agedays[1],]

          if (nrow(dop) > 0){
            # 13F
            med_diff <- abs(median(dop$tbc.sd)-s_df$tbc.sd)

            if (min(med_diff) > 2){
              rem_ids_extreme <- c(rem_ids_extreme, s_df$index)
            } else {
              de_day <- which.min(med_diff)

              # keep the value with the lowest median diff -- do not keep rest
              rem_ids <- c(rem_ids, s_df$index[-de_day])
            }
          } else if (nrow(s_df) > 3){
            #13G
            med_diff <- abs(median(s_df$tbc.sd)-s_df$tbc.sd)

            # if even and the middle two values are equidistant
            if (nrow(s_df)%%2 == 0 &
                diff(sort(med_diff)[1:2]) < 1e-8){
              # if the ageday is even, keep the lower z score; otherwise,
              # keep higher
              keep_val <-
                if (s_df$agedays[1]%%2 == 0){
                  which(s_df$v == min(s_df$v[order(med_diff)[1:2]]))
                } else {
                  which(s_df$v == max(s_df$v[order(med_diff)[1:2]]))
                }
              rem_ids <- c(rem_ids, s_df$index[-keep_val])
            }

            de_day <- which.min(med_diff)
            # keep the value with the lowest median diff -- do not keep rest
            rem_ids <- c(rem_ids, s_df$index[-de_day])
          } else {
            # 13H

            # 2 values, keep based on the age days
            keep_val <-
            if (s_df$agedays[1]%%2 == 0){
              which.min(s_df$v)
            } else {
              which.max(s_df$v)
            }
            rem_ids <- c(rem_ids, s_df$index[-keep_val])
          }
        }

        # re-include similar groups
        criteria_extreme <- subj_df$index %in% rem_ids_extreme
        criteria <- subj_df$index %in% rem_ids
        # we're going to update excludes before moving on to the rest of this
        # subject
        exclude_all[orig_idx %in% subj_df$index[criteria_extreme]] <- step_extreme
        exclude_all[orig_idx %in% subj_df$index[criteria]] <- step
      }
    }

    # replace all the "exclude temp extraneous" with includes -- other SDEs
    # were removed
    exclude_all[exclude_all == "Exclude-Temporary-Extraneous-Same-Day"] <-
      "Include"

    return(exclude_all)
  })(copy(.SD)), by = .(subjid, param), .SDcols = colnames(data.df)]

  # 15: moderate EWMA ----

  if (!quietly)
    cat(sprintf(
      "[%s] Exclude moderate EWMA...\n",
      Sys.time()
    ))

  # create the valid set
  # we only running carried forwards on valid values, non NNTE values,
  # and non single values, and non pair
  tmp <- table(paste0(data.df$subjid, "_", data.df$param))
  not_single_pairs <- paste0(data.df$subjid, "_", data.df$param) %in% names(tmp)[tmp > 2]
  valid_set <- valid(data.df, include.temporary.extraneous = FALSE) &
    !data.df$nnte_full & # does not use the "other"
    not_single_pairs

  # work in an a function order to encapsulate and not keep all the additional
  # columns

  # calculate plus/minus values

  # order just for ease later
  data.df <- data.df[order(subjid, param, agedays),]
  data.df <- data.df[valid_set, exclude := (function(df) {
    # 15A: calc plus/minus values
    df[param == "WEIGHTKG", p_plus := 1.05*v]
    df[param == "WEIGHTKG", p_minus := .95*v]

    df[param == "HEIGHTCM", p_plus := v+1]
    df[param == "HEIGHTCM", p_minus := v-1]

    df[param == "HEADCM", p_plus := v+1]
    df[param == "HEADCM", p_minus := v-1]

    #15B: smooth and recenter
    df <- calc_and_recenter_z_scores(df, "p_plus", ref.data.path)
    df <- calc_and_recenter_z_scores(df, "p_minus", ref.data.path)

    #15c: exclude agedays = 0 for ht, hc
    #15d: calculate ewma -- run within a subject/parameter
    df[(param != "HEIGHTCM" | param != "HEADCM") & agedays != 0,
             exclude := (function(df_sub) {
      # save initial exclusions to keep track
      ind_all <- copy(df_sub$index)
      exclude_all <- copy(df_sub$exclude)

      testing <- TRUE

      while (testing & nrow(df_sub) >= 3){
        df_sub[, (ewma.fields) := as.double(NaN)]

        # first, calculate which exponent we want to put through (pass a different
        # on for each exp)
        # subset df to only valid rows

        tmp <- data.frame(
          "before" = abs(df_sub$agedays - c(NA, df_sub$agedays[1:(nrow(df_sub)-1)])),
          "after" = abs(df_sub$agedays - c(df_sub$agedays[2:(nrow(df_sub))], NA))
        )
        maxdiff <- sapply(1:nrow(tmp), function(x){max(tmp[x,], na.rm = TRUE)})
        exp_vals <- rep(-1.5, nrow(tmp))
        exp_vals[maxdiff > 365.25] <- -2.5
        exp_vals[maxdiff > 730.5] <- -3.5
        df_sub[, exp_vals := exp_vals]

        # calculate ewma
        df_sub[, (ewma.fields) := ewma(agedays, tbc.sd, exp_vals, TRUE)]
        df_sub[, paste0("c.",ewma.fields) := ewma(agedays, ctbc.sd, exp_vals, TRUE)]

        # calculate dewma
        df_sub[, `:=`(
          dewma.all = tbc.sd - ewma.all,
          dewma.before = tbc.sd - ewma.before,
          dewma.after = tbc.sd - ewma.after,

          c.dewma.all = ctbc.sd - c.ewma.all
        )]

        # 15E: calculate prior and next differences
        df_sub[, tbc_diff_next := tbc.sd - c(tbc.sd[2:nrow(df_sub)], NA)]
        df_sub[, tbc_diff_prior := tbc.sd - c(NA, tbc.sd[1:(nrow(df_sub)-1)])]

        df_sub[, tbc_diff_plus_next := tbc.p_plus - c(tbc.sd[2:nrow(df_sub)], NA)]
        df_sub[, tbc_diff_plus_prior :=
                 tbc.p_plus- c(NA, tbc.sd[1:(nrow(df_sub)-1)])]

        df_sub[, tbc_diff_minus_next := tbc.p_minus - c(tbc.sd[2:nrow(df_sub)], NA)]
        df_sub[, tbc_diff_minus_prior :=
                 tbc.p_minus- c(NA, tbc.sd[1:(nrow(df_sub)-1)])]

        # 15F: all exclusion criteria
        df_sub[,
               addcrithigh :=
                 dewma.before > 1 & dewma.after > 1 &
                 ((tbc_diff_next > 1 & tbc_diff_plus_next > 1 & tbc_diff_minus_next > 1) | is.na(tbc_diff_next)) &
                 ((tbc_diff_prior > 1 & tbc_diff_plus_prior > 1 & tbc_diff_minus_prior > 1) | is.na(tbc_diff_prior))
        ]
        df_sub[,
               addcritlow :=
                 dewma.before < -1 & dewma.after < -1 &
                 ((tbc_diff_next < -1 & tbc_diff_plus_next < -1 & tbc_diff_minus_next < -1) | is.na(tbc_diff_next)) &
                 ((tbc_diff_prior < -1 & tbc_diff_plus_prior < -1 & tbc_diff_minus_prior < -1) | is.na(tbc_diff_prior))
        ]

        #15G: find all the values for the DOP
        # find the comparison -- making sure to keep only valid rows
        compare_df <- data.df[subjid == df_sub$subjid[1] &
                                param == get_dop(df_sub$param[1]) &
                                exclude == "Include",]
        if (nrow(compare_df) > 0){
          df_sub[compare_df, tbc_dop := i.tbc.sd,
                 on = c("agedays")]
          df_sub[is.na(tbc_dop), tbc_dop := median(compare_df$tbc.sd)]
        } else {
          df_sub[, tbc_dop := NA]
        }

        #15H: all exclusions
        df_sub$rowind <- 1:nrow(df_sub)

        #a
        df_sub[-c(1, nrow(df_sub)),][
          dewma.all > 1 & (c.dewma.all > 1 | is.na(c.dewma.all)) & addcrithigh,
          exclude := "Exclude-EWMA2-middle"]
        df_sub[-c(1, nrow(df_sub)),][
          dewma.all < -1 & (c.dewma.all < -1 | is.na(c.dewma.all)) & addcritlow,
          exclude := "Exclude-EWMA2-middle"]

        #b
        df_sub[agedays == 0 & agedays[2] < 365.25 &
                 dewma.all > 3 & (c.dewma.all > 3 | is.na(c.dewma.all)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-birth-WT"]
        df_sub[agedays == 0 & agedays[2] < 365.25 &
                 dewma.all < -3 & (c.dewma.all < -3 | is.na(c.dewma.all)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-birth-WT"]

        #c
        df_sub[agedays == 0 & agedays[2] >= 365.25 &
                 dewma.all > 4 & (c.dewma.all > 4 | is.na(c.dewma.all)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-birth-WT-ext"]
        df_sub[agedays == 0 & agedays[2] >= 365.25 &
                 dewma.all < -4 & (c.dewma.all < -4 | is.na(c.dewma.all)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-birth-WT-ext"]

        #d
        df_sub[rowind == 1 & agedays[1] != 0 & agedays[2]-agedays[1] < 365.25 &
                 dewma.all > 2 & (c.dewma.all > 2 | is.na(c.dewma.all)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-first"]
        df_sub[rowind == 1 & agedays[1] != 0 & agedays[2]-agedays[1] < 365.25 &
                 dewma.all < -2 & (c.dewma.all < -2 | is.na(c.dewma.all)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-first"]

        # e
        df_sub[rowind == 1 & agedays[1] != 0 & agedays[2]-agedays[1] >= 365.25 &
                 dewma.all > 3 & (c.dewma.all > 3 | is.na(c.dewma.all)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-first-ext"]
        df_sub[rowind == 1 & agedays[1] != 0 & agedays[2]-agedays[1] >= 365.25 &
                 dewma.all < -3 & (c.dewma.all < -3 | is.na(c.dewma.all)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-first-ext"]

        # f
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] < 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) < 2 &
                 dewma.all > 2 & (c.dewma.all > 2 | is.na(c.dewma.all)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-last"]
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] < 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) < 2 &
                 dewma.all < -2 & (c.dewma.all < -2 | is.na(c.dewma.all)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-last"]

        # g
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] < 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) >= 2 &
                 dewma.all > abs(tbc.sd[nrow(df_sub)-1]) &
                 (c.dewma.all > 3 | is.na(c.dewma.all)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-last-high"]
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] < 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) >= 2 &
                 dewma.all < abs(tbc.sd[nrow(df_sub)-1]) &
                 (c.dewma.all < -3 | is.na(c.dewma.all)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-last-high"]

        # h
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] >= 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) < 2 &
                 dewma.all > 3 &
                 (c.dewma.all > 3 | is.na(c.dewma.all)) &
                 (tbc.sd - tbc_dop > 4 | is.na(tbc_dop)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-last-ext"]
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] >= 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) < 2 &
                 dewma.all < -3 &
                 (c.dewma.all <- 3 | is.na(c.dewma.all)) &
                 (tbc.sd - tbc_dop < -4 | is.na(tbc_dop)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-last-ext"]

        # i
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] >= 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) >= 2 &
                 dewma.all > (1+abs(tbc.sd[nrow(df_sub)-1])) &
                 (c.dewma.all > 3 | is.na(c.dewma.all)) &
                 (tbc.sd - tbc_dop > 4 | is.na(tbc_dop)) &
                 addcrithigh
               , exclude := "Exclude-EWMA2-last-ext-high"]
        df_sub[rowind == nrow(df_sub) &
                 agedays[nrow(df_sub)]-agedays[nrow(df_sub)-1] >= 365.25*2 &
                 abs(tbc.sd[nrow(df_sub)-1]) >= 2 &
                 dewma.all < (-1-abs(tbc.sd[nrow(df_sub)-1])) &
                 (c.dewma.all <- 3 | is.na(c.dewma.all)) &
                 (tbc.sd - tbc_dop < -4 | is.na(tbc_dop)) &
                 addcritlow
               , exclude := "Exclude-EWMA2-last-ext-high"]

        # figure out if any of the exclusions hit
        count_exclude <- sum(df_sub$exclude != "Include")
        if (count_exclude > 0){
          df_sub[, abssum := abs(tbc.sd + dewma.all)]

          # choose the highest abssum for exclusion
          idx <- df_sub$index[which.max(df_sub[df_sub$exclude != "Include",
                                               abssum])]

          exclude_all[ind_all == idx] <- df_sub[index == idx, exclude]

          #set up to continue on
          testing <- TRUE

          df_sub <- df_sub[index != idx, ]
        } else {
          testing <- FALSE
        }
      }

      return(exclude_all)
    })(copy(.SD)), , by = .(subjid, param), .SDcols = colnames(df)]

    return(df$exclude)
  })(copy(.SD)), .SDcols = colnames(data.df)]

  # 16: moderate EWMA for birth HT and HC ----

  if (!quietly)
    cat(sprintf(
      "[%s] Exclude moderate EWMA for birth height and head circumference...\n",
      Sys.time()
    ))

  # create the valid set
  # we only running carried forwards on valid values, non NNTE values,
  # and non single values, and non weight
  tmp <- table(paste0(data.df$subjid, "_", data.df$param))
  not_single <- paste0(data.df$subjid, "_", data.df$param) %in% names(tmp)[tmp > 1]
  valid_set <- valid(data.df, include.temporary.extraneous = FALSE) &
    !data.df$nnte_full & # does not use the "other"
    not_single &
    data.df$param != "WEIGHTKG" # do not run on weight

  # work in an a function order to encapsulate and not keep all the additional
  # columns

  # NOTE: check NNTE na's

  # calculate plus/minus values

  # order just for ease later
  data.df <- data.df[order(subjid, param, agedays),]
  data.df <- data.df[valid_set, exclude := (function(df) {
    # =calc plus/minus values
    df[param == "HEIGHTCM", p_plus := v+1]
    df[param == "HEIGHTCM", p_minus := v-1]

    df[param == "HEADCM", p_plus := v+1]
    df[param == "HEADCM", p_minus := v-1]

    #smooth and recenter
    df <- calc_and_recenter_z_scores(df, "p_plus", ref.data.path)
    df <- calc_and_recenter_z_scores(df, "p_minus", ref.data.path)

    #16a: calculate ewma -- run within a subject/parameter
    # inly evaluate on subject/parameters with a birth ageday
    df <- df[subjid %in% subjid[agedays == 0],
      exclude := (function(df_sub) {
        # save initial exclusions to keep track
        ind_all <- copy(df_sub$index)
        exclude_all <- copy(df_sub$exclude)

        testing <- TRUE

        while (testing & nrow(df_sub) >= 3){
          df_sub[, (ewma.fields) := as.double(NaN)]

          # first, calculate which exponent we want to put through (pass a different
          # on for each exp)
          # subset df to only valid rows

          tmp <- data.frame(
            "before" = abs(df_sub$agedays - c(NA, df_sub$agedays[1:(nrow(df_sub)-1)])),
            "after" = abs(df_sub$agedays - c(df_sub$agedays[2:(nrow(df_sub))], NA))
          )
          maxdiff <- sapply(1:nrow(tmp), function(x){max(tmp[x,], na.rm = TRUE)})
          exp_vals <- rep(-1.5, nrow(tmp))
          exp_vals[maxdiff > 365.25] <- -2.5
          exp_vals[maxdiff > 730.5] <- -3.5
          df_sub[, exp_vals := exp_vals]

          # calculate ewma
          df_sub[, (ewma.fields) := ewma(agedays, tbc.sd, exp_vals, TRUE)]
          df_sub[, paste0("c.",ewma.fields) := ewma(agedays, ctbc.sd, exp_vals, TRUE)]

          # calculate dewma
          df_sub[, `:=`(
            dewma.all = tbc.sd - ewma.all,
            dewma.before = tbc.sd - ewma.before,
            dewma.after = tbc.sd - ewma.after,

            c.dewma.all = ctbc.sd - c.ewma.all
          )]

          # 16B: calculate prior and next differences
          df_sub[, tbc_diff_next := tbc.sd - c(tbc.sd[2:nrow(df_sub)], NA)]
          df_sub[, tbc_diff_prior := tbc.sd - c(NA, tbc.sd[1:(nrow(df_sub)-1)])]

          df_sub[, tbc_diff_plus_next := tbc.p_plus - c(tbc.sd[2:nrow(df_sub)], NA)]
          df_sub[, tbc_diff_plus_prior :=
                   tbc.p_plus- c(NA, tbc.sd[1:(nrow(df_sub)-1)])]

          df_sub[, tbc_diff_minus_next := tbc.p_minus - c(tbc.sd[2:nrow(df_sub)], NA)]
          df_sub[, tbc_diff_minus_prior :=
                   tbc.p_minus- c(NA, tbc.sd[1:(nrow(df_sub)-1)])]

          # 16C: all exclusion criteria
          df_sub[,
                 addcrithigh :=
                   dewma.before > 1 & dewma.after > 1 &
                   ((tbc_diff_next > 1 & tbc_diff_plus_next > 1 & tbc_diff_minus_next > 1) | is.na(tbc_diff_next)) &
                   ((tbc_diff_prior > 1 & tbc_diff_plus_prior > 1 & tbc_diff_minus_prior > 1) | is.na(tbc_diff_prior))
          ]
          df_sub[,
                 addcritlow :=
                   dewma.before < -1 & dewma.after < -1 &
                   ((tbc_diff_next < -1 & tbc_diff_plus_next < -1 & tbc_diff_minus_next < -1) | is.na(tbc_diff_next)) &
                   ((tbc_diff_prior < -1 & tbc_diff_plus_prior < -1 & tbc_diff_minus_prior < -1) | is.na(tbc_diff_prior))
          ]

          #16D: all exclusions
          df_sub$rowind <- 1:nrow(df_sub)

          #a
          df_sub[agedays == 0 & agedays[2] < 365.25 &
                   dewma.all > 3 & (c.dewma.all > 3 | is.na(c.dewma.all)) &
                   addcrithigh
                 , exclude := "Exclude-EWMA2-birth-HT-HC"]
          df_sub[agedays == 0 & agedays[2] < 365.25 &
                   dewma.all < -3 & (c.dewma.all < -3 | is.na(c.dewma.all)) &
                   addcritlow
                 , exclude := "Exclude-EWMA2-birth-HT-HC"]

          #b
          df_sub[agedays == 0 & agedays[2] >= 365.25 &
                   dewma.all > 4 & (c.dewma.all > 4 | is.na(c.dewma.all)) &
                   addcrithigh
                 , exclude := "Exclude-EWMA2-birth-HT-HC-ext"]
          df_sub[agedays == 0 & agedays[2] >= 365.25 &
                   dewma.all < -4 & (c.dewma.all < -4 | is.na(c.dewma.all)) &
                   addcritlow
                 , exclude := "Exclude-EWMA2-birth-HT-HC-ext"]

          # figure out if any of the exclusions hit
          count_exclude <- sum(df_sub$exclude != "Include")
          if (count_exclude > 0){
            df_sub[, abssum := abs(tbc.sd + dewma.all)]

            # choose the highest abssum for exclusion
            idx <- df_sub$index[which.max(df_sub[df_sub$exclude != "Include",
                                                 abssum])]

            exclude_all[ind_all == idx] <- df_sub[index == idx, exclude]

            #set up to continue on
            testing <- TRUE

            df_sub <- df_sub[index != idx, ]
          } else {
            testing <- FALSE
          }
        }
        return(exclude_all)
      })(copy(.SD)), , by = .(subjid, param), .SDcols = colnames(df)]

    return(df$exclude)
  })(copy(.SD)), .SDcols = colnames(data.df)]

  # 17: raw differences ----

  if (!quietly)
    cat(sprintf(
      "[%s] Exclude raw differences...\n",
      Sys.time()
    ))

  # read in tanner data
  tanner_ht_vel_rev_path <- ifelse(
    ref.data.path == "",
    system.file(file.path("extdata", "tanner_ht_vel.csv.gz"), package = "growthcleanr"),
    file.path(ref.data.path, "tanner_ht_vel.csv.gz")
  )

  tanner.ht.vel.rev <- fread(tanner_ht_vel_rev_path)

  setnames(tanner.ht.vel.rev,
           colnames(tanner.ht.vel.rev),
           gsub('_', '.', colnames(tanner.ht.vel.rev)))
  setkey(tanner.ht.vel.rev, sex, tanner.months)

  # read in the who height data
  who_max_ht_vel_path <- ifelse(
    ref.data.path == "",
    system.file(file.path("extdata", "who_ht_maxvel_3sd.csv.gz"), package = "growthcleanr"),
    file.path(ref.data.path, "who_ht_maxvel_3sd.csv.gz")
  )

  who_ht_vel_3sd_path <- ifelse(
    ref.data.path == "",
    system.file(file.path("extdata", "who_ht_vel_3sd.csv.gz"), package = "growthcleanr"),
    file.path(ref.data.path, "who_ht_vel_3sd.csv.gz")
  )

  who.max.ht.vel <- fread(who_max_ht_vel_path)
  who.ht.vel <- fread(who_ht_vel_3sd_path)
  setkey(who.max.ht.vel, sex, whoagegrp_ht)
  setkey(who.ht.vel, sex, whoagegrp_ht)
  who.ht.vel <- as.data.table(dplyr::full_join(who.ht.vel, who.max.ht.vel, by =
                                                 c('sex', 'whoagegrp_ht')))

  setnames(who.ht.vel, colnames(who.ht.vel), gsub('_', '.', colnames(who.ht.vel)))
  setkey(who.ht.vel, sex, whoagegrp.ht)

  # read in who hc data
  who_max_hc_vel_path <- ifelse(
    ref.data.path == "",
    system.file(file.path("extdata", "who_hc_maxvel_3sd_infants.csv.gz"), package = "growthcleanr"),
    file.path(ref.data.path, "who_hc_maxvel_3sd_infants.csv.gz")
  )

  who_hc_vel_3sd_path <- ifelse(
    ref.data.path == "",
    system.file(file.path("extdata", "who_hc_vel_3sd_infants.csv.gz"), package = "growthcleanr"),
    file.path(ref.data.path, "who_hc_vel_3sd_infants.csv.gz")
  )

  who.max.hc.vel <- fread(who_max_hc_vel_path)
  who.hc.vel <- fread(who_hc_vel_3sd_path)
  setkey(who.max.hc.vel, sex, whoagegrp_ht)
  setkey(who.hc.vel, sex, whoagegrp_ht)
  who.hc.vel <- as.data.table(dplyr::full_join(who.hc.vel, who.max.hc.vel, by =
                                                 c('sex', 'whoagegrp_ht')))

  setnames(who.hc.vel, colnames(who.hc.vel), gsub('_', '.', colnames(who.hc.vel)))
  setkey(who.hc.vel, sex, whoagegrp.ht)

  # create the valid set
  # we only running carried forwards on valid values, non NNTE values,
  # and non single values, and non weight
  tmp <- table(paste0(data.df$subjid, "_", data.df$param))
  not_single <- paste0(data.df$subjid, "_", data.df$param) %in% names(tmp)[tmp > 1]
  valid_set <- valid(data.df, include.temporary.extraneous = FALSE) &
    !data.df$nnte_full & # does not use the "other"
    not_single &
    data.df$param != "WEIGHTKG" # do not run on weight

  # order just for ease later
  data.df <- data.df[order(subjid, param, agedays),]
  data.df <- data.df[valid_set, exclude := (function(df) {
    # work inside a closure to drop added column values

    # save initial exclusions to keep track
    ind_all <- copy(df$index)
    exclude_all <- copy(df$exclude)

    testing <- TRUE

    while (testing & nrow(df) > 1){
      # sort df since it got reordered with keys
      df <- df[order(agedays),]

      # 17A
      df[, d_agedays := dplyr::lead(agedays) - agedays]

      # 17E -- only applies to height
      if (df$param[1] == "HEIGHTCM"){
        # 17B
        df[, tanner.months := 6+12*(round(.5*(agedays + dplyr::lead(agedays))/365.25))]
        # set tanner to missing for smaller values
        df[(agedays/30.4375) < 30, tanner.months := NA]

        # merge with the tanner info
        setkey(df, sex, tanner.months)
        df <- tanner.ht.vel.rev[df]

        # 17C

        # a
        df[max.ht.vel < 2.54, max.ht.vel := 2.54]
        df[d_agedays > 2*30.4375 & max.ht.vel < 2*2.54, max.ht.vel := 2*2.54]
        df[d_agedays > .5*365.25 & max.ht.vel < 4*2.54, max.ht.vel := 4*2.54]
        df[d_agedays > 365.25 & max.ht.vel < 8*2.54, max.ht.vel := 8*2.54]

        # b
        df[d_agedays < 365.25, mindiff := .5*min.ht.vel*(d_agedays/365.25)-3 ]
        df[d_agedays > 365.25, mindiff := .5*min.ht.vel-3 ]
        df[d_agedays < 365.25,
           maxdiff := 2*min.ht.vel*(d_agedays/365.25)^1.5 + 5.5 ]
        df[d_agedays > 365.25,
           maxdiff := 2*min.ht.vel*(d_agedays/365.25)^0.33 + 5.5 ]

        # 17D
        # generate the who age group variable
        df[, whoagegrp.ht := round(agedays/30.4375)]
        df[whoagegrp.ht > 24 | dplyr::lead(whoagegrp.ht) > 24,
           whoagegrp.ht := NA]


        # 17E
        df[d_agedays >= 20 & d_agedays < 46, whoinc.age.ht := 1]
        df[d_agedays >= 46 & d_agedays < 76, whoinc.age.ht := 2]
        df[d_agedays >= 76 & d_agedays < 107, whoinc.age.ht := 3]
        df[d_agedays >= 107 & d_agedays < 153, whoinc.age.ht := 4]
        df[d_agedays >= 153 & d_agedays < 199, whoinc.age.ht := 6]

        # update the edge intervals
        df[d_agedays < 20, whoinc.age.ht := 1]
        df[d_agedays > 199, d_agedays := 200]
        df[d_agedays == 200, whoinc.age.ht := 6]

        # 17F
        # merge with WHO
        # add the column name we want to grab
        m_who_ht_vel <- merge(df, who.ht.vel, by = c("sex", "whoagegrp.ht"),
                               all.x = TRUE, sort = FALSE)
        for (i in unique(df$whoinc.age.ht[!is.na(df$whoinc.age.ht) &
                                          !is.na(df$whoagegrp.ht)])){
          sub_m_who_ht_vel <- m_who_ht_vel[whoinc.age.ht == i,]

          cn <- paste0("whoinc.", i, ".ht")
          df[whoinc.age.ht == i,
             who_mindiff_ht := as.numeric(sub_m_who_ht_vel[, get(cn)])]
          cn <- paste0("max.whoinc.", i, ".ht")
          df[whoinc.age.ht == i,
             who_maxdiff_ht := as.numeric(sub_m_who_ht_vel[, get(cn)])]
        }
        # if there are none, preallocate for ease
        if (length(unique(df$whoinc.age.ht[!is.na(df$whoinc.age.ht) &
                                           !is.na(df$whoagegrp.ht)])) < 1){
          df[, who_mindiff_ht := NA_real_]
          df[, who_maxdiff_ht := NA_real_]
        }
        df[, who_mindiff_ht := as.numeric(who_mindiff_ht)]
        df[, who_maxdiff_ht := as.numeric(who_maxdiff_ht)]

        # 17G
        df[d_agedays < whoinc.age.ht*30.4375, who_mindiff_ht :=
             who_mindiff_ht * d_agedays/(whoinc.age.ht*30.4375)]
        df[d_agedays > whoinc.age.ht*30.4375, who_maxdiff_ht :=
             who_maxdiff_ht * d_agedays/(whoinc.age.ht*30.4375)]

        df[d_agedays < 9*30.4375, who_mindiff_ht := who_mindiff_ht*.5-3]
        df[d_agedays < 9*30.4375, who_maxdiff_ht := who_maxdiff_ht*2+3]

        # 17H
        # tanner is implicit
        # greater than 9 months, use tanner if available, otherwise who
        df[(d_agedays < 9*30.4375 | is.na(min.ht.vel)) &
             !is.na(who_mindiff_ht), mindiff := who_mindiff_ht]
        df[(d_agedays < 9*30.4375 | is.na(min.ht.vel)) &
             !is.na(who_mindiff_ht), maxdiff := who_maxdiff_ht]
        # otherwise, fill in
        df[is.na(mindiff), mindiff := 3]
        # for birth measurements, add allowance of 1.5cm
        df[agedays == 0, mindiff := mindiff - 1.5]
        df[agedays == 0, maxdiff := maxdiff + 1.5]

        # 17I
        # sort df since it got reordered with keys
        df <- df[order(agedays),]
        df[, mindiff_prev := dplyr::lag(mindiff)]
        df[, maxdiff_prev := dplyr::lag(maxdiff)]
      } else { # head circumference
        # 17D
        # generate the who age group variable
        df[, whoagegrp.ht := round(agedays/30.4375)]
        df[whoagegrp.ht > 24 | dplyr::lead(whoagegrp.ht) > 24,
           whoagegrp.ht := NA]

        # 17J
        df[d_agedays >= 46 & d_agedays < 76, whoinc.age.hc := 2]
        df[d_agedays >= 76 & d_agedays < 107, whoinc.age.hc := 3]
        df[d_agedays >= 107 & d_agedays < 153, whoinc.age.hc := 4]
        df[d_agedays >= 153 & d_agedays < 199, whoinc.age.hc := 6]

        # update the edge intervals to missing
        df[d_agedays < 46 | d_agedays > 199, whoinc.age.hc := NA]

        # 17K
        # merge with WHO
        # add the column name we want to grab

        m_who_hc_vel <- merge(df, who.hc.vel, by = c("sex", "whoagegrp.ht"),
                              all.x = TRUE, sort = FALSE)
        for (i in unique(df$whoinc.age.hc[!is.na(df$whoinc.age.hc) &
                                          !is.na(df$whoagegrp.hc)])){
          sub_m_who_hc_vel <- m_who_hc_vel[whoinc.age.ht == i,]

          cn <- paste0("whoinc.", i, ".ht")
          df[whoinc.age.hc == i,
             who_mindiff_hc := as.numeric(sub_m_who_ht_vel[, get(cn)])]
          cn <- paste0("max.whoinc.", i, ".ht")
          df[whoinc.age.hc == i,
             who_maxdiff_hc := as.numeric(sub_m_who_ht_vel[, get(cn)])]
        }
        # if there are none, preallocate for ease
        if (length(unique(df$whoinc.age.hc[!is.na(df$whoinc.age.hc) &
                                           !is.na(df$whoagegrp.hc)])) < 1){
          df[, who_mindiff_hc := NA_real_]
          df[, who_maxdiff_hc := NA_real_]
        }
        df[, who_mindiff_hc := as.numeric(who_mindiff_hc)]
        df[, who_maxdiff_hc := as.numeric(who_maxdiff_hc)]

        # 17L
        df[d_agedays < whoinc.age.hc*30.4375, who_mindiff_hc :=
             who_mindiff_hc * d_agedays/(whoinc.age.hc*30.4375)]
        df[d_agedays > whoinc.age.hc*30.4375, who_maxdiff_hc :=
             who_maxdiff_hc * d_agedays/(whoinc.age.hc*30.4375)]

        df[d_agedays < 9*30.4375, who_mindiff_hc := who_mindiff_hc*.5-1.5]
        df[d_agedays < 9*30.4375, who_maxdiff_hc := who_maxdiff_hc*2+1.5]

        # 17M
        # use who
        df[, mindiff := who_mindiff_hc]
        df[, maxdiff := who_maxdiff_hc]
        # otherwise, fill in
        df[is.na(mindiff), mindiff := -1.5]
        # for birth measurements, add allowance of .5cm
        df[agedays == 0, mindiff := mindiff - .5]
        df[agedays == 0, maxdiff := maxdiff + .5]

        # 17N
        # sort df since it got reordered with keys
        df <- df[order(agedays),]
        df[, mindiff_prev := dplyr::lag(mindiff)]
        df[, maxdiff_prev := dplyr::lag(maxdiff)]
      }

      # 17O: generate ewma
      df[, (ewma.fields) := as.double(NaN)]

      # first, calculate which exponent we want to put through (pass a different
      # on for each exp)
      # subset df to only valid rows
      tmp <- data.frame(
        "before" = abs(df$agedays - c(NA, df$agedays[1:(nrow(df)-1)])),
        "after" = abs(df$agedays - c(df$agedays[2:(nrow(df))], NA))
      )
      maxdiff_e <- sapply(1:nrow(tmp), function(x){max(tmp[x,], na.rm = TRUE)})
      exp_vals <- rep(-1.5, nrow(tmp))
      exp_vals[maxdiff_e > 365.25] <- -2.5
      exp_vals[maxdiff_e > 730.5] <- -3.5
      df[, exp_vals := exp_vals]

      # calculate ewma
      df[, (ewma.fields) := ewma(agedays, tbc.sd, exp_vals, TRUE)]

      # calculate dewma
      df[, `:=`(
        dewma.all = tbc.sd - ewma.all,
        dewma.before = tbc.sd - ewma.before,
        dewma.after = tbc.sd - ewma.after
      )]

      # add differences for convenience
      df[, diff_prev := (v-dplyr::lag(v))]
      df[, diff_next := (dplyr::lead(v)-v)]

      if (nrow(df) > 2){
        # 17P/R: identify pairs and calculate exclusions
        df[, pair := (v-dplyr::lag(v)) < mindiff_prev |
             (dplyr::lead(v)-v) < mindiff |
             (v-dplyr::lag(v)) > maxdiff_prev | (dplyr::lead(v)-v) > maxdiff
        ]
        df[is.na(pair), pair := FALSE]
        df[pair & abs(dewma.before) > dplyr::lag(abs(dewma.after)),
           bef.g.aftm1 := TRUE]
        df[pair & abs(dewma.after) > dplyr::lead(abs(dewma.before)),
           aft.g.aftm1 := TRUE]

        # Q
        df[, val_excl := exclude]
        df[diff_prev < mindiff_prev & bef.g.aftm1, val_excl := "Exclude-Min-diff"]
        df[diff_next < mindiff & aft.g.aftm1, val_excl := "Exclude-Min-diff"]
        df[diff_prev > maxdiff_prev & bef.g.aftm1, val_excl := "Exclude-Max-diff"]
        df[diff_next > maxdiff & aft.g.aftm1, val_excl := "Exclude-Max-diff"]



      } else { # only 2 values
        # 17Q/R -- exclusions for pairs
        df[, val_excl := exclude]
        df[diff_prev < mindiff_prev & abs(tbc.sd) > dplyr::lag(abs(tbc.sd)),
           val_excl := "Exclude-Min-diff"]
        df[diff_next < mindiff & abs(tbc.sd) > dplyr::lead(abs(tbc.sd)),
           val_excl := "Exclude-Min-diff"]
        df[diff_prev > maxdiff_prev & abs(tbc.sd) > dplyr::lag(abs(tbc.sd)),
           val_excl := "Exclude-Max-diff"]
        df[diff_next > maxdiff &  abs(tbc.sd) > dplyr::lead(abs(tbc.sd)),
           val_excl := "Exclude-Max-diff"]
      }

      # figure out if any of the exclusions hit
      count_exclude <- sum(df$val_excl != "Include")
      if (count_exclude > 0){
        if (nrow(df) > 2){
          df[, absval := abs(dewma.all)]
        } else {
          df[, absval := abs(tbc.sd)]
        }

        # choose the highest abssum for exclusion
        idx <- df$index[which.max(df[df$val_excl != "Include",
                                     absval])]

        exclude_all[ind_all == idx] <- df[index == idx, val_excl]

        #set up to continue on
        if (count_exclude > 1){
          testing <- TRUE

          df <- df[index != idx, ]
        } else {
          testing <- FALSE
        }
      } else {
        testing <- FALSE
      }
    }

    return(exclude_all)
  })(copy(.SD)), by = .(subjid, param), .SDcols = colnames(data.df)]

  # 19: 1 or 2 measurements ----

  if (!quietly)
    cat(sprintf(
      "[%s] Exclude 1 or 2 measurements...\n",
      Sys.time()
    ))

  # create the valid set
  # we only running carried forwards on valid values, non NNTE values,
  # and non single values, and non weight
  tmp <- table(paste0(data.df$subjid, "_", data.df$param))
  only_single_pairs <- paste0(data.df$subjid, "_", data.df$param) %in% names(tmp)[tmp <= 2]
  valid_set <- valid(data.df, include.temporary.extraneous = FALSE) &
    !data.df$nnte_full & # does not use the "other"
    only_single_pairs

  # order just for ease later
  data.df <- data.df[order(subjid, param, agedays),]
  data.df <- data.df[valid_set, exclude := (function(df) {
    # save initial exclusions to keep track
    ind_all <- copy(df$index)
    exclude_all <- copy(df$exclude)

    # 19A: is it a single or a pair?
    is_single <- nrow(df) == 1

    # find the DOP (designated other parameter)
    dop <-
      data.df[subjid == df$subjid[1] & param == get_dop(df$param[1]) &
                exclude == "Include",]

    # 19D: calculate the voi comparison
    if (nrow(dop) > 0){
      for (i in 1:nrow(df)){
        comp_val <-
          if (df$agedays[i] %in% dop$agedays){
            abs(dop[dop$agedays == df$agedays[i], tbc.sd] - df[i, tbc.sd])
          } else {
            abs(median(dop$tbc.sd) - df[i, tbc.sd])
          }

        df[i, comp_diff := comp_val]
      }
    } else {
      df[, comp_diff := rep(NA, nrow(df))]
    }

    # 19B/C: for pairs, calculate info
    if (!is_single){
      diff_tbc.sd <- df[2, tbc.sd] - df[1, tbc.sd]
      diff_ctbc.sd <- df[2, ctbc.sd] - df[1, ctbc.sd]
      diff_agedays <- df[2, agedays] - df[1, agedays]

      abs_tbd.sd <- abs(df[1, tbc.sd])
      abs_ctbd.sd <- abs(df[1, ctbc.sd])

      med_dop <-
        if (nrow(dop) > 0){
          median(dop$tbc.sd)
        } else {
          NA
        }
      med_cdop <-
        if (nrow(dop) > 0){
          median(dop$ctbc.sd)
        } else {
          NA
        }

      # 19E: which is larger
      max_ind <- if (!all(is.na(df$comp_diff))){
        which.max(df$comp_diff)
      } else {
        which.max(abs(df$tbc.sd))
      }

      # 19F/G: which exclusion
      if (diff_tbc.sd > 4 & (diff_ctbc.sd > 4 | is.na(diff_ctbc.sd)) &
          diff_agedays >=365.25){
        df[max_ind, exclude := "Exclude-2-meas->1-year"]
      } else if (diff_tbc.sd > 2.5 & (diff_ctbc.sd > 2.5 | is.na(diff_ctbc.sd)) &
                 diff_agedays < 365.25){
        df[max_ind, exclude := "Exclude-2-meas-<1-year"]
      }

      # save the results
      exclude_all <- df$exclude

      # 19H
      # if one needs to get removed, we want to reevaluate as a single
      df <- df[exclude == "Include",]
    }

    if (nrow(df) == 1){
      df[(abs(tbc.sd) > 3 & !is.na(comp_diff) & comp_diff > 5) |
           (abs(tbc.sd) > 5 & is.na(comp_diff)),
         exclude := "Exclude-1-meas"]
    }

    return(exclude_all)
  })(copy(.SD)), by = .(subjid, param), .SDcols = colnames(data.df)]

  # 21: error load ----

  if (!quietly)
    cat(sprintf(
      "[%s] Exclude error load...\n",
      Sys.time()
    ))

  valid_set <- !data.df$nnte_full

  data.df[valid_set,
          err_ratio := sum(!exclude %in%
                             c("Include",
                               "Exclude-SDE-Identical",
                               "Exclude-SDE-All-Exclude",
                               "Exclude-SDE-All-Extreme",
                               "Exclude-SDE-EWMA",
                               "Exclude-SDE-All-Extreme",
                               "Exclude-SDE-One-Day",
                               "Exclude-Carried-Forward",
                               "Exclude-1-CF-deltaZ-<0.05",
                               "Exclude-1-CF-deltaZ-<0.1-wholehalfimp",
                               "Exclude-Teen-2-plus-CF-deltaZ-<0.05",
                               "Exclude-Teen-2-plus-CF-deltaZ-<0.1-wholehalfimp"
                             ))/.N,
          by = c("subjid", "param")]

  data.df[valid_set & err_ratio > .4 & exclude == "Include",
          exclude := "Exclude-Error-load"]


  # end ----

  if (!quietly)
    cat(sprintf("[%s] Completed Batch #%s...\n", Sys.time(), data.df$batch[1]))
  if (!quietly & parallel)
    sink()

  return(data.df[j = .(line, exclude, tbc.sd, param)]) #debugging
}

#'
NULL

#' Example electronic health record (heightcm, weightkg) data.
#'
#' @title syngrowth
#' @description A synthetic set of measurements from ~3,500
#'   subjects generated using Synthea, with measurement errors
#'   for testing with growthcleanr. Contains both pediatric and adult data.
#'
#' @format A data frame with six variables: \code{id}, \code{subjid},
#'   \code{sex}, \code{agedays}, \code{param}, and \code{measurement}
"syngrowth"

# Supporting Adult growthcleanr functions
# Supporting functions for adult piece of algorithm, ordered by step first used (if
# not a convenience function, or EWMA)

# convenience functions ----

#' convenience function -- see if numeric vector falls between two numbers
#' returns boolean vector
#' @keywords internal
#' @noRd
check_between <- function(vect, num_low, num_high, incl = TRUE){
  return(
    if (incl){
      vect <= num_high & vect >= num_low
    } else {
      vect < num_high & vect > num_low
    }
  )
}

#' convenience function -- round to the nearest .x
#' @keywords internal
#' @noRd
round_pt <- function(val, pt){
  return(round(val/pt)*pt)
}

#' convenience function to get remainders for floats (a/b)
#' @keywords internal
#' @noRd
get_float_rem <- function(a, b){
  return(abs(round(a/b) - (a/b)))
}

# EWMA functions ----

#' function to calculate as delta matrix for adults
#' @keywords internal
#' @noRd
as.matrix.delta_dn <- function(agedays) {
  n <- length(agedays)
  delta <- abs(matrix(rep(agedays, n), n, byrow = TRUE) - agedays)

  return(delta)
}

#' Exponentially Weighted Moving Average (EWMA) (daymont implementation)
#'
#' \code{ewma} calculates the exponentially weighted moving average (EWMA) for a set of numeric observations over time.
#'
#' @param agedays Vector of age in days for each z score (potentially transformed to adjust weighting).
#'
#' @param meas Input vector of numeric MEASUREMENT data.
#'
#' @param ewma.exp Exponent to use for weighting.
#'
#' @param ewma.adjacent Specify whether EWMA values excluding adjacent measurements should be calculated.  Defaults to TRUE.
#'
#' @return Data frame with 3 variables:
#' * The first variable (ewma.all) contains the EWMA at observation time
#'   excluding only the actual observation for that time point.
#' * The second variable (ewma.before) contains the EWMA for each observation excluding both the actual observation
#'   and the immediate prior observation.
#' * The third variable (ewma.after) contains the EWMA for each observation excluding both the actual observation
#'   and the subsequent observation.
#' @keywords internal
#' @noRd
ewma_dn <- function(agedays, meas, ewma.exp = -5, ewma.adjacent = TRUE) {
  # 6.  EWMA calculation description: Most of the next steps will involve calculating the exponentially weighted moving average for each subject and parameter. I will
  #     describe how to calculate EWMASDs, and will describe how it needs to be varied in subsequent steps.
  # a.	The overall goal of the EWMASD calculation is to identify the difference between the SD-score and what we might predict that DS-score should be, in order to
  #     determine whether it should be excluded.
  # b.	Only nonmissing SD-scores for a parameter that are not designated for exclusion are included in the following calculations.
  # c.	For each SD-score SDi and associated agedaysi calculate the following for every other measurement (SDj...SDn) and associated agedays (agedaysj...agedaysn)  for the
  #     same subject and parameter
  #   i.	(delta)Agej=agedaysj-agedaysi
  #   ii.	EWMAZ=SDi=[(sigma)j->n(SDj*((5+(delta)Agej)^-1.5))]/[ (sigma)j->n((5+(delta)Agej)^-1.5)]
  #   iii.	For most EWMASD calculations, there are 3 EWMASDs that need to be calculated. I will note if not all of these need to be done for a given step.
  #     1.	EWMASDall calculated as above
  #     2.	EWMAZbef calculated excluding the SD-score just before the SD-score of interest (sorted by agedays). For the first observation for a parameter for a
  #         subject, this should be identical to EWMASDall rather than missing.
  #     3.	EWMAZaft calculated excluding the measurement just after the SD-score of interest (sorted by agedays). For the lastobservation for a parameter for a subject,
  #         this should be identical to EWMASDall rather than missing.
  #   iv.	For each of the three EWMASDs, calculate the dewma_*=SD-EWMASD
  # d.	EWMASDs and (delta)EWMASDs will change if a value is excluded or manipulated using one of the methods below, therefore EWMASDs and (delta)EWMASDs be recalculated for each
  #     step where they are needed.
  # e.	For these calculations, use variables that allow for precise storage of numbers (in Stata this is called 'double') because otherwise rounding errors can cause
  #     problems in a few circumstances

  n <- length(agedays)
  # initialize response variables
  ewma.all <- ewma.before <- ewma.after <- vector('numeric', 0)
  if (n > 0) {
    # organize into data frame and sort into order of increasing age,
    # but retain original sort order information in index
    if (!all(agedays == cummax(agedays)))
      warning("EWMA ordering is not sorted; double check") #add in a check to make sure the inputs are already sorted (they should be)
    index <- order(agedays)

    # calculate matrix of differences in age, and add 5 to each delta per Daymont algorithm
    delta <- as.matrix.delta_dn(agedays)
    delta <- ifelse(delta == 0, 0, (delta) ^ ewma.exp)

    # calculate EWMAs, and return in order of original data
    ewma.all[index] <- delta %*% meas / apply(delta, 1, sum)

    if (ewma.adjacent) {
      if (n > 2) {
        delta2 = delta
        delta2[col(delta2) == row(delta2) - 1] = 0
        ewma.before[index] = delta2 %*% meas / apply(delta2, 1, sum)
        delta3 = delta
        delta3[col(delta3) == row(delta3) + 1] = 0
        ewma.after[index] = delta3 %*% meas / apply(delta3, 1, sum)
      } else {
        ewma.before <- ewma.after <- ewma.all
      }
    }
  }
  # return all 3 EWMAs as a data frame
  return(if (ewma.adjacent)
    data.frame(ewma.all, ewma.before, ewma.after)
    else
      data.frame(ewma.all))
}

# step 1w, W BIV ----

#' function to remove BIVs, based on cutoffs for the given method
#' inputs:
#'   subj_df: data frame with measurements of a given type
#'   type: height, weight, or bmi
#'   biv_df: data frame with BIV cutoffs for the given type
#'   include: default FALSE, whether or not to include the endpoints
#' outputs:
#'   logical, true if the given record should be removed due to being a BIV
#' @keywords internal
#' @noRd
remove_biv <- function(subj_df, type, biv_df, include = FALSE){
  too_low <- remove_biv_low(subj_df, type, biv_df, include)
  too_high <- remove_biv_high(subj_df, type, biv_df, include)

  return(too_low | too_high)
}

#' function to remove only the low end of BIVs, based on cutoffs for the given
#' method (for intermediate processing only)
#' inputs:
#'   subj_df: data frame with measurements of a given type
#'   type: height, weight, or bmi
#'   biv_df: data frame with BIV cutoffs for the given type
#'   include: default FALSE, whether or not to include the endpoints
#' outputs:
#'   logical, true if the given record should be removed due to being a BIV
#' @keywords internal
#' @noRd
remove_biv_low <- function(subj_df, type, biv_df, include = FALSE){
  if (!include){
    too_low <- subj_df$measurement < biv_df[type, "low"]
  } else {
    too_low <- subj_df$measurement <= biv_df[type, "low"]
  }

  return(too_low)
}

#' function to remove only the high end of BIVs, based on cutoffs for the given
#' method (for intermediate processing only)
#' inputs:
#'   subj_df: data frame with measurements of a given type
#'   type: height, weight, or bmi
#'   biv_df: data frame with BIV cutoffs for the given type
#'   include: default FALSE, whether or not to include the endpoints
#' outputs:
#'   logical, true if the given record should be removed due to being a BIV
#' @keywords internal
#' @noRd
remove_biv_high <- function(subj_df, type, biv_df, include = FALSE){
  if (!include){
    too_high <- subj_df$measurement > biv_df[type, "high"]
  } else {
    too_high <- subj_df$measurement >= biv_df[type, "high"]
  }

  return(too_high)
}

# step 2w, W repeated values ----

#' Function to identify repeated values in WEIGHT values
#' adds two colummns to w_subj_df:
#'   is_first_rv: is it the first repeated value?
#'   is_rv: is it a repeated value that is not a first rv
#' @keywords internal
#' @noRd
identify_rv <- function(w_subj_df){
  if (nrow(w_subj_df) > 0){
    # follows a similar process to temp sde (step 3)
    # identify which of these have duplicate values
    tab_vals <- table(w_subj_df$meas_m)
    dup_vals <- names(tab_vals)[tab_vals > 1] # coerces to character

    # if there are any duplicate days, we want to identify the first
    if (length(dup_vals) > 0){
      # flag repeated values, as well as if it's a first value
      w_subj_df$is_first_rv <- w_subj_df$is_rv <- FALSE
      for (dv in dup_vals){
        first_rv <- which(as.character(w_subj_df$meas_m) == dv)[1]
        w_subj_df$is_first_rv[first_rv] <- TRUE
        w_subj_df$is_rv[which(as.character(w_subj_df$meas_m) == dv)[-1]] <- TRUE
      }
    } else {
      # no repeated values
      w_subj_df$is_first_rv <- w_subj_df$is_rv <- FALSE
    }
  }

  return(w_subj_df)
}

# step 3, temp extraneous ----

#' Function to identify temporary same days extraneous
#' adds a column to subj_df : "extraneous" designating whether or not the row is
#' temporarily extraneous (not to be considered in the future)
#' ptype: height or weight, weight excludes repeated values
#' @keywords internal
#' @noRd
temp_sde <- function(subj_df, ptype = "height"){
  # identify which of these have duplicate days
  tab_days <- table(subj_df$age_days)
  dup_days <- names(tab_days)[tab_days > 1] # coerces to character

  if (nrow(subj_df) >= 2){
    # can't do this without having no duplicate days
    med_wo <-
      if (sum(!as.character(subj_df$age_days) %in% dup_days) > 0){
        # get the median without duplicate days
        median(subj_df$measurement[
            if (ptype == "weight"){
              !subj_df$is_rv
            } else {
              rep(TRUE, nrow(subj_df))
            }
        ])
        # if it's weight, we also don't want to include RV in median calculation

        # distribute that median out duplicate days
      } else {
        # for those with other nonduplicate parameters, make the median 0
        0
      }

    subj_df$diff <- NA
    subj_df$diff[as.character(subj_df$age_days) %in% dup_days] <-
      abs(subj_df$measurement[as.character(subj_df$age_days) %in% dup_days] -
            med_wo)

    # flag extraneous values on the same day that is not the minimum distance
    # from the median sd score
    subj_df$extraneous <- FALSE
    for (dd in dup_days){
      minz <-which.min(subj_df$diff[as.character(subj_df$age_days) == dd])
      subj_df$extraneous[as.character(subj_df$age_days) == dd][-minz] <-
        TRUE
    }

  } else if (nrow(subj_df) > 0){
    # nothing is duplicate
    subj_df$extraneous <- FALSE
  }

  return(subj_df)
}

# function to redo repeated values after you've done a temporary same day extraneous
# w_subj_df: weight subject data.table
#' @keywords internal
#' @noRd
redo_identify_rv <- function(w_subj_df){
  # redo RVs just if any first RVs became extraneous
  if (nrow(w_subj_df) > 0 & any(w_subj_df$extraneous & w_subj_df$is_first_rv)){
    inc_df <- copy(w_subj_df[!w_subj_df$extraneous,])
    inc_df <- identify_rv(inc_df)

    # reassign new rvs to weight df -- ordered the same way
    w_subj_df$is_first_rv <- w_subj_df$is_rv <- FALSE
    w_subj_df$is_rv[w_subj_df$id %in% inc_df$id] <- inc_df$is_rv
    w_subj_df$is_first_rv[w_subj_df$id %in% inc_df$id] <- inc_df$is_first_rv
  }

  return(w_subj_df)
}

# step 5, hundreds ----

#' function to identify hundred exclusions
#' inc_df: subj_df with temp extraneous/first rv removed
#' dewma: delta ewma, for metric
#' meas_col: meas_im or meas_m
#' hundreds: 100/200/300, etc.
#' ptype: param type, "weight" or "height"
#' mtype: m or imp (metric or imperial)
#' returns criteria (true if implausible)
#' @keywords internal
#' @noRd
rem_hundreds <- function(inc_df, dewma, meas_col, hundreds, ptype = "weight"){
  # calculate difference between values -- ENDS ARE PROTECTED ON EITHER SIDE
  inc_df$diff_prev <- c(NA, diff(unlist(inc_df[, meas_col, with = FALSE])))
  inc_df$diff_next <- c(diff(unlist(inc_df[, meas_col, with = FALSE])), NA)

  # state upper and lower limits (hundreds +/- 2)
  # modifier for height vs weight
  modifier <- if (ptype == "height"){
    5.08
  } else {
    2
  }
  # conversion for imperial for weight (none for height)
  div_modifier <- if (grepl("_m", meas_col) | ptype == "height"){
    1
  } else {
    2.2046226
  }
  # these are metric limits
  llimit <- (hundreds / div_modifier) - modifier
  ulimit <- (hundreds / div_modifier) + modifier
  # these are imperial limits
  llimit_imp <- if (ptype == "height" | grepl("_m", meas_col)){
    llimit
  } else {
    hundreds - 2*2.2046226
  }
  ulimit_imp <- if (ptype == "height" | grepl("_m", meas_col)){
    ulimit
  } else {
    hundreds + 2*2.2046226
  }

  # identify hundred exclusions -- only hundred down
  exc_up <-
    (check_between(dewma$dewma.all, llimit, ulimit)) &
    (check_between(dewma$dewma.before, llimit, ulimit)) &
    (check_between(dewma$dewma.after, llimit, ulimit)) &
    (check_between(inc_df$diff_prev, llimit_imp, ulimit_imp) |
       check_between(inc_df$diff_next, llimit_imp, ulimit_imp))
  exc_down <-
    (check_between(dewma$dewma.all, -ulimit, -llimit)) &
    (check_between(dewma$dewma.before, -ulimit, -llimit)) &
    (check_between(dewma$dewma.after, -ulimit, -llimit)) &
    (check_between(inc_df$diff_prev, -ulimit_imp, -llimit_imp) |
       check_between(inc_df$diff_next, -ulimit_imp, -llimit_imp))
  exc_hundred <- if (ptype == "height"){
    exc_down
  } else {
    exc_down | exc_up
  }
  # end criteria depends on the number of distinct values
  criteria <-
    if ((ptype == "height" & length(unique(inc_df$meas_m)) > 2) |
        (ptype == "weight" & length(inc_df$meas_m) > 2)){
      exc_hundred
    } else {
      exc_hundred &
        if (ptype == "height"){
          inc_df$meas_m < 100
        } else {
          inc_df$meas_m < 40 | inc_df$meas_m > 182
        }
    }
  criteria[is.na(criteria)] <- FALSE

  return(criteria)
}

# step 6, unit errors ----

#' function to remove unit errors
#' inc_df: subj_df with temp extraneous/first rv removed
#' ptype: height or weight
#' returns criteria (true if implausible)
#' @keywords internal
#' @noRd
rem_unit_errors <- function(inc_df, ptype = "height"){
  # add "unit error": metric encoded as imperial
  inc_df$ue <- inc_df$meas_m * (if (ptype == "height"){ 2.54 } else {2.2046226})

  # calculate ewma (using metric)
  ewma_res <- ewma_dn(inc_df$age_days, inc_df$meas_m)
  dewma <- (inc_df$meas_m- ewma_res)
  # delta ewma with unit error
  absdewma_ue <- abs(inc_df$ue - ewma_res)
  colnames(dewma) <- colnames(absdewma_ue) <-
    paste0("d",colnames(ewma_res))

  # calculate difference between values
  inc_df$abs_ue_prev <- c(NA, abs(inc_df$ue[2:nrow(inc_df)] -
                                    inc_df$meas_m[1:(nrow(inc_df)-1)]))
  inc_df$abs_ue_next <- c(abs(inc_df$ue[1:(nrow(inc_df)-1)] -
                                inc_df$meas_m[2:nrow(inc_df)]), NA)

  # identify unit error exclusions
  exc_ue <- if (ptype == "height"){
    dewma$dewma.all < -80 &
      dewma$dewma.before < -80 &
      dewma$dewma.after < -80 &
      (absdewma_ue$dewma.all <= 2.54 | inc_df$abs_ue_prev <= 2.54 |
         inc_df$abs_ue_next <= 2.54)
  } else {
    dewma$dewma.all > 40 &
      dewma$dewma.before > 40 &
      dewma$dewma.after > 40 &
      (absdewma_ue$dewma.all <= 2 | inc_df$abs_ue_prev <= 2 |
         inc_df$abs_ue_next <= 2)
  }

  # end criteria depends on the number of distinct values
  criteria <-
    if ((ptype == "height" & length(unique(inc_df$meas_m)) > 2) &
        (ptype == "weight" & length(inc_df$meas_m) > 2)){
      exc_ue
    } else {
      exc_ue &
        if (ptype == "height"){
          inc_df$meas_m < 100
        } else {
          inc_df$meas_m > 182
        }
    }
  criteria[is.na(criteria)] <- FALSE

  return(criteria)
}

# step 7h, transpositions ----

# get ones/tens places of a given number
# returns vector of desired digits
#' @keywords internal
#' @noRd
get_num_places <- function(num, place){
  place_map <- c(
    "ones" = 1,
    "tens" = 2
  )

  # there should always be a tens place -- BIVs filter out single digits
  res <- sapply(strsplit(as.character(num), ""), function(x){
    if ("." %in% x){
      as.numeric(x[which(x == ".") - place_map[place]])
    } else {
      as.numeric(x[length(x) + 1 - place_map[place]])
    }
  })

  return(res)
}

#' function to switch the ones and tens place digit of a number
#' returns vector of switched numbers
#' @keywords internal
#' @noRd
switch_tens_ones <- function(num){
  # gets 10s and ones digits
  tens <- get_num_places(num, "tens")
  ones <- get_num_places(num, "ones")

  # get everything to the left of the tens place
  left_num <- sapply(strsplit(as.character(num), ""), function(x){
    if (length(x) == 2){
      ""
    } else if ("." %in% x){
      paste(x[1:(which(x == ".") - 3)], collapse = "")
    } else {
      paste(x[1:(length(x) + 1 - 3)], collapse = "")
    }
  })
  # get everything to right of ones place
  right_num <- sapply(strsplit(as.character(num), ""), function(x){
    if ("." %in% x){
      paste(x[(which(x == ".")):length(x)], collapse = "")
    } else {
      ""
    }
  })

  # return number with tens and ones place switched
  return(as.numeric(paste0(left_num, ones, tens, right_num)))
}

#' function to calculate transpositions
#' inc_df: subj_df with temp extraneous/first rv removed
#' ptype: height or weight
#' returns criteria (true if implausible)
#' @keywords internal
#' @noRd
rem_transpositions <- function(inc_df, ptype = "height"){
  # calculate ewma (using metric)
  ewma_res <- ewma_dn(inc_df$age_days, inc_df$meas_m)
  dewma <- (inc_df$meas_m- ewma_res)
  colnames(dewma) <- paste0("d",colnames(ewma_res))

  criteria <- rep(FALSE, nrow(inc_df))
  for (mtype in c("m", "im")){
    inc_df$transpo <- switch_tens_ones(
      unlist(inc_df[, paste0("meas_", mtype), with = FALSE])
      )

    # if imperial, we want to convert to metric
    if (mtype == "im"){
      inc_df$transpo <- inc_df$transpo /
        (if (ptype == "height"){ 2.54 } else {2.2046226})
    }

    inc_df$ones <- get_num_places(
      unlist(inc_df[, paste0("meas_", mtype), with = FALSE]), "ones"
    )
    inc_df$tens <- get_num_places(
      unlist(inc_df[, paste0("meas_", mtype), with = FALSE]), "tens"
    )

    absdewma_transpo <- abs(inc_df$transpo - ewma_res)
    colnames(absdewma_transpo) <- paste0("d",colnames(ewma_res))

    # calculate difference between values
    inc_df$abs_transpo_prev <- c(NA, abs(inc_df$transpo[2:nrow(inc_df)] -
                                           inc_df$meas_m[1:(nrow(inc_df)-1)]))
    inc_df$abs_transpo_next <- c(abs(inc_df$transpo[1:(nrow(inc_df)-1)] -
                                       inc_df$meas_m[2:nrow(inc_df)]), NA)

    # transposition cutoff
    tcut <- if (ptype == "height"){10} else {30}
    # allowance for change in height/weight
    allowance <- if (ptype == "height"){2.54} else {2}

    # identify transposition exclusions
    exc_transpo <-
      ((abs(dewma$dewma.all) > tcut &
          abs(dewma$dewma.before) > abs(.9*dewma$dewma.all) &
          abs(dewma$dewma.after) > abs(.9*dewma$dewma.all)) |
         (abs(dewma$dewma.all) < -1*tcut &
            abs(dewma$dewma.before) < abs(-.9*dewma$dewma.all) &
            abs(dewma$dewma.after) < abs(-.9*dewma$dewma.all))) &
      (absdewma_transpo$dewma.all <= allowance |
         inc_df$abs_transpo_prev <= allowance |
         inc_df$abs_transpo_next <= allowance) &
      abs(inc_df$tens - inc_df$ones) >= 3

    criteria <- criteria | exc_transpo
    criteria[is.na(criteria)] <- FALSE
  }

  return(criteria)
}

# step 10 hab, H distinct values ----

#' function to calculate height growth allowance
#' @keywords internal
#' @noRd
ht_allow <- function(velocity, ageyears1, ageyears2){
  return(
    velocity*(log(ageyears2 - 16.9)) - (velocity*log(ageyears1 - 16.9))
  )
}

#' function to generate height growth/loss groups
#' returns either empty lists or too long lists if it fails
#' @keywords internal
#' @noRd
ht_change_groups <- function(h_subj_df, cutoff, type = "loss"){
  # already ordered by age
  glist <- galist <- list()
  # keep track of some current group variables
  cg <-  1 # current group
  glist[[cg]] <- setNames(h_subj_df$meas_m[1], h_subj_df$id[1])
  galist[[cg]] <- h_subj_df$age_years[1]
  for (m in 2:nrow(h_subj_df)){
    cm <- h_subj_df$meas_m[m] # current measurement

    # find range of group with current measurement added
    crng <- max(c(glist[[cg]], cm)) - min(c(glist[[cg]], cm))

    # store temporary difference between current and minimum/maximum
    temp_mindiff <- min(glist[[cg]]) - cm
    temp_maxdiff <- max(glist[[cg]]) - cm

    # if the range is below 2 inches with the added value, add and move on
    # we're also going to set the names on the measurements as ids for ease later
    if (crng < (5.08 + .001)){
      glist[[cg]] <- setNames(c(glist[[cg]], cm),
                              c(names(glist[[cg]]), h_subj_df$id[m]))
      galist[[cg]] <- c(galist[[cg]], h_subj_df$age_years[m])
    } else {
      # otherwise, we add a new group for the current measurement
      cg <- cg + 1
      glist[[cg]] <- setNames(cm, h_subj_df$id[m])
      galist[[cg]] <- h_subj_df$age_years[m]
    }

    # if there are too many groups, we're going to stop -- we're not doing
    # this test
    if (cg > cutoff){
      break
    }

    # if you're going the wrong direction, you're out
    if (type == "loss"){
      if (temp_mindiff < -(5.08 + .001)){
        glist <- galist <- list()
        break
      }
    } else { # type is gain
      if (temp_maxdiff > (5.08 + .001)){
        glist <- galist <- list()
        break
      }
    }
  }

  return(list(
    "meas" = glist,
    "age" = galist
  ))
}

#' function to compare growth for 3D height groups
#' compare: before or first
#' returns whether or not to use original exclusions
#' @keywords internal
#' @noRd
ht_3d_growth_compare <- function(mean_ht, min_age, glist,
                                 compare = "before"){
  # preallocating on whether or not we want to go by original exclusion
  origexc <- FALSE
  for (i in 2:6){
    # if there are no members of the group, we want to skip
    if (i > length(glist)){
      next
    }

    # for ease, creating reference variables
    check_num <- if (compare == "before"){ i - 1} else {1}
    ageyears1 <- min_age[check_num]
    ageyears2 <- min_age[i]
    mh1 <- mean_ht[check_num]
    mh2 <- mean_ht[i]

    # check based on growth
    htcompare <- ifelse(ageyears2 > 25, 25, ageyears2)

    # using short circuiting
    hta <-
      if ((htcompare - ageyears1) < 1){
        ht_allow(20, ageyears1, htcompare)
      } else if ((htcompare - ageyears1) <= 3){
        ht_allow(15, ageyears1, htcompare)
      } else if ((htcompare - ageyears1) > 3){
        ht_allow(12, ageyears1, htcompare)
      }

    origexc <- origexc |
      ((mh2 - mh1) < 0 |
      (mh2 - mh1) > hta)
  }

  return(origexc)
}

# Step 9w, W extreme EWMA ----

#' Function to remove data based on exponentially-weighted moving average
#' (Daymont, et al.) for WEIGHT. Cutoff defaults adjusted for adults.
#' inputs:
#' subj_df: subject data frame, which has age in days and measurement
#' ewma_cutoff: EWMA past which considered invalid (center value). left and right
#'   are .5 less.
#' outputs:
#'  logical indicating whether to exclude a record
#' @keywords internal
#' @noRd
remove_ewma_wt <- function(subj_df, ewma_cutoff_low = 60,
                           ewma_cutoff_high = 100){
  orig_subj_df <- subj_df

  # all three need to be beyond a cutoff for exclusion
  # exclude the most extreme, then recalculate again and again
  rem_ids <- c()
  change <- TRUE
  iter <- 1
  while (change){
    # figure out time difference between points
    agedays_bef <- c(Inf, diff(subj_df$age_days))
    agedays_aft <- c(diff(subj_df$age_days), Inf)
    # both year: different cutoffs if they're at least a year apart
    both_year <- agedays_bef > 365.25 & agedays_aft > 365.25

    # calculate ewma
    ewma_res <- ewma_dn(subj_df$age_days, subj_df$meas_m)

    dewma <- subj_df$meas_m - ewma_res
    colnames(dewma) <- paste0("d",colnames(ewma_res))

    criteria_low <-
      !both_year &
      ((dewma$dewma.all > ewma_cutoff_low &
          dewma$dewma.before > .9*dewma$dewma.all &
          dewma$dewma.after > .9*dewma$dewma.all) |
         (dewma$dewma.all < -1*ewma_cutoff_low &
            dewma$dewma.before < -.9*dewma$dewma.all &
            dewma$dewma.after < -.9*dewma$dewma.all))
    criteria_high <-
      both_year &
      ((dewma$dewma.all > ewma_cutoff_high &
          dewma$dewma.before > .9*dewma$dewma.all &
          dewma$dewma.after > .9*dewma$dewma.all) |
         (dewma$dewma.all < -1*ewma_cutoff_high &
            dewma$dewma.before < -.9*dewma$dewma.all &
            dewma$dewma.after < -.9*dewma$dewma.all))
    criteria_new <- criteria_low | criteria_high
    criteria_new[is.na(criteria_new)] <- FALSE

    if (all(!criteria_new)){
      # if none of them are to be removed
      change <- FALSE
      # if using intermediate values, we want to keep some
    } else {
      # figure out the most extreme value and remove it and rerun
      to_rem <- which.max(abs(dewma$dewma.all)[criteria_new])

      # keep the ids that failed and remove
      rem_ids[length(rem_ids)+1] <- unlist(subj_df[criteria_new,][to_rem, "id"])
      subj_df <- subj_df[subj_df$id != rem_ids[length(rem_ids)],]
      # update iteration
      iter <- iter + 1

      # check if this is viable -- you need at least three points, otherwise
      # we're done
      if (nrow(subj_df) < 3){
        change <- FALSE
      }
    }
  }

  # form results into a logical vector
  criteria <- rep(FALSE, nrow(orig_subj_df))
  criteria[orig_subj_df$id %in% rem_ids] <- TRUE

  return(criteria)
}

# Step 11wb, W moderate EWMA ----

#' Function to remove data based on exponentially-weighted moving average
#' (Daymont, et al.) for WEIGHT. Moderate cutoff defaults adjusted for adults.
#' inputs:
#' full_inc_df: subject data frame, which has age in days and measurement
#' outputs:
#'  logical indicating whether to exclude a record
#' @keywords internal
#' @noRd
remove_mod_ewma_wt <- function(full_inc_df){
  inc_df <- copy(full_inc_df)

  # exclude the most extreme, then recalculate again and again
  rem_ids <- c()
  change <- TRUE
  iter <- 1
  while (change){
    # set a limit for wts to be percent of other wts, focused on lower wts
    perc_limit <- rep(.7, nrow(inc_df))
    perc_limit[inc_df$meas_m > 45] <- .4

    # figure out time difference between points
    ageyears_bef <- c(Inf, diff(inc_df$age_years))
    ageyears_aft <- c(diff(inc_df$age_years), Inf)
    minagediff <- ageyears_bef
    minagediff[ageyears_aft < ageyears_bef] <-
      ageyears_aft[ageyears_aft < ageyears_bef]
    # convert to days - rounded
    agedays_bef <- round(ageyears_bef*365.25)
    agedays_aft <- round(ageyears_aft*365.25)

    # figure out weight difference between points
    wt_bef <- c(NA, diff(inc_df$meas_m))
    wt_aft <- c(diff(inc_df$meas_m), NA)

    # 'polation (inter/extra-polation)
    # "interpolation" - between prior and next with error of 5 on either end
    binerr_interpol <-
      if (nrow(inc_df) >= 3){
        c(NA, sapply(2:(nrow(inc_df)-1), function(x){
          check_between(inc_df$meas_m[x],
                        inc_df$meas_m[x-1]-5, inc_df$meas_m[x+1]+5) |
            check_between(inc_df$meas_m[x],
                          inc_df$meas_m[x+1]-5, inc_df$meas_m[x-1]+5)
        }), NA)
      } else {
        c(NA, NA)
      }
    # extrapolation -- prior weights
    lepolate_p <- binerr_lepolate_p <- c(rep(NA,2))
    if (nrow(inc_df) >= 3){
      for (x in 3:nrow(inc_df)){
        slope <- (inc_df$meas_m[x-1] - inc_df$meas_m[x-2])/
          (inc_df$age_days[x-1] - inc_df$age_days[x-2])
        lepolate_p <- c(lepolate_p, round_pt(
          inc_df$meas_m[x-1] +
            slope*(inc_df$meas_m[x]-inc_df$meas_m[x-1]),
          .2
        ))

        # is current value between extrapolated and 2 previous
        binerr_lepolate_p <- c(
          binerr_lepolate_p,
          check_between(inc_df$meas_m[x],
                        inc_df$meas_m[x - 2] - 5, lepolate_p[x] + 5) |
            check_between(inc_df$meas_m[x],
                          lepolate_p[x] - 5, inc_df$meas_m[x - 2] + 5)
        )
      }
    }
    # extrapolation -- next weights
    lepolate_n <- binerr_lepolate_n <- c()
    if (nrow(inc_df) >= 3){
      for (x in 1:(nrow(inc_df)-2)){
        slope <- (inc_df$meas_m[x+1] - inc_df$meas_m[x+2])/
          (inc_df$age_days[x+1] - inc_df$age_days[x+2])
        lepolate_n <- c(lepolate_n, round_pt(
          inc_df$meas_m[x+1] +
            slope*(inc_df$meas_m[x+1]-inc_df$meas_m[x]),
          .2
        ))

        # is current value between extrapolated and 2 next
        binerr_lepolate_n <- c(
          binerr_lepolate_n,
          check_between(inc_df$meas_m[x],
                        inc_df$meas_m[x + 2] - 5, lepolate_n[x] + 5) |
            check_between(inc_df$meas_m[x],
                          lepolate_n[x] - 5, inc_df$meas_m[x + 2] + 5)
        )
      }
    }
    lepolate_n <- c(lepolate_n, rep(NA,2))
    binerr_lepolate_n <- c(binerr_lepolate_n, rep(NA,2))

    # compute "weight allow" how much change is allowed over time
    wta <- 4 + 18*log(1 + (minagediff*12))
    # cap at 60
    wta[wta > 60] <- 60

    # calculate ewma
    ewma_res <- ewma_dn(inc_df$age_days, inc_df$meas_m)

    dewma <- inc_df$meas_m - ewma_res
    colnames(dewma) <- paste0("d",colnames(ewma_res))

    # moderate EWMA exclusion criteria
    exc_mod_ewma <-
      (dewma$dewma.all > wta &
         dewma$dewma.before > (.75*wta) &
         dewma$dewma.after > (.75*wta)) |
      (dewma$dewma.all < -1*wta &
         dewma$dewma.before < (-1*.75*wta) &
         dewma$dewma.after < (-1*.75*wta))

    # alternate ewma criteria -- if difference with adjacent within wta and
    # difference in agedays <= 14
    # prior is close in age but far in weight
    alt_ewma_exc <-
      agedays_bef <= 14 &
      abs(wt_bef) > wta &
      (dewma$dewma.all > wta &
         dewma$dewma.after > (.75*wta)) |
      (dewma$dewma.all < -1*wta &
         dewma$dewma.after < (-1*.75*wta))
    # next is close in age but far in weight
    alt_ewma_exc <- alt_ewma_exc |
      (agedays_aft <= 14 &
         abs(wt_aft) > wta &
         (dewma$dewma.all > wta &
            dewma$dewma.before > (.75*wta)) |
         (dewma$dewma.all < -1*wta &
            dewma$dewma.before < (-1*.75*wta)))
    exc_mod_ewma[is.na(exc_mod_ewma)] <- FALSE
    alt_ewma_exc[is.na(alt_ewma_exc)] <- FALSE
    exc_mod_ewma <- exc_mod_ewma | alt_ewma_exc

    # identify binerr criteria -- edge ones are marked as true
    binerr_lepolate_n[is.na(binerr_lepolate_n)] <- FALSE
    binerr_lepolate_p[is.na(binerr_lepolate_p)] <- FALSE
    binerr_interpol[is.na(binerr_interpol)] <- FALSE
    exc_binerr <- !binerr_lepolate_p & !binerr_lepolate_n & !binerr_interpol
    # alternate binerr crtieria -- if difference with adjacent within wta and
    # difference in agedays <= 14
    alt_exc_binerr <-
      agedays_bef <= 14 &
      abs(wt_bef) > wta &
      !binerr_lepolate_n
    alt_exc_binerr <- alt_exc_binerr |
      (agedays_aft <= 14 &
         abs(wt_aft) > wta &
         !binerr_lepolate_p)
    exc_binerr <- exc_binerr | alt_exc_binerr
    exc_binerr[is.na(exc_binerr)] <- FALSE

    # ratio to ewma can also lead to exclusion
    percewma <- inc_df$meas_m/ewma_res
    exc_perc <- percewma$ewma.all < perc_limit &
      percewma$ewma.before < perc_limit &
      percewma$ewma.after < perc_limit
    exc_perc[is.na(exc_perc)] <- FALSE

    criteria_new <- (exc_mod_ewma & exc_binerr) | exc_perc

    if (all(!criteria_new)){
      change <- FALSE
    } else {
      # ewma ratio determines which to exclude -- highest ewmaratio
      ewmaratio <- abs(dewma$dewma.all)/wta
      # boost middle values -- be more strict
      ewmaratio[-c(1, length(ewmaratio))] <-
        ewmaratio[-c(1, length(ewmaratio))] + .2

      # figure out the most extreme value and remove it and rerun
      to_rem <- which.max(ewmaratio[criteria_new])

      # keep the ids that failed and remove
      rem_ids[length(rem_ids)+1] <- unlist(inc_df[criteria_new,][to_rem, "id"])
      inc_df <- inc_df[inc_df$id != rem_ids[length(rem_ids)],]

      # check if this is viable -- you need at least three points, otherwise
      # we're done
      if (nrow(inc_df) < 3){
        change <- FALSE
      }
      # update iteration
      iter <- iter + 1
    }
  }

  # form results into a logical vector
  criteria <- rep(FALSE, nrow(full_inc_df))
  criteria[full_inc_df$id %in% rem_ids] <- TRUE

  return(criteria)
}

# Infants support

# supporting (all) ---

# function that takes in a parameter name and returns the designated other
# parameter (DOP)
#' @keywords internal
#' @noRd
get_dop <- function(param_name){
  dop <- if (param_name == "WEIGHTKG"){
    "HEIGHTCM"
  } else if (param_name == "HEIGHTCM"){
    "WEIGHTKG"
  } else { # HEADCM
    "HEIGHTCM"
  }

  return(dop)
}

# temporary extraneous ----
#' Function for temporary extraneous (step 5):
#' 5.  Temporary extraneous: I use extraneous to refer to more than more than one recorded value for a parameter on the same day,
#'     and we need to select which one to include in our analysis. The overall strategy will be to select a measurement using a simple
#'     strategy that will be used temporarily, and select permanently in a later step after we have a somewhat cleaner dataset that
#'     can help us identify the best extraneous.
#' a.  For subjects/parameters with extraneous: Determine median_tbc*sd for both parameters: the median tbc*sd for each subject
#'     and parameter including only non-extraneous values with exc_*==0. The median of the same parameter as the extraneous will
#'     be referred to as median_tbc*sd, the median of the other parameter will be referred to as median_tbcOsd.
#' b.	For each subject/parameter with extraneous and at least one value for the subject/parameter on a day with no extraneous,
#'     select the value closest to the median_tbc*sd for temporary inclusion, and assign all other extraneous exc_*=2.
#'  i.	For each subject/parameter with extraneous and no values for the subject/parameter on a day with no extraneous, select the
#'     value closest to the median_tbcOsd for temporary inclusion, and assign all other extraneous exc_*=2.
#'     If median_tbcOsd is missing because there are no values for the other parameter, randomly choose one extraneous value for
#'     each subject/parameter/age to keep as exc_*=0 and replace exc_*=2 for all other extraneous for that subject/parameter/age.
#'
#' @keywords internal
#' @noRd
temporary_extraneous_infants <- function(df) {
  # avoid "no visible binding" warnings
  agedays <- delta.median.sd <- extraneous <- extraneous.this.day <- NULL
  index <- median.sd <- param <- subjid <- tbc.sd <- NULL

  # add subjid and param if needed (may be missing depending on where this is called from)
  if (is.null(df$subjid))
    df[, subjid := NA]
  if (is.null(df$param))
    df[, param := NA]
  # only operate on valid rows (but include rows that may have previously been flagged as a "temporary extraneous")
  # also do not calculate temporary extraneous for NNTEs
  valid.rows <- valid(df, include.temporary.extraneous = TRUE) &
    !df$nnte
  # make a small copy of df with only the fields we need for efficiency
  df <- df[j = .(tbc.sd), keyby = .(subjid, param, agedays, index)]
  # initialize some useful fields
  df[, `:=`(
    median.sd = as.double(NA),
    delta.median.sd = as.double(NA),
    extraneous.this.day = FALSE,
    extraneous = FALSE
  )]

  # determine on which days there are extraneous measurements (more than 1 valid measurement on that age day)
  df[valid.rows, extraneous.this.day := (.N > 1), by = .(subjid, param, agedays)]
  # calculate median of measurements on days where there is no extraneous
  df[valid.rows &
       !extraneous.this.day, median.sd := median(tbc.sd), by = .(subjid, param)]

  # distribute median to other valid or potential extraneous rows for same parameter for that subject
  df[valid.rows, median.sd := sort(median.sd)[1], by = .(subjid, param)]
  # take care of subject/parameters with more than one day with a valid observation
  # determine the absolute difference between the measurements sd score and the median for that parameter for each child
  df[valid.rows &
       extraneous.this.day, delta.median.sd := abs(tbc.sd - median.sd), by = .(subjid, param)]
  # identify subjects that have duplicates/extraneous on all days of observation for that parameter (i.e. delta.median.sd undefined)
  subj.all.dups <- df[valid.rows &
                        extraneous.this.day &
                        is.na(delta.median.sd), unique(subjid)]

  # add the "other map" -- now including head circumference
  other_p_map <- c(
    "HEIGHTCM" = "WEIGHTKG",
    "WEIGHTKG" = "HEIGHTCM",
    "HEADCM" = "HEIGHTCM"
  )

  df[valid.rows &
       subjid %in% subj.all.dups, delta.median.sd := (function(subj.df) {
         # iterate over parameters where delta.median.sd is not yet defined
         # pass 1: take median from other parameter(s)
         for (p in subj.df[is.na(delta.median.sd) &
                           extraneous.this.day, unique(param)]) {
           median.other.sd <- subj.df[param == other_p_map[p] &
                                        !extraneous.this.day, median(tbc.sd)]
           subj.df[param == p, delta.median.sd := abs(tbc.sd - median.other.sd)]
         }
         return(subj.df$delta.median.sd)
       })(copy(.SD)), by = .(subjid)]

  # Final pass: take median as zero (i.e. if no measurements from a different parameter)
  # NOTE -- this is not exactly the same as taking a random parameter
  df[valid.rows &
       extraneous.this.day &
       is.na(delta.median.sd), delta.median.sd := abs(tbc.sd)]

  # flag any extraneous value on the same day that is not the minimum distance from the median sd score
  # NOTE: in the case of exact ducplicates, "which.min" will pick the first
  df[valid.rows &
       extraneous.this.day, extraneous := seq_along(delta.median.sd) != which.min(delta.median.sd), by =
       .(subjid, param, agedays)]
  # return the extraneous valid rows (i.e. the ones that should be temporarily excluded)
  return(df$extraneous & valid.rows)
}

# evil twins ----

#' Function to calculate out of bounds (OOB) measurements for Evil Twins step
#'
#' @param df data table with all parameters
#'
#' @keywords internal
#' @noRd
calc_oob_evil_twins <- function(df){
  # # start by determining if a measurement is out of bounds (oob)

  # for differences for adjacent, we pad with a high number to default to true
  # we don't want to consider the comparison with the next subj/param
  oob <- ((
    (abs(c(df$tbc.sd[2:nrow(df)], Inf) - df$tbc.sd) > 5 &
       rev(duplicated(rev(paste(df$subjid, "_", df$param))))) |
      (abs(c(Inf, df$tbc.sd[1:(nrow(df)-1)]) - df$tbc.sd) > 5 &
         duplicated(paste(df$subjid, "_", df$param)))
  )) & ((
    (abs(c(df$ctbc.sd[2:nrow(df)], Inf) - df$ctbc.sd) > 5 &
       rev(duplicated(rev(paste(df$subjid, "_", df$param))))) |
      (abs(c(Inf, df$ctbc.sd[1:(nrow(df)-1)]) - df$ctbc.sd) > 5 &
         duplicated(paste(df$subjid, "_", df$param)))
  ))

  df[, "oob" := oob]

  return(df)
}

# moderate ewma ----

# function to calculate and recenter z scores for a given column
# df: data frame with parameter, agedays, sex, cn
# cn: column name to calculate, smooth, and recenter
# ref.data.path: reference data path
#
# returns df with additional column, tbc.(cn), which is the recentered z-score
# for the input
#' @keywords internal
#' @noRd
calc_and_recenter_z_scores <- function(df, cn, ref.data.path){
  # avoid no visible warning errors
  cn.orig_cdc <- param <- agedays <- sex <- cn.orig_who <- cn.orig <- subjid <-
    tbc.cn <- sd.median <- NULL

  # for infants, use z and who
  measurement.to.z <- read_anthro(ref.data.path, cdc.only = TRUE,
                                  prelim_infants = TRUE)
  measurement.to.z_who <- read_anthro(ref.data.path, cdc.only = FALSE,
                                      prelim_infants = TRUE)

  # calculate "standard deviation" scores
  df[, cn.orig_cdc := measurement.to.z(param, agedays, sex, get(cn), TRUE)]
  df[, cn.orig_who := measurement.to.z_who(param, agedays, sex, get(cn), TRUE)]

  # smooth z-scores/SD scores between ages 1 - 3yo using weighted scores
  # older uses cdc, younger uses who
  who_weight <- 4 - (df$agedays/365.25)
  cdc_weight <- (df$agedays/365.25) - 2

  smooth_val <- df$agedays/365.25 >= 2 &
    df$agedays/365.25 <= 4 &
    df$param != "HEADCM"
  df[smooth_val,
           cn.orig := (cn.orig_cdc[smooth_val]*cdc_weight[smooth_val] +
                         cn.orig_who[smooth_val]*who_weight[smooth_val])/2]

  # otherwise use WHO and CDC for older and younger, respectively
  who_val <- df$param == "HEADCM" |
    df$agedays/365.25 < 2
  df[who_val | (smooth_val & is.na(df$cn.orig_cdc)),
           cn.orig := df$cn.orig_who[who_val  | (smooth_val & is.na(df$cn.orig_cdc))]]

  cdc_val <- df$param != "HEADCM" |
    df$agedays/365.25 > 4
  df[cdc_val | (smooth_val & is.na(df$cn.orig_who)),
           cn.orig := df$cn.orig_cdc[cdc_val | (smooth_val & is.na(df$sd.orig_who))]]

  # now recenter -- already has the sd.median from the original recentering
  setkey(df, subjid, param, agedays)
  df[, tbc.cn := cn.orig - sd.median]

  # rename ending column
  setnames(df, "tbc.cn", paste0("tbc.", cn))

  return(df)
}

#' Function for LMS formula with modified (m) z-scores
#'
#' @keywords internal
#' @noRd
z_score <- function(var, l, m, s) {
  ls <- l * s
  invl <- 1 / l
  z <- (((var / m) ^ l) - 1) / (ls) # z-score formula
  sdp2 <- (m * (1 + 2 * ls) ^ (invl)) - m
  # modified z-score (+2)
  sdm2 <- m - (m * (1 - 2 * ls) ^ (invl))
  mz <-
    fifelse(var < m, (var - m) / (0.5 * sdm2), (var - m) / (sdp2 * 0.5))
  return(list(z, mz))
}

#' Function to reorder columns of data table
#'
#' @keywords internal
#' @noRd
set_cols_first <- function(DT, cols, intersection = TRUE)
{
  # thanks to hutils
  if (intersection) {
    return(setcolorder(DT, c(
      intersect(cols, names(DT)),
      setdiff(names(DT), cols)
    )))
  }
  else {
    return(setcolorder(DT, c(cols, setdiff(
      names(DT), cols
    ))))
  }
}

#' Calculate extended BMI measures
#'
#' \code{ext_bmiz} Calculates the sigma (scale parameter for the half-normal
#' distribution), extended BMI percentile, extended BMIz, and the CDC LMS
#' Z-scores for weight, height, and BMI for children between 2 and 19.9 years
#' of age. Note that for BMIs <= 95th percentile of the CDC growth charts, the
#' extended values for BMI are equal to the LMS values. The extended values
#' differ only for children who have a BMI > 95th percentile.
#'
#' This function should produce output equivalent to the SAS macro provided at
#' https://www.cdc.gov/growth-chart-training/hcp/computer-programs/sas.html. The macro
#' was updated in December, 2022, according to the findings of the NCHS
#' report available at https://dx.doi.org/10.15620/cdc:121711. This function has
#' been updated to match it as of growthcleanr v2.1.0.
#'
#' The extended BMIz is the inverse cumulative distribution function (CDF) of
#' the extended BMI percentile. If the extended percentile is very close to
#' 100, the \code{qnorm} function in R produces an infinite value. This occurs
#' only if the the extended BMI percentile is > 99.99999999999999. This occurs
#' infrequently, such as a 48-month-old with a BMI > 39, and it is likely that
#' these BMIs represent data entry errors. For these cases, extended BMIz is
#' set to 8.21, a value that is slightly greater than the largest value that
#' can be calculated.
#'
#' See the \code{README.md} file for descriptions of the output columns
#' generated by this function.
#'
#' \code{data} must have columns for at least age, sex, weight, height, and bmi.
#'
#' \code{age} should be coded in months, using the most precise values available.
#' To convert to months from age in years, multiply by 12. To convert to months
#' from age in days, divide by 30.4375 (365.25 / 12).
#'
#' \code{sex} is coded as 1, boys, Boys, b, B, males, Males, m, or M for male
#' subjects or 2, girls, Girls, g, G, females, Females, f, or F for female
#' subjects. Note that this is different from \code{cleangrowth}, which uses
#' 0 (Male) and 1 (Female).
#'
#' \code{wt} should be in kilograms.
#'
#' \code{ht} should be in centimeters.
#'
#' Specify the input data parameter names for \code{age}, \code{wt},
#' \code{ht}, \code{bmi} using quotation marks. See example below.
#'
#' If the parameter \code{adjust.integer.age} is \code{TRUE} (the default),
#' 0.5 will be added to all \code{age} if all input values are integers. Set to
#' \code{FALSE} to disable.
#'
#' By default, the reference data file \code{CDCref_d.csv}, made available at
#' https://www.cdc.gov/growth-chart-training/hcp/computer-programs/sas.html, is included
#' in this package for convenience. If you are developing this package, use
#' \code{ref.data.path} to adjust the path to this file from your working
#' directory if necessary.
#'
#' @param data Input data frame or data table
#'
#' @param age Name of input column containing subject age in months in quotes, default "agem"
#' @param wt Name of input column containing weight (kg) value in quotes, default "wt"
#' @param ht Name of input column containing height (cm) value in quotes, default "ht"
#' @param bmi Name of input column containing calculated BMI in quotes, default "bmi"
#' @param adjust.integer.age If age inputs are all integer, add 0.5 if TRUE;
#'   default TRUE
#' @param ref.data.path Path to directory containing reference data
#'
#' @return Expanded data frame containing computed BMI values
#'
#' @export
#' @import data.table
#' @rawNamespace import(dplyr, except = c(last, first, summarize, src, between))
#' @import labelled
#' @import magrittr
#' @importFrom stats approx pnorm qnorm
#' @rawNamespace import(R.utils, except = c(extract))
#' @examples
#' \donttest{
#' # Run on a small subset of given data
#' df <- as.data.frame(syngrowth)
#' df <- df[df$subjid %in% unique(df[, "subjid"])[1:5], ]
#' df <- cbind(df,
#'             "gcr_result" = cleangrowth(df$subjid,
#'                                        df$param,
#'                                        df$agedays,
#'                                        df$sex,
#'                                        df$measurement))
#' df_wide <- longwide(df) # convert to wide format for ext_bmiz
#' df_wide_bmi <- simple_bmi(df_wide) # compute simple BMI
#'
#' # Calling the function with default column names
#' df_bmiz <- ext_bmiz(df_wide_bmi)
#'
#' # Specifying different column names; note that quotes are used
#' dfc <- simple_bmi(df_wide)
#' colnames(dfc)[colnames(dfc) %in% c("agem", "wt", "ht")] <-
#'   c("agemos", "weightkg", "heightcm")
#' df_bmiz <- ext_bmiz(dfc, age="agemos", wt="weightkg", ht="heightcm")
#'
#' # Disabling conversion of all-integer age in months to (age + 0.5)
#' dfc <- simple_bmi(df_wide)
#' df_bmiz <- ext_bmiz(dfc, adjust.integer.age=FALSE)
#' }
ext_bmiz <- function (data,
                      age = "agem",
                      wt = "wt",
                      ht = "ht",
                      bmi = "bmi",
                      adjust.integer.age = TRUE,
                      ref.data.path = "")
{
  # avoid "no visible binding" warnings
  agemos <- agemos1 <- agemos2 <- agey <- NULL
  bmip95 <- bp <- bz <- denom <- ebp <- ebz <- haz <- l <- NULL
  lbmi1 <- lbmi2 <- lht1 <- lht2 <- lwt1 <- lwt2 <- m <- NULL
  mbmi <- mbmi1 <- mbmi2 <- mht1 <- mht2 <- mref <- mwt1 <- NULL
  mwt2 <- p95 <- s <- sbmi <- sbmi1 <- sbmi2 <- seq_ <- NULL
  sex <- sexn <- sht1 <- sht2 <- sigma <- sref <- swt1 <- NULL
  swt2 <- waz <- z1 <- `_AGEMOS1` <- NULL

  setDT(data)

  setnames(data,
           old = c(age, wt, ht, bmi),
           new = c("age", "wt", "ht", "bmi"))

  # needed for merging back with original data
  data$seq_ <- 1L:nrow(data)
  dorig <- copy(data)
  if (adjust.integer.age) {
    if (isTRUE(all.equal(data$age, round(data$age)))) {
      data[, `:=`(age, age + 0.5)]
    }
  }

  data[, sexn := toupper(substr(sex, 1, 1))]
  data[, sexn := fcase(sexn %in% c(1, 'B', 'M'), 1L,
                       sexn %in% c(2, 'G', 'F'), 2L)]

  data <- data[between(age, 24, 240) & !(is.na(wt) & is.na(ht)),
               .(seq_, sexn, age, wt, ht, bmi)]
  v1 <- c("seq_", "id", "sexn", "age", "wt", "ht", "bmi")

  dref_path <-
    ifelse(
      ref.data.path == "",
      system.file("extdata/CDCref_d.csv.gz", package = "growthcleanr"),
      paste(ref.data.path, "CDCref_d.csv.gz", sep = "")
    )
  # Note: referring to underscore-leading column as `_AGEMOS1`, i.e. with
  # backticks, results in a no visible binding warning, but vars can't start
  # with an "_", so we have to use backticks at assignment up above as well.
  dref <- fread(dref_path)[`_AGEMOS1` > 23 & denom == "age"]
  names(dref) <- tolower(names(dref))
  names(dref) <- gsub("^_", "", names(dref))
  setnames(dref, 'sex', 'sexn')

  d20 <- dref[agemos2 == 240, .(sexn,
                                agemos2,
                                lwt2,
                                mwt2,
                                swt2,
                                lbmi2,
                                mbmi2,
                                sbmi2,
                                lht2,
                                mht2,
                                sht2)]
  names(d20) <- gsub("2", "", names(d20))

  dref <- dref[, .(sexn,
                   agemos1,
                   lwt1,
                   mwt1,
                   swt1,
                   lbmi1,
                   mbmi1,
                   sbmi1,
                   lht1,
                   mht1,
                   sht1)]
  names(dref) <- gsub("1", "", names(dref))

  dref <- rbindlist(list(dref, d20))
  adj_bmi_met <- dref[agemos == 240, .(sexn, mbmi, sbmi)] %>%
    setnames(., c("sexn", "mref", "sref"))
  dref <- dref[adj_bmi_met, on = "sexn"]
  v <- c("sexn",
         "age",
         "wl",
         "wm",
         "ws",
         "bl",
         "bm",
         "bs",
         "hl",
         "hm",
         "hs",
         "mref",
         "sref")
  setnames(dref, v)
  if (length(setdiff(data$age, dref$age)) > 0) {
    uages <- unique(data$age)
    fapprox <- function(i) {
      .d <- dref[sexn == i]
      fapp <- function(vars, ...)
        approx(.d$age, vars,
               xout = uages)$y
      # Note: specifying v with "..v" gives no visible binding warning, use
      # with option
      data.frame(sapply(.d[, v, with = FALSE], fapp))
    }
    dref <- rbindlist(lapply(1:2, fapprox))
  }

  setkey(data, sexn, age)
  setkey(dref, sexn, age)
  dt <- dref[data]

  dt[, `:=`(c("waz", "mwaz"), z_score(dt$wt, dt$wl, dt$wm, dt$ws))]
  dt[, `:=`(c("haz", "mhaz"), z_score(dt$ht, dt$hl, dt$hm, dt$hs))]
  dt[, `:=`(c("bz", "mbz"), z_score(dt$bmi, dt$bl, dt$bm, dt$bs))]

  setDT(dt)
  setnames(dt, c("bl", "bm", "bs"), c("l", "m", "s"))
  dt[, `:=`(c("wl", "wm", "ws", "hl", "hm", "hs"), NULL)]

  dt <- mutate(
    dt,
    bp = 100 * pnorm(.data$bz),
    p95 = .data$m * (1 + .data$l * .data$s * qnorm(0.95)) ^ (1 / .data$l),
    p97 = .data$m * (1 + .data$l * .data$s * qnorm(0.97)) ^ (1 / .data$l),
    bmip95 = 100 * (.data$bmi / .data$p95),
    wp = 100 * pnorm(.data$waz),
    hp = 100 * pnorm(.data$haz),

    # other BMI metrics -- PMID 31439056
    z1 = ((.data$bmi / .data$m) - 1) / .data$s,
    # LMS formula when L=1: ((BMI/M)-1)/S
    dist1 = .data$z1 * .data$m * .data$s,
    # unadjusted distance from median
    adist1 = .data$z1 * .data$sref * .data$mref,
    # Adjusted (to age 20y) dist from median
    perc1 = .data$z1 * 100 * .data$s,
    # unadjusted %distance from median
    aperc1 = .data$z1 * 100 * .data$sref,
    # adj %distance from median

    obese = 1L * (.data$bmi >= .data$p95),
    sev_obese = 1L * (.data$bmip95 >= 120)
  ) %>% setDT()

  # now create Extended z-score for BMI >=95th P
  dt[, `:=`(ebz = bz,
            ebp = bp,
            agey = age / 12)]
  dt[, `:=`(
    sigma,
    fifelse(
      sexn == 1,
      0.3728 + 0.5196 * agey - 0.0091 * agey ^ 2,
      0.8334 + 0.3712 * agey - 0.0011 * agey ^ 2
    )
  )]
  dt[bp >= 95, `:=`(ebp, 90 + 10 * pnorm((bmi - p95) / sigma))]
  dt[bp >= 95, `:=`(ebz, qnorm(ebp / 100))]
  dt[bp > 99 & is.infinite(ebz), `:=`(ebz, 8.21)]

  x <- c("agey", "mref", "sref", "sexn", "wt", "ht", "bmi")
  dt[, `:=`((x), NULL)]
  setnames(
    dt,
    c(
      "adist1",
      "aperc1",
      "bp",
      "bz",
      "mbz",
      "mwaz",
      "mhaz",
      "ebp",
      "ebz",
      "l",
      "m",
      "s"
    ),
    c(
      "adj_dist1",
      "adj_perc1",
      "original_bmip",
      "original_bmiz",
      "mod_bmiz",
      "mod_waz",
      "mod_haz",
      # DF changes
      "bmip",
      "bmiz",
      "bmi_l",
      "bmi_m",
      "bmi_s"
    )
  )

  v <- c(
    "seq_",
    "bmiz",
    "bmip",
    "waz",
    "wp",
    "haz",
    "hp",
    "p95",
    "p97",
    "bmip95",
    "mod_bmiz",
    "mod_waz",
    "mod_haz",
    "sigma",
    "original_bmip",
    "original_bmiz",
    "sev_obese",
    "obese"
  )
  dt <- dt[, v, with = FALSE]

  setkey(dt, seq_)
  setkey(dorig, seq_)
  dtot <- dt[dorig]
  set_cols_first(dtot, names(dorig))
  dtot[, `:=`(c("seq_"), NULL)]

  # Add labels for convenience
  dtot <-
    dtot %>% labelled::set_variable_labels(
      age = "Age (months)",
      ht = "Height (cm)",
      wt = "Weight (kg)",
      bmi = "BMI",
      original_bmiz = "LMS BMI-for-sex/age z-score",
      original_bmip = "LMS BMI-for-sex/age percentile",
      waz = "LMS Weight-for-sex/age z-score",
      wp = "LMS Weight-for-sex/age percentile",
      haz = "LMS Height-for-sex/age z-score",
      hp = "LMS Height-for-sex/age percentile",
      p95 = "95th percentile of BMI in growth charts",
      p97 = "97th percentile of BMI in growth charts",
      bmip95 = "BMI as a percentage of the 95th percentile",
      mod_bmiz = "Modified BMI-for-age z-score",
      mod_waz = "Modified Weight-for-age z-score",
      mod_haz = "Modified Height-for-age z-score",
      sigma = "Scale parameter for half-normal distribution",
      bmip = "LMS / Extended BMI percentile",
      bmiz = "LMS / Extended BMI z-score",
      sev_obese = "BMI >= 120% of 95th percentile (0/1)",
      obese = "BMI >= 95th percentile (0/1)"
    )
  return(dtot[])
}

testthat::skip_on_cran()

test_that("ext_bmiz produces comparable output to CDC SAS implementation", {
  # run ext_bmiz on input data
  mydatapath <-
    system.file(file.path("extdata", "test_syngrowth_wide.csv.gz"), package = "growthcleanr")
  mydata <- read.csv(gzfile(mydatapath))

  # mydata hasn't changed in dimension
  expect_equal(nrow(mydata), 17191)
  expect_equal(ncol(mydata), 6)

  # run the data through ext_bmiz
  myd_bmi <- ext_bmiz(
    mydata,
    age = "agemos",
    wt = "weight",
    ht = "height",
    bmi = "bmi",
    adjust.integer.age = F
  )

  # load the SAS output
  cdcdatapath <-
    system.file(file.path("extdata", "test_syngrowth_sas_output_compare.csv.gz"),
                package = "growthcleanr")
  cdcdata <- read.csv(gzfile(cdcdatapath))

  # compare dimensions to the output of ext_bmiz
  expect_equal(nrow(myd_bmi), nrow(cdcdata))

  # check that all ids are accounted for
  expect_equal(myd_bmi$id, cdcdata$id)

  # reorder for ease of comparison
  myd_bmi <- myd_bmi[order(myd_bmi$id), ]
  cdcdata <- cdcdata[order(cdcdata$id), ]

  # map of column names between ext_bmiz to cdc
  map_to_cdc <- c(
    "bmiz" = "bmiz",
    "bmip" = "bmipct",
    "waz" = "waz",
    "wp" = "wapct",
    "haz" = "haz",
    "hp" = "hapct",
    "p95" = "bmi95",
    "bmip95" = "bmip95",
    "mod_bmiz" = "mod_bmiz",
    "mod_waz" = "mod_waz",
    "mod_haz" = "mod_haz"
  )

  # check all values within a tolerance of 1e-4
  err_df <- abs(as.data.frame(myd_bmi)[, names(map_to_cdc)] -
                  cdcdata[, unname(map_to_cdc)])
  # 99.99% of values should be within tolerance of 1e-4
  expect_gte(sum(err_df <= 1e-4) / nrow(cdcdata) / length(map_to_cdc) *
               100,
             99.99)

})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/growth.R
\name{read_anthro}
\alias{read_anthro}
\title{Function to calculate z-scores and csd-scores based on anthro tables.}
\usage{
read_anthro(path = "", cdc.only = FALSE, prelim_infants = FALSE)
}
\arguments{
\item{path}{Path to supplied reference anthro data. Defaults to package anthro tables.}

\item{cdc.only}{Whether or not only CDC data should be used. Defaults to false.}

\item{prelim_infants}{TRUE/FALSE. Run the in-development release of the infants algorithm (expands pediatric algorithm to improve performance for children 0 – 2 years). Not recommended for use in research. For more information regarding the logic of the algorithm, see the vignette 'Preliminary Infants Algorithm.' Defaults to FALSE.}
}
\value{
Function for calculating BMI based on measurement, age in days, sex, and measurement value.
}
\description{
Function to calculate z-scores and csd-scores based on anthro tables.
}
\examples{
\donttest{
# Return calculating function with all defaults
afunc <- read_anthro()

# Return calculating function while specifying a path and using only CDC data
afunc <- read_anthro(path = system.file("extdata", package = "growthcleanr"),
                     cdc.only = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{growth_cdc_ext}
\alias{growth_cdc_ext}
\title{CDC Growth Percentile Table}
\description{
Part of default CDC-derived tables
}
\details{
Contains percentiles for various ages, gender, and weights, pre-calculated by CDC
}
\section{growthfile_cdc_ext.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/growth.R
\name{cleangrowth}
\alias{cleangrowth}
\title{Clean growth measurements}
\usage{
cleangrowth(
  subjid,
  param,
  agedays,
  sex,
  measurement,
  recover.unit.error = FALSE,
  sd.extreme = 25,
  z.extreme = 25,
  lt3.exclude.mode = "default",
  height.tolerance.cm = 2.5,
  error.load.mincount = 2,
  error.load.threshold = 0.5,
  sd.recenter = NA,
  sdmedian.filename = "",
  sdrecentered.filename = "",
  include.carryforward = FALSE,
  ewma.exp = -1.5,
  ref.data.path = "",
  log.path = NA,
  parallel = FALSE,
  num.batches = NA,
  quietly = TRUE,
  adult_cutpoint = 20,
  weight_cap = Inf,
  adult_columns_filename = "",
  prelim_infants = FALSE
)
}
\arguments{
\item{subjid}{Vector of unique identifiers for each subject in the database.}

\item{param}{Vector identifying each measurement, may be 'WEIGHTKG', 'WEIGHTLBS', 'HEIGHTCM', 'HEIGHTIN', 'LENGTHCM', or 'HEADCM'.
'HEIGHTCM'/'HEIGHTIN' vs. 'LENGTHCM' only affects z-score calculations between ages 24 to 35 months (730 to 1095 days).
All linear measurements below 731 days of life (age 0-23 months) are interpreted as supine length, and
all linear measurements above 1095 days of life (age 36+ months) are interpreted as standing height.
Note: at the moment, all LENGTHCM will be converted to HEIGHTCM. In the future, the algorithm will be updated to consider this difference.
Additionally, imperial 'HEIGHTIN' and 'WEIGHTLBS' measurements are converted to
metric during algorithm calculations.}

\item{agedays}{Numeric vector containing the age in days at each measurement.}

\item{sex}{Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F', 'f' or 1 for females.}

\item{measurement}{Numeric vector containing the actual measurement data.  Weight must be in
kilograms (kg), and linear measurements (height vs. length) in centimeters (cm).}

\item{recover.unit.error}{Indicates whether the cleaning algorithm should
attempt to identify unit errors (I.e. inches vs. cm, lbs vs. kg). If unit
errors are identified, the value will be corrected and retained within the
cleaning algorithm as a valid measurement.  Defaults to FALSE.}

\item{sd.extreme}{Measurements more than sd.extreme standard deviations from
the mean (either above or below) will be flagged as invalid. Defaults to 25.}

\item{z.extreme}{Measurements with an absolute z-score greater than
z.extreme will be flagged as invalid. Defaults to 25.}

\item{lt3.exclude.mode}{Determines type of exclusion procedure to use for 1 or 2 measurements of one type without
matching same ageday measurements for the other parameter. Options include "default" (standard growthcleanr approach),
and "flag.both" (in case of two measurements of one type without matching values for the other parameter, flag both
for exclusion if beyond threshold)}

\item{height.tolerance.cm}{maximum decrease in height tolerated for sequential measurements}

\item{error.load.mincount}{minimum count of exclusions on parameter before
considering excluding all measurements. Defaults to 2.}

\item{error.load.threshold}{threshold of percentage of excluded measurement count to included measurement
count that must be exceeded before excluding all measurements of either parameter. Defaults to 0.5.}

\item{sd.recenter}{specifies how to recenter medians. May be a data frame or
table w/median SD-scores per day of life by gender and parameter, or "NHANES"
or "derive" as a character vector.
\itemize{
\item If \code{sd.recenter} is specified as a data set, use the data set
\item If \code{sd.recenter} is specified as "\code{nhanes}", use NHANES reference medians
\item If \code{sd.recenter} is specified as "\code{derive}", derive from input
\item If \code{sd.recenter} is not specified or \code{NA}:
\itemize{
\item If the input set has at least 5,000 observations, derive medians from input
\item If the input set has fewer than 5,000 observations, use NHANES
}
}

If specifying a data set, columns must include param, sex, agedays, and sd.median
(referred to elsewhere as "modified Z-score"), and those medians will be used
for recentering. A summary of how the NHANES reference medians were derived is
available in README.md. Defaults to NA.}

\item{sdmedian.filename}{Name of file to save sd.median data calculated on the input dataset to as CSV.
Defaults to "", for which this data will not be saved. Use for extracting medians for parallel processing
scenarios other than the built-in parallel option.}

\item{sdrecentered.filename}{Name of file to save re-centered data to as CSV. Defaults to "", for which this
data will not be saved. Useful for post-processing and debugging.}

\item{include.carryforward}{Determines whether Carry-Forward values are kept in the output. Defaults to False.}

\item{ewma.exp}{Exponent to use for weighting measurements in the
exponentially weighted moving average calculations. Defaults to -1.5.
This exponent should be negative in order to weight growth measurements
closer to the measurement being evaluated more strongly. Exponents that are
further from zero (e.g. -3) will increase the relative influence of
measurements close in time to the measurement being evaluated compared to
using the default exponent.}

\item{ref.data.path}{Path to reference data. If not supplied, the year 2000
Centers for Disease Control (CDC) reference data will be used.}

\item{log.path}{Path to log file output when running in parallel (non-quiet mode). Default is NA. A new
directory will be created if necessary. Set to NA to disable log files.}

\item{parallel}{Determines if function runs in parallel.  Defaults to FALSE.}

\item{num.batches}{Specify the number of batches to run in parallel. Only
applies if parallel is set to TRUE. Defaults to the number of workers
returned by the getDoParWorkers function in the foreach package.}

\item{quietly}{Determines if function messages are to be displayed and if log files (parallel only) are to be generated.
Defaults to TRUE}

\item{adult_cutpoint}{Number between 18 and 20, describing ages when the
pediatric algorithm should not be applied (< adult_cutpoint), and the adult
algorithm should apply (>= adult_cutpoint). Numbers outside this range will be
changed to the closest number within the range. Defaults to 20.}

\item{weight_cap}{Positive number, describing a weight cap in kg (rounded to the
nearest .1, +/- .1) within the adult dataset. If there is no weight cap, set
to Inf. Defaults to Inf.}

\item{adult_columns_filename}{Name of file to save original adult data, with additional output columns to
as CSV. Defaults to "", for which this data will not be saved. Useful
for post-analysis. For more information on this output, please see README.}

\item{prelim_infants}{TRUE/FALSE. Run the in-development release of the infants algorithm (expands pediatric algorithm to improve performance for children 0 – 2 years). Not recommended for use in research. For more information regarding the logic of the algorithm, see the vignette 'Preliminary Infants Algorithm.' Defaults to FALSE.}
}
\value{
Vector of exclusion codes for each of the input measurements.

Possible values for each code are:
\itemize{
\item 'Include', 'Unit-Error-High', 'Unit-Error-Low', 'Swapped-Measurements', 'Missing',
\item 'Exclude-Carried-Forward', 'Exclude-SD-Cutoff', 'Exclude-EWMA-Extreme', 'Exclude-EWMA-Extreme-Pair',
\item 'Exclude-Extraneous-Same-Day',
\item 'Exclude-EWMA-8', 'Exclude-EWMA-9', 'Exclude-EWMA-10', 'Exclude-EWMA-11', 'Exclude-EWMA-12', 'Exclude-EWMA-13', 'Exclude-EWMA-14',
\item 'Exclude-Min-Height-Change', 'Exclude-Max-Height-Change',
\item 'Exclude-Pair-Delta-17', 'Exclude-Pair-Delta-18', 'Exclude-Pair-Delta-19',
\item 'Exclude-Single-Outlier', 'Exclude-Too-Many-Errors', 'Exclude-Too-Many-Errors-Other-Parameter'
}
}
\description{
Clean growth measurements
}
\examples{
\donttest{
# Run calculation using a small subset of given data
df_stats <- as.data.frame(syngrowth)
df_stats <- df_stats[df_stats$subjid \%in\% unique(df_stats[, "subjid"])[1:5], ]

clean_stats <-cleangrowth(subjid = df_stats$subjid,
                         param = df_stats$param,
                         agedays = df_stats$agedays,
                         sex = df_stats$sex,
                         measurement = df_stats$measurement)

# Once processed you can filter data based on result value
df_stats <- cbind(df_stats, "clean_result" = clean_stats)
clean_df_stats <- df_stats[df_stats$clean_result == "Include",]

# Parallel processing: run using 2 cores and batches
clean_stats <- cleangrowth(subjid = df_stats$subjid,
                           param = df_stats$param,
                           agedays = df_stats$agedays,
                           sex = df_stats$sex,
                           measurement = df_stats$measurement,
                           parallel = TRUE,
                           num.batches = 2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{simple_bmi}
\alias{simple_bmi}
\title{Compute BMI using standard formula}
\usage{
simple_bmi(wide_df, wtcol = "wt", htcol = "ht")
}
\arguments{
\item{wide_df}{A data frame or data table containing heights and weights in
wide format, e.g., after transformation with longwide()}

\item{wtcol}{name of observation height value column, default 'wt'}

\item{htcol}{name of subject weight value column, default 'ht'}
}
\value{
Returns a data table with the added column "bmi"
}
\description{
\code{simple_bmi} Computes BMI using standard formula. Assumes input compatible with
output from longwide().
}
\examples{
\donttest{
# Simple usage
# Run on a small subset of given data
df <- as.data.frame(syngrowth)
df <- df[df$subjid \%in\% unique(df[, "subjid"])[1:2], ]
df <- cbind(df,
            "gcr_result" = cleangrowth(df$subjid,
                                       df$param,
                                       df$agedays,
                                       df$sex,
                                       df$measurement))
# Convert to wide format
wide_df <- longwide(df)
wide_df_with_bmi <- simple_bmi(wide_df)

# Specifying different column names; note that quotes are used
colnames(wide_df)[colnames(wide_df) \%in\% c("wt", "ht")] <-
  c("weight", "height")
wide_df_with_bmi <- simple_bmi(wide_df, wtcol = "weight", htcol = "height")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{syngrowth}
\alias{syngrowth}
\title{syngrowth}
\format{
A data frame with six variables: \code{id}, \code{subjid},
\code{sex}, \code{agedays}, \code{param}, and \code{measurement}
}
\usage{
syngrowth
}
\description{
A synthetic set of measurements from ~3,500
subjects generated using Synthea, with measurement errors
for testing with growthcleanr. Contains both pediatric and adult data.
}
\details{
Example electronic health record (heightcm, weightkg) data.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{nhanes-reference-medians}
\alias{nhanes-reference-medians}
\title{NHANES reference medians}
\description{
Contains reference median values for default recentering, derived from NHANES
years 2009-2018
}
\section{nhanes-reference-medians.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{rc-reference-medians}
\alias{rc-reference-medians}
\title{Infants reference medians}
\description{
Contains reference median values for default recentering in the infants
algorithm
}
\section{rcfile-2023-08-15_format.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/growth.R
\name{sd_median}
\alias{sd_median}
\title{Calculate median SD score by age for each parameter.}
\usage{
sd_median(param, sex, agedays, sd.orig)
}
\arguments{
\item{param}{Vector identifying each measurement, may be 'WEIGHTKG', or 'HEIGHTCM'.}

\item{sex}{Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F', 'f' or 1 for females.}

\item{agedays}{Numeric vector containing the age in days at each measurement.}

\item{sd.orig}{Vector of previously calculated standard deviation (SD) scores for each measurement before re-centering.}
}
\value{
Table of data with median SD-scores per day of life by gender and parameter.
}
\description{
Calculate median SD score by age for each parameter.
}
\examples{
# Run on 1 subject
df_stats <- as.data.frame(syngrowth)
df_stats <- df_stats[df_stats$subjid == df_stats$subjid[1], ]

# Get the original standard deviations
measurement_to_z <- read_anthro(cdc.only = TRUE)
sd.orig <- measurement_to_z(df_stats$param,
                       df_stats$agedays,
                       df_stats$sex,
                       df_stats$measurement,
                       TRUE)

# Calculate median standard deviations
sd.m <- sd_median(df_stats$param,
                  df_stats$sex,
                  df_stats$agedays,
                  sd.orig)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cdc.R
\name{ext_bmiz}
\alias{ext_bmiz}
\title{Calculate extended BMI measures}
\usage{
ext_bmiz(
  data,
  age = "agem",
  wt = "wt",
  ht = "ht",
  bmi = "bmi",
  adjust.integer.age = TRUE,
  ref.data.path = ""
)
}
\arguments{
\item{data}{Input data frame or data table}

\item{age}{Name of input column containing subject age in months in quotes, default "agem"}

\item{wt}{Name of input column containing weight (kg) value in quotes, default "wt"}

\item{ht}{Name of input column containing height (cm) value in quotes, default "ht"}

\item{bmi}{Name of input column containing calculated BMI in quotes, default "bmi"}

\item{adjust.integer.age}{If age inputs are all integer, add 0.5 if TRUE;
default TRUE}

\item{ref.data.path}{Path to directory containing reference data}
}
\value{
Expanded data frame containing computed BMI values
}
\description{
\code{ext_bmiz} Calculates the sigma (scale parameter for the half-normal
distribution), extended BMI percentile, extended BMIz, and the CDC LMS
Z-scores for weight, height, and BMI for children between 2 and 19.9 years
of age. Note that for BMIs <= 95th percentile of the CDC growth charts, the
extended values for BMI are equal to the LMS values. The extended values
differ only for children who have a BMI > 95th percentile.
}
\details{
This function should produce output equivalent to the SAS macro provided at
https://www.cdc.gov/growth-chart-training/hcp/computer-programs/sas.html. The macro
was updated in December, 2022, according to the findings of the NCHS
report available at https://dx.doi.org/10.15620/cdc:121711. This function has
been updated to match it as of growthcleanr v2.1.0.

The extended BMIz is the inverse cumulative distribution function (CDF) of
the extended BMI percentile. If the extended percentile is very close to
100, the \code{qnorm} function in R produces an infinite value. This occurs
only if the the extended BMI percentile is > 99.99999999999999. This occurs
infrequently, such as a 48-month-old with a BMI > 39, and it is likely that
these BMIs represent data entry errors. For these cases, extended BMIz is
set to 8.21, a value that is slightly greater than the largest value that
can be calculated.

See the \code{README.md} file for descriptions of the output columns
generated by this function.

\code{data} must have columns for at least age, sex, weight, height, and bmi.

\code{age} should be coded in months, using the most precise values available.
To convert to months from age in years, multiply by 12. To convert to months
from age in days, divide by 30.4375 (365.25 / 12).

\code{sex} is coded as 1, boys, Boys, b, B, males, Males, m, or M for male
subjects or 2, girls, Girls, g, G, females, Females, f, or F for female
subjects. Note that this is different from \code{cleangrowth}, which uses
0 (Male) and 1 (Female).

\code{wt} should be in kilograms.

\code{ht} should be in centimeters.

Specify the input data parameter names for \code{age}, \code{wt},
\code{ht}, \code{bmi} using quotation marks. See example below.

If the parameter \code{adjust.integer.age} is \code{TRUE} (the default),
0.5 will be added to all \code{age} if all input values are integers. Set to
\code{FALSE} to disable.

By default, the reference data file \code{CDCref_d.csv}, made available at
https://www.cdc.gov/growth-chart-training/hcp/computer-programs/sas.html, is included
in this package for convenience. If you are developing this package, use
\code{ref.data.path} to adjust the path to this file from your working
directory if necessary.
}
\examples{
\donttest{
# Run on a small subset of given data
df <- as.data.frame(syngrowth)
df <- df[df$subjid \%in\% unique(df[, "subjid"])[1:5], ]
df <- cbind(df,
            "gcr_result" = cleangrowth(df$subjid,
                                       df$param,
                                       df$agedays,
                                       df$sex,
                                       df$measurement))
df_wide <- longwide(df) # convert to wide format for ext_bmiz
df_wide_bmi <- simple_bmi(df_wide) # compute simple BMI

# Calling the function with default column names
df_bmiz <- ext_bmiz(df_wide_bmi)

# Specifying different column names; note that quotes are used
dfc <- simple_bmi(df_wide)
colnames(dfc)[colnames(dfc) \%in\% c("agem", "wt", "ht")] <-
  c("agemos", "weightkg", "heightcm")
df_bmiz <- ext_bmiz(dfc, age="agemos", wt="weightkg", ht="heightcm")

# Disabling conversion of all-integer age in months to (age + 0.5)
dfc <- simple_bmi(df_wide)
df_bmiz <- ext_bmiz(dfc, adjust.integer.age=FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{who_hc_vel_3sd}
\alias{who_hc_vel_3sd}
\title{WHO Head Circumference Velocity for (3\eqn{\sigma})}
\description{
Part of default WHO-derived tables
}
\details{
Contains three standard deviations for the World Health Organization values of
head circumference velocities.
}
\section{who_hc_vel_3sd_infants.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{test_syngrowth_wide}
\alias{test_syngrowth_wide}
\title{CDC SAS BMI Input}
\description{
Contains input data for CDC SAS macro for calculating BMI values.
}
\section{test_syngrowth_wide.csv.gz}{


Used to test function \code{ext_bmiz()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{test_syngrowth_sas_output_compare}
\alias{test_syngrowth_sas_output_compare}
\title{CDC SAS BMI Output}
\description{
Contains results of CDC SAS macro for calculating BMI values.
}
\section{test_syngrowth_sas_output_compare.csv.gz}{


Used to test function \code{ext_bmiz()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{longwide}
\alias{longwide}
\title{Transform data in growthcleanr format into wide structure for BMI calculation}
\usage{
longwide(
  long_df,
  id = "id",
  subjid = "subjid",
  sex = "sex",
  agedays = "agedays",
  param = "param",
  measurement = "measurement",
  gcr_result = "gcr_result",
  include_all = FALSE,
  inclusion_types = c("Include"),
  extra_cols = NULL,
  keep_unmatched_data = FALSE
)
}
\arguments{
\item{long_df}{A data frame to be transformed. Expects columns: id, subjid,
sex, agedays, param, measurement, and gcr_result.}

\item{id}{name of observation ID column}

\item{subjid}{name of subject ID column}

\item{sex}{name of sex descriptor column}

\item{agedays}{name of age (in days) descriptor column}

\item{param}{name of parameter column to identify each type of measurement}

\item{measurement}{name of measurement column containing the actual
measurement data}

\item{gcr_result}{name of column of results from growthcleanr::cleangrowth()}

\item{include_all}{Determines whether the function keeps all exclusion codes.
If TRUE, all exclusion types are kept and the inclusion_types argument is
ignored. Defaults to FALSE.}

\item{inclusion_types}{Vector indicating which exclusion codes from the
cleaning algorithm should be included in the data, given that include_all is
FALSE. For all options, see growthcleanr::cleangrowth(). Defaults to
c("Include").}

\item{extra_cols}{Vector of additional columns to include in the output. If
a column C1 differs on agedays matched height and weight values, then include
separate ht_C1 and wt_C1 columns as well as a match_C1 column that gives
booleans indicating where ht_C1 and wt_C1 are the same. If the agedays
matched height and weight columns are identical, then only include a single
version of C1. Defaults to empty vector (not keeping any additional columns).}

\item{keep_unmatched_data}{boolean indicating whether to keep height/weight
observations that do not have a matching weight/height on that day}
}
\value{
Returns a data frame transformed from long to wide. Includes only
values flagged with indicated inclusion types. Potentially includes
additional columns if arguments are passed to extra_cols. For each subject,
heights without corresponding weights for a given age (and vice versa) will
be dropped unless keep_unmatched_data is set to TRUE.
}
\description{
\code{longwide} transforms data from long to wide format. Ideal for
transforming output from growthcleanr::cleangrowth() into a format suitable
for growthcleanr::ext_bmiz().
}
\examples{
\donttest{
# Run on a small subset of given data
df <- as.data.frame(syngrowth)
df <- df[df$subjid \%in\% unique(df[, "subjid"])[1:2], ]
df <- cbind(df,
            "gcr_result" = cleangrowth(df$subjid,
                                       df$param,
                                       df$agedays,
                                       df$sex,
                                       df$measurement))
# Convert to wide format
wide_df <- longwide(df)

# Include all inclusion types
wide_df <- longwide(df, include_all = TRUE)

# Specify all inclusion codes
wide_df <- longwide(df, inclusion_types = c("Include", "Exclude-Carried-Forward"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{ref_for_corraga}
\alias{ref_for_corraga}
\title{Fenton-Based Corrected Age Reference (Weight to GA Lookup)}
\description{
Reference values for premature infant corrected-age estimation, derived from
data extracted from published Fenton (2013) growth curves. Provides a
reverse lookup from integer weight (grams, rounded to 10) to estimated
gestational age in days, by sex.
}
\section{ref_for_corraga.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{CDCref_d}
\alias{CDCref_d}
\title{CDC BMI reference data}
\description{
Used for extended BMIz computation
}
\section{CDCref_d.csv.gz}{


Used for extended BMI computation
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/growth.R
\name{ewma}
\alias{ewma}
\title{Exponentially Weighted Moving Average (EWMA)}
\usage{
ewma(agedays, z, ewma.exp, ewma.adjacent = TRUE)
}
\arguments{
\item{agedays}{Vector of age in days for each z score (potentially transformed to adjust weighting).}

\item{z}{Input vector of numeric z-score data.}

\item{ewma.exp}{Exponent to use for weighting.}

\item{ewma.adjacent}{Specify whether EWMA values excluding adjacent measurements should be calculated.  Defaults to TRUE.}
}
\value{
Data frame with 3 variables:
\itemize{
\item The first variable (ewma.all) contains the EWMA at observation time
excluding only the actual observation for that time point.
\item The second variable (ewma.before) contains the EWMA for each observation excluding both the actual observation
and the immediate prior observation.
\item The third variable (ewma.after) contains the EWMA for each observation excluding both the actual observation
and the subsequent observation.
}
}
\description{
\code{ewma} calculates the exponentially weighted moving average (EWMA) for a set of numeric observations over time.
}
\examples{
# Run on 1 subject, 1 type of parameter
df_stats <- as.data.frame(syngrowth)
df_stats <- df_stats[df_stats$subjid == df_stats$subjid[1] &
                       df_stats$param == "HEIGHTCM", ]

# Get the uncentered z-scores
measurement_to_z <- read_anthro(cdc.only = TRUE)
sd <- measurement_to_z(df_stats$param,
                       df_stats$agedays,
                       df_stats$sex,
                       df_stats$measurement,
                       TRUE)

# Calculate exponentially weighted moving average
e_df <- ewma(df_stats$agedays, sd, ewma.exp = -1.5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjustcarryforward.R
\name{acf_answers}
\alias{acf_answers}
\title{Answers for adjustcarryforward}
\usage{
acf_answers(
  subjid,
  param,
  agedays,
  sex,
  measurement,
  orig.exclude,
  sd.recenter = NA,
  ewma.exp = -1.5,
  ref.data.path = "",
  quietly = TRUE
)
}
\arguments{
\item{subjid}{Vector of unique identifiers for each subject in the database.}

\item{param}{Vector identifying each measurement, may be 'WEIGHTKG', 'HEIGHTCM', or 'LENGTHCM'
'HEIGHTCM' vs. 'LENGTHCM' only affects z-score calculations between ages 24 to 35 months (730 to 1095 days).
All linear measurements below 731 days of life (age 0-23 months) are interpreted as supine length, and
all linear measurements above 1095 days of life (age 36+ months) are interpreted as standing height.
Note: at the moment, all LENGTHCM will be converted to HEIGHTCM. In the future, the algorithm will be updated to consider this difference.}

\item{agedays}{Numeric vector containing the age in days at each measurement.}

\item{sex}{Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F',
'f' or 1 for females.}

\item{measurement}{Numeric vector containing the actual measurement data.  Weight must be in
kilograms (kg), and linear measurements (height vs. length) in centimeters (cm).}

\item{orig.exclude}{Vector of exclusion assessment results from cleangrowth()}

\item{sd.recenter}{Data frame or table with median SD-scores per day of life}

\item{ewma.exp}{Exponent to use for weighting measurements in the exponentially weighted moving
average calculations. Defaults to -1.5. This exponent should be negative in order to weight growth
measurements closer to the measurement being evaluated more strongly. Exponents that are further from
zero (e.g. -3) will increase the relative influence of measurements close in time to the measurement
being evaluated compared to using the default exponent.}

\item{ref.data.path}{Path to reference data. If not supplied, the year 2000
Centers for Disease Control (CDC) reference data will be used.}

\item{quietly}{Determines if function messages are to be displayed and if log files (parallel only)
are to be generated. Defaults to TRUE.}
}
\value{
A data frame, containing an index "n" of rows, corresponding to the
original order of the input vectors, and "acf_answers", containing the answers
on whether a height value should be kept or excluded (returns "Definitely
Exclude", "Definitely Include", or "Unknown" for height values, NA for weight
values).
}
\description{
Determines what should absolutely be reincluded or definitely excluded
for a given dataset, already run through \code{cleangrowth}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{splitinput}
\alias{splitinput}
\title{Split input data into multiple files}
\usage{
splitinput(
  df,
  fname = deparse(substitute(df)),
  fdir = NA,
  min_nrow = 10000,
  keepcol = "subjid"
)
}
\arguments{
\item{df}{data frame to split}

\item{fname}{new name for each of the split files to start with}

\item{fdir}{directory to put each of the split files (use "." for  working directory). Must be changed from default (NA), which will trigger error.}

\item{min_nrow}{minimum number of rows for each split file (default 10000)}

\item{keepcol}{the column name (default "subjid") to use to keep records with the same values together in the same single split file}
}
\value{
the count number referring to the last split file written
}
\description{
\code{splitinput} Splits input based on keepcol specified, yielding csv files each with at least the mininum
number of rows that are written and saved separately (except for the last split file written, which may be
smaller). Allows splitting input data while ensuring all records for each individual subject will stay together
in one file. Pads split filenames with zeros out to five digits for consistency, assuming < 100,000 file count
result.
}
\examples{
\donttest{
# Run on given data
df <- as.data.frame(syngrowth)

# Run with all defaults (specifying directory)
splitinput(df, fdir = tempdir())

# Specifying the name, directory and minimum row size
splitinput(df, fname = "syngrowth", fdir = tempdir(), min_nrow = 5000)

# Specifying a different subject ID column
colnames(df)[colnames(df) == "subjid"] <- "sub_id"
splitinput(df, fdir = tempdir(), keepcol = "sub_id")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adjustcarryforward.R
\name{adjustcarryforward}
\alias{adjustcarryforward}
\title{adjustcarryforward
\code{adjustcarryforward} Uses absolute height velocity to identify values
excluded as carried forward values for reinclusion.}
\usage{
adjustcarryforward(
  subjid,
  param,
  agedays,
  sex,
  measurement,
  orig.exclude,
  exclude_opt = 0,
  sd.recenter = NA,
  ewma.exp = -1.5,
  ref.data.path = "",
  quietly = TRUE,
  minfactor = 0.5,
  maxfactor = 2,
  banddiff = 3,
  banddiff_plus = 5.5,
  min_ht.exp_under = 2,
  min_ht.exp_over = 0,
  max_ht.exp_under = 0.33,
  max_ht.exp_over = 1.5
)
}
\arguments{
\item{subjid}{Vector of unique identifiers for each subject in the database.}

\item{param}{Vector identifying each measurement, may be 'WEIGHTKG', 'HEIGHTCM', or 'LENGTHCM'
'HEIGHTCM' vs. 'LENGTHCM' only affects z-score calculations between ages 24 to 35 months (730 to 1095 days).
All linear measurements below 731 days of life (age 0-23 months) are interpreted as supine length, and
all linear measurements above 1095 days of life (age 36+ months) are interpreted as standing height.
Note: at the moment, all LENGTHCM will be converted to HEIGHTCM. In the future, the algorithm will be updated to consider this difference.}

\item{agedays}{Numeric vector containing the age in days at each measurement.}

\item{sex}{Vector identifying the gender of the subject, may be 'M', 'm', or 0 for males, vs. 'F',
'f' or 1 for females.}

\item{measurement}{Numeric vector containing the actual measurement data.  Weight must be in
kilograms (kg), and linear measurements (height vs. length) in centimeters (cm).}

\item{orig.exclude}{Vector of exclusion assessment results from cleangrowth()}

\item{exclude_opt}{Number from 0 to 3 indicating which option to use to handle strings of carried-forwards:
0. no change.
\enumerate{
\item when deciding to exclude values, if we have a string of carried forwards,
drop the most deviant value, and all CFs in the same string, and move on as
normal.
\item when deciding to exclude values, if the most deviant in a
string of carried forwards is flagged, check all the CFs in that
string from 1:N. Exclude all after the first that is flagged for
exclusion when comparing to the Include before and after. Do not
remove things designated as include.
\item when deciding to exclude values, if the most deviant in a
string of carried forwards is flagged, check all the CFs in that
string from 1:N. Exclude all after the first that is flagged for
exclusion when comparing to the Include before and after. Make sure
remove things designated as include.
}}

\item{sd.recenter}{Data frame or table with median SD-scores per day of life}

\item{ewma.exp}{Exponent to use for weighting measurements in the exponentially weighted moving
average calculations. Defaults to -1.5. This exponent should be negative in order to weight growth
measurements closer to the measurement being evaluated more strongly. Exponents that are further from
zero (e.g. -3) will increase the relative influence of measurements close in time to the measurement
being evaluated compared to using the default exponent.}

\item{ref.data.path}{Path to reference data. If not supplied, the year 2000
Centers for Disease Control (CDC) reference data will be used.}

\item{quietly}{Determines if function messages are to be displayed and if log files (parallel only)
are to be generated. Defaults to TRUE.}

\item{minfactor}{Sweep variable for computing mindiff.next.ht in 15f, default 0.5}

\item{maxfactor}{Sweep variable for computing maxdiff.next.ht in 15f, default 2}

\item{banddiff}{Sweep variable for computing mindiff.next.ht in 15f, default 3}

\item{banddiff_plus}{Sweep variable for computing maxdiff.next.ht in 15, default 5.5}

\item{min_ht.exp_under}{Sweep variable for computing ht.exp in 15f, default 2}

\item{min_ht.exp_over}{Sweep variable for computing ht.exp in 15f, default 0}

\item{max_ht.exp_under}{Sweep variable for computing ht.exp in 15f, default 0.33}

\item{max_ht.exp_over}{Sweep variable for computing ht.exp in 15f, default 1.5}
}
\value{
Re-evaluated exclusion assessments based on height velocity.
}
\description{
adjustcarryforward
\code{adjustcarryforward} Uses absolute height velocity to identify values
excluded as carried forward values for reinclusion.
}
\examples{
\donttest{
# Run on a small subset of given data
df <- as.data.frame(syngrowth)
df <- df[df$subjid \%in\% unique(df[, "subjid"])[1:2], ]
clean_df <- cbind(df,
                  "gcr_result" = cleangrowth(df$subjid,
                                             df$param,
                                             df$agedays,
                                             df$sex,
                                             df$measurement))

# Adjust carry forward values in cleaned data
adj_clean <- adjustcarryforward(subjid = clean_df$subjid,
                                param = clean_df$param,
                                agedays = clean_df$agedays,
                                sex = clean_df$sex,
                                measurement = clean_df$measurement,
                                orig.exclude = clean_df$gcr_result)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{recode_sex}
\alias{recode_sex}
\title{Recode binary sex variable for compatibility}
\usage{
recode_sex(
  input_data,
  sourcecol = "sex",
  sourcem = "0",
  sourcef = "1",
  targetcol = "sex_recoded",
  targetm = 1L,
  targetf = 2L
)
}
\arguments{
\item{input_data}{a data frame or data table to be transformed. Expects a source column containing a binary sex variable.}

\item{sourcecol}{name of sex descriptor column. Defaults to "sex"}

\item{sourcem}{variable indicating "male" sex in input data. Defaults to "0"}

\item{sourcef}{variable indicating "female" sex in input data. Defaults to "1"}

\item{targetcol}{desired name of recoded sex descriptor column. Defaults to "sex_recoded"}

\item{targetm}{desired name of recoded sex variable indicating "male" sex in output data. Defaults to 1}

\item{targetf}{desired name of recoded sex variable indicating "female" sex in output data. Defaults to 2}
}
\value{
Returns a data table with recoded sex variables.
}
\description{
\code{recode_sex} recodes a binary sex variable for a given source column in a data frame or data table.
Useful in transforming output from growthcleanr::cleangrowth() into a format suitable for growthcleanr::ext_bmiz().
}
\examples{
# Run on given data
df <- as.data.frame(syngrowth)

# Run with all defaults
df_r <- recode_sex(df)

# Specify different targets
df_rt <- recode_sex(df, targetcol = "sexr", targetm = "Male", targetf = "Female")

# Specify different inputs
df_ri <- recode_sex(df_rt, sourcecol = "sexr", sourcem = "Male", sourcef = "Female")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{ref_for_corrz}
\alias{ref_for_corrz}
\title{Fenton-Based Corrected Z-Score Reference}
\description{
Reference values for premature infant corrected z-score calculation, derived
from data extracted from published Fenton (2013) growth curves. Provides
weight, length, and head circumference reference values by gestational age
and sex.
}
\section{ref_for_corrz.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{bmianthro}
\alias{bmianthro}
\title{BMI Anthro}
\description{
Part of default CDC-derived tables
}
\details{
Contains BMI data for calculating BMI
}
\section{bmianthro.txt.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{weianthro}
\alias{weianthro}
\title{Weight Anthro Table}
\description{
Part of default CDC-derived tables
}
\details{
Contains median and standard deviation for weight by age and gender
}
\section{weianthro.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{who_ht_maxvel_2sd}
\alias{who_ht_maxvel_2sd}
\title{WHO Maximum Height Velocity for (2\eqn{\sigma})}
\description{
Part of default WHO-derived tables
}
\details{
Contains two standard deviations for the World Health Organization values of
maximum height velocities.
}
\section{who_ht_maxvel_2sd.csv.gz}{


Used in function \code{acf_answers()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{tanner_ht_vel_rev}
\alias{tanner_ht_vel_rev}
\title{Tanner Growth Velocity Table for Infants}
\description{
Part of default CDC-derived tables
}
\details{
Contains velocities for growth pre-calculated by CDC, used for the infants
algorithm
}
\section{tanner_ht_vel_rev.csv.gz}{


Used in function \code{cleangrowth()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extdata.R
\name{tanner_ht_vel_with_2sd}
\alias{tanner_ht_vel_with_2sd}
\title{Tanner Growth Velocity Table with (2\eqn{\sigma})}
\description{
Part of default CDC-derived tables
}
\details{
Contains velocities for growth pre-calculated by CDC, including those 2
standard deviations away.
}
\section{tanner_ht_vel_with_2sd.csv.gz}{


Used in function \code{acf_answers()}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/testacf_func.R
\name{testacf}
\alias{testacf}
\title{Function to test adjust carried forward}
\usage{
testacf(
  infile,
  seed = 7,
  searchtype = "random",
  grid.length = 9,
  writeout = FALSE,
  outfile = paste0("test_adjustcarryforward_", format(Sys.time(),
    "\%m-\%d-\%Y_\%H-\%M-\%S")),
  quietly = FALSE,
  param = "none",
  debug = FALSE,
  maxrecs = 0,
  exclude_opt = 0,
  add_answers = TRUE
)
}
\arguments{
\item{infile}{Input data frame/data table, cleaned by cleangrowth(), with columns as described in main README.md}

\item{seed}{Numeric random seed, used only when performing random search}

\item{searchtype}{Type of search to perform: random (default), line-grid,
full-grid}

\item{grid.length}{Number of steps in grid to search}

\item{writeout}{Write output to file? Default FALSE.}

\item{outfile}{"Output file name, default 'test_adjustcarrforward_DATE_TIME', where DATE is the current system date and time"}

\item{quietly}{Verbose progress info}

\item{param}{"none", or data frame to specify which parameters to run full search on, and values to use if not, used only when performing full-grid search}

\item{debug}{Produce extra data files for debugging}

\item{maxrecs}{Limit to specified # subjects, default 0 (no limit)}

\item{exclude_opt}{Type of exclusion method for carried forward strings, 0 to 3. See adjustcarryforward documentation for more information}

\item{add_answers}{TRUE or FALSE, indicating whether or not to add answers (definely include/exclude) for the given dataset. Defaults to TRUE}
}
\value{
A list containing:
testacf_res: data frame with adjustcarryforward results for each run,
params: a data frame containing parameter values for each run.
debug_filtered_data: data frame with original data, returned if debug TRUE
}
\description{
The goal of this script is to consider the height values that growthcleanr
excludes as “carried forward” for potential re-inclusion by using a reverse
absolute height velocity check based on step 15 of the Daymont et al.
algorithm
}
